#ifndef _LENS_SYSTEM_HH
#define _LENS_SYSTEM_HH

#include "elements/SphericalSpectral.hh"
#include "elements/CylindricalSpectral.hh"
#include "elements/Cylindrical3.hh"
#include "elements/Spherical5.hh"
#include "elements/Propagation.hh"
#include "elements/OpticalMaterial.hh"
#include "spectrum.h"

#include <cfloat>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <inttypes.h>

namespace poly {


template<typename scalar, int numVarsIn, int numVarsOut>
class SpectralLensFunctor
{
public:
  static inline void transmitCylindrical(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
  {
    refractCylindricalSpectral(poly, R, A1, B1, A2, B2);
  }

  static inline void transmitSpherical(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
  {
    refractSphericalSpectral(poly, R, A1, B1, A2, B2);
  }

};

template<typename scalar, int numVarsIn, int numVarsOut>
class GreyLensFunctor
{
public:
  static inline void transmitCylindrical(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float n1, float B1, float n2, float B2)
  {
    refractCylindrical3(poly, R, n1, n2);
    poly[4].unity(4);
  }

  static inline void transmitSpherical(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float n1, float B1, float n2, float B2)
  {
    refractSpherical5(poly, R, n1, n2);
    poly[4].unity(4);
  }
};

class LensSystem
{
public:
    LensSystem(const char *filename)
    {
        strncpy(this->filename, filename, 1024);
        focal_length = 0.0f;
        interface_count = -1;
        readFile();
    }

    /* estimate back focal length. */
    float estimateFocalLength(int degree, float zoom)
    {
      PolyT5f system, aperture;
      system.truncDegree = aperture.truncDegree = degree;
      float aperture_radius;
      getSystem<float,5,5>(system, aperture, aperture_radius, .0f, .0f, degree, false);
      float light[5] = {10.0f, 10.0f, 0.0f, 0.0f, 0.550f};
      float sensor[5] = {0.0f};
      system.evaluate(light, sensor);
      // now intersect with optical axis:
      const float f_x = - sensor[0]/sensor[2];
      const float f_y = - sensor[1]/sensor[3];
      // since our input only varied in x, this should be the value we're looking for:
      return getThickness(lenses.size()-1, zoom) + (f_x + f_y)*.5f;
    }

    /*
     * returns the adjoint lens system, that is computing the polynomial starting at the sensor
     * and working its way to the outgoing pupil. this is an approximation of the inverse of
     * the system constructed the other way (but since it's a polynomial itself obviously only an
     * approximation)
     */
    template<typename scalar, int numVarsIn, int numVarsOut>
    void getAdjointSystem(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance, // XXX remove this?
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
      getAdjointSystem_internal<scalar, numVarsIn, numVarsOut, SpectralLensFunctor<scalar,numVarsIn,numVarsOut> >(
          system, aperture, aperture_radius, sensor_distance, zoom, degree, skipBFL);
    }

    template<typename scalar, int numVarsIn, int numVarsOut>
    void getSystem(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance,
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
      getSystem_internal<scalar, numVarsIn, numVarsOut, SpectralLensFunctor<scalar,numVarsIn,numVarsOut> >(
          system, aperture, aperture_radius, sensor_distance, zoom, degree, skipBFL);
    }

    template<typename scalar, int numVarsIn, int numVarsOut>
    void getAdjointSystemGrey(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance, // XXX remove this?
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
      getAdjointSystem_internal<scalar, numVarsIn, numVarsOut, GreyLensFunctor<scalar,numVarsIn,numVarsOut> >(
          system, aperture, aperture_radius, sensor_distance, zoom, degree, skipBFL);
    }

    template<typename scalar, int numVarsIn, int numVarsOut>
    void getSystemGrey(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance,
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
      getSystem_internal<scalar, numVarsIn, numVarsOut, GreyLensFunctor<scalar,numVarsIn,numVarsOut> >(
          system, aperture, aperture_radius, sensor_distance, zoom, degree, skipBFL);
    }

    template<typename scalar, int numVarsIn, int numVarsOut, class LensFunctor>
    void getAdjointSystem_internal(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance, // XXX remove this?
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
        system.truncDegree = aperture.truncDegree = degree;

        if(lenses.size() == 0) readFile();

        // temp transforms:
        truncPolySystem<scalar,numVarsIn,numVarsOut> interface, propagation;
        interface.truncDegree = propagation.truncDegree = degree;

        int interfaces = 0;

        float old_n_550 = 1.0f, old_thickness = 0.0f;
        float A1 = 1.0f, B1 = 0.0f, A2 = 1.0f, B2 = 1.0f; // cauchy: n = A + B/lambda^2
        float R1 = 0.0f, D = 0.0f, f1 = FLT_MAX;
        for(int32_t i=lenses.size()-1;i>=0;i--)
        {
            if((i < (int32_t)lenses.size() - 1) && A2 != 1.0f)
            {
                // not first run or coming from air, we can check if we have to update the focal length estimate:
                const float R2 = lenses[i].lens_radius;
                // iteratively use lensmaker's equation and compound lens equation to get focal length of the system:
                const float one_over_f = (old_n_550 - 1.0f) * (1.0f/R1 - 1.0f/R2 + (old_n_550-1.0f)*old_thickness/(old_n_550*R1*R2));
                f1 = 1.0f/(one_over_f + 1.0f/f1 - D/f1*one_over_f);
                D = 0.0f;
                // fprintf(stderr, "[LensSystem] estimated focal length %f (added %f distance %f)\n", f1, 1.0/one_over_f, D);
            }

            if(i == 0 || !strcasecmp(lenses[i-1].material, "air") || !strcasecmp(lenses[i-1].material, "IRIS"))
            {
                // only count */air and air/* but not air/air interfaces (arise for apertures in lens descriptions):
                if(A2 != 1.0f) interfaces++;
                A2 = 1.0f;
                B2 = 0.0f;
                if(i > 0) D += getThickness(i-1, zoom);
            }
            else
            {
                spectrum_cauchy_from_abbe(lenses[i-1].ior, lenses[i-1].vno, &A2, &B2);

                old_thickness = getThickness(i-1, zoom);
                old_n_550 = lenses[i-1].mat.getIndex(550.0f);

                interfaces++;
            }
            propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(i, zoom));
            if(lenses[i].anamorphic)
              LensFunctor::transmitCylindrical(interface, - lenses[i].lens_radius, A1, B1, A2, B2);
            else
              LensFunctor::transmitSpherical(interface, - lenses[i].lens_radius, A1, B1, A2, B2);
            if(skipBFL && i == (int32_t)lenses.size() - 1)
              ;// nothing
            if(i == (int32_t)lenses.size() - 1)
                system = propagation >> interface;
            else
                system = system >> propagation >> interface;

            // store aperture system after iris was processed, we're adjoint (bottom up) so we want the propagation to take place first
            // interface will be a no-op anyways.
            // respect housing_radius here and store to sample aperture.
            if(!strcasecmp(lenses[i].material, "IRIS"))
            {
                aperture = system;
                aperture_radius = lenses[i].housing_radius;
            }


            R1 = lenses[i].lens_radius;
            A1 = A2;
            B1 = B2;
        }
        interface_count = interfaces;
        focal_length = f1;
        // fprintf(stderr, "[LensSystem] estimated adjoint focal length %f\n", focal_length);
    }

    template<typename scalar, int numVarsIn, int numVarsOut, class LensFunctor>
    void getSystem_internal(truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& aperture,
            float &aperture_radius,
            const float sensor_distance,
            const float zoom,
            const int degree,
            bool skipBFL = false)
    {
        if(lenses.size() == 0) readFile();

        system.truncDegree = degree;
        aperture.truncDegree = degree;

        truncPolySystem<scalar,numVarsIn,numVarsOut> interface, propagation;
        interface.truncDegree = propagation.truncDegree = degree;

        int interfaces = 0;

        bool firstrun = true;
        float old_n_550 = 1.0f, old_thickness = 0.0f;
        float A1 = 1.0f, B1 = 0.0f, A2 = 1.0f, B2 = 0.0f;
        float R1 = 0.0f, D = 0.0f, f1 = FLT_MAX;
        for(uint32_t i=0;i<lenses.size();i++)
        {
            if(!firstrun && A2 != 1.0f)
            {
                // not first run or coming from air, we can check if we have to update the focal length estimate:
                const float R2 = lenses[i].lens_radius;
                // iteratively use lensmaker's equation and compound lens equation to get focal length of the system:
                const float one_over_f = (old_n_550 - 1.0f) * (1.0f/R1 - 1.0f/R2 + (old_n_550-1.0f)*old_thickness/(old_n_550*R1*R2));
                f1 = 1.0f/(one_over_f + 1.0f/f1 - D/f1*one_over_f);
                D = 0.0f;
                // fprintf(stderr, "[LensSystem] estimated focal length %f (added %f distance %f)\n", f1, 1.0/one_over_f, D);
            }

            // respect housing_radius here and store to sample aperture.
            if(!strcasecmp(lenses[i].material, "IRIS"))
            {
                aperture = system;
                aperture_radius = lenses[i].housing_radius;
            }

            if(!strcasecmp(lenses[i].material, "air") || !strcasecmp(lenses[i].material, "IRIS"))
            {
                // only count */air and air/* but not air/air interfaces (arise for apertures in lens descriptions):
                if(A2 != 1.0f) interfaces++;
                A2 = 1.0f;
                B2 = 0.0f;
                D += getThickness(i, zoom);
            }
            else
            {
                spectrum_cauchy_from_abbe(lenses[i].ior, lenses[i].vno, &A2, &B2);

                old_thickness = getThickness(i, zoom);
                old_n_550 = lenses[i].mat.getIndex(550.0f);

                interfaces++;
            }
            if(lenses[i].anamorphic)
              LensFunctor::transmitCylindrical(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            else
              LensFunctor::transmitSpherical(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(i, zoom));
            if(firstrun)
                system = interface >> propagation;
            else if(skipBFL && (i == lenses.size() - 1))
                system = system >> interface; // don't propagate BFL
            else
                system = system >> interface >> propagation;

            firstrun = false;
            R1 = lenses[i].lens_radius;
            A1 = A2;
            B1 = B2;
        }
        interface_count = interfaces;
        focal_length = f1;
        // fprintf(stderr, "[LensSystem] estimated focal length %f\n", focal_length);
    }

    float getFocusDistance(const float lambda, const float object_distance)
    {
        const float focus_dist = 1.0f/(1.0f/focal_length - 1.0f/object_distance);
        fprintf(stderr, "[LensSystem] estimated in-focus sensor distance %f\n", focus_dist);
        return focus_dist;
    }

    bool sampleAperture(float &x, float &y, float r1, float r2, const int radius)
    {
        const int blades = 5;
        const int tri = (int)(r1*blades);
        // rescale:
        r1 = r1*blades - tri;
        
        // sample triangle:
        float a = sqrtf(r1);
        float b = (1.0f-r2)*a;
        float c = r2*a;

        float p1[2], p2[2];

        sincosf(2.0f*M_PI/blades * (tri+1), p1, p1+1);
        sincosf(2.0f*M_PI/blades * tri, p2, p2+1);

        x = radius * (b * p1[1] + c * p2[1]);
        y = radius * (b * p1[0] + c * p2[0]);

        return true;
    }

    int getNumFlares()
    {
        // only primary ghosts, i.e. T+RT*RT+ type paths (exactly two R)
        int n = 0;
        for(uint32_t i=1;i<lenses.size();i++) n += i;
        // fprintf(stderr, "[LensSystem] number of flares: %d\n", n);
        return n;
    }

    template<typename scalar, int numVarsIn, int numVarsOut>
    void getFlareSystem(
            truncPolySystem<scalar,numVarsIn,numVarsOut>& system,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& apertureT,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& apertureTR,
            truncPolySystem<scalar,numVarsIn,numVarsOut>& apertureTRR,
            float &apertureTradius,
            float &apertureTRradius,
            float &apertureTRRradius,
            const int num,
            const float sensor_distance,
            const float zoom,
            const int degree)
    {
        if(lenses.size() == 0) readFile();

        system.truncDegree = degree;
        apertureT.truncDegree = degree;
        apertureTR.truncDegree = degree;
        apertureTRR.truncDegree = degree;

        // switch off clipping:
        apertureTradius = 0.0f;
        apertureTRradius = 0.0f;
        apertureTRRradius = 0.0f;

        truncPolySystem<scalar,numVarsIn,numVarsOut> interface, interface1, interface2;
        truncPolySystem<scalar,numVarsIn,numVarsOut> propagation, propagation1, propagation2;
        interface.truncDegree = interface1.truncDegree = interface2.truncDegree = propagation.truncDegree = propagation1.truncDegree = propagation2.truncDegree = degree;

        // walk the interfaces, take route dictated by num (requested flare type)
        // get first R interface:
        uint32_t iface1 = 1, iface2 = 0;
        for(uint32_t n=0;iface1<lenses.size();iface1++)
        {
            n += iface1;
            if((uint32_t)num < n)
            {
                // our flare is in this block of numbers, so iface1 is found.
                iface2 = n - num - 1;
                break;
            }
        }
#if 0
        // avoid log flooding:
        static uint32_t printiface1 = 0;
        if(printiface1 != iface1)
            fprintf(stderr, "[LensSystem::getFlare] using interfaces %d * (of %d)\n", iface1, (uint32_t)lenses.size());
        printiface1 = iface1;
#endif

        assert(iface1 > 0);
        assert(iface1 < lenses.size());
        assert(iface2 >= 0);
        assert(iface2 < iface1);

#if 0
        // only print anamorphic lensflares, because they are interesting:
        if(!lenses[iface1].anamorphic && !lenses[iface2].anamorphic)
        {
            system = 0;
            return;
        }
#endif
#if 0
        // don't print anamorphic lensflares, because they look silly:
        if(lenses[iface1].anamorphic || lenses[iface2].anamorphic)
        {
            system = 0;
            return;
        }
#endif
        float A1=1.0f, B1=0.0f, A2=1.0f, B2=0.0f;
        spectrum_cauchy_from_abbe(lenses[0].ior, lenses[0].vno, &A2, &B2);

        // path to iface1
        if(lenses[0].anamorphic)
            refractCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[0].lens_radius, A1, B1, A2, B2);
        else
            refractSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[0].lens_radius, A1, B1, A2, B2);
        propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(0, zoom));
        system = interface >> propagation;

        for(uint32_t i=1;i<iface1;i++)
        {
            // to be able to clip at aperture with housing_radius, remember for this flare
            if(!strcasecmp(lenses[i].material, "IRIS"))
            {
                apertureT = system;
                apertureTradius = lenses[i].housing_radius;
            }

            spectrum_cauchy_from_abbe(lenses[i-1].ior, lenses[i-1].vno, &A1, &B1);
            spectrum_cauchy_from_abbe(lenses[i].ior, lenses[i].vno, &A2, &B2);
            if(lenses[i].anamorphic)
                refractCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            else
                refractSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(i, zoom));
            system = system >> interface >> propagation;
        }

        spectrum_cauchy_from_abbe(lenses[iface1-1].ior, lenses[iface1-1].vno, &A1, &B1);
        spectrum_cauchy_from_abbe(lenses[iface1].ior, lenses[iface1].vno, &A2, &B2);
        // iface1: R
        if(lenses[iface1].anamorphic)
            reflectCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface1, lenses[iface1].lens_radius, A1, B1, A2, B2);
        else
            reflectSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface1, lenses[iface1].lens_radius, A1, B1, A2, B2);
        propagate<scalar,numVarsIn,numVarsOut>(propagation1, getThickness(iface1-1, zoom));
        system = system >> interface1 >> propagation1;

        // path to iface2 (this is reverse direction, i.e. sensor to outgoing pupil)
        for(uint32_t i=iface1-1;i>iface2;i--)
        {
            if(!strcasecmp(lenses[i].material, "IRIS"))
            {
                apertureTR = system;
                apertureTRradius = lenses[i].housing_radius;
            }
            spectrum_cauchy_from_abbe(lenses[i+1].ior, lenses[i+1].vno, &A1, &B1);
            spectrum_cauchy_from_abbe(lenses[i].ior, lenses[i].vno, &A2, &B2);
            if(lenses[i].anamorphic)
              refractCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            else
              refractSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(i-1, zoom));
            system = system >> interface >> propagation;
        }
        spectrum_cauchy_from_abbe(lenses[iface2].ior, lenses[iface2].vno, &A1, &B1);
        if(iface2 == 0)
        { A2 = 1.0; B2 = 0.0f; }
        else spectrum_cauchy_from_abbe(lenses[iface2-1].ior, lenses[iface2-1].vno, &A2, &B2);
        // iface2: R (with reverse sign)
        if(lenses[iface2].anamorphic)
          reflectCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface2, -lenses[iface2].lens_radius, A1, B1, A2, B2);
        else
          reflectSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface2, -lenses[iface2].lens_radius, A1, B1, A2, B2);
        propagate<scalar,numVarsIn,numVarsOut>(propagation2, getThickness(iface2, zoom));
        system = system >> interface2 >> propagation2;

        // ..and out to sensor
        for(uint32_t i=iface2+1;i<lenses.size();i++)
        {
            if(!strcasecmp(lenses[i].material, "IRIS"))
            {
                apertureTRR = system;
                apertureTRRradius = lenses[i].housing_radius;
            }
            spectrum_cauchy_from_abbe(lenses[i-1].ior, lenses[i-1].vno, &A1, &B1);
            spectrum_cauchy_from_abbe(lenses[i].ior, lenses[i].vno, &A2, &B2);
            if(lenses[i].anamorphic)
              refractCylindricalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            else
              refractSphericalSpectral<scalar,numVarsIn,numVarsOut>(interface, lenses[i].lens_radius, A1, B1, A2, B2);
            propagate<scalar,numVarsIn,numVarsOut>(propagation, getThickness(i, zoom));
            system = system >> interface >> propagation;
        }
    }
    struct lens_t
    {
        float lens_radius, thickness_short, thickness_mid, thickness_long, ior, vno, housing_radius;
        char material[32];
        bool anamorphic;
        OpticalMaterial mat;
    };

    float getThickness(int i, float t)
    {
        if(t < .5f) return lenses[i].thickness_short*(1.0f-2.0f*t) + lenses[i].thickness_mid*2.0f*t;
        else        return lenses[i].thickness_mid*(1.0f-2.0f*(t-.5f)) + lenses[i].thickness_long*(t-.5f)*2.0f;
    }

    void readFile()
    {
        std::ifstream is(filename);
        while (is)
        {
            lens_t lens;
            memset(&lens, 0, sizeof(lens_t));
            std::string line;
            std::getline(is, line);
            // rest is a comment..?
            if(!line.compare(0, 2, "//") || !line.compare(0, 1, "#"))
                continue;
            std::stringstream ss(line);
            if(!ss) continue;
            ss >> lens.lens_radius;
            // look for varying distance in form of short/mid/long:
            std::string thickness;
            std::string thicknesses;
            if(!ss) continue;
            ss >> thicknesses; // whitespace separated word for all thicknesses
            std::stringstream ssss(thicknesses);
            if(std::getline(ssss, thickness, '/'))
            {
                std::stringstream sss(thickness);
                sss >> lens.thickness_short;
                lens.thickness_mid = lens.thickness_long = lens.thickness_short;
            }
            if(std::getline(ssss, thickness, '/'))
            {
                std::stringstream sss(thickness);
                sss >> lens.thickness_mid;
            }
            if(std::getline(ssss, thickness, '/'))
            {
                std::stringstream sss(thickness);
                sss >> lens.thickness_long;
            }
            if(!ss) continue;
            ss >> lens.material;
            if(!strncmp(lens.material, "cx_", 3))
            {
                lens.anamorphic = true;
                char tmp[32];
                memset(tmp, 0, 32);
                // chop off prefix
                strncpy(tmp, lens.material, 32);
                strncpy(lens.material, tmp+3, 29);
            }
            if(strcasecmp(lens.material, "air") && strcasecmp(lens.material, "iris"))
            {
                // these are empty for air materials
                if(!ss) continue;
                ss >> lens.ior;
                if(!ss) continue;
                ss >> lens.vno;
            }
            else
            {
                lens.ior = 1.0f;
                lens.vno = 0.0f;
            }
            if(!ss) continue;
            ss >> lens.housing_radius;
            lens.mat.init(lens.material, lens.ior, lens.vno, false);
            lenses.push_back(lens);
            // we don't care about the rest of the line.

            // fprintf(stderr, "%f %f/%f/%f %s %f %f %f\n", lens.lens_radius, lens.thickness_short, lens.thickness_mid, lens.thickness_long, lens.material, lens.ior, lens.vno, lens.housing_radius);
        }
    }

    void GetInfo(
        float *outer_pupil_radius,
        float *inner_pupil_radius,
        float *lens_length,
        float *lens_focal_length,
        float *aperture_pos,
        float *aperture_f1_radius,
        const float zoom)
    {
      *outer_pupil_radius = lenses[0].housing_radius;
      *inner_pupil_radius = lenses.back().housing_radius;
      *aperture_f1_radius = -1.0f;
      float len = 0.0f, apos = 0.0f;
      for(size_t k=0;k<lenses.size();k++)
      {
        len += getThickness(k, zoom);
        if(apos <= 0.0f)
        {
          if(strcasecmp(lenses[k].material, "iris"))
            apos -= getThickness(k, zoom);
          else
          {
            apos = - apos;
            *aperture_f1_radius = lenses[k].housing_radius;
          }
        }
      }
      *lens_length = len;
      *aperture_pos = apos;
      // *lens_focal_length = estimateFocalLength(4, zoom);
      // rely on BFL in file, that way we can adjust it from the data files
      *lens_focal_length = getThickness(lenses.size()-1, zoom);
    }

    std::vector<lens_t> lenses;

    // number of interfaces counted in this lens, which could be causing lens flares:
    int interface_count;
    // estimated compound focal length
    float focal_length;
    // input lens description
    char filename[1024];
};

} // namespace poly

#endif
