#ifndef ELEMENTS_PROPAGATION_TANGENTSPACE_HH
#define ELEMENTS_PROPAGATION_TANGENTSPACE_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void propagateTangentspace(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float *o, float *n, float *a, float *b, float Ra0, float Rb0, float Ra1, float Rb1)
{
  assert(numVarsIn >= 5); // [x,y,dx,dy,f]
  poly.clear();
  const float ox = o[0];
  const float oy = o[1];
  const float oz = o[2];
  const float nx1 = n[0];
  const float ny1 = n[1];
  const float nz1 = n[2];
  const float ax1 = a[0];
  const float ay1 = a[1];
  const float az1 = a[2];
  const float bx1 = b[0];
  const float by1 = b[1];
  const float bz1 = b[2];

  //=================================================== 
  // taylor series degree 3 of position x 
  //=================================================== 
  // 1  :  
  poly[ 0 ].add(Term5f( -pow(nz1,-1)*((ay1*nz1-az1*ny1)*oy+(ax1*nz1-az1*nx1)*ox) ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x  :  
  poly[ 0 ].add(Term5f( (ax1*nz1-az1*nx1)*pow(nz1,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 0 ].add(Term5f( (ay1*nz1-az1*ny1)*pow(nz1,-1) ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 0 ].add(Term5f( pow(nz1,-2)*((ax1*pow(nz1,2)-az1*nx1*nz1)*oz+(ax1*ny1*nz1-az1*nx1*ny1)*oy+(ax1*nx1*nz1-az1*pow(nx1,2))*ox) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 0 ].add(Term5f( -(ax1*nx1*nz1-az1*pow(nx1,2))*pow(nz1,-2) ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*y  :  
  poly[ 0 ].add(Term5f( -(ax1*ny1*nz1-az1*nx1*ny1)*pow(nz1,-2) ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 0 ].add(Term5f( -pow(nz1,-3)*((ax1*nx1*pow(nz1,2)-az1*pow(nx1,2)*nz1)*oz+(ax1*nx1*ny1*nz1-az1*pow(nx1,2)*ny1)*oy+(ax1*pow(nx1,2)*nz1-az1*pow(nx1,3))*ox) ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 0 ].add(Term5f( (ax1*pow(nx1,2)*nz1-az1*pow(nx1,3))*pow(nz1,-3) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*y  :  
  poly[ 0 ].add(Term5f( (ax1*nx1*ny1*nz1-az1*pow(nx1,2)*ny1)*pow(nz1,-3) ,  0 ,  1 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 0 ].add(Term5f( pow(nz1,-4)*((ax1*pow(nx1,2)*pow(nz1,2)-az1*pow(nx1,3)*nz1)*oz+(ax1*pow(nx1,2)*ny1*nz1-az1*pow(nx1,3)*ny1)*oy+(ax1*pow(nx1,3)*nz1-az1*pow(nx1,4))*ox) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy  :  
  poly[ 0 ].add(Term5f( pow(nz1,-2)*((ay1*pow(nz1,2)-az1*ny1*nz1)*oz+(ay1*ny1*nz1-az1*pow(ny1,2))*oy+(ay1*nx1*nz1-az1*nx1*ny1)*ox) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x  :  
  poly[ 0 ].add(Term5f( -(ay1*nx1*nz1-az1*nx1*ny1)*pow(nz1,-2) ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*y  :  
  poly[ 0 ].add(Term5f( -(ay1*ny1*nz1-az1*pow(ny1,2))*pow(nz1,-2) ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy  :  
  poly[ 0 ].add(Term5f( -pow(nz1,-3)*(((ax1*ny1+ay1*nx1)*pow(nz1,2)-2*az1*nx1*ny1*nz1)*oz+((ax1*pow(ny1,2)+ay1*nx1*ny1)*nz1-2*az1*nx1*pow(ny1,2))*oy+((ax1*nx1*ny1+ay1*pow(nx1,2))*nz1-2*az1*pow(nx1,2)*ny1)*ox) ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 0 ].add(Term5f( ((ax1*nx1*ny1+ay1*pow(nx1,2))*nz1-2*az1*pow(nx1,2)*ny1)*pow(nz1,-3) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*y  :  
  poly[ 0 ].add(Term5f( ((ax1*pow(ny1,2)+ay1*nx1*ny1)*nz1-2*az1*nx1*pow(ny1,2))*pow(nz1,-3) ,  0 ,  1 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 0 ].add(Term5f( pow(nz1,-4)*(((2*ax1*nx1*ny1+ay1*pow(nx1,2))*pow(nz1,2)-3*az1*pow(nx1,2)*ny1*nz1)*oz+((2*ax1*nx1*pow(ny1,2)+ay1*pow(nx1,2)*ny1)*nz1-3*az1*pow(nx1,2)*pow(ny1,2))*oy+((2*ax1*pow(nx1,2)*ny1+ay1*pow(nx1,3))*nz1-3*az1*pow(nx1,3)*ny1)*ox) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 0 ].add(Term5f( -pow(nz1,-3)*((ay1*ny1*pow(nz1,2)-az1*pow(ny1,2)*nz1)*oz+(ay1*pow(ny1,2)*nz1-az1*pow(ny1,3))*oy+(ay1*nx1*ny1*nz1-az1*nx1*pow(ny1,2))*ox) ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*x  :  
  poly[ 0 ].add(Term5f( (ay1*nx1*ny1*nz1-az1*nx1*pow(ny1,2))*pow(nz1,-3) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*y  :  
  poly[ 0 ].add(Term5f( (ay1*pow(ny1,2)*nz1-az1*pow(ny1,3))*pow(nz1,-3) ,  0 ,  1 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 0 ].add(Term5f( pow(nz1,-4)*(((ax1*pow(ny1,2)+2*ay1*nx1*ny1)*pow(nz1,2)-3*az1*nx1*pow(ny1,2)*nz1)*oz+((ax1*pow(ny1,3)+2*ay1*nx1*pow(ny1,2))*nz1-3*az1*nx1*pow(ny1,3))*oy+((ax1*nx1*pow(ny1,2)+2*ay1*pow(nx1,2)*ny1)*nz1-3*az1*pow(nx1,2)*pow(ny1,2))*ox) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // dy^3  :  
  poly[ 0 ].add(Term5f( pow(nz1,-4)*((ay1*pow(ny1,2)*pow(nz1,2)-az1*pow(ny1,3)*nz1)*oz+(ay1*pow(ny1,3)*nz1-az1*pow(ny1,4))*oy+(ay1*nx1*pow(ny1,2)*nz1-az1*nx1*pow(ny1,3))*ox) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of position y 
  //=================================================== 
  // 1  :  
  poly[ 1 ].add(Term5f( -pow(nz1,-1)*((by1*nz1-bz1*ny1)*oy+(bx1*nz1-bz1*nx1)*ox) ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x  :  
  poly[ 1 ].add(Term5f( (bx1*nz1-bz1*nx1)*pow(nz1,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 1 ].add(Term5f( (by1*nz1-bz1*ny1)*pow(nz1,-1) ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 1 ].add(Term5f( pow(nz1,-2)*((bx1*pow(nz1,2)-bz1*nx1*nz1)*oz+(bx1*ny1*nz1-bz1*nx1*ny1)*oy+(bx1*nx1*nz1-bz1*pow(nx1,2))*ox) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 1 ].add(Term5f( -(bx1*nx1*nz1-bz1*pow(nx1,2))*pow(nz1,-2) ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*y  :  
  poly[ 1 ].add(Term5f( -(bx1*ny1*nz1-bz1*nx1*ny1)*pow(nz1,-2) ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 1 ].add(Term5f( -pow(nz1,-3)*((bx1*nx1*pow(nz1,2)-bz1*pow(nx1,2)*nz1)*oz+(bx1*nx1*ny1*nz1-bz1*pow(nx1,2)*ny1)*oy+(bx1*pow(nx1,2)*nz1-bz1*pow(nx1,3))*ox) ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 1 ].add(Term5f( (bx1*pow(nx1,2)*nz1-bz1*pow(nx1,3))*pow(nz1,-3) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*y  :  
  poly[ 1 ].add(Term5f( (bx1*nx1*ny1*nz1-bz1*pow(nx1,2)*ny1)*pow(nz1,-3) ,  0 ,  1 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 1 ].add(Term5f( pow(nz1,-4)*((bx1*pow(nx1,2)*pow(nz1,2)-bz1*pow(nx1,3)*nz1)*oz+(bx1*pow(nx1,2)*ny1*nz1-bz1*pow(nx1,3)*ny1)*oy+(bx1*pow(nx1,3)*nz1-bz1*pow(nx1,4))*ox) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy  :  
  poly[ 1 ].add(Term5f( pow(nz1,-2)*((by1*pow(nz1,2)-bz1*ny1*nz1)*oz+(by1*ny1*nz1-bz1*pow(ny1,2))*oy+(by1*nx1*nz1-bz1*nx1*ny1)*ox) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x  :  
  poly[ 1 ].add(Term5f( -(by1*nx1*nz1-bz1*nx1*ny1)*pow(nz1,-2) ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*y  :  
  poly[ 1 ].add(Term5f( -(by1*ny1*nz1-bz1*pow(ny1,2))*pow(nz1,-2) ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy  :  
  poly[ 1 ].add(Term5f( -pow(nz1,-3)*(((bx1*ny1+by1*nx1)*pow(nz1,2)-2*bz1*nx1*ny1*nz1)*oz+((bx1*pow(ny1,2)+by1*nx1*ny1)*nz1-2*bz1*nx1*pow(ny1,2))*oy+((bx1*nx1*ny1+by1*pow(nx1,2))*nz1-2*bz1*pow(nx1,2)*ny1)*ox) ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 1 ].add(Term5f( ((bx1*nx1*ny1+by1*pow(nx1,2))*nz1-2*bz1*pow(nx1,2)*ny1)*pow(nz1,-3) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*y  :  
  poly[ 1 ].add(Term5f( ((bx1*pow(ny1,2)+by1*nx1*ny1)*nz1-2*bz1*nx1*pow(ny1,2))*pow(nz1,-3) ,  0 ,  1 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 1 ].add(Term5f( pow(nz1,-4)*(((2*bx1*nx1*ny1+by1*pow(nx1,2))*pow(nz1,2)-3*bz1*pow(nx1,2)*ny1*nz1)*oz+((2*bx1*nx1*pow(ny1,2)+by1*pow(nx1,2)*ny1)*nz1-3*bz1*pow(nx1,2)*pow(ny1,2))*oy+((2*bx1*pow(nx1,2)*ny1+by1*pow(nx1,3))*nz1-3*bz1*pow(nx1,3)*ny1)*ox) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 1 ].add(Term5f( -pow(nz1,-3)*((by1*ny1*pow(nz1,2)-bz1*pow(ny1,2)*nz1)*oz+(by1*pow(ny1,2)*nz1-bz1*pow(ny1,3))*oy+(by1*nx1*ny1*nz1-bz1*nx1*pow(ny1,2))*ox) ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*x  :  
  poly[ 1 ].add(Term5f( (by1*nx1*ny1*nz1-bz1*nx1*pow(ny1,2))*pow(nz1,-3) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*y  :  
  poly[ 1 ].add(Term5f( (by1*pow(ny1,2)*nz1-bz1*pow(ny1,3))*pow(nz1,-3) ,  0 ,  1 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 1 ].add(Term5f( pow(nz1,-4)*(((bx1*pow(ny1,2)+2*by1*nx1*ny1)*pow(nz1,2)-3*bz1*nx1*pow(ny1,2)*nz1)*oz+((bx1*pow(ny1,3)+2*by1*nx1*pow(ny1,2))*nz1-3*bz1*nx1*pow(ny1,3))*oy+((bx1*nx1*pow(ny1,2)+2*by1*pow(nx1,2)*ny1)*nz1-3*bz1*pow(nx1,2)*pow(ny1,2))*ox) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // dy^3  :  
  poly[ 1 ].add(Term5f( pow(nz1,-4)*((by1*pow(ny1,2)*pow(nz1,2)-bz1*pow(ny1,3)*nz1)*oz+(by1*pow(ny1,3)*nz1-bz1*pow(ny1,4))*oy+(by1*nx1*pow(ny1,2)*nz1-bz1*nx1*pow(ny1,3))*ox) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of direction dx 
  //=================================================== 
  // 1  :  
  poly[ 2 ].add(Term5f( az1*pow(nz1,-1) ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( (ax1*nz1-az1*nx1)*pow(nz1,-2) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 2 ].add(Term5f( -(ax1*nx1*nz1-az1*pow(nx1,2))*pow(nz1,-3) ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term5f( (ax1*pow(nx1,2)*nz1-az1*pow(nx1,3))*pow(nz1,-4) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy  :  
  poly[ 2 ].add(Term5f( (ay1*nz1-az1*ny1)*pow(nz1,-2) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy  :  
  poly[ 2 ].add(Term5f( -((ax1*ny1+ay1*nx1)*nz1-2*az1*nx1*ny1)*pow(nz1,-3) ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 2 ].add(Term5f( ((2*ax1*nx1*ny1+ay1*pow(nx1,2))*nz1-3*az1*pow(nx1,2)*ny1)*pow(nz1,-4) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 2 ].add(Term5f( -(ay1*ny1*nz1-az1*pow(ny1,2))*pow(nz1,-3) ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term5f( ((ax1*pow(ny1,2)+2*ay1*nx1*ny1)*nz1-3*az1*nx1*pow(ny1,2))*pow(nz1,-4) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // dy^3  :  
  poly[ 2 ].add(Term5f( (ay1*pow(ny1,2)*nz1-az1*pow(ny1,3))*pow(nz1,-4) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of direction dy 
  //=================================================== 
  // 1  :  
  poly[ 3 ].add(Term5f( bz1*pow(nz1,-1) ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 3 ].add(Term5f( (bx1*nz1-bz1*nx1)*pow(nz1,-2) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 3 ].add(Term5f( -(bx1*nx1*nz1-bz1*pow(nx1,2))*pow(nz1,-3) ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 3 ].add(Term5f( (bx1*pow(nx1,2)*nz1-bz1*pow(nx1,3))*pow(nz1,-4) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy  :  
  poly[ 3 ].add(Term5f( (by1*nz1-bz1*ny1)*pow(nz1,-2) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy  :  
  poly[ 3 ].add(Term5f( -((bx1*ny1+by1*nx1)*nz1-2*bz1*nx1*ny1)*pow(nz1,-3) ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term5f( ((2*bx1*nx1*ny1+by1*pow(nx1,2))*nz1-3*bz1*pow(nx1,2)*ny1)*pow(nz1,-4) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 3 ].add(Term5f( -(by1*ny1*nz1-bz1*pow(ny1,2))*pow(nz1,-3) ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 3 ].add(Term5f( ((bx1*pow(ny1,2)+2*by1*nx1*ny1)*nz1-3*bz1*nx1*pow(ny1,2))*pow(nz1,-4) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // dy^3  :  
  poly[ 3 ].add(Term5f( (by1*pow(ny1,2)*nz1-bz1*pow(ny1,3))*pow(nz1,-4) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of measurement contribution f 
  //=================================================== 
  // f  :  
  poly[ 4 ].add(Term5f( pow(nz1,2)*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // x*f  :  
  poly[ 4 ].add(Term5f( 2*nx1*pow(nz1,2)*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // x^2*f  :  
  poly[ 4 ].add(Term5f( 3*pow(nx1,2)*pow(nz1,2)*pow(pow(nz1,4)*pow(oz,4)+(4*ny1*pow(nz1,3)*oy+4*nx1*pow(nz1,3)*ox)*pow(oz,3)+(6*pow(ny1,2)*pow(nz1,2)*pow(oy,2)+12*nx1*ny1*pow(nz1,2)*ox*oy+6*pow(nx1,2)*pow(nz1,2)*pow(ox,2))*pow(oz,2)+(4*pow(ny1,3)*nz1*pow(oy,3)+12*nx1*pow(ny1,2)*nz1*ox*pow(oy,2)+12*pow(nx1,2)*ny1*nz1*pow(ox,2)*oy+4*pow(nx1,3)*nz1*pow(ox,3))*oz+pow(ny1,4)*pow(oy,4)+4*nx1*pow(ny1,3)*ox*pow(oy,3)+6*pow(nx1,2)*pow(ny1,2)*pow(ox,2)*pow(oy,2)+4*pow(nx1,3)*ny1*pow(ox,3)*oy+pow(nx1,4)*pow(ox,4),-1) ,  2 ,  0 ,  0 ,  0 ,  1 )); 
  // y*f  :  
  poly[ 4 ].add(Term5f( 2*ny1*pow(nz1,2)*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // x*y*f  :  
  poly[ 4 ].add(Term5f( 6*nx1*ny1*pow(nz1,2)*pow(pow(nz1,4)*pow(oz,4)+(4*ny1*pow(nz1,3)*oy+4*nx1*pow(nz1,3)*ox)*pow(oz,3)+(6*pow(ny1,2)*pow(nz1,2)*pow(oy,2)+12*nx1*ny1*pow(nz1,2)*ox*oy+6*pow(nx1,2)*pow(nz1,2)*pow(ox,2))*pow(oz,2)+(4*pow(ny1,3)*nz1*pow(oy,3)+12*nx1*pow(ny1,2)*nz1*ox*pow(oy,2)+12*pow(nx1,2)*ny1*nz1*pow(ox,2)*oy+4*pow(nx1,3)*nz1*pow(ox,3))*oz+pow(ny1,4)*pow(oy,4)+4*nx1*pow(ny1,3)*ox*pow(oy,3)+6*pow(nx1,2)*pow(ny1,2)*pow(ox,2)*pow(oy,2)+4*pow(nx1,3)*ny1*pow(ox,3)*oy+pow(nx1,4)*pow(ox,4),-1) ,  1 ,  1 ,  0 ,  0 ,  1 )); 
  // y^2*f  :  
  poly[ 4 ].add(Term5f( 3*pow(ny1,2)*pow(nz1,2)*pow(pow(nz1,4)*pow(oz,4)+(4*ny1*pow(nz1,3)*oy+4*nx1*pow(nz1,3)*ox)*pow(oz,3)+(6*pow(ny1,2)*pow(nz1,2)*pow(oy,2)+12*nx1*ny1*pow(nz1,2)*ox*oy+6*pow(nx1,2)*pow(nz1,2)*pow(ox,2))*pow(oz,2)+(4*pow(ny1,3)*nz1*pow(oy,3)+12*nx1*pow(ny1,2)*nz1*ox*pow(oy,2)+12*pow(nx1,2)*ny1*nz1*pow(ox,2)*oy+4*pow(nx1,3)*nz1*pow(ox,3))*oz+pow(ny1,4)*pow(oy,4)+4*nx1*pow(ny1,3)*ox*pow(oy,3)+6*pow(nx1,2)*pow(ny1,2)*pow(ox,2)*pow(oy,2)+4*pow(nx1,3)*ny1*pow(ox,3)*oy+pow(nx1,4)*pow(ox,4),-1) ,  0 ,  2 ,  0 ,  0 ,  1 )); 
  // dx*f  :  
  poly[ 4 ].add(Term5f( 2*nx1*nz1*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  1 ,  0 ,  1 )); 
  // dx*x*f  :  
  poly[ 4 ].add(Term5f( 4*pow(nx1,2)*nz1*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  1 ,  0 ,  1 ,  0 ,  1 )); 
  // dx*y*f  :  
  poly[ 4 ].add(Term5f( 4*nx1*ny1*nz1*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  0 ,  1 ,  1 ,  0 ,  1 )); 
  // dx^2*f  :  
  poly[ 4 ].add(Term5f( (pow(nz1,2)+pow(nx1,2))*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  2 ,  0 ,  1 )); 
  // dy*f  :  
  poly[ 4 ].add(Term5f( 2*ny1*nz1*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*x*f  :  
  poly[ 4 ].add(Term5f( 4*nx1*ny1*nz1*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  1 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*y*f  :  
  poly[ 4 ].add(Term5f( 4*pow(ny1,2)*nz1*pow(pow(nz1,3)*pow(oz,3)+(3*ny1*pow(nz1,2)*oy+3*nx1*pow(nz1,2)*ox)*pow(oz,2)+(3*pow(ny1,2)*nz1*pow(oy,2)+6*nx1*ny1*nz1*ox*oy+3*pow(nx1,2)*nz1*pow(ox,2))*oz+pow(ny1,3)*pow(oy,3)+3*nx1*pow(ny1,2)*ox*pow(oy,2)+3*pow(nx1,2)*ny1*pow(ox,2)*oy+pow(nx1,3)*pow(ox,3),-1) ,  0 ,  1 ,  0 ,  1 ,  1 )); 
  // dx*dy*f  :  
  poly[ 4 ].add(Term5f( 2*nx1*ny1*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  1 ,  1 ,  1 )); 
  // dy^2*f  :  
  poly[ 4 ].add(Term5f( (pow(nz1,2)+pow(ny1,2))*pow(pow(nz1,2)*pow(oz,2)+(2*ny1*nz1*oy+2*nx1*nz1*ox)*oz+pow(ny1,2)*pow(oy,2)+2*nx1*ny1*ox*oy+pow(nx1,2)*pow(ox,2),-1) ,  0 ,  0 ,  0 ,  2 ,  1 ));
}

} // namespace poly

#endif
