#ifndef SURFACE_SCATTERING_HH
#define SURFACE_SCATTERING_HH

#include "poly/truncPolySystem.hh"
#include "poly/Print.hh"
#include "elements/pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void refractSurface(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R_1, float R_2, float n1, float n2, float roughness)
{
  assert(numVarsIn >= 7); // [x,y,dx,dy,f,h_1,h_2]
  poly.clear();
  // pass on h_1 and h_2:
  poly[5].unity(5);
  poly[6].unity(6);
  // conversion beckmann roughness -> phong exponent following walter 2007
  const float exponent = 2.0f/(roughness*roughness) - 2.0f;
  assert(!(exponent != exponent));
  if(n1 == n2)
  {
    for(int k=0;k<5;k++) poly[k].unity(k);
    return;
  }

  //=================================================== 
  // taylor series degree 3 of refracted position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term7f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term7f( -(pow(R_1,2)*n2+pow(R_1,2)*n1)*pow(n2,-1) ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x*y^2  :  
  poly[ 0 ].add(Term7f( -(R_1*R_2*n2+R_1*R_2*n1)*pow(n2,-1) ,  1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 0 ].add(Term7f( (R_1*n2-R_1*n1)*pow(n2,-1) ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*y^2  :  
  poly[ 0 ].add(Term7f( (R_2*n2-R_2*n1)*pow(n2,-1) ,  0 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // h_1*x^2  :  
  poly[ 0 ].add(Term7f( (R_1*n2+R_1*n1)*pow(n2,-1) ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*y^2  :  
  poly[ 0 ].add(Term7f( (R_2*n2+R_2*n1)*pow(n2,-1) ,  0 ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term7f( 1 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^2*y  :  
  poly[ 1 ].add(Term7f( -(R_1*R_2*n2+R_1*R_2*n1)*pow(n2,-1) ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // y^3  :  
  poly[ 1 ].add(Term7f( -(pow(R_2,2)*n2+pow(R_2,2)*n1)*pow(n2,-1) ,  0 ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 1 ].add(Term7f( (R_1*n2-R_1*n1)*pow(n2,-1) ,  2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*y^2  :  
  poly[ 1 ].add(Term7f( (R_2*n2-R_2*n1)*pow(n2,-1) ,  0 ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // h_2*x^2  :  
  poly[ 1 ].add(Term7f( (R_1*n2+R_1*n1)*pow(n2,-1) ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*y^2  :  
  poly[ 1 ].add(Term7f( (R_2*n2+R_2*n1)*pow(n2,-1) ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term7f( (R_1*n2+R_1*n1)*pow(n2,-1) ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(pow(R_1,3)*n1*pow(n2,2)+2*pow(R_1,3)*pow(n1,2)*n2+pow(R_1,3)*pow(n1,3))/2 ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x*y^2  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(R_1*pow(R_2,2)*n1*pow(n2,2)+2*R_1*pow(R_2,2)*pow(n1,2)*n2+R_1*pow(R_2,2)*pow(n1,3))/2 ,  1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term7f( n1*pow(n2,-1) ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(2*pow(R_1,2)*pow(n2,3)+3*pow(R_1,2)*n1*pow(n2,2)+4*pow(R_1,2)*pow(n1,2)*n2+3*pow(R_1,2)*pow(n1,3))/2 ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*y^2  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(2*R_1*R_2*pow(n2,3)+(pow(R_2,2)+2*R_1*R_2)*n1*pow(n2,2)+2*pow(R_2,2)*pow(n1,2)*n2+pow(R_2,2)*pow(n1,3))/2 ,  0 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(R_1*n1*pow(n2,2)-2*R_1*pow(n1,2)*n2-3*R_1*pow(n1,3))/2 ,  1 ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dy*x*y  :  
  poly[ 2 ].add(Term7f( (R_1*R_2*pow(n1,2)*n2+R_1*R_2*pow(n1,3))*pow(n2,-3) ,  1 ,  1 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*dy*y  :  
  poly[ 2 ].add(Term7f( (R_2*pow(n1,2)*n2+R_2*pow(n1,3))*pow(n2,-3) ,  0 ,  1 ,  1 ,  1 ,  0 ,  0 ,  0 )); 
  // dy^2*x  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(R_1*n1*pow(n2,2)-R_1*pow(n1,3))/2 ,  1 ,  0 ,  0 ,  2 ,  0 ,  0 ,  0 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  1 ,  2 ,  0 ,  0 ,  0 )); 
  // h_1  :  
  poly[ 2 ].add(Term7f( -(n2+n1)*pow(n2,-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*x^2  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(2*pow(R_1,2)*pow(n2,3)+5*pow(R_1,2)*n1*pow(n2,2)+6*pow(R_1,2)*pow(n1,2)*n2+3*pow(R_1,2)*pow(n1,3))/2 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*y^2  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(pow(R_2,2)*pow(n2,3)+2*pow(R_2,2)*n1*pow(n2,2)+2*pow(R_2,2)*pow(n1,2)*n2+pow(R_2,2)*pow(n1,3))/2 ,  0 ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*h_1*x  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(R_1*n1*pow(n2,2)+4*R_1*pow(n1,2)*n2+3*R_1*pow(n1,3)) ,  1 ,  0 ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dx^2*h_1  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)-2*pow(n1,2)*n2-3*pow(n1,3))/2 ,  0 ,  0 ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*h_1*y  :  
  poly[ 2 ].add(Term7f( -(R_2*pow(n1,2)*n2+R_2*pow(n1,3))*pow(n2,-3) ,  0 ,  1 ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dy^2*h_1  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  0 ,  2 ,  0 ,  1 ,  0 )); 
  // h_1^2*x  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(2*R_1*pow(n2,3)+5*R_1*n1*pow(n2,2)+6*R_1*pow(n1,2)*n2+3*R_1*pow(n1,3))/2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*h_1^2  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)+4*pow(n1,2)*n2+3*pow(n1,3))/2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // h_1^3  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(pow(n2,3)+2*n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  3 ,  0 )); 
  // h_2*x*y  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(R_1*R_2*pow(n2,3)+2*R_1*R_2*n1*pow(n2,2)+2*R_1*R_2*pow(n1,2)*n2+R_1*R_2*pow(n1,3)) ,  1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*h_2*y  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(R_2*n1*pow(n2,2)+2*R_2*pow(n1,2)*n2+R_2*pow(n1,3)) ,  0 ,  1 ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dy*h_2*x  :  
  poly[ 2 ].add(Term7f( -(R_1*pow(n1,2)*n2+R_1*pow(n1,3))*pow(n2,-3) ,  1 ,  0 ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // dx*dy*h_2  :  
  poly[ 2 ].add(Term7f( -(pow(n1,2)*n2+pow(n1,3))*pow(n2,-3) ,  0 ,  0 ,  1 ,  1 ,  0 ,  0 ,  1 )); 
  // h_1*h_2*y  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(R_2*pow(n2,3)+2*R_2*n1*pow(n2,2)+2*R_2*pow(n1,2)*n2+R_2*pow(n1,3)) ,  0 ,  1 ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*h_1*h_2  :  
  poly[ 2 ].add(Term7f( (pow(n1,2)*n2+pow(n1,3))*pow(n2,-3) ,  0 ,  0 ,  0 ,  1 ,  0 ,  1 ,  1 )); 
  // h_2^2*x  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(R_1*n1*pow(n2,2)+2*R_1*pow(n1,2)*n2+R_1*pow(n1,3))/2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  2 )); 
  // dx*h_2^2  :  
  poly[ 2 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 ,  2 )); 
  // h_1*h_2^2  :  
  poly[ 2 ].add(Term7f( -pow(n2,-3)*(pow(n2,3)+2*n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 ,  2 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted direction dy 
  //=================================================== 
  // y  :  
  poly[ 3 ].add(Term7f( (R_2*n2+R_2*n1)*pow(n2,-1) ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^2*y  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(pow(R_1,2)*R_2*n1*pow(n2,2)+2*pow(R_1,2)*R_2*pow(n1,2)*n2+pow(R_1,2)*R_2*pow(n1,3))/2 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // y^3  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(pow(R_2,3)*n1*pow(n2,2)+2*pow(R_2,3)*pow(n1,2)*n2+pow(R_2,3)*pow(n1,3))/2 ,  0 ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x*y  :  
  poly[ 3 ].add(Term7f( (R_1*R_2*pow(n1,2)*n2+R_1*R_2*pow(n1,3))*pow(n2,-3) ,  1 ,  1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx^2*y  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(R_2*n1*pow(n2,2)-R_2*pow(n1,3))/2 ,  0 ,  1 ,  2 ,  0 ,  0 ,  0 ,  0 )); 
  // dy  :  
  poly[ 3 ].add(Term7f( n1*pow(n2,-1) ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(2*R_1*R_2*pow(n2,3)+(2*R_1*R_2+pow(R_1,2))*n1*pow(n2,2)+2*pow(R_1,2)*pow(n1,2)*n2+pow(R_1,2)*pow(n1,3))/2 ,  2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*y^2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(2*pow(R_2,2)*pow(n2,3)+3*pow(R_2,2)*n1*pow(n2,2)+4*pow(R_2,2)*pow(n1,2)*n2+3*pow(R_2,2)*pow(n1,3))/2 ,  0 ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term7f( (R_1*pow(n1,2)*n2+R_1*pow(n1,3))*pow(n2,-3) ,  1 ,  0 ,  1 ,  1 ,  0 ,  0 ,  0 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  2 ,  1 ,  0 ,  0 ,  0 )); 
  // dy^2*y  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(R_2*n1*pow(n2,2)-2*R_2*pow(n1,2)*n2-3*R_2*pow(n1,3))/2 ,  0 ,  1 ,  0 ,  2 ,  0 ,  0 ,  0 )); 
  // dy^3  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  0 ,  3 ,  0 ,  0 ,  0 )); 
  // h_1*x*y  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(R_1*R_2*pow(n2,3)+2*R_1*R_2*n1*pow(n2,2)+2*R_1*R_2*pow(n1,2)*n2+R_1*R_2*pow(n1,3)) ,  1 ,  1 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*h_1*y  :  
  poly[ 3 ].add(Term7f( -(R_2*pow(n1,2)*n2+R_2*pow(n1,3))*pow(n2,-3) ,  0 ,  1 ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*h_1*x  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(R_1*n1*pow(n2,2)+2*R_1*pow(n1,2)*n2+R_1*pow(n1,3)) ,  1 ,  0 ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy*h_1  :  
  poly[ 3 ].add(Term7f( -(pow(n1,2)*n2+pow(n1,3))*pow(n2,-3) ,  0 ,  0 ,  1 ,  1 ,  0 ,  1 ,  0 )); 
  // h_1^2*y  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(R_2*n1*pow(n2,2)+2*R_2*pow(n1,2)*n2+R_2*pow(n1,3))/2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dy*h_1^2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  2 ,  0 )); 
  // h_2  :  
  poly[ 3 ].add(Term7f( -(n2+n1)*pow(n2,-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*x^2  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(pow(R_1,2)*pow(n2,3)+2*pow(R_1,2)*n1*pow(n2,2)+2*pow(R_1,2)*pow(n1,2)*n2+pow(R_1,2)*pow(n1,3))/2 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*y^2  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(2*pow(R_2,2)*pow(n2,3)+5*pow(R_2,2)*n1*pow(n2,2)+6*pow(R_2,2)*pow(n1,2)*n2+3*pow(R_2,2)*pow(n1,3))/2 ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*h_2*x  :  
  poly[ 3 ].add(Term7f( -(R_1*pow(n1,2)*n2+R_1*pow(n1,3))*pow(n2,-3) ,  1 ,  0 ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dx^2*h_2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  2 ,  0 ,  0 ,  0 ,  1 )); 
  // dy*h_2*y  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(R_2*n1*pow(n2,2)+4*R_2*pow(n1,2)*n2+3*R_2*pow(n1,3)) ,  0 ,  1 ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // dy^2*h_2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)-2*pow(n1,2)*n2-3*pow(n1,3))/2 ,  0 ,  0 ,  0 ,  2 ,  0 ,  0 ,  1 )); 
  // h_1*h_2*x  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(R_1*pow(n2,3)+2*R_1*n1*pow(n2,2)+2*R_1*pow(n1,2)*n2+R_1*pow(n1,3)) ,  1 ,  0 ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dx*h_1*h_2  :  
  poly[ 3 ].add(Term7f( (pow(n1,2)*n2+pow(n1,3))*pow(n2,-3) ,  0 ,  0 ,  1 ,  0 ,  0 ,  1 ,  1 )); 
  // h_1^2*h_2  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(pow(n2,3)+2*n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  2 ,  1 )); 
  // h_2^2*y  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(2*R_2*pow(n2,3)+5*R_2*n1*pow(n2,2)+6*R_2*pow(n1,2)*n2+3*R_2*pow(n1,3))/2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  2 )); 
  // dy*h_2^2  :  
  poly[ 3 ].add(Term7f( pow(n2,-3)*(n1*pow(n2,2)+4*pow(n1,2)*n2+3*pow(n1,3))/2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 ,  2 )); 
  // h_2^3  :  
  poly[ 3 ].add(Term7f( -pow(n2,-3)*(pow(n2,3)+2*n1*pow(n2,2)+2*pow(n1,2)*n2+pow(n1,3))/2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  3 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted measurement contribution f 
  //=================================================== 
  // f  :  
  poly[ 4 ].add(Term7f( -(2*exponent+4)*n1*n2*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // f*x^2  :  
  poly[ 4 ].add(Term7f( ((4*exponent+8)*pow(R_1,2)*n1*n2+(2*exponent+4)*pow(R_1,2)*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // f*y^2  :  
  poly[ 4 ].add(Term7f( ((4*exponent+8)*pow(R_2,2)*n1*n2+(2*exponent+4)*pow(R_2,2)*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  2 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*f*x  :  
  poly[ 4 ].add(Term7f( ((2*exponent+4)*R_1*n1*n2+(2*exponent+4)*R_1*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  1 ,  0 ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*f  :  
  poly[ 4 ].add(Term7f( (exponent+2)*n1*n2*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  0 ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dy*f*y  :  
  poly[ 4 ].add(Term7f( ((2*exponent+4)*R_2*n1*n2+(2*exponent+4)*R_2*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  1 ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dy^2*f  :  
  poly[ 4 ].add(Term7f( (exponent+2)*n1*n2*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  2 ,  1 ,  0 ,  0 )); 
  // f*h_1*x  :  
  poly[ 4 ].add(Term7f( -((2*exponent+4)*R_1*n1*n2+(2*exponent+4)*R_1*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  1 ,  0 ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // f*h_1^2  :  
  poly[ 4 ].add(Term7f( (pow(exponent,2)+2*exponent)*n1*n2*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // f*h_2*y  :  
  poly[ 4 ].add(Term7f( -((2*exponent+4)*R_2*n1*n2+(2*exponent+4)*R_2*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  1 ,  0 ,  0 ,  1 ,  0 ,  1 )); 
  // f*h_2^2  :  
  poly[ 4 ].add(Term7f( (pow(exponent,2)+2*exponent)*n1*n2*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 ,  2 )); 
}

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void reflectSurface(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R_1, float R_2, float n1, float n2, float roughness)
{
  assert(numVarsIn >= 7); // [x,y,dx,dy,f,h_1,h_2]
  poly.clear();
  // pass on h_1 and h_2:
  poly[5].unity(5);
  poly[6].unity(6);
  // conversion beckmann roughness -> phong exponent following walter 2007
  const float exponent = 2.0f/(roughness*roughness) - 2.0f;
  assert(!(exponent != exponent));
  if(n1 == n2)
  {
    // index matched material, go right through.
    for(int k=0;k<5;k++) poly[k].unity(k);
    return;
  }

  //=================================================== 
  // taylor series degree 3 of reflected position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term7f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term7f( -2*pow(R_1,2) ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x*y^2  :  
  poly[ 0 ].add(Term7f( -2*R_1*R_2 ,  1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // h_1*x^2  :  
  poly[ 0 ].add(Term7f( 2*R_1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*y^2  :  
  poly[ 0 ].add(Term7f( 2*R_2 ,  0 ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term7f( 1 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^2*y  :  
  poly[ 1 ].add(Term7f( -2*R_1*R_2 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // y^3  :  
  poly[ 1 ].add(Term7f( -2*pow(R_2,2) ,  0 ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // h_2*x^2  :  
  poly[ 1 ].add(Term7f( 2*R_1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*y^2  :  
  poly[ 1 ].add(Term7f( 2*R_2 ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term7f( 2*R_1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term7f( 2*pow(R_1,3) ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x*y^2  :  
  poly[ 2 ].add(Term7f( 2*R_1*pow(R_2,2) ,  1 ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term7f( 1 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term7f( 6*pow(R_1,2) ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*y^2  :  
  poly[ 2 ].add(Term7f( 2*pow(R_2,2)+2*R_1*R_2 ,  0 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term7f( 2*R_1 ,  1 ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 )); 
  // dy*x*y  :  
  poly[ 2 ].add(Term7f( 2*R_1*R_2 ,  1 ,  1 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*dy*y  :  
  poly[ 2 ].add(Term7f( 2*R_2 ,  0 ,  1 ,  1 ,  1 ,  0 ,  0 ,  0 )); 
  // h_1  :  
  poly[ 2 ].add(Term7f( -2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*x^2  :  
  poly[ 2 ].add(Term7f( -8*pow(R_1,2) ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // h_1*y^2  :  
  poly[ 2 ].add(Term7f( -3*pow(R_2,2) ,  0 ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*h_1*x  :  
  poly[ 2 ].add(Term7f( -8*R_1 ,  1 ,  0 ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dx^2*h_1  :  
  poly[ 2 ].add(Term7f( -2 ,  0 ,  0 ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*h_1*y  :  
  poly[ 2 ].add(Term7f( -2*R_2 ,  0 ,  1 ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // h_1^2*x  :  
  poly[ 2 ].add(Term7f( 8*R_1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*h_1^2  :  
  poly[ 2 ].add(Term7f( 4 ,  0 ,  0 ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // h_1^3  :  
  poly[ 2 ].add(Term7f( -3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  3 ,  0 )); 
  // h_2*x*y  :  
  poly[ 2 ].add(Term7f( -6*R_1*R_2 ,  1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*h_2*y  :  
  poly[ 2 ].add(Term7f( -4*R_2 ,  0 ,  1 ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dy*h_2*x  :  
  poly[ 2 ].add(Term7f( -2*R_1 ,  1 ,  0 ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // dx*dy*h_2  :  
  poly[ 2 ].add(Term7f( -2 ,  0 ,  0 ,  1 ,  1 ,  0 ,  0 ,  1 )); 
  // h_1*h_2*y  :  
  poly[ 2 ].add(Term7f( 6*R_2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*h_1*h_2  :  
  poly[ 2 ].add(Term7f( 2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  1 ,  1 )); 
  // h_2^2*x  :  
  poly[ 2 ].add(Term7f( 2*R_1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 ,  2 )); 
  // dx*h_2^2  :  
  poly[ 2 ].add(Term7f( 2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 ,  2 )); 
  // h_1*h_2^2  :  
  poly[ 2 ].add(Term7f( -3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 ,  2 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dy 
  //=================================================== 
  // y  :  
  poly[ 3 ].add(Term7f( 2*R_2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // x^2*y  :  
  poly[ 3 ].add(Term7f( 2*pow(R_1,2)*R_2 ,  2 ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // y^3  :  
  poly[ 3 ].add(Term7f( 2*pow(R_2,3) ,  0 ,  3 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x*y  :  
  poly[ 3 ].add(Term7f( 2*R_1*R_2 ,  1 ,  1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dy  :  
  poly[ 3 ].add(Term7f( 1 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term7f( 2*R_1*R_2+2*pow(R_1,2) ,  2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*y^2  :  
  poly[ 3 ].add(Term7f( 6*pow(R_2,2) ,  0 ,  2 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term7f( 2*R_1 ,  1 ,  0 ,  1 ,  1 ,  0 ,  0 ,  0 )); 
  // dy^2*y  :  
  poly[ 3 ].add(Term7f( 2*R_2 ,  0 ,  1 ,  0 ,  2 ,  0 ,  0 ,  0 )); 
  // h_1*x*y  :  
  poly[ 3 ].add(Term7f( -6*R_1*R_2 ,  1 ,  1 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*h_1*y  :  
  poly[ 3 ].add(Term7f( -2*R_2 ,  0 ,  1 ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*h_1*x  :  
  poly[ 3 ].add(Term7f( -4*R_1 ,  1 ,  0 ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy*h_1  :  
  poly[ 3 ].add(Term7f( -2 ,  0 ,  0 ,  1 ,  1 ,  0 ,  1 ,  0 )); 
  // h_1^2*y  :  
  poly[ 3 ].add(Term7f( 2*R_2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  2 ,  0 )); 
  // dy*h_1^2  :  
  poly[ 3 ].add(Term7f( 2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  2 ,  0 )); 
  // h_2  :  
  poly[ 3 ].add(Term7f( -2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*x^2  :  
  poly[ 3 ].add(Term7f( -3*pow(R_1,2) ,  2 ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // h_2*y^2  :  
  poly[ 3 ].add(Term7f( -8*pow(R_2,2) ,  0 ,  2 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*h_2*x  :  
  poly[ 3 ].add(Term7f( -2*R_1 ,  1 ,  0 ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dy*h_2*y  :  
  poly[ 3 ].add(Term7f( -8*R_2 ,  0 ,  1 ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // dy^2*h_2  :  
  poly[ 3 ].add(Term7f( -2 ,  0 ,  0 ,  0 ,  2 ,  0 ,  0 ,  1 )); 
  // h_1*h_2*x  :  
  poly[ 3 ].add(Term7f( 6*R_1 ,  1 ,  0 ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dx*h_1*h_2  :  
  poly[ 3 ].add(Term7f( 2 ,  0 ,  0 ,  1 ,  0 ,  0 ,  1 ,  1 )); 
  // h_1^2*h_2  :  
  poly[ 3 ].add(Term7f( -3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  2 ,  1 )); 
  // h_2^2*y  :  
  poly[ 3 ].add(Term7f( 8*R_2 ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 ,  2 )); 
  // dy*h_2^2  :  
  poly[ 3 ].add(Term7f( 4 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 ,  2 )); 
  // h_2^3  :  
  poly[ 3 ].add(Term7f( -3 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  3 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected measurement contribution f 
  //=================================================== 
  // f  :  
  poly[ 4 ].add(Term7f( ((exponent+2)*pow(n2,2)+(2*exponent+4)*n1*n2+(exponent+2)*pow(n1,2))*pow(2*M_PI*pow(n2,2)-4*M_PI*n1*n2+2*M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // f*x^2  :  
  poly[ 4 ].add(Term7f( -((3*exponent+6)*pow(R_1,2)*pow(n2,2)+(6*exponent+12)*pow(R_1,2)*n1*n2+(3*exponent+6)*pow(R_1,2)*pow(n1,2))*pow(2*M_PI*pow(n2,2)-4*M_PI*n1*n2+2*M_PI*pow(n1,2),-1) ,  2 ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // f*y^2  :  
  poly[ 4 ].add(Term7f( -((3*exponent+6)*pow(R_2,2)*pow(n2,2)+(6*exponent+12)*pow(R_2,2)*n1*n2+(3*exponent+6)*pow(R_2,2)*pow(n1,2))*pow(2*M_PI*pow(n2,2)-4*M_PI*n1*n2+2*M_PI*pow(n1,2),-1) ,  0 ,  2 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*f*x  :  
  poly[ 4 ].add(Term7f( -((exponent+2)*R_1*pow(n2,2)+(2*exponent+4)*R_1*n1*n2+(exponent+2)*R_1*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  1 ,  0 ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*f  :  
  poly[ 4 ].add(Term7f( -((exponent+2)*pow(n2,2)+(2*exponent+4)*n1*n2+(exponent+2)*pow(n1,2))*pow(4*M_PI*pow(n2,2)-8*M_PI*n1*n2+4*M_PI*pow(n1,2),-1) ,  0 ,  0 ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dy*f*y  :  
  poly[ 4 ].add(Term7f( -((exponent+2)*R_2*pow(n2,2)+(2*exponent+4)*R_2*n1*n2+(exponent+2)*R_2*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  1 ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dy^2*f  :  
  poly[ 4 ].add(Term7f( -((exponent+2)*pow(n2,2)+(2*exponent+4)*n1*n2+(exponent+2)*pow(n1,2))*pow(4*M_PI*pow(n2,2)-8*M_PI*n1*n2+4*M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  2 ,  1 ,  0 ,  0 )); 
  // f*h_1*x  :  
  poly[ 4 ].add(Term7f( ((exponent+2)*R_1*pow(n2,2)+(2*exponent+4)*R_1*n1*n2+(exponent+2)*R_1*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  1 ,  0 ,  0 ,  0 ,  1 ,  1 ,  0 )); 
  // f*h_1^2  :  
  poly[ 4 ].add(Term7f( -((pow(exponent,2)+2*exponent)*pow(n2,2)+(2*pow(exponent,2)+4*exponent)*n1*n2+(pow(exponent,2)+2*exponent)*pow(n1,2))*pow(4*M_PI*pow(n2,2)-8*M_PI*n1*n2+4*M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // f*h_2*y  :  
  poly[ 4 ].add(Term7f( ((exponent+2)*R_2*pow(n2,2)+(2*exponent+4)*R_2*n1*n2+(exponent+2)*R_2*pow(n1,2))*pow(M_PI*pow(n2,2)-2*M_PI*n1*n2+M_PI*pow(n1,2),-1) ,  0 ,  1 ,  0 ,  0 ,  1 ,  0 ,  1 )); 
  // f*h_2^2  :  
  poly[ 4 ].add(Term7f( -((pow(exponent,2)+2*exponent)*pow(n2,2)+(2*pow(exponent,2)+4*exponent)*n1*n2+(pow(exponent,2)+2*exponent)*pow(n1,2))*pow(4*M_PI*pow(n2,2)-8*M_PI*n1*n2+4*M_PI*pow(n1,2),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 ,  2 )); 
}

} // namespace poly

#endif
