#ifndef ELEMENTS_TANGENTSPACE_HH
#define ELEMENTS_TANGENTSPACE_HH

#include "poly/truncPolySystem.hh"
#include "elements/pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void tangentspace(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float *n, float *a, float *b)
{
  assert(numVarsIn >= 4);
  poly.clear();

  const float hx = n[0], hy = n[1], hz = n[2];
  const float ax = a[0], ay = a[1], az = a[2];
  const float bx = b[0], by = b[1], bz = b[2];

  // ======================================= 
  //  taylor series degree 3 of position x 
  // ======================================= 
  // x  :  
  poly[ 0 ].add(Term4f( (ax*hz-az*hx)*pow(hz,-1) ,  1 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 0 ].add(Term4f( (ay*hz-az*hy)*pow(hz,-1) ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 0 ].add(Term4f( -(2*ax*hx*hz-az*pow(hx,2))*pow(hz,-2) ,  1 ,  0 ,  1 ,  0 )); 
  // dx*y  :  
  poly[ 0 ].add(Term4f( -(2*ax*hy*hz-az*hx*hy)*pow(hz,-2) ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*x  :  
  poly[ 0 ].add(Term4f( pow(hz,-3)*(az*hx*pow(hz,2)+4*ax*pow(hx,2)*hz-2*az*pow(hx,3))/2 ,  1 ,  0 ,  2 ,  0 )); 
  // dx^2*y  :  
  poly[ 0 ].add(Term4f( pow(hz,-3)*(az*hy*pow(hz,2)+4*ax*hx*hy*hz-2*az*pow(hx,2)*hy)/2 ,  0 ,  1 ,  2 ,  0 )); 
  // dy*x  :  
  poly[ 0 ].add(Term4f( -(2*ay*hx*hz-az*hx*hy)*pow(hz,-2) ,  1 ,  0 ,  0 ,  1 )); 
  // dy*y  :  
  poly[ 0 ].add(Term4f( -(2*ay*hy*hz-az*pow(hy,2))*pow(hz,-2) ,  0 ,  1 ,  0 ,  1 )); 
  // dx*dy*x  :  
  poly[ 0 ].add(Term4f( ((2*ax*hx*hy+2*ay*pow(hx,2))*hz-2*az*pow(hx,2)*hy)*pow(hz,-3) ,  1 ,  0 ,  1 ,  1 )); 
  // dx*dy*y  :  
  poly[ 0 ].add(Term4f( ((2*ax*pow(hy,2)+2*ay*hx*hy)*hz-2*az*hx*pow(hy,2))*pow(hz,-3) ,  0 ,  1 ,  1 ,  1 )); 
  // dy^2*x  :  
  poly[ 0 ].add(Term4f( pow(hz,-3)*(az*hx*pow(hz,2)+4*ay*hx*hy*hz-2*az*hx*pow(hy,2))/2 ,  1 ,  0 ,  0 ,  2 )); 
  // dy^2*y  :  
  poly[ 0 ].add(Term4f( pow(hz,-3)*(az*hy*pow(hz,2)+4*ay*pow(hy,2)*hz-2*az*pow(hy,3))/2 ,  0 ,  1 ,  0 ,  2 )); 
   
  // ======================================= 
  //  taylor series degree 3 of position y 
  // ======================================= 
  // x  :  
  poly[ 1 ].add(Term4f( (bx*hz-bz*hx)*pow(hz,-1) ,  1 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 1 ].add(Term4f( (by*hz-bz*hy)*pow(hz,-1) ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 1 ].add(Term4f( -(2*bx*hx*hz-bz*pow(hx,2))*pow(hz,-2) ,  1 ,  0 ,  1 ,  0 )); 
  // dx*y  :  
  poly[ 1 ].add(Term4f( -(2*bx*hy*hz-bz*hx*hy)*pow(hz,-2) ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*x  :  
  poly[ 1 ].add(Term4f( pow(hz,-3)*(bz*hx*pow(hz,2)+4*bx*pow(hx,2)*hz-2*bz*pow(hx,3))/2 ,  1 ,  0 ,  2 ,  0 )); 
  // dx^2*y  :  
  poly[ 1 ].add(Term4f( pow(hz,-3)*(bz*hy*pow(hz,2)+4*bx*hx*hy*hz-2*bz*pow(hx,2)*hy)/2 ,  0 ,  1 ,  2 ,  0 )); 
  // dy*x  :  
  poly[ 1 ].add(Term4f( -(2*by*hx*hz-bz*hx*hy)*pow(hz,-2) ,  1 ,  0 ,  0 ,  1 )); 
  // dy*y  :  
  poly[ 1 ].add(Term4f( -(2*by*hy*hz-bz*pow(hy,2))*pow(hz,-2) ,  0 ,  1 ,  0 ,  1 )); 
  // dx*dy*x  :  
  poly[ 1 ].add(Term4f( ((2*bx*hx*hy+2*by*pow(hx,2))*hz-2*bz*pow(hx,2)*hy)*pow(hz,-3) ,  1 ,  0 ,  1 ,  1 )); 
  // dx*dy*y  :  
  poly[ 1 ].add(Term4f( ((2*bx*pow(hy,2)+2*by*hx*hy)*hz-2*bz*hx*pow(hy,2))*pow(hz,-3) ,  0 ,  1 ,  1 ,  1 )); 
  // dy^2*x  :  
  poly[ 1 ].add(Term4f( pow(hz,-3)*(bz*hx*pow(hz,2)+4*by*hx*hy*hz-2*bz*hx*pow(hy,2))/2 ,  1 ,  0 ,  0 ,  2 )); 
  // dy^2*y  :  
  poly[ 1 ].add(Term4f( pow(hz,-3)*(bz*hy*pow(hz,2)+4*by*pow(hy,2)*hz-2*bz*pow(hy,3))/2 ,  0 ,  1 ,  0 ,  2 )); 
   
  // ======================================= 
  //  taylor series degree 3 of direction dx 
  // ======================================= 
  // 1  :  
  poly[ 2 ].add(Term4f( az*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term4f( (ax*hz-az*hx)*pow(hz,-1)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  0 )); 
  // dx^2  :  
  poly[ 2 ].add(Term4f( -(ax*hx*hz-az*pow(hx,2))*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  2 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term4f( (ax*pow(hx,2)*hz-az*pow(hx,3))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  3 ,  0 )); 
  // dy  :  
  poly[ 2 ].add(Term4f( (ay*hz-az*hy)*pow(hz,-1)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  1 )); 
  // dx*dy  :  
  poly[ 2 ].add(Term4f( -((ax*hy+ay*hx)*hz-2*az*hx*hy)*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  1 )); 
  // dx^2*dy  :  
  poly[ 2 ].add(Term4f( ((2*ax*hx*hy+ay*pow(hx,2))*hz-3*az*pow(hx,2)*hy)*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  2 ,  1 )); 
  // dy^2  :  
  poly[ 2 ].add(Term4f( -(ay*hy*hz-az*pow(hy,2))*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  2 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term4f( ((ax*pow(hy,2)+2*ay*hx*hy)*hz-3*az*hx*pow(hy,2))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  2 )); 
  // dy^3  :  
  poly[ 2 ].add(Term4f( (ay*pow(hy,2)*hz-az*pow(hy,3))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  3 )); 
   
  // ======================================= 
  //  taylor series degree 3 of direction dy 
  // ======================================= 
  // 1  :  
  poly[ 3 ].add(Term4f( bz*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 3 ].add(Term4f( (bx*hz-bz*hx)*pow(hz,-1)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  0 )); 
  // dx^2  :  
  poly[ 3 ].add(Term4f( -(bx*hx*hz-bz*pow(hx,2))*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  2 ,  0 )); 
  // dx^3  :  
  poly[ 3 ].add(Term4f( (bx*pow(hx,2)*hz-bz*pow(hx,3))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  3 ,  0 )); 
  // dy  :  
  poly[ 3 ].add(Term4f( (by*hz-bz*hy)*pow(hz,-1)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  1 )); 
  // dx*dy  :  
  poly[ 3 ].add(Term4f( -((bx*hy+by*hx)*hz-2*bz*hx*hy)*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  1 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term4f( ((2*bx*hx*hy+by*pow(hx,2))*hz-3*bz*pow(hx,2)*hy)*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  2 ,  1 )); 
  // dy^2  :  
  poly[ 3 ].add(Term4f( -(by*hy*hz-bz*pow(hy,2))*pow(hz,-2)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  2 )); 
  // dx*dy^2  :  
  poly[ 3 ].add(Term4f( ((bx*pow(hy,2)+2*by*hx*hy)*hz-3*bz*hx*pow(hy,2))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  1 ,  2 )); 
  // dy^3  :  
  poly[ 3 ].add(Term4f( (by*pow(hy,2)*hz-bz*pow(hy,3))*pow(hz,-3)*pow(std::abs(hz),-1) ,  0 ,  0 ,  0 ,  3 )); 
}

} // namespace poly


#endif
