#ifndef ELEMENTS_PATHSPACE_PROPAGATION_HH
#define ELEMENTS_PATHSPACE_PROPAGATION_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void propagate(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float *n0, float *n1, float l)
{
  assert(numVarsIn >= 5); // [x,y,dx,dy,f]
  poly.clear();
  const float nx0 = n0[0];
  const float ny0 = n0[1];
  const float nz0 = n0[2];
  const float nx1 = n1[0];
  const float ny1 = n1[1];
  const float nz1 = n1[2];
  // TODO: special case on these and then do only rotation (maybe flip)
  assert(fabsf(nz0) > 0.0001f);
  assert(fabsf(nz1) > 0.0001f);
  assert(fabsf(l) > 0.0001f);

  //=================================================== 
  // taylor series degree 3 of position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term5f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 0 ].add(Term5f( l ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 0 ].add(Term5f( l/2 ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dx*dy^2  :  
  poly[ 0 ].add(Term5f( l/2 ,  0 ,  0 ,  1 ,  2 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term5f( 1 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy  :  
  poly[ 1 ].add(Term5f( l ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 1 ].add(Term5f( l/2 ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^3  :  
  poly[ 1 ].add(Term5f( l/2 ,  0 ,  0 ,  0 ,  3 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of direction dx 
  //=================================================== 
  // dx  :  
  poly[ 2 ].add(Term5f( 1 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term5f( 1 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of measurement contribution f 
  //=================================================== 
  // f  :  
  poly[ 4 ].add(Term5f( pow(l,-2) ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // x*f  :  
  poly[ 4 ].add(Term5f( -pow(nz0,-1)*(2*nx0*nz1-2*nx1*nz0)*pow(nz1,-1)*pow(l,-3) ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // x^2*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(3*pow(nx0,2)*pow(nz1,2)-6*nx0*nx1*nz0*nz1+3*pow(nx1,2)*pow(nz0,2))*pow(l,-4) ,  2 ,  0 ,  0 ,  0 ,  1 )); 
  // y*f  :  
  poly[ 4 ].add(Term5f( -pow(nz0,-1)*(2*ny0*nz1-2*ny1*nz0)*pow(nz1,-1)*pow(l,-3) ,  0 ,  1 ,  0 ,  0 ,  1 )); 
  // x*y*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(6*nx0*ny0*pow(nz1,2)+(-6*nx0*ny1-6*nx1*ny0)*nz0*nz1+6*nx1*ny1*pow(nz0,2))*pow(l,-4) ,  1 ,  1 ,  0 ,  0 ,  1 )); 
  // y^2*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(3*pow(ny0,2)*pow(nz1,2)-6*ny0*ny1*nz0*nz1+3*pow(ny1,2)*pow(nz0,2))*pow(l,-4) ,  0 ,  2 ,  0 ,  0 ,  1 )); 
  // dx*f  :  
  poly[ 4 ].add(Term5f( 2*nx1*pow(nz1,-1)*pow(l,-2) ,  0 ,  0 ,  1 ,  0 ,  1 )); 
  // dx*x*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(2*pow(nx0,2)*pow(nz1,2)-6*nx0*nx1*nz0*nz1+4*pow(nx1,2)*pow(nz0,2))*pow(l,-3) ,  1 ,  0 ,  1 ,  0 ,  1 )); 
  // dx*y*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(2*nx0*ny0*pow(nz1,2)-6*nx1*ny0*nz0*nz1+4*nx1*ny1*pow(nz0,2))*pow(l,-3) ,  0 ,  1 ,  1 ,  0 ,  1 )); 
  // dx^2*f  :  
  poly[ 4 ].add(Term5f( -pow(nz1,-2)*(pow(nz1,2)-pow(nx1,2))*pow(l,-2) ,  0 ,  0 ,  2 ,  0 ,  1 )); 
  // dy*f  :  
  poly[ 4 ].add(Term5f( 2*ny1*pow(nz1,-1)*pow(l,-2) ,  0 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*x*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(2*nx0*ny0*pow(nz1,2)-6*nx0*ny1*nz0*nz1+4*nx1*ny1*pow(nz0,2))*pow(l,-3) ,  1 ,  0 ,  0 ,  1 ,  1 )); 
  // dy*y*f  :  
  poly[ 4 ].add(Term5f( pow(nz0,-2)*pow(nz1,-2)*(2*pow(ny0,2)*pow(nz1,2)-6*ny0*ny1*nz0*nz1+4*pow(ny1,2)*pow(nz0,2))*pow(l,-3) ,  0 ,  1 ,  0 ,  1 ,  1 )); 
  // dx*dy*f  :  
  poly[ 4 ].add(Term5f( 2*nx1*ny1*pow(nz1,-2)*pow(l,-2) ,  0 ,  0 ,  1 ,  1 ,  1 )); 
  // dy^2*f  :  
  poly[ 4 ].add(Term5f( -pow(nz1,-2)*(pow(nz1,2)-pow(ny1,2))*pow(l,-2) ,  0 ,  0 ,  0 ,  2 ,  1 )); 
}

} // namespace poly

#endif
