#ifndef ELEMENTS_PATHSPACE_TANGENTSPACE_HH
#define ELEMENTS_PATHSPACE_TANGENTSPACE_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void tangentspace(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float *n, float *a, float *b)
{
  assert(numVarsIn >= 5); // [x,y,dx,dy,f]
  poly.clear();
  poly[4].unity(4); // pass on f
  const float hx = n[0];
  const float hy = n[1];
  const float hz = n[2];
  const float ax = a[0];
  const float ay = a[1];
  const float az = a[2];
  const float bx = b[0];
  const float by = b[1];
  const float bz = b[2];
  // avoid division by zero:
  assert(fabsf(hz) > 0.0001f);
   
  // ======================================= 
  //  taylor series degree 3 of position x 
  // ======================================= 
  // x  :  
  poly[ 0 ].add(Term5f( (ax*hz-az*hx)*pow(hz,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 0 ].add(Term5f( (ay*hz-az*hy)*pow(hz,-1) ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 0 ].add(Term5f( -(2*ax*hx*hz-az*pow(hx,2))*pow(hz,-2) ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*y  :  
  poly[ 0 ].add(Term5f( -(2*ax*hy*hz-az*hx*hy)*pow(hz,-2) ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 0 ].add(Term5f( (2*ax*pow(hx,2)*hz-az*pow(hx,3))*pow(hz,-3) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*y  :  
  poly[ 0 ].add(Term5f( (2*ax*hx*hy*hz-az*pow(hx,2)*hy)*pow(hz,-3) ,  0 ,  1 ,  2 ,  0 ,  0 )); 
  // dy*x  :  
  poly[ 0 ].add(Term5f( -(2*ay*hx*hz-az*hx*hy)*pow(hz,-2) ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*y  :  
  poly[ 0 ].add(Term5f( -(2*ay*hy*hz-az*pow(hy,2))*pow(hz,-2) ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 0 ].add(Term5f( ((2*ax*hx*hy+2*ay*pow(hx,2))*hz-2*az*pow(hx,2)*hy)*pow(hz,-3) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*y  :  
  poly[ 0 ].add(Term5f( ((2*ax*pow(hy,2)+2*ay*hx*hy)*hz-2*az*hx*pow(hy,2))*pow(hz,-3) ,  0 ,  1 ,  1 ,  1 ,  0 )); 
  // dy^2*x  :  
  poly[ 0 ].add(Term5f( (2*ay*hx*hy*hz-az*hx*pow(hy,2))*pow(hz,-3) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*y  :  
  poly[ 0 ].add(Term5f( (2*ay*pow(hy,2)*hz-az*pow(hy,3))*pow(hz,-3) ,  0 ,  1 ,  0 ,  2 ,  0 )); 
   
  // ======================================= 
  //  taylor series degree 3 of position y 
  // ======================================= 
  // x  :  
  poly[ 1 ].add(Term5f( (bx*hz-bz*hx)*pow(hz,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // y  :  
  poly[ 1 ].add(Term5f( (by*hz-bz*hy)*pow(hz,-1) ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dx*x  :  
  poly[ 1 ].add(Term5f( -(2*bx*hx*hz-bz*pow(hx,2))*pow(hz,-2) ,  1 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*y  :  
  poly[ 1 ].add(Term5f( -(2*bx*hy*hz-bz*hx*hy)*pow(hz,-2) ,  0 ,  1 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 1 ].add(Term5f( (2*bx*pow(hx,2)*hz-bz*pow(hx,3))*pow(hz,-3) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^2*y  :  
  poly[ 1 ].add(Term5f( (2*bx*hx*hy*hz-bz*pow(hx,2)*hy)*pow(hz,-3) ,  0 ,  1 ,  2 ,  0 ,  0 )); 
  // dy*x  :  
  poly[ 1 ].add(Term5f( -(2*by*hx*hz-bz*hx*hy)*pow(hz,-2) ,  1 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*y  :  
  poly[ 1 ].add(Term5f( -(2*by*hy*hz-bz*pow(hy,2))*pow(hz,-2) ,  0 ,  1 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 1 ].add(Term5f( ((2*bx*hx*hy+2*by*pow(hx,2))*hz-2*bz*pow(hx,2)*hy)*pow(hz,-3) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx*dy*y  :  
  poly[ 1 ].add(Term5f( ((2*bx*pow(hy,2)+2*by*hx*hy)*hz-2*bz*hx*pow(hy,2))*pow(hz,-3) ,  0 ,  1 ,  1 ,  1 ,  0 )); 
  // dy^2*x  :  
  poly[ 1 ].add(Term5f( (2*by*hx*hy*hz-bz*hx*pow(hy,2))*pow(hz,-3) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dy^2*y  :  
  poly[ 1 ].add(Term5f( (2*by*pow(hy,2)*hz-bz*pow(hy,3))*pow(hz,-3) ,  0 ,  1 ,  0 ,  2 ,  0 )); 
   
  // ======================================= 
  //  taylor series degree 3 of direction dx 
  // ======================================= 
  // 1  :  
  poly[ 2 ].add(Term5f( az ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( ax ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 2 ].add(Term5f( -az/2 ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dy  :  
  poly[ 2 ].add(Term5f( ay ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 2 ].add(Term5f( -az/2 ,  0 ,  0 ,  0 ,  2 ,  0 )); 
   
  // ======================================= 
  //  taylor series degree 3 of direction dy 
  // ======================================= 
  // 1  :  
  poly[ 3 ].add(Term5f( bz ,  0 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 3 ].add(Term5f( bx ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2  :  
  poly[ 3 ].add(Term5f( -bz/2 ,  0 ,  0 ,  2 ,  0 ,  0 )); 
  // dy  :  
  poly[ 3 ].add(Term5f( by ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy^2  :  
  poly[ 3 ].add(Term5f( -bz/2 ,  0 ,  0 ,  0 ,  2 ,  0 ));
}

} // namespace poly

#endif
