
#ifndef DUMP_CODE_H
#define DUMP_CODE_H

#include "poly/truncPolySystem.hh"

using namespace poly;

#define TERM_EPS 0//1e-12

// dump out polynomial source code to a stream
template <class scalar, int numVars>
void print_poly_term_code(FILE *f, const polyTerm<scalar,numVars> term, std::string *varNames = NULL)
{
    fprintf(f, "%g", term.coefficient);
    // faster, less precise. only suitable for lens flares really:
    // fprintf(f, "%.4ff", term.coefficient);
    if (varNames)
    {
        for(int i=0; i<numVars; i++)
        {
            int e = term.exponents[i];
            std::string &vn = varNames[i];
            // c-code friendly:
            for(int k=0;k<e;k++)
                fprintf(f, " * %s", vn.c_str());
        }
    }
    else
    {
        for(int i=0; i<numVars; i++)
        {
            int e = term.exponents[i];
            // c code
            for(int k=0;k<e;k++)
                fprintf(f, " * x%d", i);
        }
    }
}

template <class scalar, int numVars>
bool code_empty(const truncPoly<scalar,numVars> poly)
{
  int N = poly.terms.size();
  const float eps = TERM_EPS;

  if (N)
  {
      for(int i = 0; i < N - 1; i++)
      {
          if(fabsf(poly.terms[i].coefficient) > eps)
          {
              return false;
          }
      }

      if(fabsf(poly.terms[N-1].coefficient) > eps)
      {
          return false;
      }
  }
  return true;
}

template <class scalar, int numVars>
bool print_poly_code(FILE *f, const truncPoly<scalar,numVars> poly, std::string *varNames = NULL)
{
  int N = poly.terms.size();
  const float eps = TERM_EPS;
  bool empty = true;

  if (N)
  {
      for(int i = 0; i < N - 1; i++)
      {
          // if(fabsf(poly.terms[i].coefficient) > eps)
          {
              empty = false;
              print_poly_term_code(f, poly.terms[i], varNames);
          }
          if(fabsf(poly.terms[i+1].coefficient) > eps) fprintf(f, " + ");
      }

      // if(fabsf(poly.terms[N-1].coefficient) > eps)
      {
          empty = false;
          print_poly_term_code(f, poly.terms[N-1], varNames);
      }
      return empty;
  }
  else
  {
      return true;
  }
}

template <class scalar, int numVarsIn, int numVarsOut>
bool print_poly_system_code(FILE *f, const truncPolySystem<scalar,numVarsIn, numVarsOut>& tpt, std::string *varNames = NULL)
{
  bool empty = false;//true;
  // actually don't need dx/dy, only straight x, y:
  for(int i = 0; i < numVarsOut; i++)
  {
    // empty = code_empty(tpt.equations[i]);
    if(empty)
    {
        break; 
    }
    else
    {
        if (varNames)
        { 
          std::string &vn = varNames[i];
            fprintf(f, "const float out_%s = ", vn.c_str());
        }
        else
        {
            fprintf(f, "const float out_x%d = ", i);
        }
        print_poly_code(f,tpt.equations[i],varNames);
        fprintf(f, ";\n");
    }
  }
  return empty;
}

// dump out jacobian source code to a stream
template <class scalar, int numVarsIn, int numVarsOut>
void print_jacobian(FILE *f, truncPolySystem<scalar,numVarsIn,numVarsOut> &system, std::string *varnames)
{
  for (int i = 0; i < numVarsOut; ++i)
  {
    for (int j = 0; j < numVarsIn; ++j)
    {
        fprintf(f, "const float dx%d%d = ", i, j);
        print_poly_code(f, system.equations[i].getDerivative(j), varnames);
        fprintf(f, "+0.0f;\n");
    }
  }
}

template <class scalar, int numVarsIn, int numVarsOut>
void print_reverse_code(
    FILE *f,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& system,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& adjoint,
    std::string *varNames)
{
  const int numVarsMin = std::min(numVarsOut, numVarsIn);
  const int numVarsMax = std::max(numVarsOut, numVarsIn);

  std::string sensor_var[numVarsMax];
  std::string light_var[numVarsMax];
  std::string sensor_r_var[numVarsMax];
  std::string light_r_var[numVarsMax];
  for(int i=0;i<numVarsMax;i++)
  {
    sensor_var[i] = "sensor_" + varNames[i];
    light_var[i] = varNames[i];
    sensor_r_var[i] = "sensor_r_" + varNames[i];
    light_r_var[i] = "light_r_" + varNames[i];
  }

  fprintf(f, "// evaluate approximate reverse polynomial\n");
  // transport light to sensor (predict)
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "float %s = ", sensor_var[i].c_str());
    print_poly_code(f, system.equations[i], varNames);
    fprintf(f, ";\n");
  }
  // fprintf(stderr, "eval: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);

  // arbitrary iteration number, 4 should be almost floating point accuracy
  fprintf(f, "// do newton iterations to get precise answer\n");
  fprintf(f, "for(int it=0;it<8;it++)\n{\n");
  // need to flip direction [dx, dy] to transport back:
  for(int i=0;i<numVarsOut;i++)
    fprintf(f, "  float %s = %s %s;\n",
      sensor_r_var[i].c_str(),
      (i == 2 || i == 3) ? "-" : "", sensor_var[i].c_str());

  // copy over constant dimensions:
  for(int i=numVarsMin;i<numVarsMax;i++) 
    fprintf(f, "  float %s = %s;\n",
        sensor_r_var[i].c_str(), varNames[i].c_str());

  fprintf(f, "  // evaluate adjoint to get error vector\n");
  // evaluate adjoint:
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "  float %s = ", light_r_var[i].c_str());
    print_poly_code(f, adjoint.equations[i], sensor_r_var);
    fprintf(f, ";\n");
  }

  // output will be a little off, lives in light domain again.

  // have to invert light_r's [dx, dy] and copy lambda again to match system again:
  for(int k=2;k<4;k++) fprintf(f, "  %s = - %s;\n", light_r_var[k].c_str(), light_r_var[k].c_str());
  for(int i=numVarsMin;i<numVarsMax;i++) fprintf(f, "  const float %s = %s;\n", light_r_var[i].c_str(),  light_var[i].c_str());

  for(int k=0;k<numVarsMin;k++)  // In
    // XXX weird damping factor:
    // fprintf(f, "  const float d_light_%d = .5f*(%s - %s);\n", k, light_var[k].c_str(), light_r_var[k].c_str());
    fprintf(f, "  const float d_light_%d = (%s - %s);\n", k, light_var[k].c_str(), light_r_var[k].c_str());

  // XXX
#if 0 
  // error check:
  float sqr_err = 0.0f;
  for(int k=0;k<numVarsIn;k++) sqr_err += d_light[k]*d_light[k];
  // fprintf(stderr, "error = %f\n", sqrtf(sqr_err));
#endif

  fprintf(f, "  // move through jacobian and correct sensor point\n");
#define USE_INV 0
#if USE_INV==1
  fprintf(f, "  float J[5][5] = {{0.0f}};\n");
#endif
  fprintf(f, "  float JI[5][5] = {{0.0f}};\n");
  fprintf(f, "  JI[4][4] = 1.0f;\n");
  // how much do we need to adjust the point on the sensor
  // to get the corresponding reversed sensor_r -> light_r = light
  for (int i = 0; i < numVarsMin; i++) // Out
  {
    for (int j = 0; j < numVarsMin; ++j) // In
    {
#if USE_INV==1
      fprintf(f, "  J[%d][%d] = ", i, j);
      print_poly_code(f, adjoint.equations[i].getDerivative(j), sensor_r_var);
#else
      fprintf(f, "  JI[%d][%d] = ", i, j);
      print_poly_code(f, system.equations[i].getDerivative(j), light_r_var);
#endif
      fprintf(f, "+0.0f;\n");
    }
  }
#if USE_INV==1
  fprintf(f, "  lens_mat5inv(J, JI);\n");
#endif

  // move sensor point:
  for(int k=0;k<numVarsMin;k++) //numVarsOut;k++)
  {
    for(int i=0;i<numVarsMin;i++) //In;i++)
      fprintf(f, "  %s %s JI[%d][%d] * d_light_%d;\n", sensor_var[k].c_str(),
#if USE_INV==1
          "-=",
#else
          "+=",
#endif
          k, i, i);
#undef USE_INV
  }
  fprintf(f, "}\n");
}


template <class scalar, int numVarsIn, int numVarsOut>
void print_solve_omega(
    FILE *f,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& system,
    std::string *var)
{
  // input to this is [x,y,dx,dy] and `dist', which is the distance to the start of the polynomial.

  std::string begin_var[numVarsIn];
  for(int k=0;k<numVarsIn;k++) begin_var[k] = "begin_" + var[k];

  // solve the first two rows of the system x1 = P(x0, omega0, ..) for given x1 and x0.
  // 1) first guess, take omega0 = normalise(x1 - x0), leave rest of x0 input unchanged
  //    this has to be done outside, we don't know about the 3d geometry any more.
  // float sqr_err = 0.0f;
  for(int k=0;k<4;k++)
    fprintf(f, "float pred_%s;\n", var[k].c_str());
  fprintf(f, "for(int k=0;k<4;k++)\n{\n");

  // 1.5) compute input to the polynomial, begin_[x,y,dx,dy] from our initial guess and the distance:
  fprintf(f, "  const float %s = %s + dist * %s;\n", begin_var[0].c_str(), var[0].c_str(), var[2].c_str());
  fprintf(f, "  const float %s = %s + dist * %s;\n", begin_var[1].c_str(), var[1].c_str(), var[3].c_str());
  fprintf(f, "  const float %s = %s;\n", begin_var[2].c_str(), var[2].c_str());
  fprintf(f, "  const float %s = %s;\n", begin_var[3].c_str(), var[3].c_str());
  fprintf(f, "  const float %s = %s;\n", begin_var[4].c_str(), var[4].c_str());

  // 2) evaluate what we get x1' = P(x0, omega0, ..)
  for(int k=0;k<4;k++)
  {
    fprintf(f, "  pred_%s = ", var[k].c_str());
    print_poly_code(f, system.equations[k], begin_var);
    fprintf(f, ";\n");
  }

  // 3) evaluate 2x2 submatrix of jacobian dx1/domega0 and step omega0 back to our target x1.
  fprintf(f, "  float dx1_domega0[2][2];\n");
  for(int i=0;i<2;i++) for(int j=0;j<2;j++)
  {
    fprintf(f, "  dx1_domega0[%d][%d] = ", i, j);
    print_poly_code(f, system.equations[i].getDerivative(j+2), begin_var);
    fprintf(f, "+0.0f;\n");
  }

  // 4) invert jacobian (could use adjoint, but who's gonna fight over a 2x2 inversion)
  fprintf(f, "  float invJ[2][2];\n");
  fprintf(f, "  const float invdet = 1.0f/(dx1_domega0[0][0]*dx1_domega0[1][1] - dx1_domega0[0][1]*dx1_domega0[1][0]);\n");
  fprintf(f, "  invJ[0][0] =  dx1_domega0[1][1]*invdet;\n");
  fprintf(f, "  invJ[1][1] =  dx1_domega0[0][0]*invdet;\n");
  fprintf(f, "  invJ[0][1] = -dx1_domega0[0][1]*invdet;\n");
  fprintf(f, "  invJ[1][0] = -dx1_domega0[1][0]*invdet;\n");

  // 5) determine step and update omega0
  fprintf(f, "  const float dx1[2] = {out_x - pred_x, out_y - pred_y};\n");
    // sqr_err = 0.0f;
    // for(int k=0;k<2;k++) sqr_err += dx1[k]*dx1[k];
  fprintf(f, "  for(int i=0;i<2;i++)\n  {\n");
  for(int k=0;k<2;k++)
  fprintf(f, "    %s += invJ[%d][i]*dx1[i];\n", var[k+2].c_str(), k);
  fprintf(f, "  }\n");

  fprintf(f, "}\n"); // end newton iteration loop
  // now evaluate omega out
  for(int k=0;k<2;k++)
    fprintf(f, "out_%s = pred_%s;\n", var[k+2].c_str(), var[k+2].c_str());
}

// light tracer connecting to a given point on the aperture.
// idea as follows:
//
// evaluate_reverse:
//
// input : z[x,y,u,v] point on the scene geometry, [u,v] initial guess
// input : dist       distance z--y (outer pupil)
// input : a[x,y,u,v] point on aperture, [u,v] initial guess
// output: x[x,y,u,v] point on the sensor
//
// // evaluate adjoint polynomial (approximate)
// x = P’(T * z)
// // do newton iterations
// for(int it=0;it<8;it++)
// {
//   // first match scene geometry point z:
//   // forward evaluate to get error vector
//   z* = T * P(x[x,y,-u,-v])
//   delta = z - z*[x,y] // [u,v] = 0
//   // jacobian of combined adjoint system
//   Ji = jacobian’(T * z*[x,y,-u,-v])
//   x += Ji * T * delta
//
//   // also want to match aperture constraint a:
//   // forward evaluate to get error vector
//   a* = Pa(x[x,y,-u,-v])
//   delta = a[x,y] - a*[x,y] // [u,v] = 0
//   // jacobian of inverse aperture polynomial Pa^{-1}, spatial subblock 2x2
//   Ji = invert(J_a->x)(a*)
//   x += Ji * delta
// }
template <class scalar, int numVarsIn, int numVarsOut>
void print_reverse_aperture_sampling(
    FILE *f,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& system,   // light tracing direction
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& adjoint,  // path tracing direction
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& aperture, // path tracing, aperture poly
    std::string *varNames)
{
  const int numVarsMin = std::min(numVarsOut, numVarsIn);
  const int numVarsMax = std::max(numVarsOut, numVarsIn);

  std::string sensor_var[numVarsMax];
  std::string ap_it_var[numVarsMax];
  std::string ap_target_var[numVarsMax];
  std::string pupil_var[numVarsMax];
  std::string light_var[numVarsMax];
  std::string sensor_r_var[numVarsMax];
  std::string light_r_var[numVarsMax];
  for(int i=0;i<numVarsMax;i++)
  {
    sensor_var[i] = "sensor_" + varNames[i];
    light_var[i] = varNames[i];
    sensor_r_var[i] = "sensor_r_" + varNames[i];
    light_r_var[i] = "light_r_" + varNames[i];
    pupil_var[i] = "pupil_" + varNames[i];
    ap_it_var[i] = "ap_it_" + varNames[i];
    ap_target_var[i] = "ap_" + varNames[i];
  }

  fprintf(f, "// evaluate approximate reverse polynomial\n");
  // transport light to pupil
  fprintf(f, "float %s = %s + dist * %s;\n", pupil_var[0].c_str(), light_var[0].c_str(), light_var[2].c_str());
  fprintf(f, "float %s = %s + dist * %s;\n", pupil_var[1].c_str(), light_var[1].c_str(), light_var[3].c_str());
  fprintf(f, "float %s = %s;\n", pupil_var[2].c_str(), light_var[2].c_str());
  fprintf(f, "float %s = %s;\n", pupil_var[3].c_str(), light_var[3].c_str());
  fprintf(f, "float %s = %s;\n", pupil_var[4].c_str(), light_var[4].c_str());
  // transport pupil to sensor (predict)
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "float %s = ", sensor_var[i].c_str());
    print_poly_code(f, system.equations[i], pupil_var);
    fprintf(f, ";\n");
  }
  // fprintf(stderr, "eval: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);

  // arbitrary iteration number, 4 should be almost floating point accuracy
  fprintf(f, "// do newton iterations to get precise answer\n");
  fprintf(f, "for(int it=0;it<8;it++)\n{\n");
  // need to flip direction [dx, dy] to transport back:
  for(int i=0;i<numVarsOut;i++)
    fprintf(f, "  float %s = %s %s;\n",
      sensor_r_var[i].c_str(),
      (i == 2 || i == 3) ? "-" : "", sensor_var[i].c_str());

  // copy over constant dimensions:
  for(int i=numVarsMin;i<numVarsMax;i++) 
    fprintf(f, "  float %s = %s;\n",
        sensor_r_var[i].c_str(), varNames[i].c_str());

  fprintf(f, "  // evaluate adjoint to get error vector\n");
  // evaluate adjoint to get pupil point:
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "  float %s = ", pupil_var[i].c_str());
    print_poly_code(f, adjoint.equations[i], sensor_r_var);
    fprintf(f, ";\n");
  }
  // transport spatial components further to get point on scene geo
  fprintf(f, "  float %s = %s + dist * %s;\n", light_r_var[0].c_str(), pupil_var[0].c_str(), pupil_var[2].c_str());
  fprintf(f, "  float %s = %s + dist * %s;\n", light_r_var[1].c_str(), pupil_var[1].c_str(), pupil_var[3].c_str());
  fprintf(f, "  float %s = %s = -%s;\n", light_r_var[2].c_str(), pupil_var[2].c_str(), pupil_var[2].c_str());
  fprintf(f, "  float %s = %s = -%s;\n", light_r_var[3].c_str(), pupil_var[3].c_str(), pupil_var[3].c_str());
  fprintf(f, "  float %s = %s;\n", light_r_var[4].c_str(), pupil_var[4].c_str());

  // output will be a little off, lives on scene geo

  fprintf(f, "  const float d_light_0 = (%s - %s);\n", light_var[0].c_str(), light_r_var[0].c_str());
  fprintf(f, "  const float d_light_1 = (%s - %s);\n", light_var[1].c_str(), light_r_var[1].c_str());
  fprintf(f, "  const float d_light_2 = (%s - %s);\n", light_var[2].c_str(), light_r_var[2].c_str());
  fprintf(f, "  const float d_light_3 = (%s - %s);\n", light_var[3].c_str(), light_r_var[3].c_str());
  fprintf(f, "  const float d_light_4 = 0.0f;\n");

  fprintf(f, "  // move through jacobian and correct sensor point\n");
#define USE_INV 0
#if USE_INV==1
  fprintf(f, "  float J[5][5] = {{0.0f}};\n");
#endif
  fprintf(f, "  float JI[5][5] = {{0.0f}};\n");
  fprintf(f, "  JI[4][4] = 1.0f;\n");
  // how much do we need to adjust the point on the sensor
  // to get the corresponding reversed sensor_r -> light_r = light
  for (int i = 0; i < numVarsMin; i++) // Out
  {
    for (int j = 0; j < numVarsMin; ++j) // In
    {
#if USE_INV==1
      fprintf(f, "  J[%d][%d] = ", i, j);
      print_poly_code(f, adjoint.equations[i].getDerivative(j), sensor_r_var);
#else
      fprintf(f, "  JI[%d][%d] = ", i, j);
      print_poly_code(f, system.equations[i].getDerivative(j), pupil_var);
#endif
      fprintf(f, "+0.0f;\n");
    }
  }
#if USE_INV==1
  fprintf(f, "  lens_mat5inv(J, JI);\n");
#endif

  // now it remains to do Ji * T * delta. since delta[u,v]=0, T=I and we forget about it.
  // move sensor point:
  for(int k=0;k<4;k++)
  {
    for(int i=0;i<numVarsMin;i++)
      fprintf(f, "  %s %s JI[%d][%d] * d_light_%d;\n", sensor_var[k].c_str(),
#if USE_INV==1
          "-=",
#else
          "+=",
#endif
          k, i, i);
#undef USE_INV
  }
  // update direction, too
  fprintf(f, "  %s = %s;\n", varNames[2].c_str(), light_r_var[2].c_str());
  fprintf(f, "  %s = %s;\n", varNames[3].c_str(), light_r_var[3].c_str());


  // second part, optimize for point on aperture.
  // start at sensor point from before, only flip sign:
  for(int i=0;i<numVarsOut;i++)
    fprintf(f, "  %s = %s %s;\n",
      sensor_r_var[i].c_str(),
      (i == 2 || i == 3) ? "-" : "", sensor_var[i].c_str());
  // compute point on aperture
  fprintf(f, "  // evaluate adjoint aperture point\n");
  // evaluate adjoint to get pupil point:
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "  float %s = ", ap_it_var[i].c_str());
    print_poly_code(f, aperture.equations[i], sensor_r_var);
    fprintf(f, ";\n");
  }
  // get distance, spatial only
  fprintf(f, "  const float d_ap_0 = (%s - %s);\n", ap_target_var[0].c_str(), ap_it_var[0].c_str());
  fprintf(f, "  const float d_ap_1 = (%s - %s);\n", ap_target_var[1].c_str(), ap_it_var[1].c_str());
  fprintf(f, "  const float d_ap_2 = (%s - %s);\n", ap_target_var[2].c_str(), ap_it_var[2].c_str());
  fprintf(f, "  const float d_ap_3 = (%s - %s);\n", ap_target_var[3].c_str(), ap_it_var[3].c_str());
  fprintf(f, "  const float d_ap_4 = 0.0f;\n");

  fprintf(f, "  // move through jacobian and correct sensor point\n");
  fprintf(f, "  float J[5][5] = {{0.0f}};\n");
  // how much do we need to adjust the point on the sensor
  // to get the corresponding reversed sensor_r -> light_r = light
  for (int i = 0; i < numVarsMin; i++) // Out
  {
    for (int j = 0; j < numVarsMin; ++j) // In
    {
      fprintf(f, "  J[%d][%d] = ", i, j);
      print_poly_code(f, aperture.equations[i].getDerivative(j), sensor_r_var);
      fprintf(f, "+0.0f;\n");
    }
  }
  fprintf(f, "  lens_mat5inv(J, JI);\n");

  // move sensor point:
  for(int k=0;k<4;k++) // update [x,y,dx,dy] not lambda
    for(int i=0;i<numVarsMin;i++)
      fprintf(f, "  %s %s JI[%d][%d] * d_ap_%d;\n", sensor_var[k].c_str(), (k<2) ? "+=" : "-=" , k, i, i);

  fprintf(f, "  %s = %s;\n", ap_target_var[2].c_str(), ap_it_var[2].c_str());
  fprintf(f, "  %s = %s;\n", ap_target_var[3].c_str(), ap_it_var[3].c_str());

  fprintf(f, "}\n");
}


#undef TERM_EPS
#endif
