#ifndef _CORONA_LENS_MANIFOLD_H
#define _CORONA_LENS_MANIFOLD_H


#include "pathspace.h"
#include "matrix2.h"

// map tangent space delta at end of chain (last vertex v[n] in array, in respective tangent space uv)
// to world space at given vertex (v[v], v < n, delta, 3d)
static inline void manifold_map(const vertex_t *vtx, float u, float v, float *delta)
{
  const float Tx = vtx->diffgeo.Tp[0] * u + vtx->diffgeo.Tp[1] * v;
  const float Ty = vtx->diffgeo.Tp[2] * u + vtx->diffgeo.Tp[3] * v;
  for(int k=0;k<3;k++)
    delta[k] = Tx * vtx->diffgeo.dpdu[k] + Ty * vtx->diffgeo.dpdv[k];
}

// initialize some essential variables in the diffgeo struct on each vertex.
// don't actually compute the tangent space mappings yet.
static inline void manifold_sphere_init(path_t *p, int v, const float *n, const float radius)
{
  // compute dndu dndv for every vertex, depends on uv coordinates to be useful
  p->v[v].diffgeo.dpdu[0] = -p->v[v].hit.x[1] * 2*M_PI;
  p->v[v].diffgeo.dpdu[1] =  p->v[v].hit.x[0] * 2*M_PI;
  p->v[v].diffgeo.dpdu[2] =  0.0f;
  const float zrad = p->v[v].hit.x[0]*p->v[v].hit.x[0] +  p->v[v].hit.x[1]*p->v[v].hit.x[1];
  const float cos_theta = p->v[v].hit.x[2]/radius;
  const float sin_theta = sqrtf(fmaxf(0.0f, 1.0f-cos_theta*cos_theta));
  float cos_phi, sin_phi;
  if(zrad > 0.0f)
  {
    cos_phi = p->v[v].hit.x[0] / zrad;
    sin_phi = p->v[v].hit.x[1] / zrad;
  }
  else
  {
    cos_phi = 0.0f;
    sin_phi = 1.0f;
  }
  p->v[v].diffgeo.dpdv[0] = p->v[v].hit.x[2] * cos_phi * M_PI;
  p->v[v].diffgeo.dpdv[1] = p->v[v].hit.x[2] * sin_phi * M_PI;
  p->v[v].diffgeo.dpdv[2] = -sin_theta * radius * M_PI;

  for(int k=0;k<3;k++)
  {
    p->v[v].diffgeo.dndu[k] = p->v[v].diffgeo.dpdu[k] / radius;
    p->v[v].diffgeo.dndv[k] = p->v[v].diffgeo.dpdv[k] / radius;
  }

  // ortho normalise, stick to dpdu as direction:
  const float ilu = 1.0f/sqrtf(dotproduct(p->v[v].diffgeo.dpdu, p->v[v].diffgeo.dpdu));
  for(int k=0;k<3;k++) p->v[v].diffgeo.dpdu[k] *= ilu;
  for(int k=0;k<3;k++) p->v[v].diffgeo.dndu[k] *= ilu;
  const float dot = dotproduct(p->v[v].diffgeo.dpdu, p->v[v].diffgeo.dpdv);
  for(int k=0;k<3;k++) p->v[v].diffgeo.dpdv[k] -= dot*p->v[v].diffgeo.dpdu[k];
  for(int k=0;k<3;k++) p->v[v].diffgeo.dndv[k] -= dot*p->v[v].diffgeo.dndu[k];
  const float ilv = 1.0f/sqrtf(dotproduct(p->v[v].diffgeo.dpdv, p->v[v].diffgeo.dpdv));
  for(int k=0;k<3;k++) p->v[v].diffgeo.dpdv[k] *= ilv;
  for(int k=0;k<3;k++) p->v[v].diffgeo.dndv[k] *= ilv;

  // shading frame is also rotated to match dpdu:
  memcpy(p->v[v].hit.a, p->v[v].diffgeo.dpdu, sizeof(float)*3);
  memcpy(p->v[v].hit.b, p->v[v].diffgeo.dpdv, sizeof(float)*3);
  normalise(p->v[v].hit.a);
  normalise(p->v[v].hit.b);
}

// same as wenzel's code in mitsuba.
// initialize tangent space derivatives and invert matrix, so we can
// map tangent space vertex relocations through the specular chain (fixed halfvectors)
// start: diffuse vertex before the spec chain
// end: diffuse vertex at end of spec chain
static inline float manifold_compute_tangents(path_t *p, int start, int end)
{
  assert(start >= 0);
  assert(end <= p->length-1);
  const int n = end;

  mat2_set_zero(p->v[0].diffgeo.Tp);     // does not move at all
  mat2_set_identity(p->v[n].diffgeo.Tp); // moves freely along its own tangent space

  if(end-start == 1) return 1.0f; // all good.

  // assemble matrices a, b, c (how does vertex i depend on i-1, i, and i+1)
  for(int i=start;i<end;i++)
  {
    // fprintf(stderr, "mf create %d/%d\n", i, n);
    vertex_manifold_t *mf = &p->v[i].diffgeo;
    float wo[3]; // pointing from vert i to i+1
    for(int k=0;k<3;k++) wo[k] = p->v[i+1].hit.x[k] - p->v[i].hit.x[k];
    float ilo = sqrtf(dotproduct(wo, wo));
    if(ilo == 0.0f) return 0.0f;
    ilo = 1.0/ilo;
    for(int k=0;k<3;k++) wo[k] *= ilo;

    if(mf->type == s_pinned_position)
    {
      mat2_set_zero(mf->a);
      mat2_set_identity(mf->b);
      mat2_set_zero(mf->c);
      continue;
    }
    else if(mf->type == s_pinned_direction)
    {
      float dC_dnext_u[3], dC_dnext_v[3], dC_dcurr_u[3], dC_dcurr_v[3];
      const float dot_wo_v0dpdu = dotproduct(wo, p->v[i].diffgeo.dpdu);
      const float dot_wo_v0dpdv = dotproduct(wo, p->v[i].diffgeo.dpdv);
      const float dot_wo_v1dpdu = dotproduct(wo, p->v[i+1].diffgeo.dpdu);
      const float dot_wo_v1dpdv = dotproduct(wo, p->v[i+1].diffgeo.dpdv);
      for(int k=0;k<3;k++)
      {
        dC_dnext_u[k] = (p->v[i+1].diffgeo.dpdu[k] - wo[k] * dot_wo_v1dpdu) * ilo;
        dC_dnext_v[k] = (p->v[i+1].diffgeo.dpdv[k] - wo[k] * dot_wo_v1dpdv) * ilo;
        dC_dcurr_u[k] = (wo[k] * dot_wo_v0dpdu - p->v[i].diffgeo.dpdu[k]) * ilo;
        dC_dcurr_v[k] = (wo[k] * dot_wo_v0dpdv - p->v[i].diffgeo.dpdv[k]) * ilo;
      }

      mat2_set_zero(mf->a);
      mf->b[0] = dotproduct(dC_dcurr_u, p->v[0].diffgeo.dpdu);
      mf->b[1] = dotproduct(dC_dcurr_v, p->v[0].diffgeo.dpdu);
      mf->b[2] = dotproduct(dC_dcurr_u, p->v[0].diffgeo.dpdv);
      mf->b[3] = dotproduct(dC_dcurr_v, p->v[0].diffgeo.dpdv);

      mf->c[0] = dotproduct(dC_dnext_u, p->v[0].diffgeo.dpdu);
      mf->c[1] = dotproduct(dC_dnext_v, p->v[0].diffgeo.dpdu);
      mf->c[2] = dotproduct(dC_dnext_u, p->v[0].diffgeo.dpdv);
      mf->c[3] = dotproduct(dC_dnext_v, p->v[0].diffgeo.dpdv);
      continue;
    }

    assert(i > start); // first vertex should have been pinned one or the other way above.
    float wi[3]; // pointing from vert i to i-1
    for(int k=0;k<3;k++) wi[k] = p->v[i-1].hit.x[k] - p->v[i].hit.x[k];
    float ili = sqrtf(dotproduct(wi, wi));
    if(ili == 0.0f) return 0.0f;
    ili = 1.0/ili;
    for(int k=0;k<3;k++) wi[k] *= ili;

    assert(p->v[i].mode & s_transmit); // special case here for lenses
    if(p->v[i].mode & (s_reflect | s_transmit)) /* reflection or refraction */
    {
      // ratio n1/n2 where n2 is inside the shape
      // TODO: cache that on diffgeo struct
      const float eta = (p->v[i].mode & s_reflect) ? 1.0f : (1.0f/p->v[i].diffgeo.eta); // n2/n1
      // reflected path also goes through normalise_H, but happens to have eta == 1.
      const int normalise_H = !((p->v[i].mode & s_transmit) && (eta == 1.0f));

      float ilh, H[3];
      if(normalise_H)
      {
        for(int k=0;k<3;k++) H[k] = wi[k] + eta * wo[k];
        ilh = 1.0f/sqrtf(dotproduct(H, H));
        for(int k=0;k<3;k++) H[k] *= ilh;
      }
      else
      {
        // don't normalise H for index-matched transmission
        for(int k=0;k<3;k++) H[k] = wi[k] + wo[k];
        ilh = 1.0f;
      }

      float dot_H_n    = dotproduct(p->v[i].hit.n, H);
      float dot_H_dndu = dotproduct(mf->dndu, H);
      float dot_H_dndv = dotproduct(mf->dndv, H);
      float dot_dpdu_n = dotproduct(mf->dpdu, p->v[i].hit.n);
      float dot_dpdv_n = dotproduct(mf->dpdv, p->v[i].hit.n);

      // tangent frame around shading normal:
      float *s = p->v[i].hit.a, *t = p->v[i].hit.b;

      ilo *= eta * ilh; ili *= ilh;

      // derivatives of C wrt previous vertex v[i-1]
      float dH_du[3], dH_dv[3];
      const float dot_wi_dpdu = dotproduct(wi, p->v[i-1].diffgeo.dpdu);
      const float dot_wi_dpdv = dotproduct(wi, p->v[i-1].diffgeo.dpdv);
      for(int k=0;k<3;k++) dH_du[k] = (p->v[i-1].diffgeo.dpdu[k] - wi[k] * dot_wi_dpdu) * ili;
      for(int k=0;k<3;k++) dH_dv[k] = (p->v[i-1].diffgeo.dpdv[k] - wi[k] * dot_wi_dpdv) * ili;

      if(normalise_H)
      {
        const float dotu = dotproduct(dH_du, H);
        const float dotv = dotproduct(dH_dv, H);
        for(int k=0;k<3;k++)
        {
          dH_du[k] -= H[k] * dotu;
          dH_dv[k] -= H[k] * dotv;
        }
      }

      mf->a[0] = dotproduct(dH_du, s);
      mf->a[1] = dotproduct(dH_dv, s);
      mf->a[2] = dotproduct(dH_du, t);
      mf->a[3] = dotproduct(dH_dv, t);

      // derivatives of C wrt current vertex v[i]
      for(int k=0;k<3;k++)
      {
        dH_du[k] = -p->v[i].diffgeo.dpdu[k] * (ili + ilo) + wi[k] * dotproduct(wi, p->v[i].diffgeo.dpdu) * ili
                                                          + wo[k] * dotproduct(wo, p->v[i].diffgeo.dpdu) * ilo;
        dH_dv[k] = -p->v[i].diffgeo.dpdv[k] * (ili + ilo) + wi[k] * dotproduct(wi, p->v[i].diffgeo.dpdv) * ili
                                                          + wo[k] * dotproduct(wo, p->v[i].diffgeo.dpdv) * ilo;
      }
      if(normalise_H)
      {
        const float dotu = dotproduct(dH_du, H);
        const float dotv = dotproduct(dH_dv, H);
        for(int k=0;k<3;k++)
        {
          dH_du[k] -= H[k] * dotu;
          dH_dv[k] -= H[k] * dotv;
        }
      }

      mf->b[0] = dotproduct(dH_du, s) - dotproduct(mf->dpdu, mf->dndu) * dot_H_n - dot_dpdu_n * dot_H_dndu;
      mf->b[1] = dotproduct(dH_dv, s) - dotproduct(mf->dpdu, mf->dndv) * dot_H_n - dot_dpdu_n * dot_H_dndv;
      mf->b[2] = dotproduct(dH_du, t) - dotproduct(mf->dpdv, mf->dndu) * dot_H_n - dot_dpdv_n * dot_H_dndu;
      mf->b[3] = dotproduct(dH_dv, t) - dotproduct(mf->dpdv, mf->dndv) * dot_H_n - dot_dpdv_n * dot_H_dndv;

      // derivatives of C wrt v[i+1]
      for(int k=0;k<3;k++)
      {
        dH_du[k] = (p->v[i+1].diffgeo.dpdu[k] - wo[k] * dotproduct(wo, p->v[i+1].diffgeo.dpdu)) * ilo;
        dH_dv[k] = (p->v[i+1].diffgeo.dpdv[k] - wo[k] * dotproduct(wo, p->v[i+1].diffgeo.dpdv)) * ilo;
      }

      if(normalise_H)
      {
        const float dotu = dotproduct(dH_du, H);
        const float dotv = dotproduct(dH_dv, H);
        for(int k=0;k<3;k++)
        {
          dH_du[k] -= H[k] * dotu;
          dH_dv[k] -= H[k] * dotv;
        }
      }

      mf->c[0] = dotproduct(dH_du, s);
      mf->c[1] = dotproduct(dH_dv, s);
      mf->c[2] = dotproduct(dH_du, t);
      mf->c[3] = dotproduct(dH_dv, t);

      // store microfacet normal
      mf->h[0] = dotproduct(H, s);
      mf->h[1] = dotproduct(H, t);
      mf->h[2] = dotproduct(H, p->v[i].hit.n);
      // XXX, orient into same hemisphere as geo normal:
      // if(dotproduct(H, p->v[i].hit.gn) < 0.0f)
        // for(int k=0;k<3;k++) mf->h[k] = - mf->h[k];
      // fprintf(stderr, "[%d] H = %f %f %f h = %f %f %f\n", i, H[0], H[1], H[2], mf->h[0], mf->h[1], mf->h[2]);
    }
  }

  // block tridiagonal LU factorization to create transport matrices Tp
  // mapping a change at vertex v[n] to each vertex v[n-1]..v[0]. v[n].Tp = identity.
  //
  // note: could do it backwards, to map changes at v[0] to v[1]..v[n-1], by
  // inverting the loop direction and swapping a (depend on prev) and c (depend on next).
  //
  // we use Salkuyeh's way to compute the overall determinant on the way.
  // (Comments on ``A note on a three-term recurrence for a tridiagonal matrix'')
  // det M = \Prod_{k=1}^n det\Lambda_k
  // \Lambda_k = b_k - a_k \Lambda^{-1}_{k-1} c_{k-1},   \Lambda_1 = b_1
  float det = 1.0;
  float Li[4];
  det *= mat2_det(p->v[start].diffgeo.b);
  if(mat2_invert(p->v[start].diffgeo.b, Li)) return 0.0f;
  for(int i=start;i<end-1;i++)
  {
    mat2_mul(Li, p->v[i].diffgeo.c, p->v[i].diffgeo.u);
    float au[4], b_au[4]; // b - a*u
    mat2_mul(p->v[i+1].diffgeo.a, p->v[i].diffgeo.u, au);
    mat2_sub(p->v[i+1].diffgeo.b, au, b_au);
    det *= mat2_det(b_au);
    if(mat2_invert(b_au, Li))
    {
      // fprintf(stderr, "invert b_au %d failed!\n", i+1);
      return 0.0f;
    }
  }
  mat2_mul(Li, p->v[end-1].diffgeo.c, p->v[end-1].diffgeo.Tp);
  for(int k=0;k<4;k++) p->v[end-1].diffgeo.Tp[k] = -p->v[end-1].diffgeo.Tp[k];
  for(int i=end-2;i>=start;i--)
  {
    mat2_mul(p->v[i].diffgeo.u, p->v[i+1].diffgeo.Tp, p->v[i].diffgeo.Tp);
    for(int k=0;k<4;k++) p->v[i].diffgeo.Tp[k] = -p->v[i].diffgeo.Tp[k];
  }
  return det;
}

static float raytrace_intersect_sphere(const float *center, const float *p, const float *omega, const float radius)
{
  const float a = 1.0f; // ||omega||^2
  float o[3] = {p[0]-center[0],p[1]-center[1],p[2]-center[2]};
  const float b = 2.0f*raytrace_dot(o,omega);
  const float c = raytrace_dot(o,o) - radius*radius;
  if (a == 0)
  {
    if (b != 0)
      return -c / b;
    return -FLT_MAX;
  }

  float discrim = b*b - 4.0f*a*c;
  if (discrim < 0) return -FLT_MAX;

  float temp, sqrtDiscrim = sqrtf(discrim);
  if (b < 0)
    temp = -0.5f * (b - sqrtDiscrim);
  else
    temp = -0.5f * (b + sqrtDiscrim);

  const float x0 = temp / a;
  const float x1 = c / temp;

  // return minimum if not smaller than zero
  if(x0 < 0.0f) return x1;
  else if(x1 < 0.0f) return x0;
  return fminf(x0, x1);
}

static inline int evaluate_J_aperture(
    const lens_element_t *lenses,
    const int lenses_cnt,
    const float zoom,
    const float *ap,
    float *center,
    float *du,
    float *dv)
{
  // construct path from aperture v[0] to sensor v[length-1]:
  path_t path;
  path_init(&path);
  int error = 0;
  float n1 = 1.0f;
  int v = 0;
  path.v[0].mode = s_transmit | s_specular;
  path.v[0].hit.x[0] = ap[0];
  path.v[0].hit.x[1] = ap[1];
  path.v[0].hit.x[2] = 0.0f;
  path.e[1].omega[0] = -ap[2];
  path.e[1].omega[1] = -ap[3];
  path.e[1].omega[2] = 1.0f;
  raytrace_normalise(path.e[1].omega);

  // tangent space init (path_init zeros out the rest):
  path.v[0].hit.a[0] = 1.0f;
  path.v[0].hit.b[1] = 1.0f;
  path.v[0].hit.n[2] = -1.0f;
  path.v[0].diffgeo.dpdu[0] = 1.0f;
  path.v[0].diffgeo.dpdv[1] = 1.0f;
  // dndu = 0
  float pos = 0;

  // fprintf(stderr, "mf omega in %f %f\n", -ap[2], -ap[3]);

  for(int k=0;k<lenses_cnt;)
  {
    // do nothing until iris is found
    if(!v && strcasecmp(lenses[k].material, "iris")) { k++; continue; }
    v++;

    assert(!lenses[k].anamorphic); // TODO, but don't think we'll ever need it.

    // fprintf(stderr, "mf eval dist %d\n", k);
    const float dist = lens_get_thickness(lenses+k, zoom);
    pos += dist;
    if(k == lenses_cnt - 1)
    {
      path.length = v+1;
      // last propagation, intersect with sensor plane
      const float t = (pos - path.v[v-1].hit.x[2])/path.e[v].omega[2];
      for(int k=0;k<3;k++)
        path.v[v].hit.x[k] = path.v[v-1].hit.x[k] + t * path.e[v].omega[k];

      // tangent space of last vertex
      path.v[v].hit.a[0] = 1.0f;
      path.v[v].hit.b[1] = 1.0f;
      path.v[v].hit.n[2] = -1.0f;
      path.v[v].diffgeo.dpdu[0] = 1.0f;
      path.v[v].diffgeo.dpdv[1] = 1.0f;
      break;
    }
    k++; // increment k here, as we need the distance = thickness of the previous element and the spherical interface of the next.
    const float R = lenses[k].lens_radius;

    path.v[v].mode = s_transmit | s_specular;

    // fprintf(stderr, "mf eval surface %d\n", k);
    float c[3] = {0.0, 0.0, 0.0};
    c[2] = pos + R;
    const float t = raytrace_intersect_sphere(c, path.v[v-1].hit.x, path.e[v].omega, R);
#if 0
    fprintf(stderr, "intersect ray center %f %f %f pos %f %f %f dir %f %f %f R %f => t = %f\n", c[0], c[1], c[2],
        path.v[v-1].hit.x[0],
        path.v[v-1].hit.x[1],
        path.v[v-1].hit.x[2],
        path.e[v].omega[0],
        path.e[v].omega[1],
        path.e[v].omega[2],
        R,
        t);
#endif

    for(int i=0;i<3;i++)
      path.v[v].hit.x[i] = path.v[v-1].hit.x[i] + t * path.e[v].omega[i];
    // fprintf(stderr, "mf v[%d] %f %f\n", k, path.v[v].hit.x[0], path.v[v].hit.x[1]);

    // transmit through the medium.
    // mark this ray as theoretically dead:
    const float zrad2 = path.v[v].hit.x[0]*path.v[v].hit.x[0] + path.v[v].hit.x[1]*path.v[v].hit.x[1];
    if(zrad2 > lenses[k].housing_radius*lenses[k].housing_radius)
      error |= 16;

    // index of refraction and ratio current/next:
    const float n2 = spectrum_eta_from_abbe_um(lenses[k].ior, lenses[k].vno, ap[4]);
    const float eta = n1/n2;

    // get normal pointing from center of intersected sphere to intersection point:
    const float aR = fabsf(R);
    const float n[3] = {path.v[v].hit.x[0]/R, lenses[k].anamorphic ? 0.0f : path.v[v].hit.x[1]/R, -sqrtf(fmaxf(0.0f, R*R - zrad2))/aR};

    // inid tangent space
    manifold_sphere_init(&path, v, n, R);

    // transmit using snell's law:
    const float cos1 = - raytrace_dot(n, path.e[v].omega);
    const float cos2_2 = 1.0f-eta*eta*(1.0f-cos1*cos1);
    // total (inner) reflection?
    if(cos2_2 <= 0.0f)
      return error |= 8;
    const float cos2 = sqrtf(cos2_2);

    // cache eta ratio on diffgeo struct!
    path.v[v].diffgeo.eta = n1/n2;

    if(n2 != n1)
      for(int i=0;i<3;i++) path.e[v+1].omega[i] = path.e[v].omega[i]*eta + (eta*cos1-cos2)*n[i];
    else
      for(int i=0;i<3;i++) path.e[v+1].omega[i] = path.e[v].omega[i];
    // fprintf(stderr, "mf update eta %f n %f %f %f omega on surf %d %f %f\n", n2/n1, n[0], n[1], n[2], k, path.e[v+1].omega[0]/path.e[v+1].omega[2], path.e[v+1].omega[1]/path.e[v+1].omega[2]);

    // mark this ray as theoretically dead:
    if(path.e[v+1].omega[2] <= 0.0f) return error |= 2;
    // and renormalise:
    raytrace_normalise(path.e[v+1].omega);
    n1 = n2;
    // fprintf(stderr, "path v[%d] %f %f %f\n", v, path.v[v].hit.x[0],path.v[v].hit.x[1],path.v[v].hit.x[2]);
  }

  //fprintf(stderr, "last vert v[%d] %f %f %f -- %f %f\n", v,
  //    path.v[v].hit.x[0], 
  //    path.v[v].hit.x[1], 
  //    path.v[v].hit.x[2],
  //    path.e[v].omega[0]/path.e[v].omega[2],
  //    path.e[v].omega[1]/path.e[v].omega[2]);

  path.v[0].diffgeo.type = s_pinned_position;
  if(manifold_compute_tangents(&path, 0, path.length-1) == 0.0f) return 256;
  for(int k=0;k<2;k++)
    center[k] = path.v[v-1].hit.x[k];
  center[2] = path.v[v].hit.x[2] - path.v[v-1].hit.x[2]; // put origin to center of sensor. this should be > 0.
  manifold_map(path.v + v-1, 1, 0, du);
  manifold_map(path.v + v-1, 0, 1, dv);
  return error;
}


#endif
