/* Truncated Polynomial Class
 * (C) 2012, Matthias Hullin <hullin@cs.ubc.ca>
 * 
 * Defines classes polyTerm, truncPoly, truncPolySystem.
 */

#ifndef POLY_TRUNC_POLY_SYSTEM_HH
#define POLY_TRUNC_POLY_SYSTEM_HH

#include <vector>
#include <cmath>
#include <algorithm>
#include <assert.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <cstdio>
#include <cstdarg>
#include <string.h>
#include <stdint.h>

namespace poly {

// Abbreviations to reduce visual clutter of templates
#define _tS template <typename scalar>
#define _S <scalar>
#define _tO template <typename otherScalar>
#define _O <otherScalar>
#define _tSN template <typename scalar, int numVars>
#define _SN <scalar, numVars>
#define _ON <otherScalar, numVars>
#define _tSIO template <typename scalar, int numVarsIn, int numVarsOut>
#define _SIO <scalar, numVarsIn, numVarsOut>
#define _OIO <otherScalar, numVarsIn, numVarsOut>
#define _O1O <otherScalar, (numVarsIn-1), numVarsOut>

// Data type for powers. x^255 should be enough for all reasonable purposes. 
// Gives nice 8-byte packing for polynomial terms.
#define echar unsigned char    

_tSN class truncPoly;


_tS
inline scalar intpow(const scalar base, const echar exponent)
{
  if (exponent == 0) return scalar(1);
  else if (exponent == 1)  return base;
  else if (exponent == 2)  return base*base;
  else if (exponent & 1)  return base * intpow(base,exponent/2) * intpow(base,exponent/2); 
  else return intpow(base,exponent/2) * intpow(base,exponent/2);
}


// A single term of a polynomial. 
// Example: alpha * x1^3 * x2 * x4^2
// coefficient = alpha; exponents = {0, 3, 1, 0, 4}

_tSN
class polyTerm
{
public:
  scalar coefficient;    // Scalar factor
  echar exponents[numVars];    // Powers of each term -- let's assume they won't be >255
    
  // Adds up all exponents to get total degree of term. 
  // Example: alpha * x1^3 * x2 * x4^2 => degree = 8
  int degree() const;
    
  // Constructor sets everything to zero
  polyTerm();

  // initialize from other polyTerm in less variables than we are able to express.
  // just zero out all other exponents and copy what we can get.
//   template<int numVarsTerm>
//   polyTerm(const polyTerm<scalar, numVarsTerm> &t)
//   {
//     assert(numVarsTerm <= numVars);
//     coefficient = t.coefficient;
//     for(int k=0;k<numVarsTerm;k++) exponents[k] = t.exponents[k];
//     for(int k=numVarsTerm;k<numVars;k++) exponents[k] = 0.0f;
//   }

  // Popular constructors
  polyTerm(scalar coeff);                                       // constant
  polyTerm(scalar coeff, int ex, ...);                          // numVars many exponents


  // used to initialize high-dimensional polynomials (swap variables).
  // especially useful in combination with the mixed-type copy constructor above.
  inline void swapVar(int v1, int v2)
  {
    std::swap(exponents[v1], exponents[v2]);
  }

  /**************     POLYTERM ARITHMETIC     ***************/

  // Assignment operator
  //polyTerm _SN & operator= (const polyTerm _SN &rhs);

  // Assignment operator, also works for different scalar type
  // _tO
  // polyTerm _SN & operator= (const polyTerm _ON &rhs);

  // Assignment operator
  polyTerm _SN & operator= (const scalar rhs);

  // Multiplication-assignment operator: adds exponents, multiplies coefficients
  polyTerm  _SN & operator*= (const polyTerm _SN &rhs);

  // Addition operator: term + term = polynomial
  const truncPoly _SN operator+ (const polyTerm _SN &rhs) const;

  // Addition operator: term - term = polynomial
  const truncPoly _SN operator- (const polyTerm _SN &rhs) const;

  // Multiplication operator: adds exponents, multiplies coefficients
  const polyTerm  _SN  operator* (const polyTerm _SN &rhs) const;

  // Relation operator (needed for sorting). Sorting order is first to last exponent
  // (x0 * x1^4 < x0^2 * x1^3)
  const bool  operator< (const polyTerm _SN &rhs) const;

  // Check exponents for equality (NOT coefficients)
  const bool  operator== (const polyTerm _SN &rhs) const;
    
  // Same for inequality (NOT coefficients)
  inline bool  operator!= (const polyTerm _SN &rhs) const {
    return !(*this == rhs); // just invert equality!
  }

  // Access exponents by [] brackets:
  inline echar& operator[] (int var)
  {
    assert(var>=0 && var<numVars);
    return exponents[var];
  }
  const echar& operator[] (int var) const
  {
    assert(var>=0 && var<numVars);
    return exponents[var];
  } 
};

  
// A polynomial

_tSN
class truncPoly
{
public:
  /* if the absolute value of a coefficient of a term is smaller or equal than that, it will be dropped. */
#define c_discardTermThreshold 1e-15//0
    
  int truncDegree;
  std::vector<polyTerm _SN > terms;
  bool consolidated;
  
  // Return count of terms
  int getNumTerms() {return terms.size();} 

  // Default constructor: empty polynomial
  truncPoly() {truncDegree = -1; consolidated = true;} // do not truncate by default

  // Construct from scalar
  inline truncPoly(scalar constant) {
    truncDegree = -1;
    terms.push_back(polyTerm _SN(constant));
    consolidated = true;
  }
  // Construct from polyTerm

  inline truncPoly(polyTerm _SN term) {
    truncDegree = -1;
    terms.push_back(term);
    consolidated = true;
  }
  
  // Construct from truncPoly
  // _tO
  // inline truncPoly(truncPoly _ON rhs) {
    // operator=(rhs);
  // }

  template <int numVarsTerm>
  inline void add(const polyTerm <scalar, numVarsTerm> &t)
  {
    assert(numVarsTerm <= numVars);
    if(numVarsTerm < numVars)
      terms.push_back(polyTerm _SN(t));
    else
      terms.push_back(t);
    consolidated = false;
  }

  // swap the position of variable v1 and v2 in all our terms.
  inline void swapVar(int v1, int v2)
  {
    for(uint32_t k=0;k<terms.size();k++)
      terms[k].swapVar(v1, v2);
  }

  // make this poly unity, i.e. P(x) = x.
  // needs to know which variable we're talking about
  inline void unity(int i)
  {
    terms.resize(0);
    terms.push_back(polyTerm<scalar,numVars>());
    terms[0].coefficient = 1.0f;
    terms[0].exponents[i] = 1;
  }

  /********     RAW DATA DUMPING (for easy interfacing to C-like languages such as GLSL, CUDA)     *******/

  // Construct by raw data dump
  inline truncPoly(int numTerms, void* data, int elementSize);

  // Dump in raw data
  inline void assign(int numTerms, void* data, int elementSize) {truncPoly(numTerms, data, elementSize);}

  // Dump out raw data
  inline void readData(void* data, int elementSize);


  /**************     MISC ASSIGNMENT FUNCTIONS     ***************/

  // Assign truncPoly to truncPoly, also works for other scalar types
  // _tO
  // truncPoly _SN &  operator= (const truncPoly _ON &rhs);

  // Assignment polyTerm to truncPoly (a single term is also a polynomial)
  truncPoly _SN &  operator= (const polyTerm _SN &rhs);


  /**************     POLYNOMIAL ARITHMETIC     ***************/

  // Add to polynomial
  truncPoly  _SN & operator+= (const truncPoly _SN &rhs);

  // Subtract from polynomial
  truncPoly  _SN & operator-= (const truncPoly _SN &rhs);

  // Multiply with polynomial
  truncPoly  _SN & operator*= (const truncPoly _SN &rhs);

  // Add two polynomials
  const truncPoly  _SN operator+ (const truncPoly _SN &rhs) const;

  // Subtract two polynomials
  const truncPoly  _SN operator- (const truncPoly _SN &rhs) const;

  // Multiply two polynomials
  const truncPoly  _SN operator* (const truncPoly _SN &rhs) const;

  // nth power of a polynomial
  inline const truncPoly _SN operator^ (const echar power) const {
    assert(power >= 0);
    truncPoly _SN result;
    if (power == 0) result=truncPoly _SN ((scalar)1);
    else if (power == 1)  result=*this;
    else if (power == 2)  result=*this * *this;
    else if (power & 1)  result= *this * (*this^(power/2)) * (*this^(power/2)); 
    else result = (*this^(power/2)) * (*this^(power/2)); 
    return result;
  }

  // Evaluate the poly: Map input (x0,...) to output scalar
  scalar evaluate(const scalar x[numVars]);
    
  // Compute derivative with respect to var. wrt at location x.
  scalar getDerivative(int wrt, scalar x[numVars]);

  // Derive system analytically wrt ("with respect to") variable wrt.
  const truncPoly _SN  getDerivative(int wrt) const;

  // Swap columns
  void swapColumns(int col1, int col2) {
    assert(col1 >= 0 && col1 < numVars && col2 >= 0 && col2 < numVars);
    for (int i = 0; i < terms.size(); ++i) {
      echar tmp = terms[i].exponents[col1];
      terms[i].exponents[col1] = terms[i].exponents[col2];
      terms[i].exponents[col2] = tmp;
    }
  }


  // Truncates poly degree, sorts terms, and consolidates duplicates 
  // Example: 4 x0^3 + 5 x0^3 + 2 x0^2 => 9 x0^3 + 2 x0^2
  void consolidateTerms();//int degree=-1);    


  // Builds the concatenation of two polynomials P1 and P2 (P2 must be in 1 variable):
  // x'' = P12(x) = P2(P1(x));
  const truncPoly _SN plugInto(const truncPoly <scalar, 1> &P2) const;

  // Concatenate polynomials in operator syntax
  const truncPoly _SN operator>>(const truncPoly <scalar, 1> &P2) const {
    return plugInto(P2);
  }

  // Cut polynomial to desired degree
  void cut(int degree) {
    truncDegree = degree;
    consolidateTerms();
  }

  truncPoly  _SN & operator%= (int degree) {
    cut(degree); return *this;
  }

  const truncPoly  _SN operator% (int degree) const {
    truncPoly  _SN result = *this;
    result.cut(degree);
    return result;
  }

  // Needs to be friends with polyTerm+polyTerm => truncPoly addition
  template <typename s, int nv> 
    friend  truncPoly<s, nv> operator+(const polyTerm<s,nv>& lhs, const polyTerm<s,nv>& );

  // Output some trivia about the polynomial: 
  // #in variables, trunc degree, highest degree term, #terms
  void printStats(std::ostream& os) const;

};


// Polynomial system: rectangular equation system of the form
//  x0'         = P1        (x0, ... , x{nvIn-1})
//  ...           ...
//  x{nvOut-1}' = P{nvOut-1}(x0, ... , x{nvIn-1})
//
//  For systems P1 and P2 applied in sequence, x'' = P2(x') = P2(P1(x)), 
//  the combined system P12 can be obtained as P1.plugInto(P2);
//  (If P1 and P2 contain only terms of degree 1, this is equivalent to a matrix product P2P1.)

_tSIO
class truncPolySystem {

public:
  truncPoly <scalar, numVarsIn> equations[numVarsOut];
  int truncDegree;

  truncPolySystem(int degree=-1){truncDegree=degree;}

  // Construct from other scalar tps
  // _tO
  // inline truncPolySystem(const truncPolySystem _OIO &rhs) {
  //   operator=(rhs);
  // }

  int write(const char *filename)
  {
    FILE *f = fopen(filename, "wb");
    int r = 0;
    if(!f) return 1;
    for(int k=0;k<numVarsOut;k++)
    {
      int32_t num_terms = equations[k].terms.size();
      r = fwrite(&num_terms, sizeof(int32_t), 1, f);
      if(r != 1) { fclose(f); return 1; }
      for(int t=0;t<num_terms;t++)
      {
        float coeff = equations[k].terms[t].coefficient;
        fwrite(&coeff, sizeof(float), 1, f);
        for(int v=0;v<numVarsIn;v++)
        {
          int32_t exp = equations[k].terms[t].exponents[v];
          fwrite(&exp, sizeof(int32_t), 1, f);
        }
      }
    }
    fclose(f);
    return 0;
  }

  // Assign truncPoly to truncPolySystem: only defined for numVarsOut==1
  _tO
  truncPolySystem _SIO &  operator= (const truncPoly <otherScalar, numVarsIn> &rhs) {
    assert(numVarsOut==1);
    equations[0] = rhs;
    truncDegree = rhs.truncDegree;

    return *this;
  }

  // Assign truncPolySystem to truncPolySystem
//   _tO
//   truncPolySystem _SIO &  operator= (const truncPolySystem _OIO &rhs) {
//     if (this == (truncPolySystem _SIO *) &rhs)      // Self-assignment?
//       return *this;        // Yes, so skip assignment, and just return *this.
//     for (int i = 0; i < numVarsOut; ++i)
//       equations[i] = rhs.equations[i];
//     truncDegree = rhs.truncDegree;
//     return *this;
//   }


  // Access equations by [] brackets:
  truncPoly <scalar, numVarsIn>& operator[] (int i);       
  const truncPoly <scalar, numVarsIn>& operator[] (int i) const; 


  // reset to null polynomial (no terms)
  void clear()
  {
    for(int k=0;k<numVarsOut;k++)
      equations[k].terms.clear();
  }

  inline void swapVar(int v1, int v2)
  {
    equations[v1].terms.swap(equations[v2].terms);
    for(int k=0;k<numVarsOut;k++) equations[k].swapVar(v1, v2);
  }

  // Builds the concatenation of two polynomial systems P1 and P2:
  // x'' = P12(x) = P2(P1(x));
  template <int numVarsResult>
  const truncPolySystem<scalar, numVarsIn, numVarsResult> plugInto(const truncPolySystem<scalar, numVarsOut, numVarsResult> &P2) const;

  // Concatenate systems in operator syntax
  template <int numVarsResult>
  const truncPolySystem<scalar, numVarsIn, numVarsResult> operator>>(const truncPolySystem<scalar, numVarsOut, numVarsResult> &P2) const {
    return (*this).plugInto(P2);
  }

  // Builds the concatenation of a polynomial system P1 and a polynomial P2:
  // x'' = P12(x) = P2(P1(x));
  const truncPoly <scalar, numVarsIn> plugInto(const truncPoly <scalar, numVarsOut> &P2) const {
    truncPolySystem <scalar, numVarsOut, 1> proxy2;
    proxy2.equations[0] = P2;  proxy2.truncDegree = P2.truncDegree;
    truncPolySystem <scalar, numVarsIn, 1> proxyOut = plugInto(proxy2);
    return proxyOut[0];

  }

  // Concatenate system and polynomial in operator syntax
  const truncPoly <scalar, numVarsIn> operator>>(const truncPoly <scalar, numVarsOut> &P2) const {
    return plugInto(P2);
  }

  // Linearly interpolate with other system, returning new system with numVarsIn+1 input vars
  const truncPolySystem <scalar, numVarsIn+1, numVarsOut> lerpWith(  const truncPolySystem _SIO &system2, const scalar position1, const scalar position2) const;

  // Bake one input variable as constant into coefficients
  const truncPolySystem <scalar, numVarsIn-1, numVarsOut> bakeInputVariable(int varIdx, scalar value) const;

  // Drop a line from the equation system
  const truncPolySystem <scalar, numVarsIn, numVarsOut-1> dropEquation(int eqIdx) const;

  // Evaluate the system: Map input (x0,...) to output (x1,...)
  void evaluate(const scalar x0[numVarsIn], scalar x1[numVarsOut], bool print = false);

  // evaluate given the adjoint polynomial (in fact try to invert the adjoint by using this poly as starting point)
  float evaluateWithAdjoint(const scalar x0[numVarsIn], scalar x1[numVarsOut], truncPolySystem _SIO &adjoint);

  // given x0 and x1 (spatial dimensions and all additional things such as potential lambda), iterate on omega to enforce polynomial constraint.
  float solveOmega(scalar x0[numVarsIn], scalar x1[numVarsOut]);

  // Evaluate the system: Map input (x0,...) to output (x1,...)
  void evaluateDouble(const scalar x0[numVarsIn], scalar x1[numVarsOut], bool print = false);

  // Evaluate the system and return in same variable
  void evaluate(scalar x0[numVarsIn]);

  // Evaluate the i-th equation of the system
  scalar evaluateLine(const scalar x0[numVarsIn], int i);

  // Derive system analytically wrt ("with respect to") variable wrt.
  const truncPolySystem _SIO  getDerivative(int wrt) const;

  // Computes the system's Jacobian at location x, and stores it in coeffs (row-major indexing).
  void getJacobian(scalar coeffs[numVarsIn*numVarsOut], scalar x[numVarsIn]);

  // Swap rows
  void swapRows(int row1, int row2) {
    assert (row1 >= 0 && row1 < numVarsOut && row2 >= 0 && row2 < numVarsOut);
    truncPoly <scalar, numVarsIn> temp = equations[row1];
    equations[row1] = equations[row2];
    equations[row2] = temp;
  }

  // Swap columns
  void swapColumns(int col1, int col2) {
    for (int i = 0; i < numVarsOut; ++i)
      equations[i].swapColumns(col1,col2);  
  }

  // Limits poly degree, sorts terms, and consolidates duplicates 
  // Example: 4 x0^3 + 5 x0^3 + 2 x0^2 => 9 x0^3 + 2 x0^2
  inline void consolidateTerms() {
    for (int i = 0; i < numVarsOut; ++i) {
      equations[i].truncDegree = truncDegree;
      equations[i].consolidateTerms();
    }
  } 

  // Cut polynomial system to desired degree
  void cut(int degree) {
    truncDegree = degree;
    consolidateTerms();
  }

  truncPolySystem  _SIO & operator%= (int degree) {
    cut(degree);
    return *this;
  }

  const truncPolySystem  _SIO operator% (int degree) const {
    truncPolySystem _SIO result = *this;
    result.cut(degree);
    return result;
  }

  // Add to this polynomial systems
  truncPolySystem   _SIO & operator+= (const truncPolySystem  _SIO &rhs);

  // Add two polynomial systems
  const truncPolySystem   _SIO operator+ (const truncPolySystem  _SIO &rhs) const;

  // Output some trivia about the system: 
  // #in variables, #out variables, trunc degree, highest degree term, #terms

  void printStats(std::ostream& os) const;

};



/**************     FORMATTED OUTPUT     ***************/

// Print polyTerm as text
_tSN
std::ostream& operator<<(std::ostream& os, const polyTerm _SN& term);

// Print truncPoly as text (polynomial)
_tSN
std::ostream& operator<<(std::ostream& os, const truncPoly _SN& poly);

// Print truncPolySystem as text (equation system)
_tSIO
std::ostream& operator<<(std::ostream& os, const truncPolySystem _SIO& tpt);


/**************************************************************************/
/*                         IMPLEMENTATION PART                            */
/**************************************************************************/



/**************     MIXED-CLASS ARITHMETIC     ***************/




// Addition of polynomial and term (in general) gives a polynomial
template <typename s, int nv>
truncPoly<s, nv> operator+(const truncPoly<s,nv>& lhs, const polyTerm<s,nv>& rhs){

  truncPoly <s, nv> poly = lhs;
  poly.terms.push_back(rhs);
  poly.consolidateTerms();
      
  return poly;
}

// Addition of term and polynomial (in general) gives a polynomial
template <typename s, int nv>
truncPoly<s, nv> operator+(const polyTerm<s,nv>& lhs, const truncPoly<s,nv>& rhs){

  return rhs+lhs;
}

// Left-multiplication of polyTerm with a scalar
template <typename s, int nv>
polyTerm <s,nv> operator*(const s lhs, const polyTerm <s,nv> & rhs) {
  polyTerm <s,nv> result = rhs;
  result.coefficient *= lhs;
  return result;
}

// Left-multiplication of truncPoly with a scalar
template <typename s, int nv>
truncPoly <s,nv> operator*(const s lhs, const truncPoly <s,nv> & rhs) {

  if (lhs == 0) { truncPoly <s,nv> zero; return zero;} // multiplication by zero: return empty polynomial
  truncPoly <s,nv> poly = rhs;
  for (unsigned int i = 0; i < poly.terms.size(); ++i) {
    poly.terms[i].coefficient *= lhs;
  }
  return poly;
}









/**************  polyTerm CLASS FUNCTIONS  ****************/


_tSN
int polyTerm _SN::degree() const {
  int d = 0;
  for (int i = 0; i < numVars; ++i) {
    d += exponents[i];
  }
  return d;
}    

_tSN
inline polyTerm _SN::polyTerm() {
  coefficient = 0;
  for (int i = 0; i < numVars; ++i)
    exponents[i] = 0;
}    

  // Constructor from scalar
_tSN
inline polyTerm _SN::polyTerm(scalar constant)
{
  coefficient = constant;
  for (int i = 0; i < numVars; ++i)
    exponents[i] = 0;
}

// Construct a term with 1 independent variable
_tSN
inline polyTerm _SN::polyTerm(scalar coeff, int ex, ...)
{
  coefficient = coeff;
  va_list ap;
  va_start(ap, ex);
  exponents[0] = ex;
  for(int j = 1; j < numVars; j++)
    exponents[j] = va_arg(ap, int);
  va_end(ap);
}

//   _tSN
//   _tO
//   polyTerm  _SN & polyTerm _SN::operator= (const polyTerm _ON &rhs) {
//   if (this == (polyTerm _SN *) &rhs)      // Same object?
//     return *this;        // Yes, so skip assignment, and just return *this.
//   coefficient = rhs.coefficient;
//   for (int i = 0; i < numVars; ++i)
//     exponents[i] = rhs.exponents[i];
//   return *this;
// }

_tSN
polyTerm  _SN & polyTerm _SN::operator= (const scalar rhs) {
  coefficient = rhs;
  for (int i = 0; i < numVars; ++i)
    exponents[i] = 0;
  return *this;
}

_tSN
const truncPoly  _SN polyTerm _SN::operator+ (const polyTerm _SN &rhs) const {
  truncPoly _SN result;
  result = result + *this + rhs;
  return result;
} 

_tSN
const truncPoly  _SN polyTerm _SN::operator- (const polyTerm _SN &rhs) const {
  truncPoly _SN result;
  result = result + *this - rhs;
  return result;
}  

_tSN
polyTerm  _SN & polyTerm _SN::operator*= (const polyTerm _SN &rhs) {
  coefficient *= rhs.coefficient;
  for (int i = 0; i < numVars; ++i) {
    exponents[i] += rhs.exponents[i];
  }
  return *this;
} 

_tSN
const polyTerm  _SN polyTerm _SN::operator* (const polyTerm _SN &rhs) const {
  polyTerm _SN result = *this;
  result *= rhs;           
  return result;           
} 

_tSN
const bool polyTerm _SN::operator< (const polyTerm _SN &rhs) const {
  for (int i = 0; i < numVars; ++i) { 
    // The first differing exponent decides
    if (exponents[i] != rhs.exponents[i]) 
      return (exponents[i] < rhs.exponents[i]);
  }
  return false; // else equal, hence "not less than"
}

_tSN
const bool polyTerm _SN::operator== (const polyTerm _SN &rhs) const {
  
  for (int i = 0; i < numVars; ++i) { 
    // If one of the exponents differs, not equal
    if (exponents[i] != rhs.exponents[i])
      return false;
  }
  
  return true; // else equal
}



/**************  truncPoly CLASS FUNCTIONS  ****************/

_tSN
inline truncPoly _SN::truncPoly(int numTerms, void* data, int elementSize) {
  assert(elementSize==sizeof(polyTerm _SN)); // make sure the data is of the right format!
  truncDegree = -1;
  terms.resize(numTerms);
  for (int i = 0; i < numTerms; ++i) { // avoid memcpy
    terms[i] = ((polyTerm _SN*) data) [i];
  }
  consolidated = false;
}

_tSN
inline void truncPoly _SN::readData(void* data, int elementSize) {
  assert(elementSize==sizeof(polyTerm _SN)); // make sure the data is of the right format!
  for (int i = 0; i < terms.size(); ++i) { // avoid memcpy
    ((polyTerm _SN*) data) [i] = terms[i];
  }
}

// _tSN
// _tO
// truncPoly _SN & truncPoly _SN::operator= (const truncPoly _ON &rhs) {
//   
// if (this == (truncPoly _SN *)&rhs)      // Same object?
//     return *this;        // Yes, so skip assignment, and just return *this.
//   truncDegree = rhs.truncDegree;
//   terms.resize(rhs.terms.size());
//   for (int i = 0; i < terms.size(); ++i) terms[i] = rhs.terms[i];
//   consolidated = rhs.consolidated;
//   return *this;
// }

_tSN
truncPoly _SN & truncPoly _SN::operator= (const polyTerm _SN &rhs) {
  truncDegree = -1;
  terms.clear();
  terms.push_back(rhs);
  consolidated = true;
  return *this;
}

_tSN
truncPoly  _SN & truncPoly _SN::operator+= (const truncPoly _SN &rhs) {
  int lhsSize = terms.size();
  terms.resize(lhsSize + rhs.terms.size());

  // Simply add rhs's terms:
  int i = lhsSize;

  for (int j = 0; j < (int)rhs.terms.size(); ++j, ++i) 
    terms[i] = rhs.terms[j];

  consolidateTerms();  
  return *this;
}

_tSN
const truncPoly  _SN truncPoly _SN::operator+ (const truncPoly _SN &rhs) const {
  truncPoly _SN result = *this;
  result += rhs;           
  return result;           
}

_tSN
truncPoly  _SN & truncPoly _SN::operator-= (const truncPoly _SN &rhs) {
  *this += ((scalar)(-1) * rhs);
  return *this;
}

_tSN
const truncPoly  _SN truncPoly _SN::operator- (const truncPoly _SN &rhs) const {
  truncPoly _SN result = *this;
  result -= rhs;           
  return result;           
}

_tSN
truncPoly  _SN & truncPoly _SN::operator*= (const truncPoly _SN &rhs) {
  truncPoly _SN result;
  result.terms.resize(terms.size() * rhs.terms.size());
  result.truncDegree = truncDegree;
  int t = 0;
  for (int i = 0; i < (int)terms.size(); ++i)
    for (int j = 0; j < (int)rhs.terms.size(); ++j) {
      if (truncDegree < 0 || (terms[i].degree() + rhs.terms[j].degree()) <= truncDegree) {
	polyTerm _SN product = terms[i] * rhs.terms[j];
	result.terms[t++] = product;
      }
    }

  result.consolidateTerms();
  terms = result.terms;
  return *this;
  
}

_tSN
const truncPoly  _SN truncPoly _SN::operator* (const truncPoly _SN &rhs) const {
  truncPoly _SN result = *this;
  result *= rhs;           
  return result;           
}

_tSN
void truncPoly _SN::consolidateTerms()
{
  // Pass 0: Sort terms according to their exponents
  std::sort(terms.begin(), terms.end());

  // Pass 1: Only keep terms with tolerable degree and nonzero coefficient
  std::vector<polyTerm _SN> newTerms;
  for (int i = 0; i < (int)terms.size(); ++i)
    if ((truncDegree < 0 || terms[i].degree() <= truncDegree) && 
         std::abs(terms[i].coefficient) > c_discardTermThreshold)
      newTerms.push_back(terms[i]);

  // Pass 2: Add up terms with identical exponents
  if (newTerms.size() > 0)
  {
    terms[0] = newTerms[0];
    int j = 1;
    for (int i = 1; i < (int)newTerms.size(); ++i)
    {
      if (newTerms[i-1] < newTerms[i])
        terms[j++] = newTerms[i];
      else
        terms[j-1].coefficient += newTerms[i].coefficient;
    }
    terms.resize(j);
  }
  else
    terms.resize(0);

  consolidated = true;
}


// Evaluate one line (equation) of the system: Map input (x0,...) to output x_i
_tSN
scalar truncPoly _SN::evaluate(const scalar x[numVars]){
  scalar result; 

  
  for (int j = terms.size() - 1; j >= 0; --j) {
    polyTerm<scalar,numVars> term = terms[j];
    scalar termValue = term.coefficient;
    for (int k = 0; k < numVars; ++k)
      if (term.exponents[k]) 
	termValue *= intpow(x[k],term.exponents[k]);
    result += termValue;
  }
  return result;
}


_tSN
scalar truncPoly _SN::getDerivative(int wrt, scalar x[numVars]) {
  scalar der = 0;
  for (int i = 0; i < (int)terms.size(); ++i) {
    
    scalar term = terms[i].coefficient;

    for (int v = 0; v < numVars; ++v) { // go through variables
      
      if (v == wrt) {
	echar e = terms[i].exponents[v];
	if (e) {
	  term *= e * intpow(x[v], e-1);
	} else term = 0;
      } else {
	echar e = terms[i].exponents[v];
	term *= intpow(x[v], e);
      }
    }
    
    der += term;
  }
  return der;
}

_tSN
const truncPoly _SN  truncPoly _SN::getDerivative(int wrt) const {
  assert(wrt >= 0 && wrt < numVars);
  
  truncPoly _SN result;
  
  for (int i = 0; i < (int)terms.size(); ++i) {
    
    polyTerm _SN term;
    echar e = terms[i].exponents[wrt];

    if (e) {
      term.coefficient = terms[i].coefficient * (scalar)e;
      for (int j = 0; j < numVars; ++j)
	term.exponents[j] = terms[i].exponents[j];
      --term.exponents[wrt];
      if (term.coefficient!=(scalar)0) 
	result.terms.push_back(term);
    }
  }
  
  //result.consolidateTerms();
  return result;

}

_tSN
void truncPoly _SN::printStats(std::ostream& os) const {
  int numTerms = terms.size();
  int highestDegree = 0;
  for (int j = 0; j < numTerms; ++j) {
    int termDegree = terms[j].degree();
    if (termDegree > highestDegree) 
      highestDegree = termDegree;
    }

  os << "Polynomial (" 
     << numVars << "in, numTerms=" 
     << numTerms << ", highestDegree=" 
     << highestDegree <<", trunc=";
  if (truncDegree >= 0) os << truncDegree;
  else os << "NOTRUNC";
  os << ")" << std::endl;
}



  // Builds the concatenation of two polynomials P1 and P2 (P2 must be in 1 variable)
  // x'' = P12(x) = P2(P1(x));
_tSN
const truncPoly _SN
truncPoly _SN::plugInto(const truncPoly<scalar,1> &P2) const {

  truncPoly _SN P12; // start with empty polynomial 
  
  P12.truncDegree = truncDegree;

  for (int j = 0; j < (int)P2.terms.size(); ++j) { // then, term by term
    truncPoly _SN P2Term(P2.terms[j].coefficient); // keep the coefficient
    P2Term.truncDegree = truncDegree;
     
    P2Term *= (*this)^(P2.terms[j].exponents[0]);
    P12 += P2Term;
    P12.consolidateTerms();
  }
  return P12;
}



/**************  truncPolySystem CLASS FUNCTIONS  ****************/

// Constructs system by linear blending of two systems with one less input variable
_tSIO
const truncPolySystem<scalar, numVarsIn+1, numVarsOut>
truncPolySystem _SIO::lerpWith( const truncPolySystem _SIO &system2, 
							    const scalar position1,
				                const scalar position2) const {
  truncPolySystem _SIO system1 = *this;

  truncPolySystem<scalar, numVarsIn+1, numVarsOut> result;
  
  result.truncDegree = std::max(system1.truncDegree,system2.truncDegree) + 1;
  for (int i = 0; i < numVarsOut; ++i) {
	// Reserve space for terms from input systems
    int numTerms = system1[i].terms.size() + system2[i].terms.size();
	result[i].terms.resize(2 * numTerms);
	result[i].truncDegree = truncDegree;

	int j = 0;
    // Copy terms from system 1, and add linear dependency
	for (int k = 0; k < (int)system1[i].terms.size(); ++k) {
		scalar c1 = system1[i].terms[k].coefficient;
		result[i].terms[j].coefficient = c1 * (1-(position1/(position2-position1)));
		for (int ie = 0; ie < numVarsIn; ++ie)
			result[i].terms[j].exponents[ie] = system1[i].terms[k].exponents[ie];
		result[i].terms[j].exponents[numVarsIn] = 0;

		// Copy term
		result[i].terms[numTerms + j] = result[i].terms[j];
		result[i].terms[numTerms + j].coefficient = - c1 / (position2-position1);
		result[i].terms[numTerms + j].exponents[numVarsIn] = 1;
		++j;
	}



	// Copy terms from system 2
	for (int k = 0; k < (int)system2[i].terms.size(); ++k) {
		scalar c2 = system2[i].terms[k].coefficient;
		result[i].terms[j].coefficient = c2 * position1 / (position2-position1);
		for (int ie = 0; ie < numVarsIn; ++ie)
			result[i].terms[j].exponents[ie] = system2[i].terms[k].exponents[ie];
		result[i].terms[j].exponents[numVarsIn] = 0;

		// Copy term
		result[i].terms[numTerms + j] = result[i].terms[j];
		result[i].terms[numTerms + j].coefficient = c2 / (position2-position1);
		result[i].terms[numTerms + j].exponents[numVarsIn] = 1;
		++j;
	}

	result[i].consolidateTerms();
  }
  return result;
}

_tSIO
const 
truncPolySystem <scalar, numVarsIn-1, numVarsOut> truncPolySystem _SIO::bakeInputVariable(int varIdx, scalar value) const {
  truncPolySystem <scalar, numVarsIn-1, numVarsOut> result;
  assert (varIdx >= 0 && varIdx < numVarsIn);
  // Copy terms over, removing "column" varIdx
  for (int i = 0; i < numVarsOut; ++i) {
    result[i].terms.resize(equations[i].terms.size());
	for (int k = 0; k < (int)equations[i].terms.size(); ++k) {
		result[i].terms[k].coefficient = equations[i].terms[k].coefficient 
			        * intpow(value, equations[i].terms[k].exponents[varIdx]);
		int j = 0;
		for (int ie = 0; ie < numVarsIn; ++ie) {
		  if (ie != varIdx) {
		    result[i].terms[k].exponents[j] = equations[i].terms[k].exponents[ie];
		    ++j;
		  }
		}
	}
	result[i].consolidateTerms();
  }
  return result;
}

_tSIO
const 
truncPolySystem <scalar, numVarsIn, numVarsOut-1> truncPolySystem _SIO::dropEquation(int eqIdx) const {
  truncPolySystem <scalar, numVarsIn, numVarsOut-1> result;
  assert (eqIdx > 0 && eqIdx < numVarsOut);
  // Copy terms over, removing "column" varIdx
  int j = 0;

  for (int i = 0; i < numVarsOut; ++i) {
	  if (i != eqIdx) {
		result[j] = equations[i]; 
		++j;
	  }
  }
  return result;
}


_tSIO
inline
truncPoly<scalar,numVarsIn>& truncPolySystem _SIO::operator[] (int row) {
  assert(row>=0 && row<numVarsOut);
  return equations[row];
}
 
_tSIO
inline
const truncPoly<scalar,numVarsIn>& truncPolySystem _SIO::operator[] (int row) const {
  assert(row>=0 && row<numVarsOut);
  return equations[row];
} 


  // Builds the concatenation of two polynomial systems P1
  // x'' = P12(x) = P2(P1(x));
_tSIO
template <int numVarsResult>
const truncPolySystem<scalar,numVarsIn,numVarsResult> 
truncPolySystem _SIO::plugInto(const truncPolySystem<scalar,numVarsOut,numVarsResult> &P2) const {

  truncPolySystem<scalar,numVarsIn,numVarsResult> P12;

  P12.truncDegree = truncDegree;

  for (int i = 0; i < numVarsResult; ++i) { // Row by row through the equation system

    // Now, unroll the polynomial by plugging in the corresponding rows of P1:

    truncPoly <scalar, numVarsIn> P12i; // start with empty polynomial 
    P12i.truncDegree = truncDegree;

    for (int j = 0; j < (int)P2[i].terms.size(); ++j) { // then, term by term

      truncPoly <scalar, numVarsIn> P2Term(P2[i].terms[j].coefficient); // keep the coefficient

      for (int k = 0; k < numVarsOut; ++k) { // but replace all variables with their polynomials as found in P1
        truncPoly <scalar, numVarsIn> temp = equations[k] ^ P2[i].terms[j].exponents[k];
        P2Term *= temp;
      }

      P12i += P2Term;
    } 

    P12.equations[i] = P12i;
    P12.equations[i].truncDegree = truncDegree;
    P12.equations[i].consolidateTerms();
 
  }
  return P12;
}




// Evaluate the system: Map input (x0,...) to output (x1,...)
_tSIO
void truncPolySystem _SIO::evaluate(const scalar x0[numVarsIn], scalar x1[numVarsOut], bool print){
  for (int i = 0; i < numVarsOut; ++i) {
    scalar result = 0;
    for (int j = (int)equations[i].terms.size() - 1; j >= 0; --j) {
      polyTerm<scalar,numVarsIn> *term = &equations[i].terms[j];
      scalar termValue = term->coefficient;
      for (int k = 0; k < numVarsIn; ++k) 
	termValue *= intpow(x0[k],term->exponents[k]);
      
      result += termValue;
    }
    x1[i] = result;
  }
}

#if 0 // unused now
namespace
{
// fake 5x5 matrix inverse (only inverts the upper left 4x4 sub block)
void mat5inv(const float X[][5], float R[][5])
{
  const float det =
    X[0][3] * X[1][2] * X[2][1] * X[3][0]
    - X[0][2] * X[1][3] * X[2][1] * X[3][0]
    - X[0][3] * X[1][1] * X[2][2] * X[3][0]
    + X[0][1] * X[1][3] * X[2][2] * X[3][0]
    + X[0][2] * X[1][1] * X[2][3] * X[3][0]
    - X[0][1] * X[1][2] * X[2][3] * X[3][0]
    - X[0][3] * X[1][2] * X[2][0] * X[3][1]
    + X[0][2] * X[1][3] * X[2][0] * X[3][1]
    + X[0][3] * X[1][0] * X[2][2] * X[3][1]
    - X[0][0] * X[1][3] * X[2][2] * X[3][1]
    - X[0][2] * X[1][0] * X[2][3] * X[3][1]
    + X[0][0] * X[1][2] * X[2][3] * X[3][1]
    + X[0][3] * X[1][1] * X[2][0] * X[3][2]
    - X[0][1] * X[1][3] * X[2][0] * X[3][2]
    - X[0][3] * X[1][0] * X[2][1] * X[3][2]
    + X[0][0] * X[1][3] * X[2][1] * X[3][2]
    + X[0][1] * X[1][0] * X[2][3] * X[3][2]
    - X[0][0] * X[1][1] * X[2][3] * X[3][2]
    - X[0][2] * X[1][1] * X[2][0] * X[3][3]
    + X[0][1] * X[1][2] * X[2][0] * X[3][3]
    + X[0][2] * X[1][0] * X[2][1] * X[3][3]
    - X[0][0] * X[1][2] * X[2][1] * X[3][3]
    - X[0][1] * X[1][0] * X[2][2] * X[3][3]
    + X[0][0] * X[1][1] * X[2][2] * X[3][3];
  R[0][0] = ( X[1][2]*X[2][3]*X[3][1] - X[1][3]*X[2][2]*X[3][1] + X[1][3]*X[2][1]*X[3][2] - X[1][1]*X[2][3]*X[3][2] - X[1][2]*X[2][1]*X[3][3] + X[1][1]*X[2][2]*X[3][3] ) / det;
  R[1][0] = ( X[1][3]*X[2][2]*X[3][0] - X[1][2]*X[2][3]*X[3][0] - X[1][3]*X[2][0]*X[3][2] + X[1][0]*X[2][3]*X[3][2] + X[1][2]*X[2][0]*X[3][3] - X[1][0]*X[2][2]*X[3][3] ) / det;
  R[2][0] = ( X[1][1]*X[2][3]*X[3][0] - X[1][3]*X[2][1]*X[3][0] + X[1][3]*X[2][0]*X[3][1] - X[1][0]*X[2][3]*X[3][1] - X[1][1]*X[2][0]*X[3][3] + X[1][0]*X[2][1]*X[3][3] ) / det;
  R[3][0] = ( X[1][2]*X[2][1]*X[3][0] - X[1][1]*X[2][2]*X[3][0] - X[1][2]*X[2][0]*X[3][1] + X[1][0]*X[2][2]*X[3][1] + X[1][1]*X[2][0]*X[3][2] - X[1][0]*X[2][1]*X[3][2] ) / det;

  R[0][1] = ( X[0][3]*X[2][2]*X[3][1] - X[0][2]*X[2][3]*X[3][1] - X[0][3]*X[2][1]*X[3][2] + X[0][1]*X[2][3]*X[3][2] + X[0][2]*X[2][1]*X[3][3] - X[0][1]*X[2][2]*X[3][3] ) / det;
  R[1][1] = ( X[0][2]*X[2][3]*X[3][0] - X[0][3]*X[2][2]*X[3][0] + X[0][3]*X[2][0]*X[3][2] - X[0][0]*X[2][3]*X[3][2] - X[0][2]*X[2][0]*X[3][3] + X[0][0]*X[2][2]*X[3][3] ) / det;
  R[2][1] = ( X[0][3]*X[2][1]*X[3][0] - X[0][1]*X[2][3]*X[3][0] - X[0][3]*X[2][0]*X[3][1] + X[0][0]*X[2][3]*X[3][1] + X[0][1]*X[2][0]*X[3][3] - X[0][0]*X[2][1]*X[3][3] ) / det;
  R[3][1] = ( X[0][1]*X[2][2]*X[3][0] - X[0][2]*X[2][1]*X[3][0] + X[0][2]*X[2][0]*X[3][1] - X[0][0]*X[2][2]*X[3][1] - X[0][1]*X[2][0]*X[3][2] + X[0][0]*X[2][1]*X[3][2] ) / det;

  R[0][2] = ( X[0][2]*X[1][3]*X[3][1] - X[0][3]*X[1][2]*X[3][1] + X[0][3]*X[1][1]*X[3][2] - X[0][1]*X[1][3]*X[3][2] - X[0][2]*X[1][1]*X[3][3] + X[0][1]*X[1][2]*X[3][3] ) / det;
  R[1][2] = ( X[0][3]*X[1][2]*X[3][0] - X[0][2]*X[1][3]*X[3][0] - X[0][3]*X[1][0]*X[3][2] + X[0][0]*X[1][3]*X[3][2] + X[0][2]*X[1][0]*X[3][3] - X[0][0]*X[1][2]*X[3][3] ) / det;
  R[2][2] = ( X[0][1]*X[1][3]*X[3][0] - X[0][3]*X[1][1]*X[3][0] + X[0][3]*X[1][0]*X[3][1] - X[0][0]*X[1][3]*X[3][1] - X[0][1]*X[1][0]*X[3][3] + X[0][0]*X[1][1]*X[3][3] ) / det;
  R[3][2] = ( X[0][2]*X[1][1]*X[3][0] - X[0][1]*X[1][2]*X[3][0] - X[0][2]*X[1][0]*X[3][1] + X[0][0]*X[1][2]*X[3][1] + X[0][1]*X[1][0]*X[3][2] - X[0][0]*X[1][1]*X[3][2] ) / det;

  R[0][3] = ( X[0][3]*X[1][2]*X[2][1] - X[0][2]*X[1][3]*X[2][1] - X[0][3]*X[1][1]*X[2][2] + X[0][1]*X[1][3]*X[2][2] + X[0][2]*X[1][1]*X[2][3] - X[0][1]*X[1][2]*X[2][3] ) / det;
  R[1][3] = ( X[0][2]*X[1][3]*X[2][0] - X[0][3]*X[1][2]*X[2][0] + X[0][3]*X[1][0]*X[2][2] - X[0][0]*X[1][3]*X[2][2] - X[0][2]*X[1][0]*X[2][3] + X[0][0]*X[1][2]*X[2][3] ) / det;
  R[2][3] = ( X[0][3]*X[1][1]*X[2][0] - X[0][1]*X[1][3]*X[2][0] - X[0][3]*X[1][0]*X[2][1] + X[0][0]*X[1][3]*X[2][1] + X[0][1]*X[1][0]*X[2][3] - X[0][0]*X[1][1]*X[2][3] ) / det;
  R[3][3] = ( X[0][1]*X[1][2]*X[2][0] - X[0][2]*X[1][1]*X[2][0] + X[0][2]*X[1][0]*X[2][1] - X[0][0]*X[1][2]*X[2][1] - X[0][1]*X[1][0]*X[2][2] + X[0][0]*X[1][1]*X[2][2] ) / det;
}
}
#endif

_tSIO
float truncPolySystem _SIO::evaluateWithAdjoint(
    const scalar light[numVarsIn], scalar sensor[numVarsOut],
    truncPolySystem _SIO &adjoint)
{
  // this is supposed to be called on a system transforming light -> sensor
  // and should be passed an adjoint system transforming sensor -> light which will be used
  // as a ground truth to be inverted in the process.
  //
  // if numVarsIn != numVarsOut, the overhanging dimensions are ignored in the
  // inversion process (set to be constant)

  const int numVarsMin = std::min(numVarsOut, numVarsIn);
  const int numVarsMax = std::max(numVarsOut, numVarsIn);

  scalar sensor_r[numVarsMax]; // input for the adjoint evaluation
  scalar light_r [numVarsMax]; // probably slightly wrong output of reverse computation
  // transport light to sensor (predict, coarse approximation of f-1(y))
  evaluate(light, sensor);
  // fprintf(stderr, "eval: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);
  float sqr_err = 0.0f;

  // arbitrary iteration number.
  // 8 came from experiments with the CanonZoom lens to be able to invert paths well past the aperture boundary
  for(int it=0;it<8;it++)
  {
    for(int i=0;i<numVarsOut;i++) sensor_r[i] = sensor[i];
    // need to flip direction [dx, dy] to transport back:
    // (ground truth, we are trying to invert this)
    for(int i=2;i<4;i++) sensor_r[i] = -sensor[i];
    // copy over constant dimensions:
    for(int i=numVarsMin;i<numVarsMax;i++) sensor_r[i] = light[i];
    // evaluate f(x_n) = y_n
    adjoint.evaluate(sensor_r, light_r);

#if 0 // invert jacobian. should be slower, still unstable (in about the same spot)
    // actually works a fair bit better in edge cases.
    // TODO: do some real timing benchmarks on this.
    // TODO: if it shows that this is a lot slower, only switch to it in cases where the adjoint Jacobian doesn't lead to quick convergence.
    // J(x)
    scalar J[numVarsIn*numVarsOut];
    adjoint.getJacobian(J, sensor_r);
    scalar JI[numVarsIn*numVarsOut];
    memset(JI, 0, numVarsIn*numVarsOut*sizeof(float));
    JI[numVarsIn*numVarsOut-1] = 1.0f;
    // JI is J^{-1} (y_n), at the estimated position light_r
    mat5inv((float (*)[5])J, (float (*)[5])JI);

    // have to invert light_r's [dx, dy] and copy lambda to match system again:
    for(int k=2;k<4;k++) light_r[k] = -light_r[k];
    scalar d_light[numVarsIn];
    for(int k=0;k<numVarsIn;k++) d_light[k] = (light[k] - light_r[k]);
    sqr_err = 0.0f;
    for(int k=0;k<numVarsIn;k++) sqr_err += d_light[k]*d_light[k];
    // now compute x_{n+1} = x_n - J^{-1} f(x_n)
    for(int k=0; k<4; k++)
      for(int i=0; i<4; i++) sensor[k] -= JI[5*k+i] * d_light[i];
#endif

    // close to perfect at 180 iterations:
#if 1 // use adjoint jacobian directly.
    // output will be a little off, lives in light domain again:
    // fprintf(stderr, "adjt: %f %f %f %f -> %f %f %f %f\n", sensor_r[0], sensor_r[1], sensor_r[2], sensor_r[3], light_r[0], light_r[1], light_r[2], light_r[3]);

    // have to invert light_r's [dx, dy] and copy lambda again to match system again:
    for(int k=2;k<4;k++) light_r[k] = -light_r[k];
    for(int i=numVarsMin;i<numVarsMax;i++) light_r[i] = light[i];

    // error check:
    scalar d_light[numVarsIn];
    for(int k=0;k<numVarsIn;k++) d_light[k] = light[k] - light_r[k];

    sqr_err = 0.0f;
    for(int k=0;k<numVarsIn;k++) sqr_err += d_light[k]*d_light[k];
    // fprintf(stderr, "error = %f\n", sqrtf(sqr_err));

    // how much do we need to adjust the point on the sensor
    // to get the corresponding reversed sensor_r -> light_r = light
    scalar J[numVarsIn*numVarsOut];
    getJacobian(J, light_r);

    // move sensor point:
    for(int k=0;k<numVarsOut;k++)
    {
      for(int i=0;i<numVarsIn;i++)
        sensor[k] += J[numVarsIn*k + i]*d_light[i];
    }
#endif
  }
  return sqr_err;
}

_tSIO
float truncPolySystem _SIO::solveOmega(scalar x0[numVarsIn], scalar x1[numVarsOut])
{
  // solve the first two rows of the system x1 = P(x0, omega0, ..) for given x1 and x0.
  // 1) first guess, take omega0 = normalise(x1 - x0), leave rest of x0 input unchanged
  //    this has to be done outside, we don't know about the 3d geometry any more.
  float sqr_err = 0.0f;
  scalar x1p[numVarsOut];
  for(int k=0;k<4;k++)
  {
    // 2) evaluate what we get x1' = P(x0, omega0, ..)
    evaluate(x0, x1p);

    // 3) evaluate 2x2 submatrix of jacobian dx1/domega0 and step omega0 back to our target x1.
    scalar dx1_domega0[2][2];
    dx1_domega0[0][0] = equations[0].getDerivative(2, x0);
    dx1_domega0[0][1] = equations[0].getDerivative(3, x0);
    dx1_domega0[1][0] = equations[1].getDerivative(2, x0);
    dx1_domega0[1][1] = equations[1].getDerivative(3, x0);

    // 4) invert jacobian (could use adjoint, but who's gonna fight over a 2x2 inversion)
    scalar invJ[2][2];
    const scalar invdet = 1.0f/(dx1_domega0[0][0]*dx1_domega0[1][1] - dx1_domega0[0][1]*dx1_domega0[1][0]);
    invJ[0][0] =  dx1_domega0[1][1]*invdet;
    invJ[1][1] =  dx1_domega0[0][0]*invdet;
    invJ[0][1] = -dx1_domega0[0][1]*invdet;
    invJ[1][0] = -dx1_domega0[1][0]*invdet;

    // 5) determine step and update omega0
    const scalar dx1[2] = {x1[0] - x1p[0], x1[1] - x1p[1]};
    sqr_err = 0.0f;
    for(int k=0;k<2;k++) sqr_err += dx1[k]*dx1[k];
    for(int k=0;k<2;k++)
      for(int i=0;i<2;i++)
        x0[2+k] += invJ[k][i]*dx1[i];
  }
  // now evaluate omega out
  x1[0] = x1p[0];
  x1[1] = x1p[1];
  x1[2] = x1p[2];
  x1[3] = x1p[3];
  return sqr_err;
}

// Evaluate the system with intermediate steps taken in double precision: Map input (x0,...) to output (x1,...)
_tSIO
void truncPolySystem _SIO::evaluateDouble(const scalar x0[numVarsIn], scalar x1[numVarsOut], bool print){
  for (int i = 0; i < numVarsOut; ++i) {
    // Accumulate in double temp to avoid precision issues
    double result = 0;
    for (int j = (int)equations[i].terms.size() - 1; j >= 0; --j) {
      polyTerm<scalar,numVarsIn> *term = &equations[i].terms[j];
      double termValue = term->coefficient;
      for (int k = 0; k < numVarsIn; ++k) {
	double base = x0[k];
	termValue *= intpow(base,term->exponents[k]);
      }
      
      result += (double)termValue;
    }
    x1[i] = result;
  }
}

// Evaluate the system "in place" (return in same variable)
_tSIO
void truncPolySystem _SIO::evaluate(scalar x0[numVarsIn]) {
  assert(numVarsIn == numVarsOut);
  scalar x1[numVarsIn];
  evaluate(x0,x1);
  for (int i = 0; i < numVarsOut; ++i)
    x0[i] = x1[i];
}

// Evaluate one line (equation) of the system: Map input (x0,...) to output x_i
_tSIO
scalar truncPolySystem _SIO::evaluateLine(const scalar x0[numVarsIn], int i){
  assert(i >= 0 && i < numVarsOut);
  return evaluate(equations[i], x0);
  /*
  scalar x1 = 0;
  for (int j = equations[i].terms.size() - 1; j >= 0; --j) {
    polyTerm<scalar,numVarsIn> term = equations[i].terms[j];
    scalar termValue = term.coefficient;
    for (int k = 0; k < numVarsIn; ++k)
      if (term.exponents[k]) termValue *= pow(x0[k],term.exponents[k]);
    x1 += termValue;
  }
  return x1; */
}

_tSIO
const truncPolySystem _SIO truncPolySystem _SIO::getDerivative(int wrt) const {
  assert(wrt >= 0 && wrt < numVarsIn);
  truncPolySystem _SIO result;
  for (int i = 0; i < numVarsOut; ++i) {
    result.equations[i] = equations[i].getDerivative(wrt);
  }
  return result;
}

_tSIO
void truncPolySystem _SIO::getJacobian(scalar coeffs[numVarsIn*numVarsOut], scalar x[numVarsIn]) {

  for (int i = 0; i < numVarsOut; ++i) {
    for (int j = 0; j < numVarsIn; ++j) {

      coeffs[j+i*numVarsIn] = equations[i].getDerivative(j, x);

    }
  }
}

_tSIO
truncPolySystem   _SIO & truncPolySystem _SIO::operator+= 
(const truncPolySystem  _SIO &rhs) {
      
  for (int i = 0; i < numVarsOut; ++i)
    equations[i] += rhs.equations[i];

  consolidateTerms();
  return *this;
}

_tSIO
const truncPolySystem   _SIO truncPolySystem _SIO::operator+ 
(const truncPolySystem  _SIO &rhs) const {
      
  truncPolySystem  _SIO result = *this;
  result += rhs;
  return result;
}


_tSIO
void truncPolySystem _SIO::printStats(std::ostream& os) const {
  int numTerms[numVarsOut];
  int highestDegree = 0;
  for (int i = 0; i < numVarsOut; ++i) {
    numTerms[i] = equations[i].terms.size();
    for (int j = 0; j < numTerms[i]; ++j) {
      int termDegree = equations[i].terms[j].degree();
      if (termDegree > highestDegree) 
	highestDegree = termDegree;
    }
  }
  os << "Polynomial system (" 
     << numVarsIn << "in, " 
     << numVarsOut << "out, numTerms=";

  for (int i = 0; i < numVarsOut; ++i)
    os << numTerms[i] << ((i==numVarsOut-1)?"":";");
  os << ", highestDegree=" 
     << highestDegree << ", trunc=";
  if (truncDegree >= 0) os << truncDegree;
  else os << "NOTRUNC";
  os << ")" << std::endl;
      
}

}

#include "poly/Convenience.hh"

#undef c_discardTermThreshold

#endif
