#include "lensflare.h"
#include "opencl/lensflare_head.h"
#include "opencl/image_head.h"
#include "poly/truncPolySystem.hh"

#include "spectrum.h"
#include "elements/LensSystem.hh"

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

using namespace poly;

void
lf_lensflare_set_verbosity(
    lensflare_t *l,
    int verbosity)
{
    l->verbosity = verbosity;
}

lensflare_t*
lf_lensflare_init(int verbosity)
{
    // allocate a zeroed out struct:
    lensflare_t *l = (lensflare_t *)malloc(sizeof(lensflare_t));
    memset(l, 0, sizeof(lensflare_t));

    // used to speed up c++ file io during polynomial dumping:
    l->filebuffer = (char *)malloc(2<<20);

    // init default sun to something sane:
    l->sun_col[0] = l->sun_col[1] = l->sun_col[2] = 1.0f;
    l->sun_pos[2] = 1000.0f;

    l->verbosity = verbosity;

    // init global stuff
    cl_int err;
    cl_uint num_platforms = 0;
    l->platform = (cl_platform_id *)malloc(sizeof(cl_platform_id)*4);
    int selected_platform = -1;
    err = clGetPlatformIDs(4, l->platform, &num_platforms);

    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't get platform\n");
        free(l->filebuffer);
        free(l);
        return NULL;
    }

    int need_gpu = -1;
    for(uint32_t k=0;k<num_platforms;k++)
    {
        char vendor[1024];
        size_t vendor_size = 0;
        err = clGetPlatformInfo(l->platform[k], CL_PLATFORM_VENDOR, 1024, vendor, &vendor_size);
        if(l->verbosity)
            fprintf(stderr, "[lensflare] platform %u is `%s'\n", k, vendor);
        // pick cpu friendly platform:
        if(selected_platform < 0)
        {
            if(!strncmp(vendor, "Advanced", 8)) selected_platform = k;
            if(!strncmp(vendor, "NVIDIA", 6)) need_gpu = k;
        }
        // intel is preferred. comment out this line if you want to use the gpu instead.
        // if(!strncmp(vendor, "Intel", 5)) selected_platform = k;
    }

    if(selected_platform < 0 && need_gpu >= 0)
    {
        // fall back to gpu with long compile times :(
        selected_platform = need_gpu;
        err = clGetDeviceIDs(l->platform[selected_platform], CL_DEVICE_TYPE_GPU, 1, &l->device, NULL);
    }
    else if(selected_platform >= 0)
    {
        err = clGetDeviceIDs(l->platform[selected_platform], CL_DEVICE_TYPE_CPU, 1, &l->device, NULL);
    }
    else
    {
        err = 1;
    }

    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't get device: %d\n", err);
        exit(EXIT_FAILURE);
    }

    if(l->verbosity)
    {
        char vendor[1024];
        size_t vendor_size = 0;
        err = clGetPlatformInfo(l->platform[selected_platform], CL_PLATFORM_VENDOR, 1024, vendor, &vendor_size);
        fprintf(stderr, "[lensflare] using platform[%u]: `%s'\n", selected_platform, vendor);
    }
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't get device\n");
        exit(EXIT_FAILURE);
    }

    //Create the context
    l->context = clCreateContext(0, 1, &l->device, NULL, NULL, &err);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't get context\n");
        exit(EXIT_FAILURE);
    }

    // Create a command-queue
    l->command_queue = clCreateCommandQueue(l->context, l->device, 0, &err);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "couldn't get command queue\n");
        exit(EXIT_FAILURE);
    }

    return l;
}

void lf_lensflare_set_sun(lensflare_t *l, const float *pos, const float *rgb)
{
  memcpy(l->sun_pos, pos, sizeof(float)*3);
  memcpy(l->sun_col, rgb, sizeof(float)*3);
}

static double
get_wtime(void)
{
    struct timeval time;
    gettimeofday(&time, NULL);
    return time.tv_sec - 1290608000 + (1.0/1000000.0)*time.tv_usec;
}

// dump out polynomial source code to a stream
template <class scalar, int numVars>
void print_code(FILE *f, const polyTerm<scalar,numVars> term, std::string *varNames = NULL)
{
    fprintf(f, "%.4ff", term.coefficient);
    if (varNames)
    {
        for(int i=0; i<numVars; i++)
        {
            int e = term.exponents[i];
            std::string &vn = varNames[i];
            // c-code friendly:
            for(int k=0;k<e;k++)
                fprintf(f, " * %s", vn.c_str());
        }
    }
    else
    {
        for(int i=0; i<numVars; i++)
        {
            int e = term.exponents[i];
            // c code
            for(int k=0;k<e;k++)
                fprintf(f, " * x%d", i);
        }
    }
}

template <class scalar, int numVars>
bool code_empty(const truncPoly<scalar,numVars> poly)
{
  int N = poly.terms.size();
  const float eps = 0.0f;//1e-12f;

  if (N)
  {
      for(int i = 0; i < N - 1; i++)
      {
          if(fabsf(poly.terms[i].coefficient) > eps)
          {
              return false;
          }
      }

      if(fabsf(poly.terms[N-1].coefficient) > eps)
      {
          return false;
      }
  }
  return true;
}

template <class scalar, int numVars>
bool print_code(FILE *f, const truncPoly<scalar,numVars> poly, std::string *varNames = NULL)
{
  int N = poly.terms.size();
  const float eps = 0.0f;//1e-12f;
  bool empty = true;

  if (N)
  {
      for(int i = 0; i < N - 1; i++)
      {
          if(fabsf(poly.terms[i].coefficient) > eps)
          {
              empty = false;
              print_code(f, poly.terms[i], varNames);
          }
          if(fabsf(poly.terms[i+1].coefficient) >= eps) fprintf(f, " + ");
      }

      if(fabsf(poly.terms[N-1].coefficient) > eps)
      {
          empty = false;
          print_code(f, poly.terms[N-1], varNames);
      }
      return empty;
  }
  else
  {
      return true;
  }
}

template <class scalar, int numVarsIn, int numVarsOut>
bool print_code(FILE *f, const truncPolySystem<scalar,numVarsIn, numVarsOut>& tpt, std::string *varNames = NULL)
{
  bool empty = true;
  // actually don't need dx/dy, only straight x, y:
  for(int i = 0; i < 2; i++) // numVarsOut; i++)
  {
    empty = code_empty(tpt.equations[i]);
    if(empty)
    {
        break; 
    }
    else
    {
        if (varNames)
        { 
            std::string &vn = varNames[i];
            fprintf(f, "const float out_%s = ", vn.c_str());
        }
        else
        {
            fprintf(f, "const float out_x%d = ", i);
        }
        print_code(f,tpt.equations[i],varNames);
        fprintf(f, ";\n");
    }
  }
  return empty;
}

// dump out jacobian source code to a stream
template <class scalar, int numVarsIn, int numVarsOut>
void printJacobian(truncPolySystem<scalar,numVarsIn,numVarsOut> &t, FILE *f, std::string *varnames = NULL)
{
  for (int i = 0; i < 2; i++) // only interested in first 2x2 block: numVarsOut; ++i)
  {
    for (int j = 0; j < numVarsIn; ++j)
    {
        fprintf(f, "const float dx%d%d = ", i, j);
        print_code(f, t.equations[i].getDerivative(j), varnames);
        fprintf(f, "+0.0f;\n");
    }
  }
}

template <class scalar, int numVarsIn, int numVarsOut>
void print_reverse_code(
    FILE *f,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& system,
    const truncPolySystem<scalar, numVarsIn, numVarsOut>& adjoint,
    std::string *varNames)
{
  const int numVarsMin = std::min(numVarsOut, numVarsIn);
  const int numVarsMax = std::max(numVarsOut, numVarsIn);

  std::string sensor_var[numVarsMax];
  std::string light_var[numVarsMax];
  std::string sensor_r_var[numVarsMax];
  std::string light_r_var[numVarsMax];
  for(int i=0;i<numVarsMax;i++)
  {
    sensor_var[i] = "sensor_" + varNames[i];
    light_var[i] = varNames[i];
    sensor_r_var[i] = "sensor_r_" + varNames[i];
    light_r_var[i] = "light_r_" + varNames[i];
  }

  fprintf(f, "// evaluate approximate reverse polynomial\n");
  // transport light to sensor (predict)
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "float %s = ", sensor_var[i].c_str());
    print_code(f, system.equations[i], varNames);
    fprintf(f, ";\n");
  }
  // fprintf(stderr, "eval: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);

  // arbitrary iteration number, 4 should be almost floating point accuracy
  fprintf(f, "// do newton iterations to get precise answer\n");
  fprintf(f, "for(int it=0;it<4;it++)\n{\n");
  // need to flip direction [dx, dy] to transport back:
  for(int i=0;i<numVarsOut;i++)
    fprintf(f, "  float %s = %s %s;\n",
      sensor_r_var[i].c_str(),
      (i == 2 || i == 3) ? "-" : "", sensor_var[i].c_str());

  // copy over constant dimensions:
  for(int i=numVarsMin;i<numVarsMax;i++) 
    fprintf(f, "  float %s = %s;\n",
        sensor_r_var[i].c_str(), varNames[i].c_str());

  fprintf(f, "  // evaluate adjoint to get error vector\n");
  // evaluate adjoint:
  for(int i=0;i<numVarsOut;i++)
  {
    fprintf(f, "  float %s = ", light_r_var[i].c_str());
    print_code(f, adjoint.equations[i], sensor_r_var);
    fprintf(f, ";\n");
  }

  // output will be a little off, lives in light domain again.

  // have to invert light_r's [dx, dy] and copy lambda again to match system again:
  for(int k=2;k<4;k++) fprintf(f, "  %s = - %s;\n", light_r_var[k].c_str(), light_r_var[k].c_str());
  for(int i=numVarsMin;i<numVarsMax;i++) fprintf(f, "  const float %s = %s;\n", light_r_var[i].c_str(),  light_var[i].c_str());

  for(int k=0;k<numVarsMin;k++)  // In
    fprintf(f, "  const float d_light_%d = %s - %s;\n", k, light_var[k].c_str(), light_r_var[k].c_str());

  // XXX
#if 0 
  // error check:
  float sqr_err = 0.0f;
  for(int k=0;k<numVarsIn;k++) sqr_err += d_light[k]*d_light[k];
  // fprintf(stderr, "error = %f\n", sqrtf(sqr_err));
#endif

  fprintf(f, "  // move through jacobian and correct sensor point\n");
  // how much do we need to adjust the point on the sensor
  // to get the corresponding reversed sensor_r -> light_r = light
  for (int i = 0; i < numVarsMin; i++) // Out
  {
    for (int j = 0; j < numVarsMin; ++j) // In
    {
      fprintf(f, "  const float dx%d%d = ", i, j);
      print_code(f, system.equations[i].getDerivative(j), light_r_var);
      fprintf(f, "+0.0f;\n");
    }
  }

  // move sensor point:
  for(int k=0;k<numVarsMin;k++) //numVarsOut;k++)
  {
    for(int i=0;i<numVarsMin;i++) //In;i++)
      fprintf(f, "  %s += dx%d%d * d_light_%d;\n", sensor_var[k].c_str(), k, i, i);
  }
  fprintf(f, "}\n");
}


char*
dump_code(
    lensflare_t *l,
    int flarenum,
    int smooth,
    size_t *ret_size)
{
    FILE *f = tmpfile();
    setbuffer(f, l->filebuffer, 2<<20);
    size_t size = 0;
    char *header = 0;
    if(flarenum >= -1)
    {
      // computing lensflare
      size = opencl_lensflare_head_cl_len;
      header = (char *)opencl_lensflare_head_cl;
    }
    else
    {
      // computing the transmitted image
      size = opencl_image_head_cl_len;
      header = (char *)opencl_image_head_cl;
    }
    fwrite(header, sizeof(char), size-1, f);
    fprintf(f, "\n");

    // spherical element supports degree 1 and 3, add 1 for wavelength
    // so reasonable values are 1 (monochromatic linear), 2 (linear), and 4
    int degree = 4;
    const float f_number = 2.0f;
    const float zoom = 0.0f;

    float aperture_radius;
    PolyT5f system, system2, aperture;
    std::string varNames[5];
    varNames[0]="x";
    varNames[1]="y";
    varNames[2]="dx";
    varNames[3]="dy";
    varNames[4]="lambda";

    LensSystem &lens = *((LensSystem *)l->lens);


    const int num_lambda = 10;
    float fake_spectrum[num_lambda+1][3];
    const float lambda0 = 0.400f;
    const float lambda1 = 0.700f;
    fprintf(f, "const float lambda_center = 0.550f;\n");
    fprintf(f, "const float exposure = 1.0/(float)num_samples;\n");
    fprintf(f, "float spectrum[%d][3] = {\n", num_lambda+1);
    for(int i=0;i<=num_lambda;i++)
    {
        const float lambda = lambda0 + i/(float)num_lambda * (lambda1-lambda0);
        spectrum_p_to_rgb(1000.0f*lambda, 1.0f, fake_spectrum[i]); // in nanometers
        for(int k=0;k<3;k++) fake_spectrum[i][k] *= 1.0f/num_lambda;
        fprintf(f, "{ exposure*%f, exposure*%f, exposure*%f },\n", fake_spectrum[i][0], fake_spectrum[i][1], fake_spectrum[i][2]);
    }
    // spectral lerp:
    fprintf(f, "};{\n");
    float focus_dist = 1.0f;
#if 0
    float focal_length = lens.estimateFocalLength(1);
    fprintf(stderr, "estimated first-order focal length: %f\n", focal_length);
    focal_length = lens.estimateFocalLength(4);
    fprintf(stderr, "estimated 4th-order focal length: %f\n", focal_length);
    exit(0);
#endif

#if 0
    lens.getSystem(system,  aperture, aperture_radius, lambda0, focus_dist, zoom, degree, true);
    lens.getSystem(system2, aperture, aperture_radius, lambda1, focus_dist, zoom, degree, true);
    truncPolySystem <float,5,4> spectralSystem = system.lerpWith(system2, lambda0, lambda1);

    lens.getAdjointSystem(system,  aperture, aperture_radius, lambda0, focus_dist, zoom, degree, true);
    lens.getAdjointSystem(system2, aperture, aperture_radius, lambda1, focus_dist, zoom, degree, true);
    truncPolySystem <float,5,4> spectralAdjSystem = system.lerpWith(system2, lambda0, lambda1);
    print_reverse_code(stdout, spectralSystem, spectralAdjSystem, varNames);
    exit(0);
#endif

#if 0
    //==========================================================
    // test adjoint polynomial transport:
    lens.getSystem<float,5,5>(system,  aperture, aperture_radius, focus_dist, zoom, degree, false);
    lens.getAdjointSystem<float,5,5>(system2, aperture, aperture_radius, focus_dist, zoom, degree, false);
    float light[5] = {20.0f, 10.0f, 0.03f, 0.00f, 0.500f};  // fixed input
    float sensor[5] = {0.0f, 0.0f, 0.0f, 0.0f, 0.500f};
    float sensor_r[5] = {0.0f, 0.0f, 0.0f, 0.0f, 0.500f}; // reverse input
    float light_r[5] = {0.0f, 0.0f, 0.0f, 0.0f, 0.500f};  // probably slightly wrong output of reverse check
    // transport light to sensor (predict)
    system.evaluate(light, sensor);
    fprintf(stderr, "eval: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);

    for(int it=0;it<4;it++)
    {
      for(int i=0;i<2;i++) sensor_r[i] = sensor[i];
      // need to flip direction to transport back:
      // (ground truth, we are trying to invert this)
      for(int i=2;i<4;i++) sensor_r[i] = -sensor[i];
      system2.evaluate(sensor_r, light_r);

      // output will be a little off, lives in light domain again:
      fprintf(stderr, "adjt: %f %f %f %f -> %f %f %f %f\n", sensor_r[0], sensor_r[1], sensor_r[2], sensor_r[3], light_r[0], light_r[1], light_r[2], light_r[3]);

      // have to invert light_r's dx and dy again to match system again:
      for(int k=2;k<4;k++) light_r[k] = -light_r[k];

      // error check:
      float d_light[5] = {0.0f};
      for(int k=0;k<4;k++) d_light[k] = light[k] - light_r[k];

      const float sqr_err = d_light[0]*d_light[0] +
                            d_light[1]*d_light[1] +
                            d_light[2]*d_light[2] +
                            d_light[3]*d_light[3];
      fprintf(stderr, "error = %f\n", sqrtf(sqr_err));

      // how much do we need to adjust the point on the sensor
      // to get the corresponding reversed sensor_r -> light_r = light
      float J[25];
      system.getJacobian(J, light_r);
      // move sensor point:
      for(int k=0;k<4;k++)
      {
        for(int i=0;i<4;i++)
            sensor[k] += J[5*k + i]*d_light[i];
      }
    }
    system.evaluateWithAdjoint(light, sensor, system2);
    fprintf(stderr, "eval precise: %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], sensor[0], sensor[1], sensor[2], sensor[3]);

#if 0
    // test wavelength dependent stuff :)

    // also get the adjoint thing
    system.evaluateWithAdjoint(light, sensor, system2);
    fprintf(stderr, "eval spectral: %f %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], light[4], sensor[0], sensor[1], sensor[2], sensor[3]);
    sensor[2] = - sensor[2];
    sensor[3] = - sensor[3];
    system2.evaluate(sensor, light);
    light[2] = - light[2];
    light[3] = - light[3];
    fprintf(stderr, "eval spectral: %f %f %f %f %f -> %f %f %f %f\n", light[0], light[1], light[2], light[3], light[4], sensor[0], sensor[1], sensor[2], sensor[3]);
#endif

    exit(1);
    //==========================================================
#endif

    // only directly visible stuff if no particular one was requested:
    lens.getSystem<float,5,5>(system, aperture, aperture_radius, focus_dist, zoom, degree);

    if(flarenum < -1)
    {
        fprintf(f, "const float lambda = %f + rnd2 * (%f - %f);\n", lambda0, lambda1, lambda0);
        print_code(f, system, varNames);
        fprintf(f, "const int k = %d * rnd2;\n", num_lambda);
        fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
        fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
        fprintf(f, "float col[3] = {spectrum[k][0] * L, spectrum[k][1] * L, spectrum[k][2] * L};\n");
        fprintf(f, "splat_single(output, x, y, col, width, height);\n");
        // fprintf(f, "splat(output, x, y, col, dx00, dx01, dx02, dx03, dx10, dx11, dx12, dx13, sample_ratio_xy, sample_ratio_uv, width, height);\n");
        fprintf(f, "}{\n");
        fprintf(f, "const float lambda = %f + rnd3 * (%f - %f);\n", lambda0, lambda1, lambda0);
        print_code(f, system, varNames);
        fprintf(f, "const int k = %d * rnd3;\n", num_lambda);
        fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
        fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
        fprintf(f, "float col[3] = {spectrum[k][0] * L, spectrum[k][1] * L, spectrum[k][2] * L};\n");
        fprintf(f, "splat_single(output, x, y, col, width, height);\n");
        // fprintf(f, "splat(output, x, y, col, dx00, dx01, dx02, dx03, dx10, dx11, dx12, dx13, sample_ratio_xy, sample_ratio_uv, width, height);\n");
    }
    fprintf(f, "}\n");
    if(flarenum >= -1)
    {
    const int numFlares = lens.getNumFlares();
    if(l->verbosity)
        fprintf(stderr, "[lensflare] found %d valid flares\n", numFlares);

    const int flare_beg = (flarenum >= 0) ? flarenum : 0;
    const int flare_end = (flarenum >= 0) ? flarenum+1 : numFlares;//interesting_cnt;
    for(int ii=flare_beg;ii<flare_end;ii++)
    {
        int i = ii;
        // if(flarenum < 0) i = interesting[ii];

        PolyT5f flareAT, flareATR, flareATRR, system;
        float flareATrad, flareATRrad, flareATRRrad;
        lens.getFlareSystem(system, flareAT, flareATR, flareATRR, flareATrad, flareATRrad, flareATRRrad,
                i, focus_dist, zoom, degree);

        if(flarenum < 0)
        {
          // jacobian tests only if all flares have been requested
          float J[25];
          float input[5] = {0.0f, 0.0f, 0.0f, 0.0f, 0.5f};
          system.getJacobian(J, input);
          // check if ghost would only be overall contrast loss:
          // if(fabsf(J[0]) * fabsf(J[6]) > 16.0f) continue;
          // vertical lines:
          if(fabsf(J[0]/J[6]) > 2.0f)
          {
              // fprintf(stderr, "[lensflare] discarding flare %d due to vertical striping!\n", i);
              continue;
          }
          if(fmin(J[0], J[6]) > .1f)
          {
              // fprintf(stderr, "[lensflare] discarding flare %d due to contrast loss!\n", i);
              continue;
          }
        }

#if 0   // ignore aperture
        fprintf(f, "{{{\n");
#else   // clip at aperture
        fprintf(f, "{\n");
        fprintf(f, "const float lambda = %f + rnd2 * (%f - %f);\n", lambda0, lambda1, lambda0);
        if(flareATrad > 0.0f)
        {
            print_code(f, flareAT, varNames);
            fprintf(f, "if(clip_aperture(out_x, out_y, %f))\n", flareATrad/f_number);
        }
        fprintf(f, "{\n");
        if(flareATRrad > 0.0f)
        {
            print_code(f, flareATR, varNames);
            fprintf(f, "if(clip_aperture(out_x, out_y, %f))\n", flareATrad/f_number);
        }
        fprintf(f, "{\n");
        if(flareATRRrad > 0.0f)
        {
            print_code(f, flareATRR, varNames);
            fprintf(f, "if(clip_aperture(out_x, out_y, %f))\n", flareATrad/f_number);
        }
#endif
        fprintf(f, "{\n");
        // fprintf(f, "const float lambda = %f + rnd2 * (%f - %f);\n", lambda0, lambda1, lambda0);
        if(!print_code(f, system, varNames))
        {
          if(smooth)
          {
            printJacobian(system, f, varNames);
            fprintf(f, "const int k = %d * rnd2;\n", num_lambda);
            fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
            fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
            fprintf(f, "splat(output, x, y, spectrum[k], dx00, dx01, dx02, dx03, dx10, dx11, dx12, dx13, sample_ratio_xy, sample_ratio_uv, width, height);\n");
          }
          else
          {
            fprintf(f, "const int k = %d * rnd2;\n", num_lambda);
            fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
            fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
            fprintf(f, "splat_single(output, x, y, spectrum[k], width, height);\n");
          }
        }
        fprintf(f, "}{\n");
#if 0 // this would be nice, but would clip at the wrong aperture (wrong lambda)
        fprintf(f, "const float lambda = %f + rnd3 * (%f - %f);\n", lambda0, lambda1, lambda0);
        if(!print_code(f, system, varNames))
        {
          if(smooth)
          {
            printJacobian(system, f, varNames);
            fprintf(f, "const int k = %d * rnd3;\n", num_lambda);
            fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
            fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
            fprintf(f, "splat(output, x, y, spectrum[k], dx00, dx01, dx02, dx03, dx10, dx11, dx12, dx13, sample_ratio_xy, sample_ratio_uv, width, height);\n");
          }
          else
          {
            fprintf(f, "const int k = %d * rnd3;\n", num_lambda);
            fprintf(f, "const float x = out_x *scale + width/2.0f;\n");
            fprintf(f, "const float y = out_y *scale + height/2.0f;\n");
            fprintf(f, "splat_single(output, x, y, spectrum[k], width, height);\n");
          }
        }
#endif
        fprintf(f, "}\n");
        fprintf(f, "}}}\n");
    }
    }

    // close the for-samples loop
    fprintf(f, "}\n");
    // final closing bracket.
    fprintf(f, "}\n");

    size_t sz = ftell(f);
    fseek(f, 0, SEEK_SET);
    if(ret_size) *ret_size = sz + 1;
    char *buf = (char *)malloc(sz+1);
    size_t res = fread(buf, sz, 1, f);
    if(res != 1)
    {
        free(buf);
        return NULL;
    }
    // NULL-terminate, in case size isn't used by the caller
    buf[sz] = 0;
    fclose(f);
    return buf;
}


int
lf_lensflare_setup(
    lensflare_t *l,
    const char  *lensfile)
{
    double start = get_wtime();
    LensSystem *ll = new LensSystem(lensfile);
    l->lens = (void *)ll;
    int num_flares = ll->getNumFlares();
    double end = get_wtime();
    if(l->verbosity)
        fprintf(stderr, "[lensflare] read lens description in %.3f s\n", end-start);

    return num_flares;
}


void
lf_lensflare_compile(
    lensflare_t *l,
    int          flare_num,
    int          smooth)
{
    double start = get_wtime();
    size_t size = 0;
    char *string = dump_code(l, flare_num, smooth, &size);
#if 0
    // XXX DEBUG write out code so we can see what's going on.
    FILE *f = fopen("/tmp/code.cl", "wb");
    fwrite(string, size, 1, f);
    fclose(f);
#endif
    double end = get_wtime();
    if(l->verbosity)
        fprintf(stderr, "[lensflare] code generated in %.3f s\n", end-start);
    start = get_wtime();
    cl_int err;
    l->program = clCreateProgramWithSource (l->context,
            1,         // number of files
            (const char**)&string,   // array of strings, each one is a file
            &size,     // our code is NULL terminated, but maybe this is more efficient
            &err);
    free(string);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "couldn't create program %d\n", err);
        exit(EXIT_FAILURE);
    }

    const char* flags = "-cl-fast-relaxed-math";// MH: doesn't run here: -cl-std=CL1.1";
    err = clBuildProgram(l->program, 1, &l->device, flags, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "couldn't build program\n");

        cl_build_status build_status;
        clGetProgramBuildInfo(l->program, l->device, CL_PROGRAM_BUILD_STATUS, sizeof(cl_build_status), &build_status, NULL);
        if (build_status != CL_SUCCESS)
        {
            char *build_log;
            size_t ret_val_size;
            clGetProgramBuildInfo(l->program, l->device, CL_PROGRAM_BUILD_LOG, 0, NULL, &ret_val_size);
            build_log = (char *)malloc(sizeof(char)*(ret_val_size+1));
            clGetProgramBuildInfo(l->program, l->device, CL_PROGRAM_BUILD_LOG, ret_val_size, build_log, NULL);
            build_log[ret_val_size] = '\0';
            fprintf(stderr, "%s\n", build_log);
            free(build_log);
        }
         
        exit(EXIT_FAILURE);
    }

    l->kernel = clCreateKernel(l->program, "render", &err);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "couldn't create kernel: %d\n", err);
        exit(EXIT_FAILURE);
    }
    end = get_wtime();
    if(l->verbosity)
        fprintf(stderr, "[lensflare] compiled program in %.3f s\n", end-start);
}

void lf_lensflare_set_input_image(
    lensflare_t *l,
    int width,
    int height,
    const float *input_image)
{
    cl_int err1, err2;

    // TODO: decouple in and out width/height
    l->cl_input  = clCreateBuffer(l->context, CL_MEM_READ_ONLY,  sizeof(cl_float) * 3 * width * height, NULL, &err1);
    l->cl_output = clCreateBuffer(l->context, CL_MEM_READ_WRITE, sizeof(cl_float) * width * height * 3, NULL, &err2);
    if ((err1|err2) != CL_SUCCESS)
    {
        fprintf(stderr, "[image] couldn't alloc device buffers\n");
        exit(EXIT_FAILURE);
    }

    const int num_samples = 10;
    const float dist = 10.f;
    const float pupil_radius = 14.f;

    if(input_image)
    {
        err1  = clSetKernelArg(l->kernel, 0, sizeof(cl_mem), (void*)&l->cl_output);
        err1 |= clSetKernelArg(l->kernel, 1, sizeof(cl_mem), (void*)&l->cl_input);
        err1 |= clSetKernelArg(l->kernel, 2, sizeof(cl_float), (void*)&dist);
        err1 |= clSetKernelArg(l->kernel, 3, sizeof(cl_float), (void*)&pupil_radius);
        err1 |= clSetKernelArg(l->kernel, 4, sizeof(cl_int), (void*)&num_samples);
        err1 |= clSetKernelArg(l->kernel, 5, sizeof(cl_int), (void*)&width);
        err1 |= clSetKernelArg(l->kernel, 6, sizeof(cl_int), (void*)&height);
        err1 |= clSetKernelArg(l->kernel, 7, sizeof(cl_float), (void*)&dist); // dummy
        if (err1 != CL_SUCCESS)
        {
            fprintf(stderr, "[image] couldn't set arguments\n");
            exit(EXIT_FAILURE);
        }
    }

    // upload input image:
    err1 = clEnqueueWriteBuffer(l->command_queue, l->cl_input, CL_FALSE, 0, sizeof(cl_float) * 3 * width*height, input_image, 0, NULL, NULL);
    if (err1 != CL_SUCCESS)
    {
        fprintf(stderr, "[image] couldn't upload source image\n");
        exit(EXIT_FAILURE);
    }
}

// run the cl kernel
void lf_lensflare_render_image(
    lensflare_t *l,
    float *buffer,
    int width,
    int height,
    int num_samples,
    float dist,
    float sensordist)
{
    cl_int err;
    err = clEnqueueWriteBuffer(l->command_queue, l->cl_output, CL_TRUE, 0, sizeof(cl_float) * 3 * width * height, buffer, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[image] couldn't upload buffer\n");
        exit(EXIT_FAILURE);
    }

    // set new distance to object
    err  = clSetKernelArg(l->kernel, 2, sizeof(cl_float), (void*)&dist);
    err  = clSetKernelArg(l->kernel, 4, sizeof(cl_int),   (void*)&num_samples);
    err |= clSetKernelArg(l->kernel, 7, sizeof(cl_float), (void*)&sensordist);

    // launch kernel
    const size_t sizes[2] = {(size_t)width, (size_t)height};
    err = clEnqueueNDRangeKernel(l->command_queue, l->kernel, 2, NULL, sizes, NULL /*let driver choose local size*/, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[image] couldn't enqueue kernel: %d\n", err);
        exit(EXIT_FAILURE);
    }

    // synchronous/blocking read of results, and check accumulated errors
    err = clEnqueueReadBuffer(l->command_queue, l->cl_output, CL_TRUE, 0, sizeof(cl_float) * 3 * width * height, buffer, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[image] couldn't enqueue copy back: %d\n", err);
        exit(EXIT_FAILURE);
    } 
}

void
lf_lensflare_render(
    lensflare_t *l,
    float *buffer,
    int width,
    int height,
    int num_samples)
{
    double start = get_wtime();
    cl_int err;
    int properly_aligned = 0;

    // check if buffer alignment makes opencl happy?
    size_t align = 0, align_size = 0;
    err = clGetDeviceInfo(l->device, CL_DEVICE_MEM_BASE_ADDR_ALIGN, sizeof(size_t), &align, &align_size);
    if(err == CL_SUCCESS)
    {
        if(!((long int)buffer & (align-1)))
            properly_aligned = 1;
        else
            fprintf(stderr, "[lensflare] WARN: if the buffer was aligned to %zd bytes, we could save two memcopies!\n", align);
    }

    // if not aligned correctly, issue copies.
    cl_mem cl_output = NULL;
    // we will let the opencl implementation decide whether to make a copy or not.
    // the intel cpu sdk should just use this buffer directly:
    if(properly_aligned)
        cl_output = clCreateBuffer(l->context, CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
                sizeof(cl_float) * width * height * 3, buffer, &err);
    else
        cl_output = clCreateBuffer(l->context, CL_MEM_READ_WRITE,
                sizeof(cl_float) * width * height * 3, NULL, &err);

    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't alloc device buffers\n");
        exit(EXIT_FAILURE);
    }

    err  = clSetKernelArg(l->kernel, 0, sizeof(cl_mem), (void*)&cl_output);
    err |= clSetKernelArg(l->kernel, 1, sizeof(cl_int), (void*)&num_samples);
    err |= clSetKernelArg(l->kernel, 2, sizeof(cl_float), (void*)&l->sun_pos[0]);
    err |= clSetKernelArg(l->kernel, 3, sizeof(cl_float), (void*)&l->sun_pos[1]);
    err |= clSetKernelArg(l->kernel, 4, sizeof(cl_float), (void*)&l->sun_pos[2]);
    err |= clSetKernelArg(l->kernel, 5, sizeof(cl_int), (void*)&width);
    err |= clSetKernelArg(l->kernel, 6, sizeof(cl_int), (void*)&height);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't set arguments\n");
        exit(EXIT_FAILURE);
    }

    if(!properly_aligned)
    {
        // asynchronous write of data to GPU device
        err  = clEnqueueWriteBuffer(l->command_queue, cl_output, CL_FALSE, 0, sizeof(cl_float) * 3 * width * height, buffer, 0, NULL, NULL);
        if (err != CL_SUCCESS)
        {
            fprintf(stderr, "[lensflare] couldn't upload source image\n");
            exit(EXIT_FAILURE);
        }
    }

    // launch kernel
    const size_t sizes[2] = {(size_t)width, (size_t)height};
    err = clEnqueueNDRangeKernel(l->command_queue, l->kernel, 2, NULL, sizes, NULL /*let driver choose local size*/, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        fprintf(stderr, "[lensflare] couldn't enqueue kernel: %d\n", err);
        exit(EXIT_FAILURE);
    }

    if(!properly_aligned)
    {
        // synchronous/blocking read of results, and check accumulated errors
        err = clEnqueueReadBuffer(l->command_queue, cl_output, CL_TRUE, 0, sizeof(cl_float) * 3 * width * height, buffer, 0, NULL, NULL);
        if (err != CL_SUCCESS)
        {
            fprintf(stderr, "[lensflare] couldn't copy back output buffer: %d\n", err);
            exit(EXIT_FAILURE);
        } 
    }
    else
    {
        // could assert mapped_buf == buffer
        // float *mapped_buf = (float *)
        clEnqueueMapBuffer(
                l->command_queue,
                cl_output,
                CL_TRUE,
                CL_MAP_READ,
                0,
                sizeof(cl_float) * 3 * width * height,
                0, NULL, NULL, &err);

        if (err != CL_SUCCESS)
        {
            fprintf(stderr, "[lensflare] couldn't map output buffer: %d\n", err);
            exit(EXIT_FAILURE);
        } 
    }

    clReleaseMemObject(cl_output);
    double end = get_wtime();
    if(l->verbosity)
        fprintf(stderr, "[lensflare] rendered frame in %.3f s\n", end-start);
}


void
lf_lensflare_cleanup(
    lensflare_t *l)
{
    // destroy everything
    free(l->filebuffer);
    l->filebuffer = 0;
    // shutdown opencl
    if(l->kernel)        clReleaseKernel(l->kernel);
    if(l->program)       clReleaseProgram(l->program);
    if(l->command_queue) clReleaseCommandQueue(l->command_queue);
    if(l->context)       clReleaseContext(l->context);
    if(l->platform)      free(l->platform);
    if(l->lens)          delete (LensSystem *)l->lens;
    free(l);
}

