#ifndef ELEMENTS_PATHSPACE_SURFACEX_HH
#define ELEMENTS_PATHSPACE_SURFACEX_HH
#include "truncPolySystem.hh"
#include "pow.hh"
#include "Print.hh"

namespace poly {

// surface reflection parameterized over three vertices x0 x1 x2 in their respective tangent spaces.
// the only output is f (measurement contribution)
template<typename scalar, int numVars>
static inline void surfacex_reflect(
    truncPoly<scalar, numVars> &poly,
    float k1,              // curvature in direction of tagent space T1's a vector
    float k2,              // curvature in direction of tagent space T1's b vector
    float n1,              // current ior
    float n2,              // ior of medium
    float roughness,       // beckmann roughness
    float *T0,             // 3x3 tangent space [a,b,n] of v0, z is up
    float *T1,             // v1
    float *T2,             // v2
    float *v0,             // vertex position at v0
    float *v1,             // vertex position at v1
    float *v2)             // vertex position at v2
{
  assert(numVars >= 6); // [x0,y0,x1,y1,x2,y2]
  poly.terms.clear();

  if(n1 == n2)
  {
    // index matched material, measurement contribution is 1.0f;
    poly.add(Term6f(1,0,0,0,0,0,0));
    return;
  }

  // init tangent space [abn][xyz][012] by normalizing away v1 and T1:
  float TI1[9];
  for(int k=0;k<3;k++) for(int j=0;j<3;j++)
    TI1[3*k+j] = T1[3*j+k];
  float vx0 = 0.0f, vy0 = 0.0f, vz0 = 0.0f;
  float vx2 = 0.0f, vy2 = 0.0f, vz2 = 0.0f;

  float ax0 = 0.0f, ay0 = 0.0f, az0 = 0.0f;
  float bx0 = 0.0f, by0 = 0.0f, bz0 = 0.0f;
  float nx0 = 0.0f, ny0 = 0.0f, nz0 = 0.0f;

  float ax2 = 0.0f, ay2 = 0.0f, az2 = 0.0f;
  float bx2 = 0.0f, by2 = 0.0f, bz2 = 0.0f;
  float nx2 = 0.0f, ny2 = 0.0f, nz2 = 0.0f;

  for(int k=0;k<3;k++)
  {
    vx0 += TI1[3*0 + k] * (v0[k] - v1[k]);
    vy0 += TI1[3*1 + k] * (v0[k] - v1[k]);
    vz0 += TI1[3*2 + k] * (v0[k] - v1[k]);
    vx2 += TI1[3*0 + k] * (v2[k] - v1[k]);
    vy2 += TI1[3*1 + k] * (v2[k] - v1[k]);
    vz2 += TI1[3*2 + k] * (v2[k] - v1[k]);

    ax0 += TI1[3*0 + k] * T0[3*k + 0];
    bx0 += TI1[3*0 + k] * T0[3*k + 1];
    nx0 += TI1[3*0 + k] * T0[3*k + 2];
    ay0 += TI1[3*1 + k] * T0[3*k + 0];
    by0 += TI1[3*1 + k] * T0[3*k + 1];
    ny0 += TI1[3*1 + k] * T0[3*k + 2];
    az0 += TI1[3*2 + k] * T0[3*k + 0];
    bz0 += TI1[3*2 + k] * T0[3*k + 1];
    nz0 += TI1[3*2 + k] * T0[3*k + 2];

    ax2 += TI1[3*0 + k] * T2[3*k + 0];
    bx2 += TI1[3*0 + k] * T2[3*k + 1];
    nx2 += TI1[3*0 + k] * T2[3*k + 2];
    ay2 += TI1[3*1 + k] * T2[3*k + 0];
    by2 += TI1[3*1 + k] * T2[3*k + 1];
    ny2 += TI1[3*1 + k] * T2[3*k + 2];
    az2 += TI1[3*2 + k] * T2[3*k + 0];
    bz2 += TI1[3*2 + k] * T2[3*k + 1];
    nz2 += TI1[3*2 + k] * T2[3*k + 2];
  }

  //=================================================== 
  // taylor series degree 1 of reflected measurement contribution f 
  //=================================================== 

#if 1 // fake beckmann, no propagation
// 1  :  
poly.add(Term6f( (((160*pow(roughness,2)*pow(vz0,5)+((240*pow(roughness,2)-12)*pow(vy0,2)+(240*pow(roughness,2)-12)*pow(vx0,2))*pow(vz0,3)+((80*pow(roughness,2)-9)*pow(vy0,4)+(160*pow(roughness,2)-18)*pow(vx0,2)*pow(vy0,2)+(80*pow(roughness,2)-9)*pow(vx0,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+(((320*pow(roughness,2)-24)*vy0*vy2+(320*pow(roughness,2)-24)*vx0*vx2)*pow(vz0,4)+(((400*pow(roughness,2)-48)*pow(vy0,3)+(400*pow(roughness,2)-48)*pow(vx0,2)*vy0)*vy2+(400*pow(roughness,2)-48)*vx0*vx2*pow(vy0,2)+(400*pow(roughness,2)-48)*pow(vx0,3)*vx2)*pow(vz0,2)+((80*pow(roughness,2)-15)*pow(vy0,5)+(160*pow(roughness,2)-30)*pow(vx0,2)*pow(vy0,3)+(80*pow(roughness,2)-15)*pow(vx0,4)*vy0)*vy2+(80*pow(roughness,2)-15)*vx0*vx2*pow(vy0,4)+(160*pow(roughness,2)-30)*pow(vx0,3)*vx2*pow(vy0,2)+(80*pow(roughness,2)-15)*pow(vx0,5)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+(((240*pow(roughness,2)-12)*pow(vy2,2)+(240*pow(roughness,2)-12)*pow(vx2,2))*pow(vz0,5)+(((640*pow(roughness,2)-78)*pow(vy0,2)+(400*pow(roughness,2)-42)*pow(vx0,2))*pow(vy2,2)+(480*pow(roughness,2)-72)*vx0*vx2*vy0*vy2+(400*pow(roughness,2)-42)*pow(vx2,2)*pow(vy0,2)+(640*pow(roughness,2)-78)*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+(((400*pow(roughness,2)-72)*pow(vy0,4)+(560*pow(roughness,2)-99)*pow(vx0,2)*pow(vy0,2)+(160*pow(roughness,2)-27)*pow(vx0,4))*pow(vy2,2)+((480*pow(roughness,2)-90)*vx0*vx2*pow(vy0,3)+(480*pow(roughness,2)-90)*pow(vx0,3)*vx2*vy0)*vy2+(160*pow(roughness,2)-27)*pow(vx2,2)*pow(vy0,4)+(560*pow(roughness,2)-99)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(400*pow(roughness,2)-72)*pow(vx0,4)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+(((400*pow(roughness,2)-48)*vy0*pow(vy2,3)+(400*pow(roughness,2)-48)*vx0*vx2*pow(vy2,2)+(400*pow(roughness,2)-48)*pow(vx2,2)*vy0*vy2+(400*pow(roughness,2)-48)*vx0*pow(vx2,3))*pow(vz0,4)+(((640*pow(roughness,2)-114)*pow(vy0,3)+(560*pow(roughness,2)-96)*pow(vx0,2)*vy0)*pow(vy2,3)+((800*pow(roughness,2)-150)*vx0*vx2*pow(vy0,2)+(560*pow(roughness,2)-96)*pow(vx0,3)*vx2)*pow(vy2,2)+((560*pow(roughness,2)-96)*pow(vx2,2)*pow(vy0,3)+(800*pow(roughness,2)-150)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+(560*pow(roughness,2)-96)*vx0*pow(vx2,3)*pow(vy0,2)+(640*pow(roughness,2)-114)*pow(vx0,3)*pow(vx2,3))*pow(vz0,2)+((240*pow(roughness,2)-60)*pow(vy0,5)+(400*pow(roughness,2)-99)*pow(vx0,2)*pow(vy0,3)+(160*pow(roughness,2)-39)*pow(vx0,4)*vy0)*pow(vy2,3)+((400*pow(roughness,2)-102)*vx0*vx2*pow(vy0,4)+(560*pow(roughness,2)-141)*pow(vx0,3)*vx2*pow(vy0,2)+(160*pow(roughness,2)-39)*pow(vx0,5)*vx2)*pow(vy2,2)+((160*pow(roughness,2)-39)*pow(vx2,2)*pow(vy0,5)+(560*pow(roughness,2)-141)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+(400*pow(roughness,2)-102)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+(160*pow(roughness,2)-39)*vx0*pow(vx2,3)*pow(vy0,4)+(400*pow(roughness,2)-99)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+(240*pow(roughness,2)-60)*pow(vx0,5)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+(((80*pow(roughness,2)-9)*pow(vy2,4)+(160*pow(roughness,2)-18)*pow(vx2,2)*pow(vy2,2)+(80*pow(roughness,2)-9)*pow(vx2,4))*pow(vz0,5)+(((400*pow(roughness,2)-72)*pow(vy0,2)+(160*pow(roughness,2)-27)*pow(vx0,2))*pow(vy2,4)+(480*pow(roughness,2)-90)*vx0*vx2*vy0*pow(vy2,3)+((560*pow(roughness,2)-99)*pow(vx2,2)*pow(vy0,2)+(560*pow(roughness,2)-99)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)-90)*vx0*pow(vx2,3)*vy0*vy2+(160*pow(roughness,2)-27)*pow(vx2,4)*pow(vy0,2)+(400*pow(roughness,2)-72)*pow(vx0,2)*pow(vx2,4))*pow(vz0,3)+(((320*pow(roughness,2)-72)*pow(vy0,4)+(400*pow(roughness,2)-90)*pow(vx0,2)*pow(vy0,2)+(80*pow(roughness,2)-18)*pow(vx0,4))*pow(vy2,4)+((480*pow(roughness,2)-108)*vx0*vx2*pow(vy0,3)+(480*pow(roughness,2)-108)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+((400*pow(roughness,2)-90)*pow(vx2,2)*pow(vy0,4)+(800*pow(roughness,2)-180)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(400*pow(roughness,2)-90)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+((480*pow(roughness,2)-108)*vx0*pow(vx2,3)*pow(vy0,3)+(480*pow(roughness,2)-108)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+(80*pow(roughness,2)-18)*pow(vx2,4)*pow(vy0,4)+(400*pow(roughness,2)-90)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(320*pow(roughness,2)-72)*pow(vx0,4)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+(((80*pow(roughness,2)-15)*vy0*pow(vy2,5)+(80*pow(roughness,2)-15)*vx0*vx2*pow(vy2,4)+(160*pow(roughness,2)-30)*pow(vx2,2)*vy0*pow(vy2,3)+(160*pow(roughness,2)-30)*vx0*pow(vx2,3)*pow(vy2,2)+(80*pow(roughness,2)-15)*pow(vx2,4)*vy0*vy2+(80*pow(roughness,2)-15)*vx0*pow(vx2,5))*pow(vz0,4)+(((240*pow(roughness,2)-60)*pow(vy0,3)+(160*pow(roughness,2)-39)*pow(vx0,2)*vy0)*pow(vy2,5)+((400*pow(roughness,2)-102)*vx0*vx2*pow(vy0,2)+(160*pow(roughness,2)-39)*pow(vx0,3)*vx2)*pow(vy2,4)+((400*pow(roughness,2)-99)*pow(vx2,2)*pow(vy0,3)+(560*pow(roughness,2)-141)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+((560*pow(roughness,2)-141)*vx0*pow(vx2,3)*pow(vy0,2)+(400*pow(roughness,2)-99)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+((160*pow(roughness,2)-39)*pow(vx2,4)*pow(vy0,3)+(400*pow(roughness,2)-102)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+(160*pow(roughness,2)-39)*vx0*pow(vx2,5)*pow(vy0,2)+(240*pow(roughness,2)-60)*pow(vx0,3)*pow(vx2,5))*pow(vz0,2)+((160*pow(roughness,2)-48)*pow(vy0,5)+(240*pow(roughness,2)-72)*pow(vx0,2)*pow(vy0,3)+(80*pow(roughness,2)-24)*pow(vx0,4)*vy0)*pow(vy2,5)+((320*pow(roughness,2)-96)*vx0*vx2*pow(vy0,4)+(400*pow(roughness,2)-120)*pow(vx0,3)*vx2*pow(vy0,2)+(80*pow(roughness,2)-24)*pow(vx0,5)*vx2)*pow(vy2,4)+((240*pow(roughness,2)-72)*pow(vx2,2)*pow(vy0,5)+(640*pow(roughness,2)-192)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+(400*pow(roughness,2)-120)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+((400*pow(roughness,2)-120)*vx0*pow(vx2,3)*pow(vy0,4)+(640*pow(roughness,2)-192)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+(240*pow(roughness,2)-72)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+((80*pow(roughness,2)-24)*pow(vx2,4)*pow(vy0,5)+(400*pow(roughness,2)-120)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+(320*pow(roughness,2)-96)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+(80*pow(roughness,2)-24)*vx0*pow(vx2,5)*pow(vy0,4)+(240*pow(roughness,2)-72)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+(160*pow(roughness,2)-48)*pow(vx0,5)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,6)+((320*pow(roughness,2)-12)*pow(vy0,2)+(320*pow(roughness,2)-12)*pow(vx0,2))*pow(vz0,4)+((180*pow(roughness,2)-15)*pow(vy0,4)+(360*pow(roughness,2)-30)*pow(vx0,2)*pow(vy0,2)+(180*pow(roughness,2)-15)*pow(vx0,4))*pow(vz0,2)+(20*pow(roughness,2)-3)*pow(vy0,6)+(60*pow(roughness,2)-9)*pow(vx0,2)*pow(vy0,4)+(60*pow(roughness,2)-9)*pow(vx0,4)*pow(vy0,2)+(20*pow(roughness,2)-3)*pow(vx0,6))*pow(vz2,6)+(((320*pow(roughness,2)-24)*vy0*vy2+(320*pow(roughness,2)-24)*vx0*vx2)*pow(vz0,5)+(((560*pow(roughness,2)-60)*pow(vy0,3)+(560*pow(roughness,2)-60)*pow(vx0,2)*vy0)*vy2+(560*pow(roughness,2)-60)*vx0*vx2*pow(vy0,2)+(560*pow(roughness,2)-60)*pow(vx0,3)*vx2)*pow(vz0,3)+(((240*pow(roughness,2)-36)*pow(vy0,5)+(480*pow(roughness,2)-72)*pow(vx0,2)*pow(vy0,3)+(240*pow(roughness,2)-36)*pow(vx0,4)*vy0)*vy2+(240*pow(roughness,2)-36)*vx0*vx2*pow(vy0,4)+(480*pow(roughness,2)-72)*pow(vx0,3)*vx2*pow(vy0,2)+(240*pow(roughness,2)-36)*pow(vx0,5)*vx2)*vz0)*pow(vz2,5)+(((320*pow(roughness,2)-12)*pow(vy2,2)+(320*pow(roughness,2)-12)*pow(vx2,2))*pow(vz0,6)+(((920*pow(roughness,2)-90)*pow(vy0,2)+(680*pow(roughness,2)-54)*pow(vx0,2))*pow(vy2,2)+(480*pow(roughness,2)-72)*vx0*vx2*vy0*vy2+(680*pow(roughness,2)-54)*pow(vx2,2)*pow(vy0,2)+(920*pow(roughness,2)-90)*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+(((780*pow(roughness,2)-117)*pow(vy0,4)+(1200*pow(roughness,2)-171)*pow(vx0,2)*pow(vy0,2)+(420*pow(roughness,2)-54)*pow(vx0,4))*pow(vy2,2)+((720*pow(roughness,2)-126)*vx0*vx2*pow(vy0,3)+(720*pow(roughness,2)-126)*pow(vx0,3)*vx2*vy0)*vy2+(420*pow(roughness,2)-54)*pow(vx2,2)*pow(vy0,4)+(1200*pow(roughness,2)-171)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(780*pow(roughness,2)-117)*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((180*pow(roughness,2)-39)*pow(vy0,6)+(420*pow(roughness,2)-90)*pow(vx0,2)*pow(vy0,4)+(300*pow(roughness,2)-63)*pow(vx0,4)*pow(vy0,2)+(60*pow(roughness,2)-12)*pow(vx0,6))*pow(vy2,2)+((240*pow(roughness,2)-54)*vx0*vx2*pow(vy0,5)+(480*pow(roughness,2)-108)*pow(vx0,3)*vx2*pow(vy0,3)+(240*pow(roughness,2)-54)*pow(vx0,5)*vx2*vy0)*vy2+(60*pow(roughness,2)-12)*pow(vx2,2)*pow(vy0,6)+(300*pow(roughness,2)-63)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+(420*pow(roughness,2)-90)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+(180*pow(roughness,2)-39)*pow(vx0,6)*pow(vx2,2))*pow(vz2,4)+(((560*pow(roughness,2)-60)*vy0*pow(vy2,3)+(560*pow(roughness,2)-60)*vx0*vx2*pow(vy2,2)+(560*pow(roughness,2)-60)*pow(vx2,2)*vy0*vy2+(560*pow(roughness,2)-60)*vx0*pow(vx2,3))*pow(vz0,5)+(((1120*pow(roughness,2)-168)*pow(vy0,3)+(1040*pow(roughness,2)-150)*pow(vx0,2)*vy0)*pow(vy2,3)+((1280*pow(roughness,2)-204)*vx0*vx2*pow(vy0,2)+(1040*pow(roughness,2)-150)*pow(vx0,3)*vx2)*pow(vy2,2)+((1040*pow(roughness,2)-150)*pow(vx2,2)*pow(vy0,3)+(1280*pow(roughness,2)-204)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+(1040*pow(roughness,2)-150)*vx0*pow(vx2,3)*pow(vy0,2)+(1120*pow(roughness,2)-168)*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+(((560*pow(roughness,2)-108)*pow(vy0,5)+(1040*pow(roughness,2)-198)*pow(vx0,2)*pow(vy0,3)+(480*pow(roughness,2)-90)*pow(vx0,4)*vy0)*pow(vy2,3)+((720*pow(roughness,2)-144)*vx0*vx2*pow(vy0,4)+(1200*pow(roughness,2)-234)*pow(vx0,3)*vx2*pow(vy0,2)+(480*pow(roughness,2)-90)*pow(vx0,5)*vx2)*pow(vy2,2)+((480*pow(roughness,2)-90)*pow(vx2,2)*pow(vy0,5)+(1200*pow(roughness,2)-234)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+(720*pow(roughness,2)-144)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+(480*pow(roughness,2)-90)*vx0*pow(vx2,3)*pow(vy0,4)+(1040*pow(roughness,2)-198)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+(560*pow(roughness,2)-108)*pow(vx0,5)*pow(vx2,3))*vz0)*pow(vz2,3)+(((180*pow(roughness,2)-15)*pow(vy2,4)+(360*pow(roughness,2)-30)*pow(vx2,2)*pow(vy2,2)+(180*pow(roughness,2)-15)*pow(vx2,4))*pow(vz0,6)+(((780*pow(roughness,2)-117)*pow(vy0,2)+(420*pow(roughness,2)-54)*pow(vx0,2))*pow(vy2,4)+(720*pow(roughness,2)-126)*vx0*vx2*vy0*pow(vy2,3)+((1200*pow(roughness,2)-171)*pow(vx2,2)*pow(vy0,2)+(1200*pow(roughness,2)-171)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(720*pow(roughness,2)-126)*vx0*pow(vx2,3)*vy0*vy2+(420*pow(roughness,2)-54)*pow(vx2,4)*pow(vy0,2)+(780*pow(roughness,2)-117)*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+(((920*pow(roughness,2)-186)*pow(vy0,4)+(1200*pow(roughness,2)-234)*pow(vx0,2)*pow(vy0,2)+(300*pow(roughness,2)-54)*pow(vx0,4))*pow(vy2,4)+((1280*pow(roughness,2)-276)*vx0*vx2*pow(vy0,3)+(1200*pow(roughness,2)-252)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+((1200*pow(roughness,2)-234)*pow(vx2,2)*pow(vy0,4)+(2520*pow(roughness,2)-504)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(1200*pow(roughness,2)-234)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+((1200*pow(roughness,2)-252)*vx0*pow(vx2,3)*pow(vy0,3)+(1280*pow(roughness,2)-276)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+(300*pow(roughness,2)-54)*pow(vx2,4)*pow(vy0,4)+(1200*pow(roughness,2)-234)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(920*pow(roughness,2)-186)*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+((320*pow(roughness,2)-84)*pow(vy0,6)+(680*pow(roughness,2)-177)*pow(vx0,2)*pow(vy0,4)+(420*pow(roughness,2)-108)*pow(vx0,4)*pow(vy0,2)+(60*pow(roughness,2)-15)*pow(vx0,6))*pow(vy2,4)+((560*pow(roughness,2)-150)*vx0*vx2*pow(vy0,5)+(1040*pow(roughness,2)-276)*pow(vx0,3)*vx2*pow(vy0,3)+(480*pow(roughness,2)-126)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+((360*pow(roughness,2)-93)*pow(vx2,2)*pow(vy0,6)+(1200*pow(roughness,2)-315)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+(1200*pow(roughness,2)-315)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+(360*pow(roughness,2)-93)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+((480*pow(roughness,2)-126)*vx0*pow(vx2,3)*pow(vy0,5)+(1040*pow(roughness,2)-276)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+(560*pow(roughness,2)-150)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+(60*pow(roughness,2)-15)*pow(vx2,4)*pow(vy0,6)+(420*pow(roughness,2)-108)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+(680*pow(roughness,2)-177)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+(320*pow(roughness,2)-84)*pow(vx0,6)*pow(vx2,4))*pow(vz2,2)+(((240*pow(roughness,2)-36)*vy0*pow(vy2,5)+(240*pow(roughness,2)-36)*vx0*vx2*pow(vy2,4)+(480*pow(roughness,2)-72)*pow(vx2,2)*vy0*pow(vy2,3)+(480*pow(roughness,2)-72)*vx0*pow(vx2,3)*pow(vy2,2)+(240*pow(roughness,2)-36)*pow(vx2,4)*vy0*vy2+(240*pow(roughness,2)-36)*vx0*pow(vx2,5))*pow(vz0,5)+(((560*pow(roughness,2)-108)*pow(vy0,3)+(480*pow(roughness,2)-90)*pow(vx0,2)*vy0)*pow(vy2,5)+((720*pow(roughness,2)-144)*vx0*vx2*pow(vy0,2)+(480*pow(roughness,2)-90)*pow(vx0,3)*vx2)*pow(vy2,4)+((1040*pow(roughness,2)-198)*pow(vx2,2)*pow(vy0,3)+(1200*pow(roughness,2)-234)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+((1200*pow(roughness,2)-234)*vx0*pow(vx2,3)*pow(vy0,2)+(1040*pow(roughness,2)-198)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+((480*pow(roughness,2)-90)*pow(vx2,4)*pow(vy0,3)+(720*pow(roughness,2)-144)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+(480*pow(roughness,2)-90)*vx0*pow(vx2,5)*pow(vy0,2)+(560*pow(roughness,2)-108)*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+(((320*pow(roughness,2)-72)*pow(vy0,5)+(560*pow(roughness,2)-126)*pow(vx0,2)*pow(vy0,3)+(240*pow(roughness,2)-54)*pow(vx0,4)*vy0)*pow(vy2,5)+((480*pow(roughness,2)-108)*vx0*vx2*pow(vy0,4)+(720*pow(roughness,2)-162)*pow(vx0,3)*vx2*pow(vy0,2)+(240*pow(roughness,2)-54)*pow(vx0,5)*vx2)*pow(vy2,4)+((560*pow(roughness,2)-126)*pow(vx2,2)*pow(vy0,5)+(1280*pow(roughness,2)-288)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+(720*pow(roughness,2)-162)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+((720*pow(roughness,2)-162)*vx0*pow(vx2,3)*pow(vy0,4)+(1280*pow(roughness,2)-288)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+(560*pow(roughness,2)-126)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+((240*pow(roughness,2)-54)*pow(vx2,4)*pow(vy0,5)+(720*pow(roughness,2)-162)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+(480*pow(roughness,2)-108)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+(240*pow(roughness,2)-54)*vx0*pow(vx2,5)*pow(vy0,4)+(560*pow(roughness,2)-126)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+(320*pow(roughness,2)-72)*pow(vx0,5)*pow(vx2,5))*vz0)*vz2+((20*pow(roughness,2)-3)*pow(vy2,6)+(60*pow(roughness,2)-9)*pow(vx2,2)*pow(vy2,4)+(60*pow(roughness,2)-9)*pow(vx2,4)*pow(vy2,2)+(20*pow(roughness,2)-3)*pow(vx2,6))*pow(vz0,6)+(((180*pow(roughness,2)-39)*pow(vy0,2)+(60*pow(roughness,2)-12)*pow(vx0,2))*pow(vy2,6)+(240*pow(roughness,2)-54)*vx0*vx2*vy0*pow(vy2,5)+((420*pow(roughness,2)-90)*pow(vx2,2)*pow(vy0,2)+(300*pow(roughness,2)-63)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+(480*pow(roughness,2)-108)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+((300*pow(roughness,2)-63)*pow(vx2,4)*pow(vy0,2)+(420*pow(roughness,2)-90)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+(240*pow(roughness,2)-54)*vx0*pow(vx2,5)*vy0*vy2+(60*pow(roughness,2)-12)*pow(vx2,6)*pow(vy0,2)+(180*pow(roughness,2)-39)*pow(vx0,2)*pow(vx2,6))*pow(vz0,4)+(((320*pow(roughness,2)-84)*pow(vy0,4)+(360*pow(roughness,2)-93)*pow(vx0,2)*pow(vy0,2)+(60*pow(roughness,2)-15)*pow(vx0,4))*pow(vy2,6)+((560*pow(roughness,2)-150)*vx0*vx2*pow(vy0,3)+(480*pow(roughness,2)-126)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+((680*pow(roughness,2)-177)*pow(vx2,2)*pow(vy0,4)+(1200*pow(roughness,2)-315)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(420*pow(roughness,2)-108)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+((1040*pow(roughness,2)-276)*vx0*pow(vx2,3)*pow(vy0,3)+(1040*pow(roughness,2)-276)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+((420*pow(roughness,2)-108)*pow(vx2,4)*pow(vy0,4)+(1200*pow(roughness,2)-315)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(680*pow(roughness,2)-177)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+((480*pow(roughness,2)-126)*vx0*pow(vx2,5)*pow(vy0,3)+(560*pow(roughness,2)-150)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+(60*pow(roughness,2)-15)*pow(vx2,6)*pow(vy0,4)+(360*pow(roughness,2)-93)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+(320*pow(roughness,2)-84)*pow(vx0,4)*pow(vx2,6))*pow(vz0,2)+((160*pow(roughness,2)-48)*pow(vy0,6)+(320*pow(roughness,2)-96)*pow(vx0,2)*pow(vy0,4)+(180*pow(roughness,2)-54)*pow(vx0,4)*pow(vy0,2)+(20*pow(roughness,2)-6)*pow(vx0,6))*pow(vy2,6)+((320*pow(roughness,2)-96)*vx0*vx2*pow(vy0,5)+(560*pow(roughness,2)-168)*pow(vx0,3)*vx2*pow(vy0,3)+(240*pow(roughness,2)-72)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+((320*pow(roughness,2)-96)*pow(vx2,2)*pow(vy0,6)+(920*pow(roughness,2)-276)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+(780*pow(roughness,2)-234)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+(180*pow(roughness,2)-54)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+((560*pow(roughness,2)-168)*vx0*pow(vx2,3)*pow(vy0,5)+(1120*pow(roughness,2)-336)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+(560*pow(roughness,2)-168)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+((180*pow(roughness,2)-54)*pow(vx2,4)*pow(vy0,6)+(780*pow(roughness,2)-234)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+(920*pow(roughness,2)-276)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+(320*pow(roughness,2)-96)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+((240*pow(roughness,2)-72)*vx0*pow(vx2,5)*pow(vy0,5)+(560*pow(roughness,2)-168)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+(320*pow(roughness,2)-96)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+(20*pow(roughness,2)-6)*pow(vx2,6)*pow(vy0,6)+(180*pow(roughness,2)-54)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+(320*pow(roughness,2)-96)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+(160*pow(roughness,2)-48)*pow(vx0,6)*pow(vx2,6))*pow(((160*pow(roughness,2)*pow(vz0,5)+(240*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,4)+160*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+400*pow(roughness,2)*vx0*vx2*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,5)+160*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,4)+160*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,5)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((240*pow(roughness,2)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,2))*pow(vz0,5)+((640*pow(roughness,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+400*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+((400*pow(roughness,2)*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+160*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((400*pow(roughness,2)*vy0*pow(vy2,3)+400*pow(roughness,2)*vx0*vx2*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2)*vy0*vy2+400*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,4)+((640*pow(roughness,2)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+560*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,2)+(240*pow(roughness,2)*pow(vy0,5)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+160*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*vx2*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(160*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4))*pow(vz0,5)+((400*pow(roughness,2)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+160*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,5)+80*pow(roughness,2)*vx0*vx2*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,4)+((240*pow(roughness,2)*pow(vy0,3)+160*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(400*pow(roughness,2)*vx0*vx2*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(560*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(160*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+160*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,6)+(320*pow(roughness,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,2))*pow(vz0,4)+(180*pow(roughness,2)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,4))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,6)+60*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+60*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,6))*pow(vz2,6)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,5)+((560*pow(roughness,2)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+560*pow(roughness,2)*vx0*vx2*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,5)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*vz0)*pow(vz2,5)+((320*pow(roughness,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,2))*pow(vz0,6)+((920*pow(roughness,2)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+680*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((780*pow(roughness,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+420*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+420*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+780*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+(180*pow(roughness,2)*pow(vy0,6)+420*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+300*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,5)+480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+60*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+420*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz2,4)+((560*pow(roughness,2)*vy0*pow(vy2,3)+560*pow(roughness,2)*vx0*vx2*pow(vy2,2)+560*pow(roughness,2)*pow(vx2,2)*vy0*vy2+560*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,5)+((1120*pow(roughness,2)*pow(vy0,3)+1040*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1280*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1040*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1120*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+((560*pow(roughness,2)*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*vz0)*pow(vz2,3)+((180*pow(roughness,2)*pow(vy2,4)+360*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+180*pow(roughness,2)*pow(vx2,4))*pow(vz0,6)+((780*pow(roughness,2)*pow(vy0,2)+420*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+720*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+420*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+780*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+((920*pow(roughness,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+300*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(1280*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1280*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+300*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,6)+680*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+420*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(560*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+420*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*vx2*pow(vy2,4)+480*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,4)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,5)+((560*pow(roughness,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(480*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1280*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,6)+60*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+60*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,6))*pow(vz0,6)+((180*pow(roughness,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(420*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(300*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+420*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+60*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,4)+((320*pow(roughness,2)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(560*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(680*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+420*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(1040*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(420*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+180*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(320*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+780*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(560*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1120*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(180*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+780*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,6)*pow(vx2,6),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// x0  :  
poly.add(Term6f( -((((24*bx0*vy0+24*ax0*vx0)*pow(vz0,5)+(-24*nx0*pow(vy0,2)-24*nx0*pow(vx0,2))*pow(vz0,4)+(24*bx0*pow(vy0,3)+24*ax0*vx0*pow(vy0,2)+24*bx0*pow(vx0,2)*vy0+24*ax0*pow(vx0,3))*pow(vz0,3)+(-24*nx0*pow(vy0,4)-48*nx0*pow(vx0,2)*pow(vy0,2)-24*nx0*pow(vx0,4))*pow(vz0,2)+(3*bx0*pow(vy0,5)+3*ax0*vx0*pow(vy0,4)+6*bx0*pow(vx0,2)*pow(vy0,3)+6*ax0*pow(vx0,3)*pow(vy0,2)+3*bx0*pow(vx0,4)*vy0+3*ax0*pow(vx0,5))*vz0-3*nx0*pow(vy0,6)-9*nx0*pow(vx0,2)*pow(vy0,4)-9*nx0*pow(vx0,4)*pow(vy0,2)-3*nx0*pow(vx0,6))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((24*bx0*vy2+24*ax0*vx2)*pow(vz0,6)+(-24*nx0*vy0*vy2-24*nx0*vx0*vx2)*pow(vz0,5)+((72*bx0*pow(vy0,2)+24*ax0*vx0*vy0+48*bx0*pow(vx0,2))*vy2+48*ax0*vx2*pow(vy0,2)+24*bx0*vx0*vx2*vy0+72*ax0*pow(vx0,2)*vx2)*pow(vz0,4)+((-72*nx0*pow(vy0,3)-72*nx0*pow(vx0,2)*vy0)*vy2-72*nx0*vx0*vx2*pow(vy0,2)-72*nx0*pow(vx0,3)*vx2)*pow(vz0,3)+((39*bx0*pow(vy0,4)+12*ax0*vx0*pow(vy0,3)+66*bx0*pow(vx0,2)*pow(vy0,2)+12*ax0*pow(vx0,3)*vy0+27*bx0*pow(vx0,4))*vy2+27*ax0*vx2*pow(vy0,4)+12*bx0*vx0*vx2*pow(vy0,3)+66*ax0*pow(vx0,2)*vx2*pow(vy0,2)+12*bx0*pow(vx0,3)*vx2*vy0+39*ax0*pow(vx0,4)*vx2)*pow(vz0,2)+((-39*nx0*pow(vy0,5)-78*nx0*pow(vx0,2)*pow(vy0,3)-39*nx0*pow(vx0,4)*vy0)*vy2-39*nx0*vx0*vx2*pow(vy0,4)-78*nx0*pow(vx0,3)*vx2*pow(vy0,2)-39*nx0*pow(vx0,5)*vx2)*vz0+(-3*ax0*vx0*pow(vy0,5)+3*bx0*pow(vx0,2)*pow(vy0,4)-6*ax0*pow(vx0,3)*pow(vy0,3)+6*bx0*pow(vx0,4)*pow(vy0,2)-3*ax0*pow(vx0,5)*vy0+3*bx0*pow(vx0,6))*vy2+3*ax0*vx2*pow(vy0,6)-3*bx0*vx0*vx2*pow(vy0,5)+6*ax0*pow(vx0,2)*vx2*pow(vy0,4)-6*bx0*pow(vx0,3)*vx2*pow(vy0,3)+3*ax0*pow(vx0,4)*vx2*pow(vy0,2)-3*bx0*pow(vx0,5)*vx2*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+(((72*bx0*vy0+48*ax0*vx0)*pow(vy2,2)+(24*ax0*vx2*vy0+24*bx0*vx0*vx2)*vy2+48*bx0*pow(vx2,2)*vy0+72*ax0*vx0*pow(vx2,2))*pow(vz0,5)+((-72*nx0*pow(vy0,2)-48*nx0*pow(vx0,2))*pow(vy2,2)-48*nx0*vx0*vx2*vy0*vy2-48*nx0*pow(vx2,2)*pow(vy0,2)-72*nx0*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((102*bx0*pow(vy0,3)+60*ax0*vx0*pow(vy0,2)+96*bx0*pow(vx0,2)*vy0+54*ax0*pow(vx0,3))*pow(vy2,2)+(42*ax0*vx2*pow(vy0,3)+54*bx0*vx0*vx2*pow(vy0,2)+54*ax0*pow(vx0,2)*vx2*vy0+42*bx0*pow(vx0,3)*vx2)*vy2+54*bx0*pow(vx2,2)*pow(vy0,3)+96*ax0*vx0*pow(vx2,2)*pow(vy0,2)+60*bx0*pow(vx0,2)*pow(vx2,2)*vy0+102*ax0*pow(vx0,3)*pow(vx2,2))*pow(vz0,3)+((-102*nx0*pow(vy0,4)-156*nx0*pow(vx0,2)*pow(vy0,2)-54*nx0*pow(vx0,4))*pow(vy2,2)+(-96*nx0*vx0*vx2*pow(vy0,3)-96*nx0*pow(vx0,3)*vx2*vy0)*vy2-54*nx0*pow(vx2,2)*pow(vy0,4)-156*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-102*nx0*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((24*bx0*pow(vy0,5)+6*ax0*vx0*pow(vy0,4)+51*bx0*pow(vx0,2)*pow(vy0,3)+15*ax0*pow(vx0,3)*pow(vy0,2)+27*bx0*pow(vx0,4)*vy0+9*ax0*pow(vx0,5))*pow(vy2,2)+(18*ax0*vx2*pow(vy0,5)+12*bx0*vx0*vx2*pow(vy0,4)+30*ax0*pow(vx0,2)*vx2*pow(vy0,3)+30*bx0*pow(vx0,3)*vx2*pow(vy0,2)+12*ax0*pow(vx0,4)*vx2*vy0+18*bx0*pow(vx0,5)*vx2)*vy2+9*bx0*pow(vx2,2)*pow(vy0,5)+27*ax0*vx0*pow(vx2,2)*pow(vy0,4)+15*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+51*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+6*bx0*pow(vx0,4)*pow(vx2,2)*vy0+24*ax0*pow(vx0,5)*pow(vx2,2))*vz0+(-24*nx0*pow(vy0,6)-57*nx0*pow(vx0,2)*pow(vy0,4)-42*nx0*pow(vx0,4)*pow(vy0,2)-9*nx0*pow(vx0,6))*pow(vy2,2)+(-30*nx0*vx0*vx2*pow(vy0,5)-60*nx0*pow(vx0,3)*vx2*pow(vy0,3)-30*nx0*pow(vx0,5)*vx2*vy0)*vy2-9*nx0*pow(vx2,2)*pow(vy0,6)-42*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-57*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-24*nx0*pow(vx0,6)*pow(vx2,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((24*bx0*pow(vy2,3)+24*ax0*vx2*pow(vy2,2)+24*bx0*pow(vx2,2)*vy2+24*ax0*pow(vx2,3))*pow(vz0,6)+(-24*nx0*vy0*pow(vy2,3)-24*nx0*vx0*vx2*pow(vy2,2)-24*nx0*pow(vx2,2)*vy0*vy2-24*nx0*vx0*pow(vx2,3))*pow(vz0,5)+((102*bx0*pow(vy0,2)+48*ax0*vx0*vy0+48*bx0*pow(vx0,2))*pow(vy2,3)+(54*ax0*vx2*pow(vy0,2)+60*bx0*vx0*vx2*vy0+96*ax0*pow(vx0,2)*vx2)*pow(vy2,2)+(96*bx0*pow(vx2,2)*pow(vy0,2)+60*ax0*vx0*pow(vx2,2)*vy0+54*bx0*pow(vx0,2)*pow(vx2,2))*vy2+48*ax0*pow(vx2,3)*pow(vy0,2)+48*bx0*vx0*pow(vx2,3)*vy0+102*ax0*pow(vx0,2)*pow(vx2,3))*pow(vz0,4)+((-102*nx0*pow(vy0,3)-96*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+(-114*nx0*vx0*vx2*pow(vy0,2)-96*nx0*pow(vx0,3)*vx2)*pow(vy2,2)+(-96*nx0*pow(vx2,2)*pow(vy0,3)-114*nx0*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-96*nx0*vx0*pow(vx2,3)*pow(vy0,2)-102*nx0*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+((72*bx0*pow(vy0,4)+36*ax0*vx0*pow(vy0,3)+99*bx0*pow(vx0,2)*pow(vy0,2)+36*ax0*pow(vx0,3)*vy0+27*bx0*pow(vx0,4))*pow(vy2,3)+(36*ax0*vx2*pow(vy0,4)+54*bx0*vx0*vx2*pow(vy0,3)+99*ax0*pow(vx0,2)*vx2*pow(vy0,2)+54*bx0*pow(vx0,3)*vx2*vy0+63*ax0*pow(vx0,4)*vx2)*pow(vy2,2)+(63*bx0*pow(vx2,2)*pow(vy0,4)+54*ax0*vx0*pow(vx2,2)*pow(vy0,3)+99*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+54*ax0*pow(vx0,3)*pow(vx2,2)*vy0+36*bx0*pow(vx0,4)*pow(vx2,2))*vy2+27*ax0*pow(vx2,3)*pow(vy0,4)+36*bx0*vx0*pow(vx2,3)*pow(vy0,3)+99*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+36*bx0*pow(vx0,3)*pow(vx2,3)*vy0+72*ax0*pow(vx0,4)*pow(vx2,3))*pow(vz0,2)+((-72*nx0*pow(vy0,5)-135*nx0*pow(vx0,2)*pow(vy0,3)-63*nx0*pow(vx0,4)*vy0)*pow(vy2,3)+(-90*nx0*vx0*vx2*pow(vy0,4)-153*nx0*pow(vx0,3)*vx2*pow(vy0,2)-63*nx0*pow(vx0,5)*vx2)*pow(vy2,2)+(-63*nx0*pow(vx2,2)*pow(vy0,5)-153*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-90*nx0*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-63*nx0*vx0*pow(vx2,3)*pow(vy0,4)-135*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-72*nx0*pow(vx0,5)*pow(vx2,3))*vz0+(-6*ax0*vx0*pow(vy0,5)+6*bx0*pow(vx0,2)*pow(vy0,4)-9*ax0*pow(vx0,3)*pow(vy0,3)+9*bx0*pow(vx0,4)*pow(vy0,2)-3*ax0*pow(vx0,5)*vy0+3*bx0*pow(vx0,6))*pow(vy2,3)+(6*ax0*vx2*pow(vy0,6)-6*bx0*vx0*vx2*pow(vy0,5)+3*ax0*pow(vx0,2)*vx2*pow(vy0,4)-3*bx0*pow(vx0,3)*vx2*pow(vy0,3)-3*ax0*pow(vx0,4)*vx2*pow(vy0,2)+3*bx0*pow(vx0,5)*vx2*vy0)*pow(vy2,2)+(3*ax0*vx0*pow(vx2,2)*pow(vy0,5)-3*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-3*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+3*bx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-6*ax0*pow(vx0,5)*pow(vx2,2)*vy0+6*bx0*pow(vx0,6)*pow(vx2,2))*vy2+3*ax0*pow(vx2,3)*pow(vy0,6)-3*bx0*vx0*pow(vx2,3)*pow(vy0,5)+9*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)-9*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+6*ax0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)-6*bx0*pow(vx0,5)*pow(vx2,3)*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+(((39*bx0*vy0+21*ax0*vx0)*pow(vy2,4)+(18*ax0*vx2*vy0+18*bx0*vx0*vx2)*pow(vy2,3)+(60*bx0*pow(vx2,2)*vy0+60*ax0*vx0*pow(vx2,2))*pow(vy2,2)+(18*ax0*pow(vx2,3)*vy0+18*bx0*vx0*pow(vx2,3))*vy2+21*bx0*pow(vx2,4)*vy0+39*ax0*vx0*pow(vx2,4))*pow(vz0,5)+((-39*nx0*pow(vy0,2)-21*nx0*pow(vx0,2))*pow(vy2,4)-36*nx0*vx0*vx2*vy0*pow(vy2,3)+(-60*nx0*pow(vx2,2)*pow(vy0,2)-60*nx0*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-36*nx0*vx0*pow(vx2,3)*vy0*vy2-21*nx0*pow(vx2,4)*pow(vy0,2)-39*nx0*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+((72*bx0*pow(vy0,3)+42*ax0*vx0*pow(vy0,2)+57*bx0*pow(vx0,2)*vy0+27*ax0*pow(vx0,3))*pow(vy2,4)+(30*ax0*vx2*pow(vy0,3)+60*bx0*vx0*vx2*pow(vy0,2)+60*ax0*pow(vx0,2)*vx2*vy0+30*bx0*pow(vx0,3)*vx2)*pow(vy2,3)+(99*bx0*pow(vx2,2)*pow(vy0,3)+99*ax0*vx0*pow(vx2,2)*pow(vy0,2)+99*bx0*pow(vx0,2)*pow(vx2,2)*vy0+99*ax0*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(30*ax0*pow(vx2,3)*pow(vy0,3)+60*bx0*vx0*pow(vx2,3)*pow(vy0,2)+60*ax0*pow(vx0,2)*pow(vx2,3)*vy0+30*bx0*pow(vx0,3)*pow(vx2,3))*vy2+27*bx0*pow(vx2,4)*pow(vy0,3)+57*ax0*vx0*pow(vx2,4)*pow(vy0,2)+42*bx0*pow(vx0,2)*pow(vx2,4)*vy0+72*ax0*pow(vx0,3)*pow(vx2,4))*pow(vz0,3)+((-72*nx0*pow(vy0,4)-99*nx0*pow(vx0,2)*pow(vy0,2)-27*nx0*pow(vx0,4))*pow(vy2,4)+(-90*nx0*vx0*vx2*pow(vy0,3)-90*nx0*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-99*nx0*pow(vx2,2)*pow(vy0,4)-198*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-99*nx0*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-90*nx0*vx0*pow(vx2,3)*pow(vy0,3)-90*nx0*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-27*nx0*pow(vx2,4)*pow(vy0,4)-99*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-72*nx0*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+((24*bx0*pow(vy0,5)+12*ax0*vx0*pow(vy0,4)+42*bx0*pow(vx0,2)*pow(vy0,3)+18*ax0*pow(vx0,3)*pow(vy0,2)+18*bx0*pow(vx0,4)*vy0+6*ax0*pow(vx0,5))*pow(vy2,4)+(12*ax0*vx2*pow(vy0,5)+24*bx0*vx0*vx2*pow(vy0,4)+36*ax0*pow(vx0,2)*vx2*pow(vy0,3)+36*bx0*pow(vx0,3)*vx2*pow(vy0,2)+24*ax0*pow(vx0,4)*vx2*vy0+12*bx0*pow(vx0,5)*vx2)*pow(vy2,3)+(30*bx0*pow(vx2,2)*pow(vy0,5)+30*ax0*vx0*pow(vx2,2)*pow(vy0,4)+60*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+60*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+30*bx0*pow(vx0,4)*pow(vx2,2)*vy0+30*ax0*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(12*ax0*pow(vx2,3)*pow(vy0,5)+24*bx0*vx0*pow(vx2,3)*pow(vy0,4)+36*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+36*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+24*ax0*pow(vx0,4)*pow(vx2,3)*vy0+12*bx0*pow(vx0,5)*pow(vx2,3))*vy2+6*bx0*pow(vx2,4)*pow(vy0,5)+18*ax0*vx0*pow(vx2,4)*pow(vy0,4)+18*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+42*ax0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+12*bx0*pow(vx0,4)*pow(vx2,4)*vy0+24*ax0*pow(vx0,5)*pow(vx2,4))*vz0+(-24*nx0*pow(vy0,6)-54*nx0*pow(vx0,2)*pow(vy0,4)-36*nx0*pow(vx0,4)*pow(vy0,2)-6*nx0*pow(vx0,6))*pow(vy2,4)+(-36*nx0*vx0*vx2*pow(vy0,5)-72*nx0*pow(vx0,3)*vx2*pow(vy0,3)-36*nx0*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(-30*nx0*pow(vx2,2)*pow(vy0,6)-90*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-90*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-30*nx0*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(-36*nx0*vx0*pow(vx2,3)*pow(vy0,5)-72*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-36*nx0*pow(vx0,5)*pow(vx2,3)*vy0)*vy2-6*nx0*pow(vx2,4)*pow(vy0,6)-36*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-54*nx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-24*nx0*pow(vx0,6)*pow(vx2,4))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((3*bx0*pow(vy2,5)+3*ax0*vx2*pow(vy2,4)+6*bx0*pow(vx2,2)*pow(vy2,3)+6*ax0*pow(vx2,3)*pow(vy2,2)+3*bx0*pow(vx2,4)*vy2+3*ax0*pow(vx2,5))*pow(vz0,6)+(-3*nx0*vy0*pow(vy2,5)-3*nx0*vx0*vx2*pow(vy2,4)-6*nx0*pow(vx2,2)*vy0*pow(vy2,3)-6*nx0*vx0*pow(vx2,3)*pow(vy2,2)-3*nx0*pow(vx2,4)*vy0*vy2-3*nx0*vx0*pow(vx2,5))*pow(vz0,5)+((24*bx0*pow(vy0,2)+15*ax0*vx0*vy0+6*bx0*pow(vx0,2))*pow(vy2,5)+(9*ax0*vx2*pow(vy0,2)+21*bx0*vx0*vx2*vy0+21*ax0*pow(vx0,2)*vx2)*pow(vy2,4)+(45*bx0*pow(vx2,2)*pow(vy0,2)+36*ax0*vx0*pow(vx2,2)*vy0+15*bx0*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(15*ax0*pow(vx2,3)*pow(vy0,2)+36*bx0*vx0*pow(vx2,3)*vy0+45*ax0*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(21*bx0*pow(vx2,4)*pow(vy0,2)+21*ax0*vx0*pow(vx2,4)*vy0+9*bx0*pow(vx0,2)*pow(vx2,4))*vy2+6*ax0*pow(vx2,5)*pow(vy0,2)+15*bx0*vx0*pow(vx2,5)*vy0+24*ax0*pow(vx0,2)*pow(vx2,5))*pow(vz0,4)+((-24*nx0*pow(vy0,3)-21*nx0*pow(vx0,2)*vy0)*pow(vy2,5)+(-30*nx0*vx0*vx2*pow(vy0,2)-21*nx0*pow(vx0,3)*vx2)*pow(vy2,4)+(-45*nx0*pow(vx2,2)*pow(vy0,3)-51*nx0*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-51*nx0*vx0*pow(vx2,3)*pow(vy0,2)-45*nx0*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-21*nx0*pow(vx2,4)*pow(vy0,3)-30*nx0*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-21*nx0*vx0*pow(vx2,5)*pow(vy0,2)-24*nx0*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+((24*bx0*pow(vy0,4)+18*ax0*vx0*pow(vy0,3)+24*bx0*pow(vx0,2)*pow(vy0,2)+15*ax0*pow(vx0,3)*vy0+3*bx0*pow(vx0,4))*pow(vy2,5)+(6*ax0*vx2*pow(vy0,4)+30*bx0*vx0*vx2*pow(vy0,3)+33*ax0*pow(vx0,2)*vx2*pow(vy0,2)+21*bx0*pow(vx0,3)*vx2*vy0+18*ax0*pow(vx0,4)*vx2)*pow(vy2,4)+(42*bx0*pow(vx2,2)*pow(vy0,4)+39*ax0*vx0*pow(vx2,2)*pow(vy0,3)+57*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+45*ax0*pow(vx0,3)*pow(vx2,2)*vy0+9*bx0*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(9*ax0*pow(vx2,3)*pow(vy0,4)+45*bx0*vx0*pow(vx2,3)*pow(vy0,3)+57*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+39*bx0*pow(vx0,3)*pow(vx2,3)*vy0+42*ax0*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(18*bx0*pow(vx2,4)*pow(vy0,4)+21*ax0*vx0*pow(vx2,4)*pow(vy0,3)+33*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+30*ax0*pow(vx0,3)*pow(vx2,4)*vy0+6*bx0*pow(vx0,4)*pow(vx2,4))*vy2+3*ax0*pow(vx2,5)*pow(vy0,4)+15*bx0*vx0*pow(vx2,5)*pow(vy0,3)+24*ax0*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+18*bx0*pow(vx0,3)*pow(vx2,5)*vy0+24*ax0*pow(vx0,4)*pow(vx2,5))*pow(vz0,2)+((-24*nx0*pow(vy0,5)-42*nx0*pow(vx0,2)*pow(vy0,3)-18*nx0*pow(vx0,4)*vy0)*pow(vy2,5)+(-36*nx0*vx0*vx2*pow(vy0,4)-54*nx0*pow(vx0,3)*vx2*pow(vy0,2)-18*nx0*pow(vx0,5)*vx2)*pow(vy2,4)+(-42*nx0*pow(vx2,2)*pow(vy0,5)-96*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-54*nx0*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-54*nx0*vx0*pow(vx2,3)*pow(vy0,4)-96*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-42*nx0*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-18*nx0*pow(vx2,4)*pow(vy0,5)-54*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-36*nx0*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-18*nx0*vx0*pow(vx2,5)*pow(vy0,4)-42*nx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-24*nx0*pow(vx0,5)*pow(vx2,5))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+((24*bx0*vy0+24*ax0*vx0)*pow(vz0,6)+(-24*nx0*pow(vy0,2)-24*nx0*pow(vx0,2))*pow(vz0,5)+(36*bx0*pow(vy0,3)+36*ax0*vx0*pow(vy0,2)+36*bx0*pow(vx0,2)*vy0+36*ax0*pow(vx0,3))*pow(vz0,4)+(-36*nx0*pow(vy0,4)-72*nx0*pow(vx0,2)*pow(vy0,2)-36*nx0*pow(vx0,4))*pow(vz0,3)+(12*bx0*pow(vy0,5)+12*ax0*vx0*pow(vy0,4)+24*bx0*pow(vx0,2)*pow(vy0,3)+24*ax0*pow(vx0,3)*pow(vy0,2)+12*bx0*pow(vx0,4)*vy0+12*ax0*pow(vx0,5))*pow(vz0,2)+(-12*nx0*pow(vy0,6)-36*nx0*pow(vx0,2)*pow(vy0,4)-36*nx0*pow(vx0,4)*pow(vy0,2)-12*nx0*pow(vx0,6))*vz0)*pow(vz2,6)+((24*bx0*vy2+24*ax0*vx2)*pow(vz0,7)+(-24*nx0*vy0*vy2-24*nx0*vx0*vx2)*pow(vz0,6)+((84*bx0*pow(vy0,2)+24*ax0*vx0*vy0+60*bx0*pow(vx0,2))*vy2+60*ax0*vx2*pow(vy0,2)+24*bx0*vx0*vx2*vy0+84*ax0*pow(vx0,2)*vx2)*pow(vz0,5)+((-84*nx0*pow(vy0,3)-84*nx0*pow(vx0,2)*vy0)*vy2-84*nx0*vx0*vx2*pow(vy0,2)-84*nx0*pow(vx0,3)*vx2)*pow(vz0,4)+((72*bx0*pow(vy0,4)+24*ax0*vx0*pow(vy0,3)+120*bx0*pow(vx0,2)*pow(vy0,2)+24*ax0*pow(vx0,3)*vy0+48*bx0*pow(vx0,4))*vy2+48*ax0*vx2*pow(vy0,4)+24*bx0*vx0*vx2*pow(vy0,3)+120*ax0*pow(vx0,2)*vx2*pow(vy0,2)+24*bx0*pow(vx0,3)*vx2*vy0+72*ax0*pow(vx0,4)*vx2)*pow(vz0,3)+((-72*nx0*pow(vy0,5)-144*nx0*pow(vx0,2)*pow(vy0,3)-72*nx0*pow(vx0,4)*vy0)*vy2-72*nx0*vx0*vx2*pow(vy0,4)-144*nx0*pow(vx0,3)*vx2*pow(vy0,2)-72*nx0*pow(vx0,5)*vx2)*pow(vz0,2)+((12*bx0*pow(vy0,6)+36*bx0*pow(vx0,2)*pow(vy0,4)+36*bx0*pow(vx0,4)*pow(vy0,2)+12*bx0*pow(vx0,6))*vy2+12*ax0*vx2*pow(vy0,6)+36*ax0*pow(vx0,2)*vx2*pow(vy0,4)+36*ax0*pow(vx0,4)*vx2*pow(vy0,2)+12*ax0*pow(vx0,6)*vx2)*vz0+(-12*nx0*pow(vy0,7)-36*nx0*pow(vx0,2)*pow(vy0,5)-36*nx0*pow(vx0,4)*pow(vy0,3)-12*nx0*pow(vx0,6)*vy0)*vy2-12*nx0*vx0*vx2*pow(vy0,6)-36*nx0*pow(vx0,3)*vx2*pow(vy0,4)-36*nx0*pow(vx0,5)*vx2*pow(vy0,2)-12*nx0*pow(vx0,7)*vx2)*pow(vz2,5)+(((84*bx0*vy0+60*ax0*vx0)*pow(vy2,2)+(24*ax0*vx2*vy0+24*bx0*vx0*vx2)*vy2+60*bx0*pow(vx2,2)*vy0+84*ax0*vx0*pow(vx2,2))*pow(vz0,6)+((-84*nx0*pow(vy0,2)-60*nx0*pow(vx0,2))*pow(vy2,2)-48*nx0*vx0*vx2*vy0*vy2-60*nx0*pow(vx2,2)*pow(vy0,2)-84*nx0*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((156*bx0*pow(vy0,3)+102*ax0*vx0*pow(vy0,2)+150*bx0*pow(vx0,2)*vy0+96*ax0*pow(vx0,3))*pow(vy2,2)+(54*ax0*vx2*pow(vy0,3)+66*bx0*vx0*vx2*pow(vy0,2)+66*ax0*pow(vx0,2)*vx2*vy0+54*bx0*pow(vx0,3)*vx2)*vy2+96*bx0*pow(vx2,2)*pow(vy0,3)+150*ax0*vx0*pow(vx2,2)*pow(vy0,2)+102*bx0*pow(vx0,2)*pow(vx2,2)*vy0+156*ax0*pow(vx0,3)*pow(vx2,2))*pow(vz0,4)+((-156*nx0*pow(vy0,4)-252*nx0*pow(vx0,2)*pow(vy0,2)-96*nx0*pow(vx0,4))*pow(vy2,2)+(-120*nx0*vx0*vx2*pow(vy0,3)-120*nx0*pow(vx0,3)*vx2*vy0)*vy2-96*nx0*pow(vx2,2)*pow(vy0,4)-252*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-156*nx0*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((72*bx0*pow(vy0,5)+36*ax0*vx0*pow(vy0,4)+144*bx0*pow(vx0,2)*pow(vy0,3)+72*ax0*pow(vx0,3)*pow(vy0,2)+72*bx0*pow(vx0,4)*vy0+36*ax0*pow(vx0,5))*pow(vy2,2)+(36*ax0*vx2*pow(vy0,5)+36*bx0*vx0*vx2*pow(vy0,4)+72*ax0*pow(vx0,2)*vx2*pow(vy0,3)+72*bx0*pow(vx0,3)*vx2*pow(vy0,2)+36*ax0*pow(vx0,4)*vx2*vy0+36*bx0*pow(vx0,5)*vx2)*vy2+36*bx0*pow(vx2,2)*pow(vy0,5)+72*ax0*vx0*pow(vx2,2)*pow(vy0,4)+72*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+144*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+36*bx0*pow(vx0,4)*pow(vx2,2)*vy0+72*ax0*pow(vx0,5)*pow(vx2,2))*pow(vz0,2)+((-72*nx0*pow(vy0,6)-180*nx0*pow(vx0,2)*pow(vy0,4)-144*nx0*pow(vx0,4)*pow(vy0,2)-36*nx0*pow(vx0,6))*pow(vy2,2)+(-72*nx0*vx0*vx2*pow(vy0,5)-144*nx0*pow(vx0,3)*vx2*pow(vy0,3)-72*nx0*pow(vx0,5)*vx2*vy0)*vy2-36*nx0*pow(vx2,2)*pow(vy0,6)-144*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-180*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-72*nx0*pow(vx0,6)*pow(vx2,2))*vz0+(-6*ax0*vx0*pow(vy0,6)+6*bx0*pow(vx0,2)*pow(vy0,5)-12*ax0*pow(vx0,3)*pow(vy0,4)+12*bx0*pow(vx0,4)*pow(vy0,3)-6*ax0*pow(vx0,5)*pow(vy0,2)+6*bx0*pow(vx0,6)*vy0)*pow(vy2,2)+(6*ax0*vx2*pow(vy0,7)-6*bx0*vx0*vx2*pow(vy0,6)+6*ax0*pow(vx0,2)*vx2*pow(vy0,5)-6*bx0*pow(vx0,3)*vx2*pow(vy0,4)-6*ax0*pow(vx0,4)*vx2*pow(vy0,3)+6*bx0*pow(vx0,5)*vx2*pow(vy0,2)-6*ax0*pow(vx0,6)*vx2*vy0+6*bx0*pow(vx0,7)*vx2)*vy2+6*ax0*vx0*pow(vx2,2)*pow(vy0,6)-6*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+12*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,4)-12*bx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+6*ax0*pow(vx0,5)*pow(vx2,2)*pow(vy0,2)-6*bx0*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vz2,4)+((36*bx0*pow(vy2,3)+36*ax0*vx2*pow(vy2,2)+36*bx0*pow(vx2,2)*vy2+36*ax0*pow(vx2,3))*pow(vz0,7)+(-36*nx0*vy0*pow(vy2,3)-36*nx0*vx0*vx2*pow(vy2,2)-36*nx0*pow(vx2,2)*vy0*vy2-36*nx0*vx0*pow(vx2,3))*pow(vz0,6)+((156*bx0*pow(vy0,2)+60*ax0*vx0*vy0+90*bx0*pow(vx0,2))*pow(vy2,3)+(96*ax0*vx2*pow(vy0,2)+72*bx0*vx0*vx2*vy0+150*ax0*pow(vx0,2)*vx2)*pow(vy2,2)+(150*bx0*pow(vx2,2)*pow(vy0,2)+72*ax0*vx0*pow(vx2,2)*vy0+96*bx0*pow(vx0,2)*pow(vx2,2))*vy2+90*ax0*pow(vx2,3)*pow(vy0,2)+60*bx0*vx0*pow(vx2,3)*vy0+156*ax0*pow(vx0,2)*pow(vx2,3))*pow(vz0,5)+((-156*nx0*pow(vy0,3)-150*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+(-168*nx0*vx0*vx2*pow(vy0,2)-150*nx0*pow(vx0,3)*vx2)*pow(vy2,2)+(-150*nx0*pow(vx2,2)*pow(vy0,3)-168*nx0*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-150*nx0*vx0*pow(vx2,3)*pow(vy0,2)-156*nx0*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((156*bx0*pow(vy0,4)+72*ax0*vx0*pow(vy0,3)+228*bx0*pow(vx0,2)*pow(vy0,2)+72*ax0*pow(vx0,3)*vy0+72*bx0*pow(vx0,4))*pow(vy2,3)+(84*ax0*vx2*pow(vy0,4)+96*bx0*vx0*vx2*pow(vy0,3)+228*ax0*pow(vx0,2)*vx2*pow(vy0,2)+96*bx0*pow(vx0,3)*vx2*vy0+144*ax0*pow(vx0,4)*vx2)*pow(vy2,2)+(144*bx0*pow(vx2,2)*pow(vy0,4)+96*ax0*vx0*pow(vx2,2)*pow(vy0,3)+228*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+96*ax0*pow(vx0,3)*pow(vx2,2)*vy0+84*bx0*pow(vx0,4)*pow(vx2,2))*vy2+72*ax0*pow(vx2,3)*pow(vy0,4)+72*bx0*vx0*pow(vx2,3)*pow(vy0,3)+228*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+72*bx0*pow(vx0,3)*pow(vx2,3)*vy0+156*ax0*pow(vx0,4)*pow(vx2,3))*pow(vz0,3)+((-156*nx0*pow(vy0,5)-300*nx0*pow(vx0,2)*pow(vy0,3)-144*nx0*pow(vx0,4)*vy0)*pow(vy2,3)+(-180*nx0*vx0*vx2*pow(vy0,4)-324*nx0*pow(vx0,3)*vx2*pow(vy0,2)-144*nx0*pow(vx0,5)*vx2)*pow(vy2,2)+(-144*nx0*pow(vx2,2)*pow(vy0,5)-324*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-180*nx0*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-144*nx0*vx0*pow(vx2,3)*pow(vy0,4)-300*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-156*nx0*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+((36*bx0*pow(vy0,6)+12*ax0*vx0*pow(vy0,5)+90*bx0*pow(vx0,2)*pow(vy0,4)+24*ax0*pow(vx0,3)*pow(vy0,3)+72*bx0*pow(vx0,4)*pow(vy0,2)+12*ax0*pow(vx0,5)*vy0+18*bx0*pow(vx0,6))*pow(vy2,3)+(24*ax0*vx2*pow(vy0,6)+24*bx0*vx0*vx2*pow(vy0,5)+78*ax0*pow(vx0,2)*vx2*pow(vy0,4)+48*bx0*pow(vx0,3)*vx2*pow(vy0,3)+84*ax0*pow(vx0,4)*vx2*pow(vy0,2)+24*bx0*pow(vx0,5)*vx2*vy0+30*ax0*pow(vx0,6)*vx2)*pow(vy2,2)+(30*bx0*pow(vx2,2)*pow(vy0,6)+24*ax0*vx0*pow(vx2,2)*pow(vy0,5)+84*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+48*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+78*bx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+24*ax0*pow(vx0,5)*pow(vx2,2)*vy0+24*bx0*pow(vx0,6)*pow(vx2,2))*vy2+18*ax0*pow(vx2,3)*pow(vy0,6)+12*bx0*vx0*pow(vx2,3)*pow(vy0,5)+72*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+24*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+90*ax0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+12*bx0*pow(vx0,5)*pow(vx2,3)*vy0+36*ax0*pow(vx0,6)*pow(vx2,3))*vz0+(-36*nx0*pow(vy0,7)-102*nx0*pow(vx0,2)*pow(vy0,5)-96*nx0*pow(vx0,4)*pow(vy0,3)-30*nx0*pow(vx0,6)*vy0)*pow(vy2,3)+(-48*nx0*vx0*vx2*pow(vy0,6)-126*nx0*pow(vx0,3)*vx2*pow(vy0,4)-108*nx0*pow(vx0,5)*vx2*pow(vy0,2)-30*nx0*pow(vx0,7)*vx2)*pow(vy2,2)+(-30*nx0*pow(vx2,2)*pow(vy0,7)-108*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)-126*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)-48*nx0*pow(vx0,6)*pow(vx2,2)*vy0)*vy2-30*nx0*vx0*pow(vx2,3)*pow(vy0,6)-96*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-102*nx0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-36*nx0*pow(vx0,7)*pow(vx2,3))*pow(vz2,3)+(((72*bx0*vy0+42*ax0*vx0)*pow(vy2,4)+(30*ax0*vx2*vy0+30*bx0*vx0*vx2)*pow(vy2,3)+(114*bx0*pow(vx2,2)*vy0+114*ax0*vx0*pow(vx2,2))*pow(vy2,2)+(30*ax0*pow(vx2,3)*vy0+30*bx0*vx0*pow(vx2,3))*vy2+42*bx0*pow(vx2,4)*vy0+72*ax0*vx0*pow(vx2,4))*pow(vz0,6)+((-72*nx0*pow(vy0,2)-42*nx0*pow(vx0,2))*pow(vy2,4)-60*nx0*vx0*vx2*vy0*pow(vy2,3)+(-114*nx0*pow(vx2,2)*pow(vy0,2)-114*nx0*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-60*nx0*vx0*pow(vx2,3)*vy0*vy2-42*nx0*pow(vx2,4)*pow(vy0,2)-72*nx0*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((156*bx0*pow(vy0,3)+90*ax0*vx0*pow(vy0,2)+138*bx0*pow(vx0,2)*vy0+72*ax0*pow(vx0,3))*pow(vy2,4)+(66*ax0*vx2*pow(vy0,3)+102*bx0*vx0*vx2*pow(vy0,2)+102*ax0*pow(vx0,2)*vx2*vy0+66*bx0*pow(vx0,3)*vx2)*pow(vy2,3)+(228*bx0*pow(vx2,2)*pow(vy0,3)+228*ax0*vx0*pow(vx2,2)*pow(vy0,2)+228*bx0*pow(vx0,2)*pow(vx2,2)*vy0+228*ax0*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(66*ax0*pow(vx2,3)*pow(vy0,3)+102*bx0*vx0*pow(vx2,3)*pow(vy0,2)+102*ax0*pow(vx0,2)*pow(vx2,3)*vy0+66*bx0*pow(vx0,3)*pow(vx2,3))*vy2+72*bx0*pow(vx2,4)*pow(vy0,3)+138*ax0*vx0*pow(vx2,4)*pow(vy0,2)+90*bx0*pow(vx0,2)*pow(vx2,4)*vy0+156*ax0*pow(vx0,3)*pow(vx2,4))*pow(vz0,4)+((-156*nx0*pow(vy0,4)-228*nx0*pow(vx0,2)*pow(vy0,2)-72*nx0*pow(vx0,4))*pow(vy2,4)+(-168*nx0*vx0*vx2*pow(vy0,3)-168*nx0*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-228*nx0*pow(vx2,2)*pow(vy0,4)-456*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-228*nx0*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-168*nx0*vx0*pow(vx2,3)*pow(vy0,3)-168*nx0*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-72*nx0*pow(vx2,4)*pow(vy0,4)-228*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-156*nx0*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((84*bx0*pow(vy0,5)+42*ax0*vx0*pow(vy0,4)+156*bx0*pow(vx0,2)*pow(vy0,3)+72*ax0*pow(vx0,3)*pow(vy0,2)+72*bx0*pow(vx0,4)*vy0+30*ax0*pow(vx0,5))*pow(vy2,4)+(42*ax0*vx2*pow(vy0,5)+66*bx0*vx0*vx2*pow(vy0,4)+108*ax0*pow(vx0,2)*vx2*pow(vy0,3)+108*bx0*pow(vx0,3)*vx2*pow(vy0,2)+66*ax0*pow(vx0,4)*vx2*vy0+42*bx0*pow(vx0,5)*vx2)*pow(vy2,3)+(114*bx0*pow(vx2,2)*pow(vy0,5)+114*ax0*vx0*pow(vx2,2)*pow(vy0,4)+228*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+228*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+114*bx0*pow(vx0,4)*pow(vx2,2)*vy0+114*ax0*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(42*ax0*pow(vx2,3)*pow(vy0,5)+66*bx0*vx0*pow(vx2,3)*pow(vy0,4)+108*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+108*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+66*ax0*pow(vx0,4)*pow(vx2,3)*vy0+42*bx0*pow(vx0,5)*pow(vx2,3))*vy2+30*bx0*pow(vx2,4)*pow(vy0,5)+72*ax0*vx0*pow(vx2,4)*pow(vy0,4)+72*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+156*ax0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+42*bx0*pow(vx0,4)*pow(vx2,4)*vy0+84*ax0*pow(vx0,5)*pow(vx2,4))*pow(vz0,2)+((-84*nx0*pow(vy0,6)-198*nx0*pow(vx0,2)*pow(vy0,4)-144*nx0*pow(vx0,4)*pow(vy0,2)-30*nx0*pow(vx0,6))*pow(vy2,4)+(-108*nx0*vx0*vx2*pow(vy0,5)-216*nx0*pow(vx0,3)*vx2*pow(vy0,3)-108*nx0*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(-114*nx0*pow(vx2,2)*pow(vy0,6)-342*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-342*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-114*nx0*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(-108*nx0*vx0*pow(vx2,3)*pow(vy0,5)-216*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-108*nx0*pow(vx0,5)*pow(vx2,3)*vy0)*vy2-30*nx0*pow(vx2,4)*pow(vy0,6)-144*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-198*nx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-84*nx0*pow(vx0,6)*pow(vx2,4))*vz0+(-6*ax0*vx0*pow(vy0,6)+6*bx0*pow(vx0,2)*pow(vy0,5)-12*ax0*pow(vx0,3)*pow(vy0,4)+12*bx0*pow(vx0,4)*pow(vy0,3)-6*ax0*pow(vx0,5)*pow(vy0,2)+6*bx0*pow(vx0,6)*vy0)*pow(vy2,4)+(6*ax0*vx2*pow(vy0,7)-6*bx0*vx0*vx2*pow(vy0,6)+6*ax0*pow(vx0,2)*vx2*pow(vy0,5)-6*bx0*pow(vx0,3)*vx2*pow(vy0,4)-6*ax0*pow(vx0,4)*vx2*pow(vy0,3)+6*bx0*pow(vx0,5)*vx2*pow(vy0,2)-6*ax0*pow(vx0,6)*vx2*vy0+6*bx0*pow(vx0,7)*vx2)*pow(vy2,3)+(6*ax0*pow(vx2,3)*pow(vy0,7)-6*bx0*vx0*pow(vx2,3)*pow(vy0,6)+6*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,5)-6*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-6*ax0*pow(vx0,4)*pow(vx2,3)*pow(vy0,3)+6*bx0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-6*ax0*pow(vx0,6)*pow(vx2,3)*vy0+6*bx0*pow(vx0,7)*pow(vx2,3))*vy2+6*ax0*vx0*pow(vx2,4)*pow(vy0,6)-6*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+12*ax0*pow(vx0,3)*pow(vx2,4)*pow(vy0,4)-12*bx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+6*ax0*pow(vx0,5)*pow(vx2,4)*pow(vy0,2)-6*bx0*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vz2,2)+((12*bx0*pow(vy2,5)+12*ax0*vx2*pow(vy2,4)+24*bx0*pow(vx2,2)*pow(vy2,3)+24*ax0*pow(vx2,3)*pow(vy2,2)+12*bx0*pow(vx2,4)*vy2+12*ax0*pow(vx2,5))*pow(vz0,7)+(-12*nx0*vy0*pow(vy2,5)-12*nx0*vx0*vx2*pow(vy2,4)-24*nx0*pow(vx2,2)*vy0*pow(vy2,3)-24*nx0*vx0*pow(vx2,3)*pow(vy2,2)-12*nx0*pow(vx2,4)*vy0*vy2-12*nx0*vx0*pow(vx2,5))*pow(vz0,6)+((72*bx0*pow(vy0,2)+36*ax0*vx0*vy0+30*bx0*pow(vx0,2))*pow(vy2,5)+(36*ax0*vx2*pow(vy0,2)+48*bx0*vx0*vx2*vy0+66*ax0*pow(vx0,2)*vx2)*pow(vy2,4)+(138*bx0*pow(vx2,2)*pow(vy0,2)+84*ax0*vx0*pow(vx2,2)*vy0+66*bx0*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(66*ax0*pow(vx2,3)*pow(vy0,2)+84*bx0*vx0*pow(vx2,3)*vy0+138*ax0*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(66*bx0*pow(vx2,4)*pow(vy0,2)+48*ax0*vx0*pow(vx2,4)*vy0+36*bx0*pow(vx0,2)*pow(vx2,4))*vy2+30*ax0*pow(vx2,5)*pow(vy0,2)+36*bx0*vx0*pow(vx2,5)*vy0+72*ax0*pow(vx0,2)*pow(vx2,5))*pow(vz0,5)+((-72*nx0*pow(vy0,3)-66*nx0*pow(vx0,2)*vy0)*pow(vy2,5)+(-84*nx0*vx0*vx2*pow(vy0,2)-66*nx0*pow(vx0,3)*vx2)*pow(vy2,4)+(-138*nx0*pow(vx2,2)*pow(vy0,3)-150*nx0*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-150*nx0*vx0*pow(vx2,3)*pow(vy0,2)-138*nx0*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-66*nx0*pow(vx2,4)*pow(vy0,3)-84*nx0*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-66*nx0*vx0*pow(vx2,5)*pow(vy0,2)-72*nx0*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((84*bx0*pow(vy0,4)+48*ax0*vx0*pow(vy0,3)+108*bx0*pow(vx0,2)*pow(vy0,2)+48*ax0*pow(vx0,3)*vy0+24*bx0*pow(vx0,4))*pow(vy2,5)+(36*ax0*vx2*pow(vy0,4)+72*bx0*vx0*vx2*pow(vy0,3)+108*ax0*pow(vx0,2)*vx2*pow(vy0,2)+72*bx0*pow(vx0,3)*vx2*vy0+72*ax0*pow(vx0,4)*vx2)*pow(vy2,4)+(156*bx0*pow(vx2,2)*pow(vy0,4)+120*ax0*vx0*pow(vx2,2)*pow(vy0,3)+216*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+120*ax0*pow(vx0,3)*pow(vx2,2)*vy0+60*bx0*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(60*ax0*pow(vx2,3)*pow(vy0,4)+120*bx0*vx0*pow(vx2,3)*pow(vy0,3)+216*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+120*bx0*pow(vx0,3)*pow(vx2,3)*vy0+156*ax0*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(72*bx0*pow(vx2,4)*pow(vy0,4)+72*ax0*vx0*pow(vx2,4)*pow(vy0,3)+108*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+72*ax0*pow(vx0,3)*pow(vx2,4)*vy0+36*bx0*pow(vx0,4)*pow(vx2,4))*vy2+24*ax0*pow(vx2,5)*pow(vy0,4)+48*bx0*vx0*pow(vx2,5)*pow(vy0,3)+108*ax0*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+48*bx0*pow(vx0,3)*pow(vx2,5)*vy0+84*ax0*pow(vx0,4)*pow(vx2,5))*pow(vz0,3)+((-84*nx0*pow(vy0,5)-156*nx0*pow(vx0,2)*pow(vy0,3)-72*nx0*pow(vx0,4)*vy0)*pow(vy2,5)+(-108*nx0*vx0*vx2*pow(vy0,4)-180*nx0*pow(vx0,3)*vx2*pow(vy0,2)-72*nx0*pow(vx0,5)*vx2)*pow(vy2,4)+(-156*nx0*pow(vx2,2)*pow(vy0,5)-336*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-180*nx0*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-180*nx0*vx0*pow(vx2,3)*pow(vy0,4)-336*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-156*nx0*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-72*nx0*pow(vx2,4)*pow(vy0,5)-180*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-108*nx0*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-72*nx0*vx0*pow(vx2,5)*pow(vy0,4)-156*nx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-84*nx0*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+((24*bx0*pow(vy0,6)+12*ax0*vx0*pow(vy0,5)+54*bx0*pow(vx0,2)*pow(vy0,4)+24*ax0*pow(vx0,3)*pow(vy0,3)+36*bx0*pow(vx0,4)*pow(vy0,2)+12*ax0*pow(vx0,5)*vy0+6*bx0*pow(vx0,6))*pow(vy2,5)+(12*ax0*vx2*pow(vy0,6)+24*bx0*vx0*vx2*pow(vy0,5)+42*ax0*pow(vx0,2)*vx2*pow(vy0,4)+48*bx0*pow(vx0,3)*vx2*pow(vy0,3)+48*ax0*pow(vx0,4)*vx2*pow(vy0,2)+24*bx0*pow(vx0,5)*vx2*vy0+18*ax0*pow(vx0,6)*vx2)*pow(vy2,4)+(42*bx0*pow(vx2,2)*pow(vy0,6)+36*ax0*vx0*pow(vx2,2)*pow(vy0,5)+102*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+72*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+78*bx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+36*ax0*pow(vx0,5)*pow(vx2,2)*vy0+18*bx0*pow(vx0,6)*pow(vx2,2))*pow(vy2,3)+(18*ax0*pow(vx2,3)*pow(vy0,6)+36*bx0*vx0*pow(vx2,3)*pow(vy0,5)+78*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+72*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+102*ax0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+36*bx0*pow(vx0,5)*pow(vx2,3)*vy0+42*ax0*pow(vx0,6)*pow(vx2,3))*pow(vy2,2)+(18*bx0*pow(vx2,4)*pow(vy0,6)+24*ax0*vx0*pow(vx2,4)*pow(vy0,5)+48*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+48*ax0*pow(vx0,3)*pow(vx2,4)*pow(vy0,3)+42*bx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+24*ax0*pow(vx0,5)*pow(vx2,4)*vy0+12*bx0*pow(vx0,6)*pow(vx2,4))*vy2+6*ax0*pow(vx2,5)*pow(vy0,6)+12*bx0*vx0*pow(vx2,5)*pow(vy0,5)+36*ax0*pow(vx0,2)*pow(vx2,5)*pow(vy0,4)+24*bx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+54*ax0*pow(vx0,4)*pow(vx2,5)*pow(vy0,2)+12*bx0*pow(vx0,5)*pow(vx2,5)*vy0+24*ax0*pow(vx0,6)*pow(vx2,5))*vz0+(-24*nx0*pow(vy0,7)-66*nx0*pow(vx0,2)*pow(vy0,5)-60*nx0*pow(vx0,4)*pow(vy0,3)-18*nx0*pow(vx0,6)*vy0)*pow(vy2,5)+(-36*nx0*vx0*vx2*pow(vy0,6)-90*nx0*pow(vx0,3)*vx2*pow(vy0,4)-72*nx0*pow(vx0,5)*vx2*pow(vy0,2)-18*nx0*pow(vx0,7)*vx2)*pow(vy2,4)+(-42*nx0*pow(vx2,2)*pow(vy0,7)-138*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)-150*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)-54*nx0*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(-54*nx0*vx0*pow(vx2,3)*pow(vy0,6)-150*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-138*nx0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-42*nx0*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(-18*nx0*pow(vx2,4)*pow(vy0,7)-72*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)-90*nx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)-36*nx0*pow(vx0,6)*pow(vx2,4)*vy0)*vy2-18*nx0*vx0*pow(vx2,5)*pow(vy0,6)-60*nx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)-66*nx0*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)-24*nx0*pow(vx0,7)*pow(vx2,5))*vz2+((12*bx0*vy0+6*ax0*vx0)*pow(vy2,6)+(6*ax0*vx2*vy0+6*bx0*vx0*vx2)*pow(vy2,5)+(30*bx0*pow(vx2,2)*vy0+24*ax0*vx0*pow(vx2,2))*pow(vy2,4)+(12*ax0*pow(vx2,3)*vy0+12*bx0*vx0*pow(vx2,3))*pow(vy2,3)+(24*bx0*pow(vx2,4)*vy0+30*ax0*vx0*pow(vx2,4))*pow(vy2,2)+(6*ax0*pow(vx2,5)*vy0+6*bx0*vx0*pow(vx2,5))*vy2+6*bx0*pow(vx2,6)*vy0+12*ax0*vx0*pow(vx2,6))*pow(vz0,6)+((-12*nx0*pow(vy0,2)-6*nx0*pow(vx0,2))*pow(vy2,6)-12*nx0*vx0*vx2*vy0*pow(vy2,5)+(-30*nx0*pow(vx2,2)*pow(vy0,2)-24*nx0*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)-24*nx0*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(-24*nx0*pow(vx2,4)*pow(vy0,2)-30*nx0*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)-12*nx0*vx0*pow(vx2,5)*vy0*vy2-6*nx0*pow(vx2,6)*pow(vy0,2)-12*nx0*pow(vx0,2)*pow(vx2,6))*pow(vz0,5)+((36*bx0*pow(vy0,3)+24*ax0*vx0*pow(vy0,2)+24*bx0*pow(vx0,2)*vy0+12*ax0*pow(vx0,3))*pow(vy2,6)+(12*ax0*vx2*pow(vy0,3)+36*bx0*vx0*vx2*pow(vy0,2)+36*ax0*pow(vx0,2)*vx2*vy0+12*bx0*pow(vx0,3)*vx2)*pow(vy2,5)+(84*bx0*pow(vx2,2)*pow(vy0,3)+72*ax0*vx0*pow(vx2,2)*pow(vy0,2)+72*bx0*pow(vx0,2)*pow(vx2,2)*vy0+60*ax0*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+(24*ax0*pow(vx2,3)*pow(vy0,3)+72*bx0*vx0*pow(vx2,3)*pow(vy0,2)+72*ax0*pow(vx0,2)*pow(vx2,3)*vy0+24*bx0*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(60*bx0*pow(vx2,4)*pow(vy0,3)+72*ax0*vx0*pow(vx2,4)*pow(vy0,2)+72*bx0*pow(vx0,2)*pow(vx2,4)*vy0+84*ax0*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+(12*ax0*pow(vx2,5)*pow(vy0,3)+36*bx0*vx0*pow(vx2,5)*pow(vy0,2)+36*ax0*pow(vx0,2)*pow(vx2,5)*vy0+12*bx0*pow(vx0,3)*pow(vx2,5))*vy2+12*bx0*pow(vx2,6)*pow(vy0,3)+24*ax0*vx0*pow(vx2,6)*pow(vy0,2)+24*bx0*pow(vx0,2)*pow(vx2,6)*vy0+36*ax0*pow(vx0,3)*pow(vx2,6))*pow(vz0,4)+((-36*nx0*pow(vy0,4)-48*nx0*pow(vx0,2)*pow(vy0,2)-12*nx0*pow(vx0,4))*pow(vy2,6)+(-48*nx0*vx0*vx2*pow(vy0,3)-48*nx0*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(-84*nx0*pow(vx2,2)*pow(vy0,4)-144*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-60*nx0*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(-96*nx0*vx0*pow(vx2,3)*pow(vy0,3)-96*nx0*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(-60*nx0*pow(vx2,4)*pow(vy0,4)-144*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-84*nx0*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(-48*nx0*vx0*pow(vx2,5)*pow(vy0,3)-48*nx0*pow(vx0,3)*pow(vx2,5)*vy0)*vy2-12*nx0*pow(vx2,6)*pow(vy0,4)-48*nx0*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)-36*nx0*pow(vx0,4)*pow(vx2,6))*pow(vz0,3)+((24*bx0*pow(vy0,5)+18*ax0*vx0*pow(vy0,4)+36*bx0*pow(vx0,2)*pow(vy0,3)+24*ax0*pow(vx0,3)*pow(vy0,2)+12*bx0*pow(vx0,4)*vy0+6*ax0*pow(vx0,5))*pow(vy2,6)+(6*ax0*vx2*pow(vy0,5)+30*bx0*vx0*vx2*pow(vy0,4)+36*ax0*pow(vx0,2)*vx2*pow(vy0,3)+36*bx0*pow(vx0,3)*vx2*pow(vy0,2)+30*ax0*pow(vx0,4)*vx2*vy0+6*bx0*pow(vx0,5)*vx2)*pow(vy2,5)+(54*bx0*pow(vx2,2)*pow(vy0,5)+48*ax0*vx0*pow(vx2,2)*pow(vy0,4)+96*bx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+84*ax0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+42*bx0*pow(vx0,4)*pow(vx2,2)*vy0+36*ax0*pow(vx0,5)*pow(vx2,2))*pow(vy2,4)+(12*ax0*pow(vx2,3)*pow(vy0,5)+60*bx0*vx0*pow(vx2,3)*pow(vy0,4)+72*ax0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+72*bx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+60*ax0*pow(vx0,4)*pow(vx2,3)*vy0+12*bx0*pow(vx0,5)*pow(vx2,3))*pow(vy2,3)+(36*bx0*pow(vx2,4)*pow(vy0,5)+42*ax0*vx0*pow(vx2,4)*pow(vy0,4)+84*bx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+96*ax0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+48*bx0*pow(vx0,4)*pow(vx2,4)*vy0+54*ax0*pow(vx0,5)*pow(vx2,4))*pow(vy2,2)+(6*ax0*pow(vx2,5)*pow(vy0,5)+30*bx0*vx0*pow(vx2,5)*pow(vy0,4)+36*ax0*pow(vx0,2)*pow(vx2,5)*pow(vy0,3)+36*bx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+30*ax0*pow(vx0,4)*pow(vx2,5)*vy0+6*bx0*pow(vx0,5)*pow(vx2,5))*vy2+6*bx0*pow(vx2,6)*pow(vy0,5)+12*ax0*vx0*pow(vx2,6)*pow(vy0,4)+24*bx0*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+36*ax0*pow(vx0,3)*pow(vx2,6)*pow(vy0,2)+18*bx0*pow(vx0,4)*pow(vx2,6)*vy0+24*ax0*pow(vx0,5)*pow(vx2,6))*pow(vz0,2)+((-24*nx0*pow(vy0,6)-54*nx0*pow(vx0,2)*pow(vy0,4)-36*nx0*pow(vx0,4)*pow(vy0,2)-6*nx0*pow(vx0,6))*pow(vy2,6)+(-36*nx0*vx0*vx2*pow(vy0,5)-72*nx0*pow(vx0,3)*vx2*pow(vy0,3)-36*nx0*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(-54*nx0*pow(vx2,2)*pow(vy0,6)-144*nx0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-126*nx0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-36*nx0*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(-72*nx0*vx0*pow(vx2,3)*pow(vy0,5)-144*nx0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-72*nx0*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(-36*nx0*pow(vx2,4)*pow(vy0,6)-126*nx0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-144*nx0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-54*nx0*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(-36*nx0*vx0*pow(vx2,5)*pow(vy0,5)-72*nx0*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)-36*nx0*pow(vx0,5)*pow(vx2,5)*vy0)*vy2-6*nx0*pow(vx2,6)*pow(vy0,6)-36*nx0*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)-54*nx0*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)-24*nx0*pow(vx0,6)*pow(vx2,6))*vz0)*pow(((160*pow(roughness,2)*pow(vz0,7)+(400*pow(roughness,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2))*pow(vz0,5)+(320*pow(roughness,2)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,6)+((720*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,7)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((240*pow(roughness,2)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,2))*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+640*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((1040*pow(roughness,2)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+560*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((400*pow(roughness,2)*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+160*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((400*pow(roughness,2)*vy0*pow(vy2,3)+400*pow(roughness,2)*vx0*vx2*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2)*vy0*vy2+400*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,6)+((1040*pow(roughness,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+(240*pow(roughness,2)*pow(vy0,7)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+160*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*vx2*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(160*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4))*pow(vz0,7)+((480*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((720*pow(roughness,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,5)+80*pow(roughness,2)*vx0*vx2*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,6)+((320*pow(roughness,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,7)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,8)+(480*pow(roughness,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2))*pow(vz0,6)+(500*pow(roughness,2)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4))*pow(vz0,4)+(200*pow(roughness,2)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,8)+80*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+120*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+80*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vz2,6)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+880*pow(roughness,2)*vx0*vx2*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,5)+((800*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*vz0)*pow(vz2,5)+((320*pow(roughness,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,2))*pow(vz0,8)+((1240*pow(roughness,2)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+((1700*pow(roughness,2)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+1100*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+((960*pow(roughness,2)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+480*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz0,2)+(180*pow(roughness,2)*pow(vy0,8)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,8))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*vy2+60*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vz2,4)+((560*pow(roughness,2)*vy0*pow(vy2,3)+560*pow(roughness,2)*vx0*vx2*pow(vy2,2)+560*pow(roughness,2)*pow(vx2,2)*vy0*vy2+560*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,7)+((1680*pow(roughness,2)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,5)+((1680*pow(roughness,2)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,4)+3520*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(1520*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+1520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,3)+((560*pow(roughness,2)*pow(vy0,7)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*vz0)*pow(vz2,3)+((180*pow(roughness,2)*pow(vy2,4)+360*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+180*pow(roughness,2)*pow(vx2,4))*pow(vz0,8)+((960*pow(roughness,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+720*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(1560*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+600*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+((1700*pow(roughness,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(2400*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+720*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,4)+((1240*pow(roughness,2)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(1560*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+1560*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+360*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,8)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1100*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,8))*pow(vy2,4)+(560*pow(roughness,2)*vx0*vx2*pow(vy0,7)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*vx2*pow(vy2,4)+480*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,4)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,7)+((800*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(1520*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,5)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1840*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,6)+60*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+60*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,6))*pow(vz0,8)+((200*pow(roughness,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,6)+((500*pow(roughness,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(1100*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(1520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,6)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+200*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vy2,6)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,5)+(320*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,4)+(560*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,3)+(180*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1240*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,8)*pow(vx2,6),-1) ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// y0  :  
poly.add(Term6f( -((((24*by0*vy0+24*ay0*vx0)*pow(vz0,5)+(-24*ny0*pow(vy0,2)-24*ny0*pow(vx0,2))*pow(vz0,4)+(24*by0*pow(vy0,3)+24*ay0*vx0*pow(vy0,2)+24*by0*pow(vx0,2)*vy0+24*ay0*pow(vx0,3))*pow(vz0,3)+(-24*ny0*pow(vy0,4)-48*ny0*pow(vx0,2)*pow(vy0,2)-24*ny0*pow(vx0,4))*pow(vz0,2)+(3*by0*pow(vy0,5)+3*ay0*vx0*pow(vy0,4)+6*by0*pow(vx0,2)*pow(vy0,3)+6*ay0*pow(vx0,3)*pow(vy0,2)+3*by0*pow(vx0,4)*vy0+3*ay0*pow(vx0,5))*vz0-3*ny0*pow(vy0,6)-9*ny0*pow(vx0,2)*pow(vy0,4)-9*ny0*pow(vx0,4)*pow(vy0,2)-3*ny0*pow(vx0,6))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((24*by0*vy2+24*ay0*vx2)*pow(vz0,6)+(-24*ny0*vy0*vy2-24*ny0*vx0*vx2)*pow(vz0,5)+((72*by0*pow(vy0,2)+24*ay0*vx0*vy0+48*by0*pow(vx0,2))*vy2+48*ay0*vx2*pow(vy0,2)+24*by0*vx0*vx2*vy0+72*ay0*pow(vx0,2)*vx2)*pow(vz0,4)+((-72*ny0*pow(vy0,3)-72*ny0*pow(vx0,2)*vy0)*vy2-72*ny0*vx0*vx2*pow(vy0,2)-72*ny0*pow(vx0,3)*vx2)*pow(vz0,3)+((39*by0*pow(vy0,4)+12*ay0*vx0*pow(vy0,3)+66*by0*pow(vx0,2)*pow(vy0,2)+12*ay0*pow(vx0,3)*vy0+27*by0*pow(vx0,4))*vy2+27*ay0*vx2*pow(vy0,4)+12*by0*vx0*vx2*pow(vy0,3)+66*ay0*pow(vx0,2)*vx2*pow(vy0,2)+12*by0*pow(vx0,3)*vx2*vy0+39*ay0*pow(vx0,4)*vx2)*pow(vz0,2)+((-39*ny0*pow(vy0,5)-78*ny0*pow(vx0,2)*pow(vy0,3)-39*ny0*pow(vx0,4)*vy0)*vy2-39*ny0*vx0*vx2*pow(vy0,4)-78*ny0*pow(vx0,3)*vx2*pow(vy0,2)-39*ny0*pow(vx0,5)*vx2)*vz0+(-3*ay0*vx0*pow(vy0,5)+3*by0*pow(vx0,2)*pow(vy0,4)-6*ay0*pow(vx0,3)*pow(vy0,3)+6*by0*pow(vx0,4)*pow(vy0,2)-3*ay0*pow(vx0,5)*vy0+3*by0*pow(vx0,6))*vy2+3*ay0*vx2*pow(vy0,6)-3*by0*vx0*vx2*pow(vy0,5)+6*ay0*pow(vx0,2)*vx2*pow(vy0,4)-6*by0*pow(vx0,3)*vx2*pow(vy0,3)+3*ay0*pow(vx0,4)*vx2*pow(vy0,2)-3*by0*pow(vx0,5)*vx2*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+(((72*by0*vy0+48*ay0*vx0)*pow(vy2,2)+(24*ay0*vx2*vy0+24*by0*vx0*vx2)*vy2+48*by0*pow(vx2,2)*vy0+72*ay0*vx0*pow(vx2,2))*pow(vz0,5)+((-72*ny0*pow(vy0,2)-48*ny0*pow(vx0,2))*pow(vy2,2)-48*ny0*vx0*vx2*vy0*vy2-48*ny0*pow(vx2,2)*pow(vy0,2)-72*ny0*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((102*by0*pow(vy0,3)+60*ay0*vx0*pow(vy0,2)+96*by0*pow(vx0,2)*vy0+54*ay0*pow(vx0,3))*pow(vy2,2)+(42*ay0*vx2*pow(vy0,3)+54*by0*vx0*vx2*pow(vy0,2)+54*ay0*pow(vx0,2)*vx2*vy0+42*by0*pow(vx0,3)*vx2)*vy2+54*by0*pow(vx2,2)*pow(vy0,3)+96*ay0*vx0*pow(vx2,2)*pow(vy0,2)+60*by0*pow(vx0,2)*pow(vx2,2)*vy0+102*ay0*pow(vx0,3)*pow(vx2,2))*pow(vz0,3)+((-102*ny0*pow(vy0,4)-156*ny0*pow(vx0,2)*pow(vy0,2)-54*ny0*pow(vx0,4))*pow(vy2,2)+(-96*ny0*vx0*vx2*pow(vy0,3)-96*ny0*pow(vx0,3)*vx2*vy0)*vy2-54*ny0*pow(vx2,2)*pow(vy0,4)-156*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-102*ny0*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((24*by0*pow(vy0,5)+6*ay0*vx0*pow(vy0,4)+51*by0*pow(vx0,2)*pow(vy0,3)+15*ay0*pow(vx0,3)*pow(vy0,2)+27*by0*pow(vx0,4)*vy0+9*ay0*pow(vx0,5))*pow(vy2,2)+(18*ay0*vx2*pow(vy0,5)+12*by0*vx0*vx2*pow(vy0,4)+30*ay0*pow(vx0,2)*vx2*pow(vy0,3)+30*by0*pow(vx0,3)*vx2*pow(vy0,2)+12*ay0*pow(vx0,4)*vx2*vy0+18*by0*pow(vx0,5)*vx2)*vy2+9*by0*pow(vx2,2)*pow(vy0,5)+27*ay0*vx0*pow(vx2,2)*pow(vy0,4)+15*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+51*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+6*by0*pow(vx0,4)*pow(vx2,2)*vy0+24*ay0*pow(vx0,5)*pow(vx2,2))*vz0+(-24*ny0*pow(vy0,6)-57*ny0*pow(vx0,2)*pow(vy0,4)-42*ny0*pow(vx0,4)*pow(vy0,2)-9*ny0*pow(vx0,6))*pow(vy2,2)+(-30*ny0*vx0*vx2*pow(vy0,5)-60*ny0*pow(vx0,3)*vx2*pow(vy0,3)-30*ny0*pow(vx0,5)*vx2*vy0)*vy2-9*ny0*pow(vx2,2)*pow(vy0,6)-42*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-57*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-24*ny0*pow(vx0,6)*pow(vx2,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((24*by0*pow(vy2,3)+24*ay0*vx2*pow(vy2,2)+24*by0*pow(vx2,2)*vy2+24*ay0*pow(vx2,3))*pow(vz0,6)+(-24*ny0*vy0*pow(vy2,3)-24*ny0*vx0*vx2*pow(vy2,2)-24*ny0*pow(vx2,2)*vy0*vy2-24*ny0*vx0*pow(vx2,3))*pow(vz0,5)+((102*by0*pow(vy0,2)+48*ay0*vx0*vy0+48*by0*pow(vx0,2))*pow(vy2,3)+(54*ay0*vx2*pow(vy0,2)+60*by0*vx0*vx2*vy0+96*ay0*pow(vx0,2)*vx2)*pow(vy2,2)+(96*by0*pow(vx2,2)*pow(vy0,2)+60*ay0*vx0*pow(vx2,2)*vy0+54*by0*pow(vx0,2)*pow(vx2,2))*vy2+48*ay0*pow(vx2,3)*pow(vy0,2)+48*by0*vx0*pow(vx2,3)*vy0+102*ay0*pow(vx0,2)*pow(vx2,3))*pow(vz0,4)+((-102*ny0*pow(vy0,3)-96*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+(-114*ny0*vx0*vx2*pow(vy0,2)-96*ny0*pow(vx0,3)*vx2)*pow(vy2,2)+(-96*ny0*pow(vx2,2)*pow(vy0,3)-114*ny0*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-96*ny0*vx0*pow(vx2,3)*pow(vy0,2)-102*ny0*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+((72*by0*pow(vy0,4)+36*ay0*vx0*pow(vy0,3)+99*by0*pow(vx0,2)*pow(vy0,2)+36*ay0*pow(vx0,3)*vy0+27*by0*pow(vx0,4))*pow(vy2,3)+(36*ay0*vx2*pow(vy0,4)+54*by0*vx0*vx2*pow(vy0,3)+99*ay0*pow(vx0,2)*vx2*pow(vy0,2)+54*by0*pow(vx0,3)*vx2*vy0+63*ay0*pow(vx0,4)*vx2)*pow(vy2,2)+(63*by0*pow(vx2,2)*pow(vy0,4)+54*ay0*vx0*pow(vx2,2)*pow(vy0,3)+99*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+54*ay0*pow(vx0,3)*pow(vx2,2)*vy0+36*by0*pow(vx0,4)*pow(vx2,2))*vy2+27*ay0*pow(vx2,3)*pow(vy0,4)+36*by0*vx0*pow(vx2,3)*pow(vy0,3)+99*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+36*by0*pow(vx0,3)*pow(vx2,3)*vy0+72*ay0*pow(vx0,4)*pow(vx2,3))*pow(vz0,2)+((-72*ny0*pow(vy0,5)-135*ny0*pow(vx0,2)*pow(vy0,3)-63*ny0*pow(vx0,4)*vy0)*pow(vy2,3)+(-90*ny0*vx0*vx2*pow(vy0,4)-153*ny0*pow(vx0,3)*vx2*pow(vy0,2)-63*ny0*pow(vx0,5)*vx2)*pow(vy2,2)+(-63*ny0*pow(vx2,2)*pow(vy0,5)-153*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-90*ny0*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-63*ny0*vx0*pow(vx2,3)*pow(vy0,4)-135*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-72*ny0*pow(vx0,5)*pow(vx2,3))*vz0+(-6*ay0*vx0*pow(vy0,5)+6*by0*pow(vx0,2)*pow(vy0,4)-9*ay0*pow(vx0,3)*pow(vy0,3)+9*by0*pow(vx0,4)*pow(vy0,2)-3*ay0*pow(vx0,5)*vy0+3*by0*pow(vx0,6))*pow(vy2,3)+(6*ay0*vx2*pow(vy0,6)-6*by0*vx0*vx2*pow(vy0,5)+3*ay0*pow(vx0,2)*vx2*pow(vy0,4)-3*by0*pow(vx0,3)*vx2*pow(vy0,3)-3*ay0*pow(vx0,4)*vx2*pow(vy0,2)+3*by0*pow(vx0,5)*vx2*vy0)*pow(vy2,2)+(3*ay0*vx0*pow(vx2,2)*pow(vy0,5)-3*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-3*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+3*by0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-6*ay0*pow(vx0,5)*pow(vx2,2)*vy0+6*by0*pow(vx0,6)*pow(vx2,2))*vy2+3*ay0*pow(vx2,3)*pow(vy0,6)-3*by0*vx0*pow(vx2,3)*pow(vy0,5)+9*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)-9*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+6*ay0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)-6*by0*pow(vx0,5)*pow(vx2,3)*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+(((39*by0*vy0+21*ay0*vx0)*pow(vy2,4)+(18*ay0*vx2*vy0+18*by0*vx0*vx2)*pow(vy2,3)+(60*by0*pow(vx2,2)*vy0+60*ay0*vx0*pow(vx2,2))*pow(vy2,2)+(18*ay0*pow(vx2,3)*vy0+18*by0*vx0*pow(vx2,3))*vy2+21*by0*pow(vx2,4)*vy0+39*ay0*vx0*pow(vx2,4))*pow(vz0,5)+((-39*ny0*pow(vy0,2)-21*ny0*pow(vx0,2))*pow(vy2,4)-36*ny0*vx0*vx2*vy0*pow(vy2,3)+(-60*ny0*pow(vx2,2)*pow(vy0,2)-60*ny0*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-36*ny0*vx0*pow(vx2,3)*vy0*vy2-21*ny0*pow(vx2,4)*pow(vy0,2)-39*ny0*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+((72*by0*pow(vy0,3)+42*ay0*vx0*pow(vy0,2)+57*by0*pow(vx0,2)*vy0+27*ay0*pow(vx0,3))*pow(vy2,4)+(30*ay0*vx2*pow(vy0,3)+60*by0*vx0*vx2*pow(vy0,2)+60*ay0*pow(vx0,2)*vx2*vy0+30*by0*pow(vx0,3)*vx2)*pow(vy2,3)+(99*by0*pow(vx2,2)*pow(vy0,3)+99*ay0*vx0*pow(vx2,2)*pow(vy0,2)+99*by0*pow(vx0,2)*pow(vx2,2)*vy0+99*ay0*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(30*ay0*pow(vx2,3)*pow(vy0,3)+60*by0*vx0*pow(vx2,3)*pow(vy0,2)+60*ay0*pow(vx0,2)*pow(vx2,3)*vy0+30*by0*pow(vx0,3)*pow(vx2,3))*vy2+27*by0*pow(vx2,4)*pow(vy0,3)+57*ay0*vx0*pow(vx2,4)*pow(vy0,2)+42*by0*pow(vx0,2)*pow(vx2,4)*vy0+72*ay0*pow(vx0,3)*pow(vx2,4))*pow(vz0,3)+((-72*ny0*pow(vy0,4)-99*ny0*pow(vx0,2)*pow(vy0,2)-27*ny0*pow(vx0,4))*pow(vy2,4)+(-90*ny0*vx0*vx2*pow(vy0,3)-90*ny0*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-99*ny0*pow(vx2,2)*pow(vy0,4)-198*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-99*ny0*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-90*ny0*vx0*pow(vx2,3)*pow(vy0,3)-90*ny0*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-27*ny0*pow(vx2,4)*pow(vy0,4)-99*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-72*ny0*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+((24*by0*pow(vy0,5)+12*ay0*vx0*pow(vy0,4)+42*by0*pow(vx0,2)*pow(vy0,3)+18*ay0*pow(vx0,3)*pow(vy0,2)+18*by0*pow(vx0,4)*vy0+6*ay0*pow(vx0,5))*pow(vy2,4)+(12*ay0*vx2*pow(vy0,5)+24*by0*vx0*vx2*pow(vy0,4)+36*ay0*pow(vx0,2)*vx2*pow(vy0,3)+36*by0*pow(vx0,3)*vx2*pow(vy0,2)+24*ay0*pow(vx0,4)*vx2*vy0+12*by0*pow(vx0,5)*vx2)*pow(vy2,3)+(30*by0*pow(vx2,2)*pow(vy0,5)+30*ay0*vx0*pow(vx2,2)*pow(vy0,4)+60*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+60*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+30*by0*pow(vx0,4)*pow(vx2,2)*vy0+30*ay0*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(12*ay0*pow(vx2,3)*pow(vy0,5)+24*by0*vx0*pow(vx2,3)*pow(vy0,4)+36*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+36*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+24*ay0*pow(vx0,4)*pow(vx2,3)*vy0+12*by0*pow(vx0,5)*pow(vx2,3))*vy2+6*by0*pow(vx2,4)*pow(vy0,5)+18*ay0*vx0*pow(vx2,4)*pow(vy0,4)+18*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+42*ay0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+12*by0*pow(vx0,4)*pow(vx2,4)*vy0+24*ay0*pow(vx0,5)*pow(vx2,4))*vz0+(-24*ny0*pow(vy0,6)-54*ny0*pow(vx0,2)*pow(vy0,4)-36*ny0*pow(vx0,4)*pow(vy0,2)-6*ny0*pow(vx0,6))*pow(vy2,4)+(-36*ny0*vx0*vx2*pow(vy0,5)-72*ny0*pow(vx0,3)*vx2*pow(vy0,3)-36*ny0*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(-30*ny0*pow(vx2,2)*pow(vy0,6)-90*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-90*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-30*ny0*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(-36*ny0*vx0*pow(vx2,3)*pow(vy0,5)-72*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-36*ny0*pow(vx0,5)*pow(vx2,3)*vy0)*vy2-6*ny0*pow(vx2,4)*pow(vy0,6)-36*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-54*ny0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-24*ny0*pow(vx0,6)*pow(vx2,4))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((3*by0*pow(vy2,5)+3*ay0*vx2*pow(vy2,4)+6*by0*pow(vx2,2)*pow(vy2,3)+6*ay0*pow(vx2,3)*pow(vy2,2)+3*by0*pow(vx2,4)*vy2+3*ay0*pow(vx2,5))*pow(vz0,6)+(-3*ny0*vy0*pow(vy2,5)-3*ny0*vx0*vx2*pow(vy2,4)-6*ny0*pow(vx2,2)*vy0*pow(vy2,3)-6*ny0*vx0*pow(vx2,3)*pow(vy2,2)-3*ny0*pow(vx2,4)*vy0*vy2-3*ny0*vx0*pow(vx2,5))*pow(vz0,5)+((24*by0*pow(vy0,2)+15*ay0*vx0*vy0+6*by0*pow(vx0,2))*pow(vy2,5)+(9*ay0*vx2*pow(vy0,2)+21*by0*vx0*vx2*vy0+21*ay0*pow(vx0,2)*vx2)*pow(vy2,4)+(45*by0*pow(vx2,2)*pow(vy0,2)+36*ay0*vx0*pow(vx2,2)*vy0+15*by0*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(15*ay0*pow(vx2,3)*pow(vy0,2)+36*by0*vx0*pow(vx2,3)*vy0+45*ay0*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(21*by0*pow(vx2,4)*pow(vy0,2)+21*ay0*vx0*pow(vx2,4)*vy0+9*by0*pow(vx0,2)*pow(vx2,4))*vy2+6*ay0*pow(vx2,5)*pow(vy0,2)+15*by0*vx0*pow(vx2,5)*vy0+24*ay0*pow(vx0,2)*pow(vx2,5))*pow(vz0,4)+((-24*ny0*pow(vy0,3)-21*ny0*pow(vx0,2)*vy0)*pow(vy2,5)+(-30*ny0*vx0*vx2*pow(vy0,2)-21*ny0*pow(vx0,3)*vx2)*pow(vy2,4)+(-45*ny0*pow(vx2,2)*pow(vy0,3)-51*ny0*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-51*ny0*vx0*pow(vx2,3)*pow(vy0,2)-45*ny0*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-21*ny0*pow(vx2,4)*pow(vy0,3)-30*ny0*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-21*ny0*vx0*pow(vx2,5)*pow(vy0,2)-24*ny0*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+((24*by0*pow(vy0,4)+18*ay0*vx0*pow(vy0,3)+24*by0*pow(vx0,2)*pow(vy0,2)+15*ay0*pow(vx0,3)*vy0+3*by0*pow(vx0,4))*pow(vy2,5)+(6*ay0*vx2*pow(vy0,4)+30*by0*vx0*vx2*pow(vy0,3)+33*ay0*pow(vx0,2)*vx2*pow(vy0,2)+21*by0*pow(vx0,3)*vx2*vy0+18*ay0*pow(vx0,4)*vx2)*pow(vy2,4)+(42*by0*pow(vx2,2)*pow(vy0,4)+39*ay0*vx0*pow(vx2,2)*pow(vy0,3)+57*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+45*ay0*pow(vx0,3)*pow(vx2,2)*vy0+9*by0*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(9*ay0*pow(vx2,3)*pow(vy0,4)+45*by0*vx0*pow(vx2,3)*pow(vy0,3)+57*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+39*by0*pow(vx0,3)*pow(vx2,3)*vy0+42*ay0*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(18*by0*pow(vx2,4)*pow(vy0,4)+21*ay0*vx0*pow(vx2,4)*pow(vy0,3)+33*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+30*ay0*pow(vx0,3)*pow(vx2,4)*vy0+6*by0*pow(vx0,4)*pow(vx2,4))*vy2+3*ay0*pow(vx2,5)*pow(vy0,4)+15*by0*vx0*pow(vx2,5)*pow(vy0,3)+24*ay0*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+18*by0*pow(vx0,3)*pow(vx2,5)*vy0+24*ay0*pow(vx0,4)*pow(vx2,5))*pow(vz0,2)+((-24*ny0*pow(vy0,5)-42*ny0*pow(vx0,2)*pow(vy0,3)-18*ny0*pow(vx0,4)*vy0)*pow(vy2,5)+(-36*ny0*vx0*vx2*pow(vy0,4)-54*ny0*pow(vx0,3)*vx2*pow(vy0,2)-18*ny0*pow(vx0,5)*vx2)*pow(vy2,4)+(-42*ny0*pow(vx2,2)*pow(vy0,5)-96*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-54*ny0*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-54*ny0*vx0*pow(vx2,3)*pow(vy0,4)-96*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-42*ny0*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-18*ny0*pow(vx2,4)*pow(vy0,5)-54*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-36*ny0*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-18*ny0*vx0*pow(vx2,5)*pow(vy0,4)-42*ny0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-24*ny0*pow(vx0,5)*pow(vx2,5))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+((24*by0*vy0+24*ay0*vx0)*pow(vz0,6)+(-24*ny0*pow(vy0,2)-24*ny0*pow(vx0,2))*pow(vz0,5)+(36*by0*pow(vy0,3)+36*ay0*vx0*pow(vy0,2)+36*by0*pow(vx0,2)*vy0+36*ay0*pow(vx0,3))*pow(vz0,4)+(-36*ny0*pow(vy0,4)-72*ny0*pow(vx0,2)*pow(vy0,2)-36*ny0*pow(vx0,4))*pow(vz0,3)+(12*by0*pow(vy0,5)+12*ay0*vx0*pow(vy0,4)+24*by0*pow(vx0,2)*pow(vy0,3)+24*ay0*pow(vx0,3)*pow(vy0,2)+12*by0*pow(vx0,4)*vy0+12*ay0*pow(vx0,5))*pow(vz0,2)+(-12*ny0*pow(vy0,6)-36*ny0*pow(vx0,2)*pow(vy0,4)-36*ny0*pow(vx0,4)*pow(vy0,2)-12*ny0*pow(vx0,6))*vz0)*pow(vz2,6)+((24*by0*vy2+24*ay0*vx2)*pow(vz0,7)+(-24*ny0*vy0*vy2-24*ny0*vx0*vx2)*pow(vz0,6)+((84*by0*pow(vy0,2)+24*ay0*vx0*vy0+60*by0*pow(vx0,2))*vy2+60*ay0*vx2*pow(vy0,2)+24*by0*vx0*vx2*vy0+84*ay0*pow(vx0,2)*vx2)*pow(vz0,5)+((-84*ny0*pow(vy0,3)-84*ny0*pow(vx0,2)*vy0)*vy2-84*ny0*vx0*vx2*pow(vy0,2)-84*ny0*pow(vx0,3)*vx2)*pow(vz0,4)+((72*by0*pow(vy0,4)+24*ay0*vx0*pow(vy0,3)+120*by0*pow(vx0,2)*pow(vy0,2)+24*ay0*pow(vx0,3)*vy0+48*by0*pow(vx0,4))*vy2+48*ay0*vx2*pow(vy0,4)+24*by0*vx0*vx2*pow(vy0,3)+120*ay0*pow(vx0,2)*vx2*pow(vy0,2)+24*by0*pow(vx0,3)*vx2*vy0+72*ay0*pow(vx0,4)*vx2)*pow(vz0,3)+((-72*ny0*pow(vy0,5)-144*ny0*pow(vx0,2)*pow(vy0,3)-72*ny0*pow(vx0,4)*vy0)*vy2-72*ny0*vx0*vx2*pow(vy0,4)-144*ny0*pow(vx0,3)*vx2*pow(vy0,2)-72*ny0*pow(vx0,5)*vx2)*pow(vz0,2)+((12*by0*pow(vy0,6)+36*by0*pow(vx0,2)*pow(vy0,4)+36*by0*pow(vx0,4)*pow(vy0,2)+12*by0*pow(vx0,6))*vy2+12*ay0*vx2*pow(vy0,6)+36*ay0*pow(vx0,2)*vx2*pow(vy0,4)+36*ay0*pow(vx0,4)*vx2*pow(vy0,2)+12*ay0*pow(vx0,6)*vx2)*vz0+(-12*ny0*pow(vy0,7)-36*ny0*pow(vx0,2)*pow(vy0,5)-36*ny0*pow(vx0,4)*pow(vy0,3)-12*ny0*pow(vx0,6)*vy0)*vy2-12*ny0*vx0*vx2*pow(vy0,6)-36*ny0*pow(vx0,3)*vx2*pow(vy0,4)-36*ny0*pow(vx0,5)*vx2*pow(vy0,2)-12*ny0*pow(vx0,7)*vx2)*pow(vz2,5)+(((84*by0*vy0+60*ay0*vx0)*pow(vy2,2)+(24*ay0*vx2*vy0+24*by0*vx0*vx2)*vy2+60*by0*pow(vx2,2)*vy0+84*ay0*vx0*pow(vx2,2))*pow(vz0,6)+((-84*ny0*pow(vy0,2)-60*ny0*pow(vx0,2))*pow(vy2,2)-48*ny0*vx0*vx2*vy0*vy2-60*ny0*pow(vx2,2)*pow(vy0,2)-84*ny0*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((156*by0*pow(vy0,3)+102*ay0*vx0*pow(vy0,2)+150*by0*pow(vx0,2)*vy0+96*ay0*pow(vx0,3))*pow(vy2,2)+(54*ay0*vx2*pow(vy0,3)+66*by0*vx0*vx2*pow(vy0,2)+66*ay0*pow(vx0,2)*vx2*vy0+54*by0*pow(vx0,3)*vx2)*vy2+96*by0*pow(vx2,2)*pow(vy0,3)+150*ay0*vx0*pow(vx2,2)*pow(vy0,2)+102*by0*pow(vx0,2)*pow(vx2,2)*vy0+156*ay0*pow(vx0,3)*pow(vx2,2))*pow(vz0,4)+((-156*ny0*pow(vy0,4)-252*ny0*pow(vx0,2)*pow(vy0,2)-96*ny0*pow(vx0,4))*pow(vy2,2)+(-120*ny0*vx0*vx2*pow(vy0,3)-120*ny0*pow(vx0,3)*vx2*vy0)*vy2-96*ny0*pow(vx2,2)*pow(vy0,4)-252*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-156*ny0*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((72*by0*pow(vy0,5)+36*ay0*vx0*pow(vy0,4)+144*by0*pow(vx0,2)*pow(vy0,3)+72*ay0*pow(vx0,3)*pow(vy0,2)+72*by0*pow(vx0,4)*vy0+36*ay0*pow(vx0,5))*pow(vy2,2)+(36*ay0*vx2*pow(vy0,5)+36*by0*vx0*vx2*pow(vy0,4)+72*ay0*pow(vx0,2)*vx2*pow(vy0,3)+72*by0*pow(vx0,3)*vx2*pow(vy0,2)+36*ay0*pow(vx0,4)*vx2*vy0+36*by0*pow(vx0,5)*vx2)*vy2+36*by0*pow(vx2,2)*pow(vy0,5)+72*ay0*vx0*pow(vx2,2)*pow(vy0,4)+72*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+144*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+36*by0*pow(vx0,4)*pow(vx2,2)*vy0+72*ay0*pow(vx0,5)*pow(vx2,2))*pow(vz0,2)+((-72*ny0*pow(vy0,6)-180*ny0*pow(vx0,2)*pow(vy0,4)-144*ny0*pow(vx0,4)*pow(vy0,2)-36*ny0*pow(vx0,6))*pow(vy2,2)+(-72*ny0*vx0*vx2*pow(vy0,5)-144*ny0*pow(vx0,3)*vx2*pow(vy0,3)-72*ny0*pow(vx0,5)*vx2*vy0)*vy2-36*ny0*pow(vx2,2)*pow(vy0,6)-144*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-180*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-72*ny0*pow(vx0,6)*pow(vx2,2))*vz0+(-6*ay0*vx0*pow(vy0,6)+6*by0*pow(vx0,2)*pow(vy0,5)-12*ay0*pow(vx0,3)*pow(vy0,4)+12*by0*pow(vx0,4)*pow(vy0,3)-6*ay0*pow(vx0,5)*pow(vy0,2)+6*by0*pow(vx0,6)*vy0)*pow(vy2,2)+(6*ay0*vx2*pow(vy0,7)-6*by0*vx0*vx2*pow(vy0,6)+6*ay0*pow(vx0,2)*vx2*pow(vy0,5)-6*by0*pow(vx0,3)*vx2*pow(vy0,4)-6*ay0*pow(vx0,4)*vx2*pow(vy0,3)+6*by0*pow(vx0,5)*vx2*pow(vy0,2)-6*ay0*pow(vx0,6)*vx2*vy0+6*by0*pow(vx0,7)*vx2)*vy2+6*ay0*vx0*pow(vx2,2)*pow(vy0,6)-6*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+12*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,4)-12*by0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+6*ay0*pow(vx0,5)*pow(vx2,2)*pow(vy0,2)-6*by0*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vz2,4)+((36*by0*pow(vy2,3)+36*ay0*vx2*pow(vy2,2)+36*by0*pow(vx2,2)*vy2+36*ay0*pow(vx2,3))*pow(vz0,7)+(-36*ny0*vy0*pow(vy2,3)-36*ny0*vx0*vx2*pow(vy2,2)-36*ny0*pow(vx2,2)*vy0*vy2-36*ny0*vx0*pow(vx2,3))*pow(vz0,6)+((156*by0*pow(vy0,2)+60*ay0*vx0*vy0+90*by0*pow(vx0,2))*pow(vy2,3)+(96*ay0*vx2*pow(vy0,2)+72*by0*vx0*vx2*vy0+150*ay0*pow(vx0,2)*vx2)*pow(vy2,2)+(150*by0*pow(vx2,2)*pow(vy0,2)+72*ay0*vx0*pow(vx2,2)*vy0+96*by0*pow(vx0,2)*pow(vx2,2))*vy2+90*ay0*pow(vx2,3)*pow(vy0,2)+60*by0*vx0*pow(vx2,3)*vy0+156*ay0*pow(vx0,2)*pow(vx2,3))*pow(vz0,5)+((-156*ny0*pow(vy0,3)-150*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+(-168*ny0*vx0*vx2*pow(vy0,2)-150*ny0*pow(vx0,3)*vx2)*pow(vy2,2)+(-150*ny0*pow(vx2,2)*pow(vy0,3)-168*ny0*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-150*ny0*vx0*pow(vx2,3)*pow(vy0,2)-156*ny0*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((156*by0*pow(vy0,4)+72*ay0*vx0*pow(vy0,3)+228*by0*pow(vx0,2)*pow(vy0,2)+72*ay0*pow(vx0,3)*vy0+72*by0*pow(vx0,4))*pow(vy2,3)+(84*ay0*vx2*pow(vy0,4)+96*by0*vx0*vx2*pow(vy0,3)+228*ay0*pow(vx0,2)*vx2*pow(vy0,2)+96*by0*pow(vx0,3)*vx2*vy0+144*ay0*pow(vx0,4)*vx2)*pow(vy2,2)+(144*by0*pow(vx2,2)*pow(vy0,4)+96*ay0*vx0*pow(vx2,2)*pow(vy0,3)+228*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+96*ay0*pow(vx0,3)*pow(vx2,2)*vy0+84*by0*pow(vx0,4)*pow(vx2,2))*vy2+72*ay0*pow(vx2,3)*pow(vy0,4)+72*by0*vx0*pow(vx2,3)*pow(vy0,3)+228*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+72*by0*pow(vx0,3)*pow(vx2,3)*vy0+156*ay0*pow(vx0,4)*pow(vx2,3))*pow(vz0,3)+((-156*ny0*pow(vy0,5)-300*ny0*pow(vx0,2)*pow(vy0,3)-144*ny0*pow(vx0,4)*vy0)*pow(vy2,3)+(-180*ny0*vx0*vx2*pow(vy0,4)-324*ny0*pow(vx0,3)*vx2*pow(vy0,2)-144*ny0*pow(vx0,5)*vx2)*pow(vy2,2)+(-144*ny0*pow(vx2,2)*pow(vy0,5)-324*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-180*ny0*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-144*ny0*vx0*pow(vx2,3)*pow(vy0,4)-300*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-156*ny0*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+((36*by0*pow(vy0,6)+12*ay0*vx0*pow(vy0,5)+90*by0*pow(vx0,2)*pow(vy0,4)+24*ay0*pow(vx0,3)*pow(vy0,3)+72*by0*pow(vx0,4)*pow(vy0,2)+12*ay0*pow(vx0,5)*vy0+18*by0*pow(vx0,6))*pow(vy2,3)+(24*ay0*vx2*pow(vy0,6)+24*by0*vx0*vx2*pow(vy0,5)+78*ay0*pow(vx0,2)*vx2*pow(vy0,4)+48*by0*pow(vx0,3)*vx2*pow(vy0,3)+84*ay0*pow(vx0,4)*vx2*pow(vy0,2)+24*by0*pow(vx0,5)*vx2*vy0+30*ay0*pow(vx0,6)*vx2)*pow(vy2,2)+(30*by0*pow(vx2,2)*pow(vy0,6)+24*ay0*vx0*pow(vx2,2)*pow(vy0,5)+84*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+48*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+78*by0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+24*ay0*pow(vx0,5)*pow(vx2,2)*vy0+24*by0*pow(vx0,6)*pow(vx2,2))*vy2+18*ay0*pow(vx2,3)*pow(vy0,6)+12*by0*vx0*pow(vx2,3)*pow(vy0,5)+72*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+24*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+90*ay0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+12*by0*pow(vx0,5)*pow(vx2,3)*vy0+36*ay0*pow(vx0,6)*pow(vx2,3))*vz0+(-36*ny0*pow(vy0,7)-102*ny0*pow(vx0,2)*pow(vy0,5)-96*ny0*pow(vx0,4)*pow(vy0,3)-30*ny0*pow(vx0,6)*vy0)*pow(vy2,3)+(-48*ny0*vx0*vx2*pow(vy0,6)-126*ny0*pow(vx0,3)*vx2*pow(vy0,4)-108*ny0*pow(vx0,5)*vx2*pow(vy0,2)-30*ny0*pow(vx0,7)*vx2)*pow(vy2,2)+(-30*ny0*pow(vx2,2)*pow(vy0,7)-108*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)-126*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)-48*ny0*pow(vx0,6)*pow(vx2,2)*vy0)*vy2-30*ny0*vx0*pow(vx2,3)*pow(vy0,6)-96*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-102*ny0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-36*ny0*pow(vx0,7)*pow(vx2,3))*pow(vz2,3)+(((72*by0*vy0+42*ay0*vx0)*pow(vy2,4)+(30*ay0*vx2*vy0+30*by0*vx0*vx2)*pow(vy2,3)+(114*by0*pow(vx2,2)*vy0+114*ay0*vx0*pow(vx2,2))*pow(vy2,2)+(30*ay0*pow(vx2,3)*vy0+30*by0*vx0*pow(vx2,3))*vy2+42*by0*pow(vx2,4)*vy0+72*ay0*vx0*pow(vx2,4))*pow(vz0,6)+((-72*ny0*pow(vy0,2)-42*ny0*pow(vx0,2))*pow(vy2,4)-60*ny0*vx0*vx2*vy0*pow(vy2,3)+(-114*ny0*pow(vx2,2)*pow(vy0,2)-114*ny0*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-60*ny0*vx0*pow(vx2,3)*vy0*vy2-42*ny0*pow(vx2,4)*pow(vy0,2)-72*ny0*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((156*by0*pow(vy0,3)+90*ay0*vx0*pow(vy0,2)+138*by0*pow(vx0,2)*vy0+72*ay0*pow(vx0,3))*pow(vy2,4)+(66*ay0*vx2*pow(vy0,3)+102*by0*vx0*vx2*pow(vy0,2)+102*ay0*pow(vx0,2)*vx2*vy0+66*by0*pow(vx0,3)*vx2)*pow(vy2,3)+(228*by0*pow(vx2,2)*pow(vy0,3)+228*ay0*vx0*pow(vx2,2)*pow(vy0,2)+228*by0*pow(vx0,2)*pow(vx2,2)*vy0+228*ay0*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(66*ay0*pow(vx2,3)*pow(vy0,3)+102*by0*vx0*pow(vx2,3)*pow(vy0,2)+102*ay0*pow(vx0,2)*pow(vx2,3)*vy0+66*by0*pow(vx0,3)*pow(vx2,3))*vy2+72*by0*pow(vx2,4)*pow(vy0,3)+138*ay0*vx0*pow(vx2,4)*pow(vy0,2)+90*by0*pow(vx0,2)*pow(vx2,4)*vy0+156*ay0*pow(vx0,3)*pow(vx2,4))*pow(vz0,4)+((-156*ny0*pow(vy0,4)-228*ny0*pow(vx0,2)*pow(vy0,2)-72*ny0*pow(vx0,4))*pow(vy2,4)+(-168*ny0*vx0*vx2*pow(vy0,3)-168*ny0*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-228*ny0*pow(vx2,2)*pow(vy0,4)-456*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-228*ny0*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-168*ny0*vx0*pow(vx2,3)*pow(vy0,3)-168*ny0*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-72*ny0*pow(vx2,4)*pow(vy0,4)-228*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-156*ny0*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((84*by0*pow(vy0,5)+42*ay0*vx0*pow(vy0,4)+156*by0*pow(vx0,2)*pow(vy0,3)+72*ay0*pow(vx0,3)*pow(vy0,2)+72*by0*pow(vx0,4)*vy0+30*ay0*pow(vx0,5))*pow(vy2,4)+(42*ay0*vx2*pow(vy0,5)+66*by0*vx0*vx2*pow(vy0,4)+108*ay0*pow(vx0,2)*vx2*pow(vy0,3)+108*by0*pow(vx0,3)*vx2*pow(vy0,2)+66*ay0*pow(vx0,4)*vx2*vy0+42*by0*pow(vx0,5)*vx2)*pow(vy2,3)+(114*by0*pow(vx2,2)*pow(vy0,5)+114*ay0*vx0*pow(vx2,2)*pow(vy0,4)+228*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+228*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+114*by0*pow(vx0,4)*pow(vx2,2)*vy0+114*ay0*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(42*ay0*pow(vx2,3)*pow(vy0,5)+66*by0*vx0*pow(vx2,3)*pow(vy0,4)+108*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+108*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+66*ay0*pow(vx0,4)*pow(vx2,3)*vy0+42*by0*pow(vx0,5)*pow(vx2,3))*vy2+30*by0*pow(vx2,4)*pow(vy0,5)+72*ay0*vx0*pow(vx2,4)*pow(vy0,4)+72*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+156*ay0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+42*by0*pow(vx0,4)*pow(vx2,4)*vy0+84*ay0*pow(vx0,5)*pow(vx2,4))*pow(vz0,2)+((-84*ny0*pow(vy0,6)-198*ny0*pow(vx0,2)*pow(vy0,4)-144*ny0*pow(vx0,4)*pow(vy0,2)-30*ny0*pow(vx0,6))*pow(vy2,4)+(-108*ny0*vx0*vx2*pow(vy0,5)-216*ny0*pow(vx0,3)*vx2*pow(vy0,3)-108*ny0*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(-114*ny0*pow(vx2,2)*pow(vy0,6)-342*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-342*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-114*ny0*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(-108*ny0*vx0*pow(vx2,3)*pow(vy0,5)-216*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-108*ny0*pow(vx0,5)*pow(vx2,3)*vy0)*vy2-30*ny0*pow(vx2,4)*pow(vy0,6)-144*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-198*ny0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-84*ny0*pow(vx0,6)*pow(vx2,4))*vz0+(-6*ay0*vx0*pow(vy0,6)+6*by0*pow(vx0,2)*pow(vy0,5)-12*ay0*pow(vx0,3)*pow(vy0,4)+12*by0*pow(vx0,4)*pow(vy0,3)-6*ay0*pow(vx0,5)*pow(vy0,2)+6*by0*pow(vx0,6)*vy0)*pow(vy2,4)+(6*ay0*vx2*pow(vy0,7)-6*by0*vx0*vx2*pow(vy0,6)+6*ay0*pow(vx0,2)*vx2*pow(vy0,5)-6*by0*pow(vx0,3)*vx2*pow(vy0,4)-6*ay0*pow(vx0,4)*vx2*pow(vy0,3)+6*by0*pow(vx0,5)*vx2*pow(vy0,2)-6*ay0*pow(vx0,6)*vx2*vy0+6*by0*pow(vx0,7)*vx2)*pow(vy2,3)+(6*ay0*pow(vx2,3)*pow(vy0,7)-6*by0*vx0*pow(vx2,3)*pow(vy0,6)+6*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,5)-6*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-6*ay0*pow(vx0,4)*pow(vx2,3)*pow(vy0,3)+6*by0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-6*ay0*pow(vx0,6)*pow(vx2,3)*vy0+6*by0*pow(vx0,7)*pow(vx2,3))*vy2+6*ay0*vx0*pow(vx2,4)*pow(vy0,6)-6*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+12*ay0*pow(vx0,3)*pow(vx2,4)*pow(vy0,4)-12*by0*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+6*ay0*pow(vx0,5)*pow(vx2,4)*pow(vy0,2)-6*by0*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vz2,2)+((12*by0*pow(vy2,5)+12*ay0*vx2*pow(vy2,4)+24*by0*pow(vx2,2)*pow(vy2,3)+24*ay0*pow(vx2,3)*pow(vy2,2)+12*by0*pow(vx2,4)*vy2+12*ay0*pow(vx2,5))*pow(vz0,7)+(-12*ny0*vy0*pow(vy2,5)-12*ny0*vx0*vx2*pow(vy2,4)-24*ny0*pow(vx2,2)*vy0*pow(vy2,3)-24*ny0*vx0*pow(vx2,3)*pow(vy2,2)-12*ny0*pow(vx2,4)*vy0*vy2-12*ny0*vx0*pow(vx2,5))*pow(vz0,6)+((72*by0*pow(vy0,2)+36*ay0*vx0*vy0+30*by0*pow(vx0,2))*pow(vy2,5)+(36*ay0*vx2*pow(vy0,2)+48*by0*vx0*vx2*vy0+66*ay0*pow(vx0,2)*vx2)*pow(vy2,4)+(138*by0*pow(vx2,2)*pow(vy0,2)+84*ay0*vx0*pow(vx2,2)*vy0+66*by0*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(66*ay0*pow(vx2,3)*pow(vy0,2)+84*by0*vx0*pow(vx2,3)*vy0+138*ay0*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(66*by0*pow(vx2,4)*pow(vy0,2)+48*ay0*vx0*pow(vx2,4)*vy0+36*by0*pow(vx0,2)*pow(vx2,4))*vy2+30*ay0*pow(vx2,5)*pow(vy0,2)+36*by0*vx0*pow(vx2,5)*vy0+72*ay0*pow(vx0,2)*pow(vx2,5))*pow(vz0,5)+((-72*ny0*pow(vy0,3)-66*ny0*pow(vx0,2)*vy0)*pow(vy2,5)+(-84*ny0*vx0*vx2*pow(vy0,2)-66*ny0*pow(vx0,3)*vx2)*pow(vy2,4)+(-138*ny0*pow(vx2,2)*pow(vy0,3)-150*ny0*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-150*ny0*vx0*pow(vx2,3)*pow(vy0,2)-138*ny0*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-66*ny0*pow(vx2,4)*pow(vy0,3)-84*ny0*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-66*ny0*vx0*pow(vx2,5)*pow(vy0,2)-72*ny0*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((84*by0*pow(vy0,4)+48*ay0*vx0*pow(vy0,3)+108*by0*pow(vx0,2)*pow(vy0,2)+48*ay0*pow(vx0,3)*vy0+24*by0*pow(vx0,4))*pow(vy2,5)+(36*ay0*vx2*pow(vy0,4)+72*by0*vx0*vx2*pow(vy0,3)+108*ay0*pow(vx0,2)*vx2*pow(vy0,2)+72*by0*pow(vx0,3)*vx2*vy0+72*ay0*pow(vx0,4)*vx2)*pow(vy2,4)+(156*by0*pow(vx2,2)*pow(vy0,4)+120*ay0*vx0*pow(vx2,2)*pow(vy0,3)+216*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+120*ay0*pow(vx0,3)*pow(vx2,2)*vy0+60*by0*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(60*ay0*pow(vx2,3)*pow(vy0,4)+120*by0*vx0*pow(vx2,3)*pow(vy0,3)+216*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+120*by0*pow(vx0,3)*pow(vx2,3)*vy0+156*ay0*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(72*by0*pow(vx2,4)*pow(vy0,4)+72*ay0*vx0*pow(vx2,4)*pow(vy0,3)+108*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+72*ay0*pow(vx0,3)*pow(vx2,4)*vy0+36*by0*pow(vx0,4)*pow(vx2,4))*vy2+24*ay0*pow(vx2,5)*pow(vy0,4)+48*by0*vx0*pow(vx2,5)*pow(vy0,3)+108*ay0*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+48*by0*pow(vx0,3)*pow(vx2,5)*vy0+84*ay0*pow(vx0,4)*pow(vx2,5))*pow(vz0,3)+((-84*ny0*pow(vy0,5)-156*ny0*pow(vx0,2)*pow(vy0,3)-72*ny0*pow(vx0,4)*vy0)*pow(vy2,5)+(-108*ny0*vx0*vx2*pow(vy0,4)-180*ny0*pow(vx0,3)*vx2*pow(vy0,2)-72*ny0*pow(vx0,5)*vx2)*pow(vy2,4)+(-156*ny0*pow(vx2,2)*pow(vy0,5)-336*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-180*ny0*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-180*ny0*vx0*pow(vx2,3)*pow(vy0,4)-336*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-156*ny0*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-72*ny0*pow(vx2,4)*pow(vy0,5)-180*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-108*ny0*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-72*ny0*vx0*pow(vx2,5)*pow(vy0,4)-156*ny0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-84*ny0*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+((24*by0*pow(vy0,6)+12*ay0*vx0*pow(vy0,5)+54*by0*pow(vx0,2)*pow(vy0,4)+24*ay0*pow(vx0,3)*pow(vy0,3)+36*by0*pow(vx0,4)*pow(vy0,2)+12*ay0*pow(vx0,5)*vy0+6*by0*pow(vx0,6))*pow(vy2,5)+(12*ay0*vx2*pow(vy0,6)+24*by0*vx0*vx2*pow(vy0,5)+42*ay0*pow(vx0,2)*vx2*pow(vy0,4)+48*by0*pow(vx0,3)*vx2*pow(vy0,3)+48*ay0*pow(vx0,4)*vx2*pow(vy0,2)+24*by0*pow(vx0,5)*vx2*vy0+18*ay0*pow(vx0,6)*vx2)*pow(vy2,4)+(42*by0*pow(vx2,2)*pow(vy0,6)+36*ay0*vx0*pow(vx2,2)*pow(vy0,5)+102*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+72*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+78*by0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+36*ay0*pow(vx0,5)*pow(vx2,2)*vy0+18*by0*pow(vx0,6)*pow(vx2,2))*pow(vy2,3)+(18*ay0*pow(vx2,3)*pow(vy0,6)+36*by0*vx0*pow(vx2,3)*pow(vy0,5)+78*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+72*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+102*ay0*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+36*by0*pow(vx0,5)*pow(vx2,3)*vy0+42*ay0*pow(vx0,6)*pow(vx2,3))*pow(vy2,2)+(18*by0*pow(vx2,4)*pow(vy0,6)+24*ay0*vx0*pow(vx2,4)*pow(vy0,5)+48*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+48*ay0*pow(vx0,3)*pow(vx2,4)*pow(vy0,3)+42*by0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+24*ay0*pow(vx0,5)*pow(vx2,4)*vy0+12*by0*pow(vx0,6)*pow(vx2,4))*vy2+6*ay0*pow(vx2,5)*pow(vy0,6)+12*by0*vx0*pow(vx2,5)*pow(vy0,5)+36*ay0*pow(vx0,2)*pow(vx2,5)*pow(vy0,4)+24*by0*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+54*ay0*pow(vx0,4)*pow(vx2,5)*pow(vy0,2)+12*by0*pow(vx0,5)*pow(vx2,5)*vy0+24*ay0*pow(vx0,6)*pow(vx2,5))*vz0+(-24*ny0*pow(vy0,7)-66*ny0*pow(vx0,2)*pow(vy0,5)-60*ny0*pow(vx0,4)*pow(vy0,3)-18*ny0*pow(vx0,6)*vy0)*pow(vy2,5)+(-36*ny0*vx0*vx2*pow(vy0,6)-90*ny0*pow(vx0,3)*vx2*pow(vy0,4)-72*ny0*pow(vx0,5)*vx2*pow(vy0,2)-18*ny0*pow(vx0,7)*vx2)*pow(vy2,4)+(-42*ny0*pow(vx2,2)*pow(vy0,7)-138*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)-150*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)-54*ny0*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(-54*ny0*vx0*pow(vx2,3)*pow(vy0,6)-150*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)-138*ny0*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)-42*ny0*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(-18*ny0*pow(vx2,4)*pow(vy0,7)-72*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)-90*ny0*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)-36*ny0*pow(vx0,6)*pow(vx2,4)*vy0)*vy2-18*ny0*vx0*pow(vx2,5)*pow(vy0,6)-60*ny0*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)-66*ny0*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)-24*ny0*pow(vx0,7)*pow(vx2,5))*vz2+((12*by0*vy0+6*ay0*vx0)*pow(vy2,6)+(6*ay0*vx2*vy0+6*by0*vx0*vx2)*pow(vy2,5)+(30*by0*pow(vx2,2)*vy0+24*ay0*vx0*pow(vx2,2))*pow(vy2,4)+(12*ay0*pow(vx2,3)*vy0+12*by0*vx0*pow(vx2,3))*pow(vy2,3)+(24*by0*pow(vx2,4)*vy0+30*ay0*vx0*pow(vx2,4))*pow(vy2,2)+(6*ay0*pow(vx2,5)*vy0+6*by0*vx0*pow(vx2,5))*vy2+6*by0*pow(vx2,6)*vy0+12*ay0*vx0*pow(vx2,6))*pow(vz0,6)+((-12*ny0*pow(vy0,2)-6*ny0*pow(vx0,2))*pow(vy2,6)-12*ny0*vx0*vx2*vy0*pow(vy2,5)+(-30*ny0*pow(vx2,2)*pow(vy0,2)-24*ny0*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)-24*ny0*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(-24*ny0*pow(vx2,4)*pow(vy0,2)-30*ny0*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)-12*ny0*vx0*pow(vx2,5)*vy0*vy2-6*ny0*pow(vx2,6)*pow(vy0,2)-12*ny0*pow(vx0,2)*pow(vx2,6))*pow(vz0,5)+((36*by0*pow(vy0,3)+24*ay0*vx0*pow(vy0,2)+24*by0*pow(vx0,2)*vy0+12*ay0*pow(vx0,3))*pow(vy2,6)+(12*ay0*vx2*pow(vy0,3)+36*by0*vx0*vx2*pow(vy0,2)+36*ay0*pow(vx0,2)*vx2*vy0+12*by0*pow(vx0,3)*vx2)*pow(vy2,5)+(84*by0*pow(vx2,2)*pow(vy0,3)+72*ay0*vx0*pow(vx2,2)*pow(vy0,2)+72*by0*pow(vx0,2)*pow(vx2,2)*vy0+60*ay0*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+(24*ay0*pow(vx2,3)*pow(vy0,3)+72*by0*vx0*pow(vx2,3)*pow(vy0,2)+72*ay0*pow(vx0,2)*pow(vx2,3)*vy0+24*by0*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(60*by0*pow(vx2,4)*pow(vy0,3)+72*ay0*vx0*pow(vx2,4)*pow(vy0,2)+72*by0*pow(vx0,2)*pow(vx2,4)*vy0+84*ay0*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+(12*ay0*pow(vx2,5)*pow(vy0,3)+36*by0*vx0*pow(vx2,5)*pow(vy0,2)+36*ay0*pow(vx0,2)*pow(vx2,5)*vy0+12*by0*pow(vx0,3)*pow(vx2,5))*vy2+12*by0*pow(vx2,6)*pow(vy0,3)+24*ay0*vx0*pow(vx2,6)*pow(vy0,2)+24*by0*pow(vx0,2)*pow(vx2,6)*vy0+36*ay0*pow(vx0,3)*pow(vx2,6))*pow(vz0,4)+((-36*ny0*pow(vy0,4)-48*ny0*pow(vx0,2)*pow(vy0,2)-12*ny0*pow(vx0,4))*pow(vy2,6)+(-48*ny0*vx0*vx2*pow(vy0,3)-48*ny0*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(-84*ny0*pow(vx2,2)*pow(vy0,4)-144*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-60*ny0*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(-96*ny0*vx0*pow(vx2,3)*pow(vy0,3)-96*ny0*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(-60*ny0*pow(vx2,4)*pow(vy0,4)-144*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-84*ny0*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(-48*ny0*vx0*pow(vx2,5)*pow(vy0,3)-48*ny0*pow(vx0,3)*pow(vx2,5)*vy0)*vy2-12*ny0*pow(vx2,6)*pow(vy0,4)-48*ny0*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)-36*ny0*pow(vx0,4)*pow(vx2,6))*pow(vz0,3)+((24*by0*pow(vy0,5)+18*ay0*vx0*pow(vy0,4)+36*by0*pow(vx0,2)*pow(vy0,3)+24*ay0*pow(vx0,3)*pow(vy0,2)+12*by0*pow(vx0,4)*vy0+6*ay0*pow(vx0,5))*pow(vy2,6)+(6*ay0*vx2*pow(vy0,5)+30*by0*vx0*vx2*pow(vy0,4)+36*ay0*pow(vx0,2)*vx2*pow(vy0,3)+36*by0*pow(vx0,3)*vx2*pow(vy0,2)+30*ay0*pow(vx0,4)*vx2*vy0+6*by0*pow(vx0,5)*vx2)*pow(vy2,5)+(54*by0*pow(vx2,2)*pow(vy0,5)+48*ay0*vx0*pow(vx2,2)*pow(vy0,4)+96*by0*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+84*ay0*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+42*by0*pow(vx0,4)*pow(vx2,2)*vy0+36*ay0*pow(vx0,5)*pow(vx2,2))*pow(vy2,4)+(12*ay0*pow(vx2,3)*pow(vy0,5)+60*by0*vx0*pow(vx2,3)*pow(vy0,4)+72*ay0*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+72*by0*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+60*ay0*pow(vx0,4)*pow(vx2,3)*vy0+12*by0*pow(vx0,5)*pow(vx2,3))*pow(vy2,3)+(36*by0*pow(vx2,4)*pow(vy0,5)+42*ay0*vx0*pow(vx2,4)*pow(vy0,4)+84*by0*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+96*ay0*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+48*by0*pow(vx0,4)*pow(vx2,4)*vy0+54*ay0*pow(vx0,5)*pow(vx2,4))*pow(vy2,2)+(6*ay0*pow(vx2,5)*pow(vy0,5)+30*by0*vx0*pow(vx2,5)*pow(vy0,4)+36*ay0*pow(vx0,2)*pow(vx2,5)*pow(vy0,3)+36*by0*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+30*ay0*pow(vx0,4)*pow(vx2,5)*vy0+6*by0*pow(vx0,5)*pow(vx2,5))*vy2+6*by0*pow(vx2,6)*pow(vy0,5)+12*ay0*vx0*pow(vx2,6)*pow(vy0,4)+24*by0*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+36*ay0*pow(vx0,3)*pow(vx2,6)*pow(vy0,2)+18*by0*pow(vx0,4)*pow(vx2,6)*vy0+24*ay0*pow(vx0,5)*pow(vx2,6))*pow(vz0,2)+((-24*ny0*pow(vy0,6)-54*ny0*pow(vx0,2)*pow(vy0,4)-36*ny0*pow(vx0,4)*pow(vy0,2)-6*ny0*pow(vx0,6))*pow(vy2,6)+(-36*ny0*vx0*vx2*pow(vy0,5)-72*ny0*pow(vx0,3)*vx2*pow(vy0,3)-36*ny0*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(-54*ny0*pow(vx2,2)*pow(vy0,6)-144*ny0*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-126*ny0*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-36*ny0*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(-72*ny0*vx0*pow(vx2,3)*pow(vy0,5)-144*ny0*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-72*ny0*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(-36*ny0*pow(vx2,4)*pow(vy0,6)-126*ny0*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-144*ny0*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-54*ny0*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(-36*ny0*vx0*pow(vx2,5)*pow(vy0,5)-72*ny0*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)-36*ny0*pow(vx0,5)*pow(vx2,5)*vy0)*vy2-6*ny0*pow(vx2,6)*pow(vy0,6)-36*ny0*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)-54*ny0*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)-24*ny0*pow(vx0,6)*pow(vx2,6))*vz0)*pow(((160*pow(roughness,2)*pow(vz0,7)+(400*pow(roughness,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2))*pow(vz0,5)+(320*pow(roughness,2)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,6)+((720*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,7)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((240*pow(roughness,2)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,2))*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+640*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((1040*pow(roughness,2)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+560*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((400*pow(roughness,2)*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+160*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((400*pow(roughness,2)*vy0*pow(vy2,3)+400*pow(roughness,2)*vx0*vx2*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2)*vy0*vy2+400*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,6)+((1040*pow(roughness,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+(240*pow(roughness,2)*pow(vy0,7)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+160*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*vx2*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(160*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4))*pow(vz0,7)+((480*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((720*pow(roughness,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,5)+80*pow(roughness,2)*vx0*vx2*pow(vy2,4)+160*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+160*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,4)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,6)+((320*pow(roughness,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,7)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(400*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,8)+(480*pow(roughness,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2))*pow(vz0,6)+(500*pow(roughness,2)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4))*pow(vz0,4)+(200*pow(roughness,2)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,8)+80*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+120*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+80*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vz2,6)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+880*pow(roughness,2)*vx0*vx2*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,5)+((800*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*vz0)*pow(vz2,5)+((320*pow(roughness,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,2))*pow(vz0,8)+((1240*pow(roughness,2)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+((1700*pow(roughness,2)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+1100*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+((960*pow(roughness,2)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+480*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz0,2)+(180*pow(roughness,2)*pow(vy0,8)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,8))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*vy2+60*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vz2,4)+((560*pow(roughness,2)*vy0*pow(vy2,3)+560*pow(roughness,2)*vx0*vx2*pow(vy2,2)+560*pow(roughness,2)*pow(vx2,2)*vy0*vy2+560*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,7)+((1680*pow(roughness,2)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,5)+((1680*pow(roughness,2)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,4)+3520*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(1520*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+1520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,3)+((560*pow(roughness,2)*pow(vy0,7)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*vz0)*pow(vz2,3)+((180*pow(roughness,2)*pow(vy2,4)+360*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+180*pow(roughness,2)*pow(vx2,4))*pow(vz0,8)+((960*pow(roughness,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+720*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(1560*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+600*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+((1700*pow(roughness,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(2400*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+720*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,4)+((1240*pow(roughness,2)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(1560*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+1560*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+360*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,8)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1100*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,8))*pow(vy2,4)+(560*pow(roughness,2)*vx0*vx2*pow(vy0,7)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*vx2*pow(vy2,4)+480*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,4)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,7)+((800*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(1520*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,5)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1840*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,6)+60*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+60*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,6))*pow(vz0,8)+((200*pow(roughness,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+480*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,6)+((500*pow(roughness,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(1100*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(1520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,6)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+200*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vy2,6)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,5)+(320*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,4)+(560*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,3)+(180*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1240*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,8)*pow(vx2,6),-1) ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
// x1  :  
poly.add(Term6f( (((24*vx0*pow(vz0,5)+(24*vx0*pow(vy0,2)+24*pow(vx0,3))*pow(vz0,3)+(3*vx0*pow(vy0,4)+6*pow(vx0,3)*pow(vy0,2)+3*pow(vx0,5))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,7)+((24*vx2+24*vx0)*pow(vz0,6)+(24*vx0*vy0*vy2+(48*vx2+48*vx0)*pow(vy0,2)+72*pow(vx0,2)*vx2+48*pow(vx0,3))*pow(vz0,4)+((12*vx0*pow(vy0,3)+12*pow(vx0,3)*vy0)*vy2+(27*vx2+27*vx0)*pow(vy0,4)+(66*pow(vx0,2)*vx2+54*pow(vx0,3))*pow(vy0,2)+39*pow(vx0,4)*vx2+27*pow(vx0,5))*pow(vz0,2)+(-3*vx0*pow(vy0,5)-6*pow(vx0,3)*pow(vy0,3)-3*pow(vx0,5)*vy0)*vy2+(3*vx2+3*vx0)*pow(vy0,6)+(6*pow(vx0,2)*vx2+9*pow(vx0,3))*pow(vy0,4)+(3*pow(vx0,4)*vx2+9*pow(vx0,5))*pow(vy0,2)+3*pow(vx0,7))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+(24*vx2*pow(vz0,7)+(72*vx0*pow(vy2,2)+(24*vx2+24*vx0)*vy0*vy2+72*vx2*pow(vy0,2)+96*vx0*pow(vx2,2)+96*pow(vx0,2)*vx2)*pow(vz0,5)+((84*vx0*pow(vy0,2)+78*pow(vx0,3))*pow(vy2,2)+((42*vx2+42*vx0)*pow(vy0,3)+(54*pow(vx0,2)*vx2+42*pow(vx0,3))*vy0)*vy2+69*vx2*pow(vy0,4)+(120*vx0*pow(vx2,2)+180*pow(vx0,2)*vx2)*pow(vy0,2)+126*pow(vx0,3)*pow(vx2,2)+111*pow(vx0,4)*vx2)*pow(vz0,3)+((9*vx0*pow(vy0,4)+21*pow(vx0,3)*pow(vy0,2)+12*pow(vx0,5))*pow(vy2,2)+((18*vx2+18*vx0)*pow(vy0,5)+(30*pow(vx0,2)*vx2+36*pow(vx0,3))*pow(vy0,3)+(12*pow(vx0,4)*vx2+18*pow(vx0,5))*vy0)*vy2+21*vx2*pow(vy0,6)+(30*vx0*pow(vx2,2)+81*pow(vx0,2)*vx2)*pow(vy0,4)+(57*pow(vx0,3)*pow(vx2,2)+99*pow(vx0,4)*vx2)*pow(vy0,2)+27*pow(vx0,5)*pow(vx2,2)+39*pow(vx0,6)*vx2)*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+(((48*vx2+48*vx0)*pow(vy2,2)+24*vx2*vy0*vy2+48*pow(vx2,3)+72*vx0*pow(vx2,2))*pow(vz0,6)+(72*vx0*vy0*pow(vy2,3)+((102*vx2+102*vx0)*pow(vy0,2)+168*pow(vx0,2)*vx2+96*pow(vx0,3))*pow(vy2,2)+(72*vx2*pow(vy0,3)+(84*vx0*pow(vx2,2)+84*pow(vx0,2)*vx2)*vy0)*vy2+(96*pow(vx2,3)+168*vx0*pow(vx2,2))*pow(vy0,2)+174*pow(vx0,2)*pow(vx2,3)+174*pow(vx0,3)*pow(vx2,2))*pow(vz0,4)+((48*vx0*pow(vy0,3)+48*pow(vx0,3)*vy0)*pow(vy2,3)+((63*vx2+63*vx0)*pow(vy0,4)+(165*pow(vx0,2)*vx2+117*pow(vx0,3))*pow(vy0,2)+102*pow(vx0,4)*vx2+54*pow(vx0,5))*pow(vy2,2)+(63*vx2*pow(vy0,5)+(66*vx0*pow(vx2,2)+144*pow(vx0,2)*vx2)*pow(vy0,3)+(66*pow(vx0,3)*pow(vx2,2)+81*pow(vx0,4)*vx2)*vy0)*vy2+(54*pow(vx2,3)+117*vx0*pow(vx2,2))*pow(vy0,4)+(165*pow(vx0,2)*pow(vx2,3)+243*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+111*pow(vx0,4)*pow(vx2,3)+126*pow(vx0,5)*pow(vx2,2))*pow(vz0,2)+(-9*vx0*pow(vy0,5)-15*pow(vx0,3)*pow(vy0,3)-6*pow(vx0,5)*vy0)*pow(vy2,3)+((9*vx2+9*vx0)*pow(vy0,6)+(9*pow(vx0,2)*vx2+24*pow(vx0,3))*pow(vy0,4)+21*pow(vx0,5)*pow(vy0,2)+6*pow(vx0,7))*pow(vy2,2)+(15*vx2*pow(vy0,7)+51*pow(vx0,2)*vx2*pow(vy0,5)+(57*pow(vx0,4)*vx2-9*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(21*pow(vx0,6)*vx2-9*pow(vx0,5)*pow(vx2,2))*vy0)*vy2+(6*pow(vx2,3)+21*vx0*pow(vx2,2))*pow(vy0,6)+(15*pow(vx0,2)*pow(vx2,3)+66*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(9*pow(vx0,4)*pow(vx2,3)+69*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+24*pow(vx0,7)*pow(vx2,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((24*vx2*pow(vy2,2)+24*pow(vx2,3))*pow(vz0,7)+(69*vx0*pow(vy2,4)+(42*vx2+42*vx0)*vy0*pow(vy2,3)+(84*vx2*pow(vy0,2)+180*vx0*pow(vx2,2)+120*pow(vx0,2)*vx2)*pow(vy2,2)+(42*pow(vx2,3)+54*vx0*pow(vx2,2))*vy0*vy2+78*pow(vx2,3)*pow(vy0,2)+111*vx0*pow(vx2,4)+126*pow(vx0,2)*pow(vx2,3))*pow(vz0,5)+((102*vx0*pow(vy0,2)+81*pow(vx0,3))*pow(vy2,4)+((72*vx2+72*vx0)*pow(vy0,3)+(114*pow(vx0,2)*vx2+72*pow(vx0,3))*vy0)*pow(vy2,3)+(102*vx2*pow(vy0,4)+(255*vx0*pow(vx2,2)+255*pow(vx0,2)*vx2)*pow(vy0,2)+255*pow(vx0,3)*pow(vx2,2)+153*pow(vx0,4)*vx2)*pow(vy2,2)+((72*pow(vx2,3)+114*vx0*pow(vx2,2))*pow(vy0,3)+(114*pow(vx0,2)*pow(vx2,3)+114*pow(vx0,3)*pow(vx2,2))*vy0)*vy2+81*pow(vx2,3)*pow(vy0,4)+(153*vx0*pow(vx2,4)+255*pow(vx0,2)*pow(vx2,3))*pow(vy0,2)+174*pow(vx0,3)*pow(vx2,4)+174*pow(vx0,4)*pow(vx2,3))*pow(vz0,3)+((18*vx0*pow(vy0,4)+33*pow(vx0,3)*pow(vy0,2)+15*pow(vx0,5))*pow(vy2,4)+((30*vx2+30*vx0)*pow(vy0,5)+(66*pow(vx0,2)*vx2+60*pow(vx0,3))*pow(vy0,3)+(36*pow(vx0,4)*vx2+30*pow(vx0,5))*vy0)*pow(vy2,3)+(42*vx2*pow(vy0,6)+(63*vx0*pow(vx2,2)+141*pow(vx0,2)*vx2)*pow(vy0,4)+(126*pow(vx0,3)*pow(vx2,2)+156*pow(vx0,4)*vx2)*pow(vy0,2)+63*pow(vx0,5)*pow(vx2,2)+57*pow(vx0,6)*vx2)*pow(vy2,2)+((30*pow(vx2,3)+60*vx0*pow(vx2,2))*pow(vy0,5)+(66*pow(vx0,2)*pow(vx2,3)+120*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(36*pow(vx0,4)*pow(vx2,3)+60*pow(vx0,5)*pow(vx2,2))*vy0)*vy2+27*pow(vx2,3)*pow(vy0,6)+(45*vx0*pow(vx2,4)+126*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+(93*pow(vx0,3)*pow(vx2,4)+171*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+48*pow(vx0,5)*pow(vx2,4)+72*pow(vx0,6)*pow(vx2,3))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+(((27*vx2+27*vx0)*pow(vy2,4)+12*vx2*vy0*pow(vy2,3)+(54*pow(vx2,3)+66*vx0*pow(vx2,2))*pow(vy2,2)+12*pow(vx2,3)*vy0*vy2+27*pow(vx2,5)+39*vx0*pow(vx2,4))*pow(vz0,6)+(63*vx0*vy0*pow(vy2,5)+((63*vx2+63*vx0)*pow(vy0,2)+117*pow(vx0,2)*vx2+54*pow(vx0,3))*pow(vy2,4)+(48*vx2*pow(vy0,3)+(144*vx0*pow(vx2,2)+66*pow(vx0,2)*vx2)*vy0)*pow(vy2,3)+((117*pow(vx2,3)+165*vx0*pow(vx2,2))*pow(vy0,2)+243*pow(vx0,2)*pow(vx2,3)+165*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(48*pow(vx2,3)*pow(vy0,3)+(81*vx0*pow(vx2,4)+66*pow(vx0,2)*pow(vx2,3))*vy0)*vy2+(54*pow(vx2,5)+102*vx0*pow(vx2,4))*pow(vy0,2)+126*pow(vx0,2)*pow(vx2,5)+111*pow(vx0,3)*pow(vx2,4))*pow(vz0,4)+((54*vx0*pow(vy0,3)+51*pow(vx0,3)*vy0)*pow(vy2,5)+((42*vx2+42*vx0)*pow(vy0,4)+(132*pow(vx0,2)*vx2+72*pow(vx0,3))*pow(vy0,2)+81*pow(vx0,4)*vx2+30*pow(vx0,5))*pow(vy2,4)+(54*vx2*pow(vy0,5)+(129*vx0*pow(vx2,2)+129*pow(vx0,2)*vx2)*pow(vy0,3)+(135*pow(vx0,3)*pow(vx2,2)+75*pow(vx0,4)*vx2)*vy0)*pow(vy2,3)+((72*pow(vx2,3)+132*vx0*pow(vx2,2))*pow(vy0,4)+(255*pow(vx0,2)*pow(vx2,3)+255*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+177*pow(vx0,4)*pow(vx2,3)+123*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(51*pow(vx2,3)*pow(vy0,5)+(75*vx0*pow(vx2,4)+135*pow(vx0,2)*pow(vx2,3))*pow(vy0,3)+(84*pow(vx0,3)*pow(vx2,4)+84*pow(vx0,4)*pow(vx2,3))*vy0)*vy2+(30*pow(vx2,5)+81*vx0*pow(vx2,4))*pow(vy0,4)+(123*pow(vx0,2)*pow(vx2,5)+177*pow(vx0,3)*pow(vx2,4))*pow(vy0,2)+96*pow(vx0,4)*pow(vx2,5)+96*pow(vx0,5)*pow(vx2,4))*pow(vz0,2)+(-6*vx0*pow(vy0,5)-9*pow(vx0,3)*pow(vy0,3)-3*pow(vx0,5)*vy0)*pow(vy2,5)+((6*vx2+6*vx0)*pow(vy0,6)+(3*pow(vx0,2)*vx2+15*pow(vx0,3))*pow(vy0,4)+(12*pow(vx0,5)-3*pow(vx0,4)*vx2)*pow(vy0,2)+3*pow(vx0,7))*pow(vy2,4)+(18*vx2*pow(vy0,7)+(57*pow(vx0,2)*vx2-3*vx0*pow(vx2,2))*pow(vy0,5)+(60*pow(vx0,4)*vx2-12*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(21*pow(vx0,6)*vx2-9*pow(vx0,5)*pow(vx2,2))*vy0)*pow(vy2,3)+((9*pow(vx2,3)+33*vx0*pow(vx2,2))*pow(vy0,6)+(12*pow(vx0,2)*pow(vx2,3)+90*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(3*pow(vx0,4)*pow(vx2,3)+81*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+24*pow(vx0,7)*pow(vx2,2))*pow(vy2,2)+(15*pow(vx2,3)*pow(vy0,7)+(3*vx0*pow(vx2,4)+60*pow(vx0,2)*pow(vx2,3))*pow(vy0,5)+(75*pow(vx0,4)*pow(vx2,3)-3*pow(vx0,3)*pow(vx2,4))*pow(vy0,3)+(30*pow(vx0,6)*pow(vx2,3)-6*pow(vx0,5)*pow(vx2,4))*vy0)*vy2+(3*pow(vx2,5)+18*vx0*pow(vx2,4))*pow(vy0,6)+(9*pow(vx0,2)*pow(vx2,5)+60*pow(vx0,3)*pow(vx2,4))*pow(vy0,4)+(6*pow(vx0,4)*pow(vx2,5)+66*pow(vx0,5)*pow(vx2,4))*pow(vy0,2)+24*pow(vx0,7)*pow(vx2,4))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((3*vx2*pow(vy2,4)+6*pow(vx2,3)*pow(vy2,2)+3*pow(vx2,5))*pow(vz0,7)+(21*vx0*pow(vy2,6)+(18*vx2+18*vx0)*vy0*pow(vy2,5)+(9*vx2*pow(vy0,2)+81*vx0*pow(vx2,2)+30*pow(vx0,2)*vx2)*pow(vy2,4)+(36*pow(vx2,3)+30*vx0*pow(vx2,2))*vy0*pow(vy2,3)+(21*pow(vx2,3)*pow(vy0,2)+99*vx0*pow(vx2,4)+57*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(18*pow(vx2,5)+12*vx0*pow(vx2,4))*vy0*vy2+12*pow(vx2,5)*pow(vy0,2)+39*vx0*pow(vx2,6)+27*pow(vx0,2)*pow(vx2,5))*pow(vz0,5)+((42*vx0*pow(vy0,2)+27*pow(vx0,3))*pow(vy2,6)+((30*vx2+30*vx0)*pow(vy0,3)+(60*pow(vx0,2)*vx2+30*pow(vx0,3))*vy0)*pow(vy2,5)+(18*vx2*pow(vy0,4)+(141*vx0*pow(vx2,2)+63*pow(vx0,2)*vx2)*pow(vy0,2)+126*pow(vx0,3)*pow(vx2,2)+45*pow(vx0,4)*vx2)*pow(vy2,4)+((60*pow(vx2,3)+66*vx0*pow(vx2,2))*pow(vy0,3)+(120*pow(vx0,2)*pow(vx2,3)+66*pow(vx0,3)*pow(vx2,2))*vy0)*pow(vy2,3)+(33*pow(vx2,3)*pow(vy0,4)+(156*vx0*pow(vx2,4)+126*pow(vx0,2)*pow(vx2,3))*pow(vy0,2)+171*pow(vx0,3)*pow(vx2,4)+93*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+((30*pow(vx2,5)+36*vx0*pow(vx2,4))*pow(vy0,3)+(60*pow(vx0,2)*pow(vx2,5)+36*pow(vx0,3)*pow(vx2,4))*vy0)*vy2+15*pow(vx2,5)*pow(vy0,4)+(57*vx0*pow(vx2,6)+63*pow(vx0,2)*pow(vx2,5))*pow(vy0,2)+72*pow(vx0,3)*pow(vx2,6)+48*pow(vx0,4)*pow(vx2,5))*pow(vz0,3)+((12*vx0*pow(vy0,4)+18*pow(vx0,3)*pow(vy0,2)+6*pow(vx0,5))*pow(vy2,6)+((12*vx2+12*vx0)*pow(vy0,5)+(36*pow(vx0,2)*vx2+24*pow(vx0,3))*pow(vy0,3)+(24*pow(vx0,4)*vx2+12*pow(vx0,5))*vy0)*pow(vy2,5)+(12*vx2*pow(vy0,6)+(42*vx0*pow(vx2,2)+42*pow(vx0,2)*vx2)*pow(vy0,4)+(78*pow(vx0,3)*pow(vx2,2)+48*pow(vx0,4)*vx2)*pow(vy0,2)+36*pow(vx0,5)*pow(vx2,2)+18*pow(vx0,6)*vx2)*pow(vy2,4)+((24*pow(vx2,3)+36*vx0*pow(vx2,2))*pow(vy0,5)+(72*pow(vx0,2)*pow(vx2,3)+72*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(48*pow(vx0,4)*pow(vx2,3)+36*pow(vx0,5)*pow(vx2,2))*vy0)*pow(vy2,3)+(18*pow(vx2,3)*pow(vy0,6)+(48*vx0*pow(vx2,4)+78*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+(102*pow(vx0,3)*pow(vx2,4)+102*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+54*pow(vx0,5)*pow(vx2,4)+42*pow(vx0,6)*pow(vx2,3))*pow(vy2,2)+((12*pow(vx2,5)+24*vx0*pow(vx2,4))*pow(vy0,5)+(36*pow(vx0,2)*pow(vx2,5)+48*pow(vx0,3)*pow(vx2,4))*pow(vy0,3)+(24*pow(vx0,4)*pow(vx2,5)+24*pow(vx0,5)*pow(vx2,4))*vy0)*vy2+6*pow(vx2,5)*pow(vy0,6)+(18*vx0*pow(vx2,6)+36*pow(vx0,2)*pow(vx2,5))*pow(vy0,4)+(42*pow(vx0,3)*pow(vx2,6)+54*pow(vx0,4)*pow(vx2,5))*pow(vy0,2)+24*pow(vx0,5)*pow(vx2,6)+24*pow(vx0,6)*pow(vx2,5))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+(((3*vx2+3*vx0)*pow(vy2,6)-3*vx2*vy0*pow(vy2,5)+(9*pow(vx2,3)+6*vx0*pow(vx2,2))*pow(vy2,4)-6*pow(vx2,3)*vy0*pow(vy2,3)+(9*pow(vx2,5)+3*vx0*pow(vx2,4))*pow(vy2,2)-3*pow(vx2,5)*vy0*vy2+3*pow(vx2,7))*pow(vz0,6)+(15*vx0*vy0*pow(vy2,7)+((9*vx2+9*vx0)*pow(vy0,2)+21*pow(vx0,2)*vx2+6*pow(vx0,3))*pow(vy2,6)+(51*vx0*pow(vx2,2)*vy0-9*vx2*pow(vy0,3))*pow(vy2,5)+((24*pow(vx2,3)+9*vx0*pow(vx2,2))*pow(vy0,2)+66*pow(vx0,2)*pow(vx2,3)+15*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+((57*vx0*pow(vx2,4)-9*pow(vx0,2)*pow(vx2,3))*vy0-15*pow(vx2,3)*pow(vy0,3))*pow(vy2,3)+(21*pow(vx2,5)*pow(vy0,2)+69*pow(vx0,2)*pow(vx2,5)+9*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+((21*vx0*pow(vx2,6)-9*pow(vx0,2)*pow(vx2,5))*vy0-6*pow(vx2,5)*pow(vy0,3))*vy2+6*pow(vx2,7)*pow(vy0,2)+24*pow(vx0,2)*pow(vx2,7))*pow(vz0,4)+((18*vx0*pow(vy0,3)+15*pow(vx0,3)*vy0)*pow(vy2,7)+((6*vx2+6*vx0)*pow(vy0,4)+(33*pow(vx0,2)*vx2+9*pow(vx0,3))*pow(vy0,2)+18*pow(vx0,4)*vx2+3*pow(vx0,5))*pow(vy2,6)+(-6*vx2*pow(vy0,5)+(57*vx0*pow(vx2,2)-3*pow(vx0,2)*vx2)*pow(vy0,3)+(60*pow(vx0,3)*pow(vx2,2)+3*pow(vx0,4)*vx2)*vy0)*pow(vy2,5)+((15*pow(vx2,3)+3*vx0*pow(vx2,2))*pow(vy0,4)+(90*pow(vx0,2)*pow(vx2,3)+12*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+60*pow(vx0,4)*pow(vx2,3)+9*pow(vx0,5)*pow(vx2,2))*pow(vy2,4)+(-9*pow(vx2,3)*pow(vy0,5)+(60*vx0*pow(vx2,4)-12*pow(vx0,2)*pow(vx2,3))*pow(vy0,3)+(75*pow(vx0,3)*pow(vx2,4)-3*pow(vx0,4)*pow(vx2,3))*vy0)*pow(vy2,3)+((12*pow(vx2,5)-3*vx0*pow(vx2,4))*pow(vy0,4)+(81*pow(vx0,2)*pow(vx2,5)+3*pow(vx0,3)*pow(vx2,4))*pow(vy0,2)+66*pow(vx0,4)*pow(vx2,5)+6*pow(vx0,5)*pow(vx2,4))*pow(vy2,2)+(-3*pow(vx2,5)*pow(vy0,5)+(21*vx0*pow(vx2,6)-9*pow(vx0,2)*pow(vx2,5))*pow(vy0,3)+(30*pow(vx0,3)*pow(vx2,6)-6*pow(vx0,4)*pow(vx2,5))*vy0)*vy2+3*pow(vx2,7)*pow(vy0,4)+24*pow(vx0,2)*pow(vx2,7)*pow(vy0,2)+24*pow(vx0,4)*pow(vx2,7))*pow(vz0,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(24*vx0*pow(vz0,6)+(36*vx0*pow(vy0,2)+36*pow(vx0,3))*pow(vz0,4)+(12*vx0*pow(vy0,4)+24*pow(vx0,3)*pow(vy0,2)+12*pow(vx0,5))*pow(vz0,2))*pow(vz2,8)+((24*vx2+24*vx0)*pow(vz0,7)+(24*vx0*vy0*vy2+(60*vx2+60*vx0)*pow(vy0,2)+84*pow(vx0,2)*vx2+60*pow(vx0,3))*pow(vz0,5)+((24*vx0*pow(vy0,3)+24*pow(vx0,3)*vy0)*vy2+(48*vx2+48*vx0)*pow(vy0,4)+(120*pow(vx0,2)*vx2+96*pow(vx0,3))*pow(vy0,2)+72*pow(vx0,4)*vx2+48*pow(vx0,5))*pow(vz0,3)+((12*vx2+12*vx0)*pow(vy0,6)+(36*pow(vx0,2)*vx2+36*pow(vx0,3))*pow(vy0,4)+(36*pow(vx0,4)*vx2+36*pow(vx0,5))*pow(vy0,2)+12*pow(vx0,6)*vx2+12*pow(vx0,7))*vz0)*pow(vz2,7)+(24*vx2*pow(vz0,8)+(84*vx0*pow(vy2,2)+(24*vx2+24*vx0)*vy0*vy2+84*vx2*pow(vy0,2)+108*vx0*pow(vx2,2)+108*pow(vx0,2)*vx2)*pow(vz0,6)+((138*vx0*pow(vy0,2)+132*pow(vx0,3))*pow(vy2,2)+((54*vx2+54*vx0)*pow(vy0,3)+(66*pow(vx0,2)*vx2+54*pow(vx0,3))*vy0)*vy2+102*vx2*pow(vy0,4)+(186*vx0*pow(vx2,2)+258*pow(vx0,2)*vx2)*pow(vy0,2)+192*pow(vx0,3)*pow(vx2,2)+156*pow(vx0,4)*vx2)*pow(vz0,4)+((48*vx0*pow(vy0,4)+96*pow(vx0,3)*pow(vy0,2)+48*pow(vx0,5))*pow(vy2,2)+((36*vx2+36*vx0)*pow(vy0,5)+(72*pow(vx0,2)*vx2+72*pow(vx0,3))*pow(vy0,3)+(36*pow(vx0,4)*vx2+36*pow(vx0,5))*vy0)*vy2+48*vx2*pow(vy0,6)+(84*vx0*pow(vx2,2)+180*pow(vx0,2)*vx2)*pow(vy0,4)+(168*pow(vx0,3)*pow(vx2,2)+216*pow(vx0,4)*vx2)*pow(vy0,2)+84*pow(vx0,5)*pow(vx2,2)+84*pow(vx0,6)*vx2)*pow(vz0,2)+(-6*vx0*pow(vy0,6)-12*pow(vx0,3)*pow(vy0,4)-6*pow(vx0,5)*pow(vy0,2))*pow(vy2,2)+((6*vx2+6*vx0)*pow(vy0,7)+(6*pow(vx0,2)*vx2+18*pow(vx0,3))*pow(vy0,5)+(18*pow(vx0,5)-6*pow(vx0,4)*vx2)*pow(vy0,3)+(6*pow(vx0,7)-6*pow(vx0,6)*vx2)*vy0)*vy2+6*vx2*pow(vy0,8)+(6*vx0*pow(vx2,2)+30*pow(vx0,2)*vx2)*pow(vy0,6)+(12*pow(vx0,3)*pow(vx2,2)+54*pow(vx0,4)*vx2)*pow(vy0,4)+(6*pow(vx0,5)*pow(vx2,2)+42*pow(vx0,6)*vx2)*pow(vy0,2)+12*pow(vx0,8)*vx2)*pow(vz2,6)+(((60*vx2+60*vx0)*pow(vy2,2)+24*vx2*vy0*vy2+60*pow(vx2,3)+84*vx0*pow(vx2,2))*pow(vz0,7)+(84*vx0*vy0*pow(vy2,3)+((156*vx2+156*vx0)*pow(vy0,2)+234*pow(vx0,2)*vx2+150*pow(vx0,3))*pow(vy2,2)+(84*vx2*pow(vy0,3)+(96*vx0*pow(vx2,2)+96*pow(vx0,2)*vx2)*vy0)*vy2+(150*pow(vx2,3)+234*vx0*pow(vx2,2))*pow(vy0,2)+240*pow(vx0,2)*pow(vx2,3)+240*pow(vx0,3)*pow(vx2,2))*pow(vz0,5)+((96*vx0*pow(vy0,3)+96*pow(vx0,3)*vy0)*pow(vy2,3)+((132*vx2+132*vx0)*pow(vy0,4)+(348*pow(vx0,2)*vx2+252*pow(vx0,3))*pow(vy0,2)+216*pow(vx0,4)*vx2+120*pow(vx0,5))*pow(vy2,2)+(96*vx2*pow(vy0,5)+(120*vx0*pow(vx2,2)+216*pow(vx0,2)*vx2)*pow(vy0,3)+(120*pow(vx0,3)*pow(vx2,2)+120*pow(vx0,4)*vx2)*vy0)*vy2+(120*pow(vx2,3)+216*vx0*pow(vx2,2))*pow(vy0,4)+(348*pow(vx0,2)*pow(vx2,3)+444*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+228*pow(vx0,4)*pow(vx2,3)+228*pow(vx0,5)*pow(vx2,2))*pow(vz0,3)+((12*vx0*pow(vy0,5)+24*pow(vx0,3)*pow(vy0,3)+12*pow(vx0,5)*vy0)*pow(vy2,3)+((36*vx2+36*vx0)*pow(vy0,6)+(114*pow(vx0,2)*vx2+102*pow(vx0,3))*pow(vy0,4)+(120*pow(vx0,4)*vx2+96*pow(vx0,5))*pow(vy0,2)+42*pow(vx0,6)*vx2+30*pow(vx0,7))*pow(vy2,2)+(36*vx2*pow(vy0,7)+(24*vx0*pow(vx2,2)+120*pow(vx0,2)*vx2)*pow(vy0,5)+(48*pow(vx0,3)*pow(vx2,2)+132*pow(vx0,4)*vx2)*pow(vy0,3)+(24*pow(vx0,5)*pow(vx2,2)+48*pow(vx0,6)*vx2)*vy0)*vy2+(30*pow(vx2,3)+66*vx0*pow(vx2,2))*pow(vy0,6)+(108*pow(vx0,2)*pow(vx2,3)+204*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(126*pow(vx0,4)*pow(vx2,3)+210*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+48*pow(vx0,6)*pow(vx2,3)+72*pow(vx0,7)*pow(vx2,2))*vz0)*pow(vz2,5)+((36*vx2*pow(vy2,2)+36*pow(vx2,3))*pow(vz0,8)+(102*vx0*pow(vy2,4)+(54*vx2+54*vx0)*vy0*pow(vy2,3)+(138*vx2*pow(vy0,2)+258*vx0*pow(vx2,2)+186*pow(vx0,2)*vx2)*pow(vy2,2)+(54*pow(vx2,3)+66*vx0*pow(vx2,2))*vy0*vy2+132*pow(vx2,3)*pow(vy0,2)+156*vx0*pow(vx2,4)+192*pow(vx0,2)*pow(vx2,3))*pow(vz0,6)+((192*vx0*pow(vy0,2)+168*pow(vx0,3))*pow(vy2,4)+((120*vx2+120*vx0)*pow(vy0,3)+(168*pow(vx0,2)*vx2+120*pow(vx0,3))*vy0)*pow(vy2,3)+(192*vx2*pow(vy0,4)+(480*vx0*pow(vx2,2)+480*pow(vx0,2)*vx2)*pow(vy0,2)+480*pow(vx0,3)*pow(vx2,2)+288*pow(vx0,4)*vx2)*pow(vy2,2)+((120*pow(vx2,3)+168*vx0*pow(vx2,2))*pow(vy0,3)+(168*pow(vx0,2)*pow(vx2,3)+168*pow(vx0,3)*pow(vx2,2))*vy0)*vy2+168*pow(vx2,3)*pow(vy0,4)+(288*vx0*pow(vx2,4)+480*pow(vx0,2)*pow(vx2,3))*pow(vy0,2)+312*pow(vx0,3)*pow(vx2,4)+312*pow(vx0,4)*pow(vx2,3))*pow(vz0,4)+((78*vx0*pow(vy0,4)+144*pow(vx0,3)*pow(vy0,2)+66*pow(vx0,5))*pow(vy2,4)+((78*vx2+78*vx0)*pow(vy0,5)+(180*pow(vx0,2)*vx2+156*pow(vx0,3))*pow(vy0,3)+(102*pow(vx0,4)*vx2+78*pow(vx0,5))*vy0)*pow(vy2,3)+(114*vx2*pow(vy0,6)+(222*vx0*pow(vx2,2)+390*pow(vx0,2)*vx2)*pow(vy0,4)+(444*pow(vx0,3)*pow(vx2,2)+438*pow(vx0,4)*vx2)*pow(vy0,2)+222*pow(vx0,5)*pow(vx2,2)+162*pow(vx0,6)*vx2)*pow(vy2,2)+((78*pow(vx2,3)+138*vx0*pow(vx2,2))*pow(vy0,5)+(180*pow(vx0,2)*pow(vx2,3)+276*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(102*pow(vx0,4)*pow(vx2,3)+138*pow(vx0,5)*pow(vx2,2))*vy0)*vy2+84*pow(vx2,3)*pow(vy0,6)+(144*vx0*pow(vx2,4)+360*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+(300*pow(vx0,3)*pow(vx2,4)+468*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+156*pow(vx0,5)*pow(vx2,4)+192*pow(vx0,6)*pow(vx2,3))*pow(vz0,2)+(-12*vx0*pow(vy0,6)-24*pow(vx0,3)*pow(vy0,4)-12*pow(vx0,5)*pow(vy0,2))*pow(vy2,4)+((12*vx2+12*vx0)*pow(vy0,7)+(12*pow(vx0,2)*vx2+36*pow(vx0,3))*pow(vy0,5)+(36*pow(vx0,5)-12*pow(vx0,4)*vx2)*pow(vy0,3)+(12*pow(vx0,7)-12*pow(vx0,6)*vx2)*vy0)*pow(vy2,3)+(24*vx2*pow(vy0,8)+96*pow(vx0,2)*vx2*pow(vy0,6)+144*pow(vx0,4)*vx2*pow(vy0,4)+96*pow(vx0,6)*vx2*pow(vy0,2)+24*pow(vx0,8)*vx2)*pow(vy2,2)+((12*pow(vx2,3)+36*vx0*pow(vx2,2))*pow(vy0,7)+(12*pow(vx0,2)*pow(vx2,3)+108*pow(vx0,3)*pow(vx2,2))*pow(vy0,5)+(108*pow(vx0,5)*pow(vx2,2)-12*pow(vx0,4)*pow(vx2,3))*pow(vy0,3)+(36*pow(vx0,7)*pow(vx2,2)-12*pow(vx0,6)*pow(vx2,3))*vy0)*vy2+12*pow(vx2,3)*pow(vy0,8)+(12*vx0*pow(vx2,4)+72*pow(vx0,2)*pow(vx2,3))*pow(vy0,6)+(24*pow(vx0,3)*pow(vx2,4)+144*pow(vx0,4)*pow(vx2,3))*pow(vy0,4)+(12*pow(vx0,5)*pow(vx2,4)+120*pow(vx0,6)*pow(vx2,3))*pow(vy0,2)+36*pow(vx0,8)*pow(vx2,3))*pow(vz2,4)+(((48*vx2+48*vx0)*pow(vy2,4)+24*vx2*vy0*pow(vy2,3)+(96*pow(vx2,3)+120*vx0*pow(vx2,2))*pow(vy2,2)+24*pow(vx2,3)*vy0*vy2+48*pow(vx2,5)+72*vx0*pow(vx2,4))*pow(vz0,7)+(96*vx0*vy0*pow(vy2,5)+((132*vx2+132*vx0)*pow(vy0,2)+216*pow(vx0,2)*vx2+120*pow(vx0,3))*pow(vy2,4)+(96*vx2*pow(vy0,3)+(216*vx0*pow(vx2,2)+120*pow(vx0,2)*vx2)*vy0)*pow(vy2,3)+((252*pow(vx2,3)+348*vx0*pow(vx2,2))*pow(vy0,2)+444*pow(vx0,2)*pow(vx2,3)+348*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(96*pow(vx2,3)*pow(vy0,3)+(120*vx0*pow(vx2,4)+120*pow(vx0,2)*pow(vx2,3))*vy0)*vy2+(120*pow(vx2,5)+216*vx0*pow(vx2,4))*pow(vy0,2)+228*pow(vx0,2)*pow(vx2,5)+228*pow(vx0,3)*pow(vx2,4))*pow(vz0,5)+((120*vx0*pow(vy0,3)+120*pow(vx0,3)*vy0)*pow(vy2,5)+((120*vx2+120*vx0)*pow(vy0,4)+(336*pow(vx0,2)*vx2+216*pow(vx0,3))*pow(vy0,2)+216*pow(vx0,4)*vx2+96*pow(vx0,5))*pow(vy2,4)+(120*vx2*pow(vy0,5)+(288*vx0*pow(vx2,2)+288*pow(vx0,2)*vx2)*pow(vy0,3)+(288*pow(vx0,3)*pow(vx2,2)+168*pow(vx0,4)*vx2)*vy0)*pow(vy2,3)+((216*pow(vx2,3)+336*vx0*pow(vx2,2))*pow(vy0,4)+(672*pow(vx0,2)*pow(vx2,3)+672*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+456*pow(vx0,4)*pow(vx2,3)+336*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(120*pow(vx2,3)*pow(vy0,5)+(168*vx0*pow(vx2,4)+288*pow(vx0,2)*pow(vx2,3))*pow(vy0,3)+(168*pow(vx0,3)*pow(vx2,4)+168*pow(vx0,4)*pow(vx2,3))*vy0)*vy2+(96*pow(vx2,5)+216*vx0*pow(vx2,4))*pow(vy0,4)+(336*pow(vx0,2)*pow(vx2,5)+456*pow(vx0,3)*pow(vx2,4))*pow(vy0,2)+240*pow(vx0,4)*pow(vx2,5)+240*pow(vx0,5)*pow(vx2,4))*pow(vz0,3)+((24*vx0*pow(vy0,5)+48*pow(vx0,3)*pow(vy0,3)+24*pow(vx0,5)*vy0)*pow(vy2,5)+((36*vx2+36*vx0)*pow(vy0,6)+(120*pow(vx0,2)*vx2+96*pow(vx0,3))*pow(vy0,4)+(132*pow(vx0,4)*vx2+84*pow(vx0,5))*pow(vy0,2)+48*pow(vx0,6)*vx2+24*pow(vx0,7))*pow(vy2,4)+(48*vx2*pow(vy0,7)+(72*vx0*pow(vx2,2)+168*pow(vx0,2)*vx2)*pow(vy0,5)+(144*pow(vx0,3)*pow(vx2,2)+192*pow(vx0,4)*vx2)*pow(vy0,3)+(72*pow(vx0,5)*pow(vx2,2)+72*pow(vx0,6)*vx2)*vy0)*pow(vy2,3)+((60*pow(vx2,3)+108*vx0*pow(vx2,2))*pow(vy0,6)+(228*pow(vx0,2)*pow(vx2,3)+324*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(276*pow(vx0,4)*pow(vx2,3)+324*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+108*pow(vx0,6)*pow(vx2,3)+108*pow(vx0,7)*pow(vx2,2))*pow(vy2,2)+(48*pow(vx2,3)*pow(vy0,7)+(48*vx0*pow(vx2,4)+168*pow(vx0,2)*pow(vx2,3))*pow(vy0,5)+(96*pow(vx0,3)*pow(vx2,4)+192*pow(vx0,4)*pow(vx2,3))*pow(vy0,3)+(48*pow(vx0,5)*pow(vx2,4)+72*pow(vx0,6)*pow(vx2,3))*vy0)*vy2+(24*pow(vx2,5)+72*vx0*pow(vx2,4))*pow(vy0,6)+(108*pow(vx0,2)*pow(vx2,5)+228*pow(vx0,3)*pow(vx2,4))*pow(vy0,4)+(144*pow(vx0,4)*pow(vx2,5)+240*pow(vx0,5)*pow(vx2,4))*pow(vy0,2)+60*pow(vx0,6)*pow(vx2,5)+84*pow(vx0,7)*pow(vx2,4))*vz0)*pow(vz2,3)+((12*vx2*pow(vy2,4)+24*pow(vx2,3)*pow(vy2,2)+12*pow(vx2,5))*pow(vz0,8)+(48*vx0*pow(vy2,6)+(36*vx2+36*vx0)*vy0*pow(vy2,5)+(48*vx2*pow(vy0,2)+180*vx0*pow(vx2,2)+84*pow(vx0,2)*vx2)*pow(vy2,4)+(72*pow(vx2,3)+72*vx0*pow(vx2,2))*vy0*pow(vy2,3)+(96*pow(vx2,3)*pow(vy0,2)+216*vx0*pow(vx2,4)+168*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(36*pow(vx2,5)+36*vx0*pow(vx2,4))*vy0*vy2+48*pow(vx2,5)*pow(vy0,2)+84*vx0*pow(vx2,6)+84*pow(vx0,2)*pow(vx2,5))*pow(vz0,6)+((114*vx0*pow(vy0,2)+84*pow(vx0,3))*pow(vy2,6)+((78*vx2+78*vx0)*pow(vy0,3)+(138*pow(vx0,2)*vx2+78*pow(vx0,3))*vy0)*pow(vy2,5)+(78*vx2*pow(vy0,4)+(390*vx0*pow(vx2,2)+222*pow(vx0,2)*vx2)*pow(vy0,2)+360*pow(vx0,3)*pow(vx2,2)+144*pow(vx0,4)*vx2)*pow(vy2,4)+((156*pow(vx2,3)+180*vx0*pow(vx2,2))*pow(vy0,3)+(276*pow(vx0,2)*pow(vx2,3)+180*pow(vx0,3)*pow(vx2,2))*vy0)*pow(vy2,3)+(144*pow(vx2,3)*pow(vy0,4)+(438*vx0*pow(vx2,4)+444*pow(vx0,2)*pow(vx2,3))*pow(vy0,2)+468*pow(vx0,3)*pow(vx2,4)+300*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+((78*pow(vx2,5)+102*vx0*pow(vx2,4))*pow(vy0,3)+(138*pow(vx0,2)*pow(vx2,5)+102*pow(vx0,3)*pow(vx2,4))*vy0)*vy2+66*pow(vx2,5)*pow(vy0,4)+(162*vx0*pow(vx2,6)+222*pow(vx0,2)*pow(vx2,5))*pow(vy0,2)+192*pow(vx0,3)*pow(vx2,6)+156*pow(vx0,4)*pow(vx2,5))*pow(vz0,4)+((60*vx0*pow(vy0,4)+96*pow(vx0,3)*pow(vy0,2)+36*pow(vx0,5))*pow(vy2,6)+((48*vx2+48*vx0)*pow(vy0,5)+(144*pow(vx0,2)*vx2+96*pow(vx0,3))*pow(vy0,3)+(96*pow(vx0,4)*vx2+48*pow(vx0,5))*vy0)*pow(vy2,5)+(60*vx2*pow(vy0,6)+(204*vx0*pow(vx2,2)+204*pow(vx0,2)*vx2)*pow(vy0,4)+(384*pow(vx0,3)*pow(vx2,2)+228*pow(vx0,4)*vx2)*pow(vy0,2)+180*pow(vx0,5)*pow(vx2,2)+84*pow(vx0,6)*vx2)*pow(vy2,4)+((96*pow(vx2,3)+144*vx0*pow(vx2,2))*pow(vy0,5)+(288*pow(vx0,2)*pow(vx2,3)+288*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(192*pow(vx0,4)*pow(vx2,3)+144*pow(vx0,5)*pow(vx2,2))*vy0)*pow(vy2,3)+(96*pow(vx2,3)*pow(vy0,6)+(228*vx0*pow(vx2,4)+384*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+(480*pow(vx0,3)*pow(vx2,4)+480*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+252*pow(vx0,5)*pow(vx2,4)+192*pow(vx0,6)*pow(vx2,3))*pow(vy2,2)+((48*pow(vx2,5)+96*vx0*pow(vx2,4))*pow(vy0,5)+(144*pow(vx0,2)*pow(vx2,5)+192*pow(vx0,3)*pow(vx2,4))*pow(vy0,3)+(96*pow(vx0,4)*pow(vx2,5)+96*pow(vx0,5)*pow(vx2,4))*vy0)*vy2+36*pow(vx2,5)*pow(vy0,6)+(84*vx0*pow(vx2,6)+180*pow(vx0,2)*pow(vx2,5))*pow(vy0,4)+(192*pow(vx0,3)*pow(vx2,6)+252*pow(vx0,4)*pow(vx2,5))*pow(vy0,2)+108*pow(vx0,5)*pow(vx2,6)+108*pow(vx0,6)*pow(vx2,5))*pow(vz0,2)+(-6*vx0*pow(vy0,6)-12*pow(vx0,3)*pow(vy0,4)-6*pow(vx0,5)*pow(vy0,2))*pow(vy2,6)+((6*vx2+6*vx0)*pow(vy0,7)+(6*pow(vx0,2)*vx2+18*pow(vx0,3))*pow(vy0,5)+(18*pow(vx0,5)-6*pow(vx0,4)*vx2)*pow(vy0,3)+(6*pow(vx0,7)-6*pow(vx0,6)*vx2)*vy0)*pow(vy2,5)+(18*vx2*pow(vy0,8)+(66*pow(vx0,2)*vx2-6*vx0*pow(vx2,2))*pow(vy0,6)+(90*pow(vx0,4)*vx2-12*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(54*pow(vx0,6)*vx2-6*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+12*pow(vx0,8)*vx2)*pow(vy2,4)+((12*pow(vx2,3)+36*vx0*pow(vx2,2))*pow(vy0,7)+(12*pow(vx0,2)*pow(vx2,3)+108*pow(vx0,3)*pow(vx2,2))*pow(vy0,5)+(108*pow(vx0,5)*pow(vx2,2)-12*pow(vx0,4)*pow(vx2,3))*pow(vy0,3)+(36*pow(vx0,7)*pow(vx2,2)-12*pow(vx0,6)*pow(vx2,3))*vy0)*pow(vy2,3)+(24*pow(vx2,3)*pow(vy0,8)+(6*vx0*pow(vx2,4)+108*pow(vx0,2)*pow(vx2,3))*pow(vy0,6)+(12*pow(vx0,3)*pow(vx2,4)+180*pow(vx0,4)*pow(vx2,3))*pow(vy0,4)+(6*pow(vx0,5)*pow(vx2,4)+132*pow(vx0,6)*pow(vx2,3))*pow(vy0,2)+36*pow(vx0,8)*pow(vx2,3))*pow(vy2,2)+((6*pow(vx2,5)+30*vx0*pow(vx2,4))*pow(vy0,7)+(6*pow(vx0,2)*pow(vx2,5)+90*pow(vx0,3)*pow(vx2,4))*pow(vy0,5)+(90*pow(vx0,5)*pow(vx2,4)-6*pow(vx0,4)*pow(vx2,5))*pow(vy0,3)+(30*pow(vx0,7)*pow(vx2,4)-6*pow(vx0,6)*pow(vx2,5))*vy0)*vy2+6*pow(vx2,5)*pow(vy0,8)+(6*vx0*pow(vx2,6)+42*pow(vx0,2)*pow(vx2,5))*pow(vy0,6)+(12*pow(vx0,3)*pow(vx2,6)+90*pow(vx0,4)*pow(vx2,5))*pow(vy0,4)+(6*pow(vx0,5)*pow(vx2,6)+78*pow(vx0,6)*pow(vx2,5))*pow(vy0,2)+24*pow(vx0,8)*pow(vx2,5))*pow(vz2,2)+(((12*vx2+12*vx0)*pow(vy2,6)+(36*pow(vx2,3)+36*vx0*pow(vx2,2))*pow(vy2,4)+(36*pow(vx2,5)+36*vx0*pow(vx2,4))*pow(vy2,2)+12*pow(vx2,7)+12*vx0*pow(vx2,6))*pow(vz0,7)+(36*vx0*vy0*pow(vy2,7)+((36*vx2+36*vx0)*pow(vy0,2)+66*pow(vx0,2)*vx2+30*pow(vx0,3))*pow(vy2,6)+(12*vx2*pow(vy0,3)+(120*vx0*pow(vx2,2)+24*pow(vx0,2)*vx2)*vy0)*pow(vy2,5)+((102*pow(vx2,3)+114*vx0*pow(vx2,2))*pow(vy0,2)+204*pow(vx0,2)*pow(vx2,3)+108*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+(24*pow(vx2,3)*pow(vy0,3)+(132*vx0*pow(vx2,4)+48*pow(vx0,2)*pow(vx2,3))*vy0)*pow(vy2,3)+((96*pow(vx2,5)+120*vx0*pow(vx2,4))*pow(vy0,2)+210*pow(vx0,2)*pow(vx2,5)+126*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+(12*pow(vx2,5)*pow(vy0,3)+(48*vx0*pow(vx2,6)+24*pow(vx0,2)*pow(vx2,5))*vy0)*vy2+(30*pow(vx2,7)+42*vx0*pow(vx2,6))*pow(vy0,2)+72*pow(vx0,2)*pow(vx2,7)+48*pow(vx0,3)*pow(vx2,6))*pow(vz0,5)+((48*vx0*pow(vy0,3)+48*pow(vx0,3)*vy0)*pow(vy2,7)+((36*vx2+36*vx0)*pow(vy0,4)+(108*pow(vx0,2)*vx2+60*pow(vx0,3))*pow(vy0,2)+72*pow(vx0,4)*vx2+24*pow(vx0,5))*pow(vy2,6)+(24*vx2*pow(vy0,5)+(168*vx0*pow(vx2,2)+72*pow(vx0,2)*vx2)*pow(vy0,3)+(168*pow(vx0,3)*pow(vx2,2)+48*pow(vx0,4)*vx2)*vy0)*pow(vy2,5)+((96*pow(vx2,3)+120*vx0*pow(vx2,2))*pow(vy0,4)+(324*pow(vx0,2)*pow(vx2,3)+228*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+228*pow(vx0,4)*pow(vx2,3)+108*pow(vx0,5)*pow(vx2,2))*pow(vy2,4)+(48*pow(vx2,3)*pow(vy0,5)+(192*vx0*pow(vx2,4)+144*pow(vx0,2)*pow(vx2,3))*pow(vy0,3)+(192*pow(vx0,3)*pow(vx2,4)+96*pow(vx0,4)*pow(vx2,3))*vy0)*pow(vy2,3)+((84*pow(vx2,5)+132*vx0*pow(vx2,4))*pow(vy0,4)+(324*pow(vx0,2)*pow(vx2,5)+276*pow(vx0,3)*pow(vx2,4))*pow(vy0,2)+240*pow(vx0,4)*pow(vx2,5)+144*pow(vx0,5)*pow(vx2,4))*pow(vy2,2)+(24*pow(vx2,5)*pow(vy0,5)+(72*vx0*pow(vx2,6)+72*pow(vx0,2)*pow(vx2,5))*pow(vy0,3)+(72*pow(vx0,3)*pow(vx2,6)+48*pow(vx0,4)*pow(vx2,5))*vy0)*vy2+(24*pow(vx2,7)+48*vx0*pow(vx2,6))*pow(vy0,4)+(108*pow(vx0,2)*pow(vx2,7)+108*pow(vx0,3)*pow(vx2,6))*pow(vy0,2)+84*pow(vx0,4)*pow(vx2,7)+60*pow(vx0,5)*pow(vx2,6))*pow(vz0,3)+((12*vx0*pow(vy0,5)+24*pow(vx0,3)*pow(vy0,3)+12*pow(vx0,5)*vy0)*pow(vy2,7)+((12*vx2+12*vx0)*pow(vy0,6)+(42*pow(vx0,2)*vx2+30*pow(vx0,3))*pow(vy0,4)+(48*pow(vx0,4)*vx2+24*pow(vx0,5))*pow(vy0,2)+18*pow(vx0,6)*vx2+6*pow(vx0,7))*pow(vy2,6)+(12*vx2*pow(vy0,7)+(48*vx0*pow(vx2,2)+48*pow(vx0,2)*vx2)*pow(vy0,5)+(96*pow(vx0,3)*pow(vx2,2)+60*pow(vx0,4)*vx2)*pow(vy0,3)+(48*pow(vx0,5)*pow(vx2,2)+24*pow(vx0,6)*vx2)*vy0)*pow(vy2,5)+((30*pow(vx2,3)+42*vx0*pow(vx2,2))*pow(vy0,6)+(120*pow(vx0,2)*pow(vx2,3)+120*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+(150*pow(vx0,4)*pow(vx2,3)+114*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+60*pow(vx0,6)*pow(vx2,3)+36*pow(vx0,7)*pow(vx2,2))*pow(vy2,4)+(24*pow(vx2,3)*pow(vy0,7)+(60*vx0*pow(vx2,4)+96*pow(vx0,2)*pow(vx2,3))*pow(vy0,5)+(120*pow(vx0,3)*pow(vx2,4)+120*pow(vx0,4)*pow(vx2,3))*pow(vy0,3)+(60*pow(vx0,5)*pow(vx2,4)+48*pow(vx0,6)*pow(vx2,3))*vy0)*pow(vy2,3)+((24*pow(vx2,5)+48*vx0*pow(vx2,4))*pow(vy0,6)+(114*pow(vx0,2)*pow(vx2,5)+150*pow(vx0,3)*pow(vx2,4))*pow(vy0,4)+(156*pow(vx0,4)*pow(vx2,5)+156*pow(vx0,5)*pow(vx2,4))*pow(vy0,2)+66*pow(vx0,6)*pow(vx2,5)+54*pow(vx0,7)*pow(vx2,4))*pow(vy2,2)+(12*pow(vx2,5)*pow(vy0,7)+(24*vx0*pow(vx2,6)+48*pow(vx0,2)*pow(vx2,5))*pow(vy0,5)+(48*pow(vx0,3)*pow(vx2,6)+60*pow(vx0,4)*pow(vx2,5))*pow(vy0,3)+(24*pow(vx0,5)*pow(vx2,6)+24*pow(vx0,6)*pow(vx2,5))*vy0)*vy2+(6*pow(vx2,7)+18*vx0*pow(vx2,6))*pow(vy0,6)+(36*pow(vx0,2)*pow(vx2,7)+60*pow(vx0,3)*pow(vx2,6))*pow(vy0,4)+(54*pow(vx0,4)*pow(vx2,7)+66*pow(vx0,5)*pow(vx2,6))*pow(vy0,2)+24*pow(vx0,6)*pow(vx2,7)+24*pow(vx0,7)*pow(vx2,6))*vz0)*vz2+(6*vx0*pow(vy2,8)+(6*vx2+6*vx0)*vy0*pow(vy2,7)+(-6*vx2*pow(vy0,2)+30*vx0*pow(vx2,2)+6*pow(vx0,2)*vx2)*pow(vy2,6)+(18*pow(vx2,3)+6*vx0*pow(vx2,2))*vy0*pow(vy2,5)+(-12*pow(vx2,3)*pow(vy0,2)+54*vx0*pow(vx2,4)+12*pow(vx0,2)*pow(vx2,3))*pow(vy2,4)+(18*pow(vx2,5)-6*vx0*pow(vx2,4))*vy0*pow(vy2,3)+(-6*pow(vx2,5)*pow(vy0,2)+42*vx0*pow(vx2,6)+6*pow(vx0,2)*pow(vx2,5))*pow(vy2,2)+(6*pow(vx2,7)-6*vx0*pow(vx2,6))*vy0*vy2+12*vx0*pow(vx2,8))*pow(vz0,6)+((24*vx0*pow(vy0,2)+12*pow(vx0,3))*pow(vy2,8)+((12*vx2+12*vx0)*pow(vy0,3)+(36*pow(vx0,2)*vx2+12*pow(vx0,3))*vy0)*pow(vy2,7)+(-12*vx2*pow(vy0,4)+96*vx0*pow(vx2,2)*pow(vy0,2)+72*pow(vx0,3)*pow(vx2,2)+12*pow(vx0,4)*vx2)*pow(vy2,6)+((36*pow(vx2,3)+12*vx0*pow(vx2,2))*pow(vy0,3)+(108*pow(vx0,2)*pow(vx2,3)+12*pow(vx0,3)*pow(vx2,2))*vy0)*pow(vy2,5)+(-24*pow(vx2,3)*pow(vy0,4)+144*vx0*pow(vx2,4)*pow(vy0,2)+144*pow(vx0,3)*pow(vx2,4)+24*pow(vx0,4)*pow(vx2,3))*pow(vy2,4)+((36*pow(vx2,5)-12*vx0*pow(vx2,4))*pow(vy0,3)+(108*pow(vx0,2)*pow(vx2,5)-12*pow(vx0,3)*pow(vx2,4))*vy0)*pow(vy2,3)+(-12*pow(vx2,5)*pow(vy0,4)+96*vx0*pow(vx2,6)*pow(vy0,2)+120*pow(vx0,3)*pow(vx2,6)+12*pow(vx0,4)*pow(vx2,5))*pow(vy2,2)+((12*pow(vx2,7)-12*vx0*pow(vx2,6))*pow(vy0,3)+(36*pow(vx0,2)*pow(vx2,7)-12*pow(vx0,3)*pow(vx2,6))*vy0)*vy2+24*vx0*pow(vx2,8)*pow(vy0,2)+36*pow(vx0,3)*pow(vx2,8))*pow(vz0,4)+((18*vx0*pow(vy0,4)+24*pow(vx0,3)*pow(vy0,2)+6*pow(vx0,5))*pow(vy2,8)+((6*vx2+6*vx0)*pow(vy0,5)+(36*pow(vx0,2)*vx2+12*pow(vx0,3))*pow(vy0,3)+(30*pow(vx0,4)*vx2+6*pow(vx0,5))*vy0)*pow(vy2,7)+(-6*vx2*pow(vy0,6)+(66*vx0*pow(vx2,2)-6*pow(vx0,2)*vx2)*pow(vy0,4)+(108*pow(vx0,3)*pow(vx2,2)+6*pow(vx0,4)*vx2)*pow(vy0,2)+42*pow(vx0,5)*pow(vx2,2)+6*pow(vx0,6)*vx2)*pow(vy2,6)+((18*pow(vx2,3)+6*vx0*pow(vx2,2))*pow(vy0,5)+(108*pow(vx0,2)*pow(vx2,3)+12*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+(90*pow(vx0,4)*pow(vx2,3)+6*pow(vx0,5)*pow(vx2,2))*vy0)*pow(vy2,5)+(-12*pow(vx2,3)*pow(vy0,6)+(90*vx0*pow(vx2,4)-12*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+(180*pow(vx0,3)*pow(vx2,4)+12*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+90*pow(vx0,5)*pow(vx2,4)+12*pow(vx0,6)*pow(vx2,3))*pow(vy2,4)+((18*pow(vx2,5)-6*vx0*pow(vx2,4))*pow(vy0,5)+(108*pow(vx0,2)*pow(vx2,5)-12*pow(vx0,3)*pow(vx2,4))*pow(vy0,3)+(90*pow(vx0,4)*pow(vx2,5)-6*pow(vx0,5)*pow(vx2,4))*vy0)*pow(vy2,3)+(-6*pow(vx2,5)*pow(vy0,6)+(54*vx0*pow(vx2,6)-6*pow(vx0,2)*pow(vx2,5))*pow(vy0,4)+(132*pow(vx0,3)*pow(vx2,6)+6*pow(vx0,4)*pow(vx2,5))*pow(vy0,2)+78*pow(vx0,5)*pow(vx2,6)+6*pow(vx0,6)*pow(vx2,5))*pow(vy2,2)+((6*pow(vx2,7)-6*vx0*pow(vx2,6))*pow(vy0,5)+(36*pow(vx0,2)*pow(vx2,7)-12*pow(vx0,3)*pow(vx2,6))*pow(vy0,3)+(30*pow(vx0,4)*pow(vx2,7)-6*pow(vx0,5)*pow(vx2,6))*vy0)*vy2+12*vx0*pow(vx2,8)*pow(vy0,4)+36*pow(vx0,3)*pow(vx2,8)*pow(vy0,2)+24*pow(vx0,5)*pow(vx2,8))*pow(vz0,2))*pow(((160*pow(roughness,2)*pow(vz0,7)+(400*pow(roughness,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2))*pow(vz0,5)+(320*pow(roughness,2)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,7)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,6)+((720*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,7)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+((400*pow(roughness,2)*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2))*pow(vz0,7)+((1280*pow(roughness,2)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((1360*pow(roughness,2)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+880*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((480*pow(roughness,2)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+240*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((720*pow(roughness,2)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*vx2*pow(vy2,2)+720*pow(roughness,2)*pow(vx2,2)*vy0*vy2+720*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,6)+((1760*pow(roughness,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1680*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1760*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((1360*pow(roughness,2)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(1680*pow(roughness,2)*vx0*vx2*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((320*pow(roughness,2)*pow(vy2,4)+640*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,4))*pow(vz0,7)+((1360*pow(roughness,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(2240*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+880*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((1760*pow(roughness,2)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(2560*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+800*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1760*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((720*pow(roughness,2)*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((480*pow(roughness,2)*vy0*pow(vy2,5)+480*pow(roughness,2)*vx0*vx2*pow(vy2,4)+960*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,4)*vy0*vy2+480*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,6)+((1360*pow(roughness,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(2560*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(2880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(1200*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+1200*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((1280*pow(roughness,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(2240*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2880*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(2880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(960*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+1280*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+(400*pow(roughness,2)*pow(vy0,7)+1040*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+1040*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6))*pow(vz0,7)+((480*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+240*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,5)+((720*pow(roughness,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,7)+80*pow(roughness,2)*vx0*vx2*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,6)+((320*pow(roughness,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(880*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(800*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(880*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,7)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,7)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,6)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1360*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1760*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1360*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,4)+(320*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+1360*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1760*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+1360*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+1280*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,6)*pow(vy0,7)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*pow(vx2,7))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,8)+(480*pow(roughness,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2))*pow(vz0,6)+(500*pow(roughness,2)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4))*pow(vz0,4)+(200*pow(roughness,2)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,8)+80*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+120*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+80*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vz2,8)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+880*pow(roughness,2)*vx0*vx2*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,5)+((800*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*vz0)*pow(vz2,7)+((480*pow(roughness,2)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,2))*pow(vz0,8)+((1720*pow(roughness,2)*pow(vy0,2)+1480*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1480*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+((2200*pow(roughness,2)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+((1160*pow(roughness,2)*pow(vy0,6)+3000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+680*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+1160*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz0,2)+(200*pow(roughness,2)*pow(vy0,8)+680*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+840*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+440*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,8))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*vy2+80*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+440*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+680*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vz2,6)+((880*pow(roughness,2)*vy0*pow(vy2,3)+880*pow(roughness,2)*vx0*vx2*pow(vy2,2)+880*pow(roughness,2)*pow(vx2,2)*vy0*vy2+880*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,7)+((2560*pow(roughness,2)*pow(vy0,3)+2480*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(2720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(2480*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+2480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,5)+((2480*pow(roughness,2)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(2320*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+2320*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,3)+((800*pow(roughness,2)*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,6)+2640*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2640*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*vz0)*pow(vz2,5)+((500*pow(roughness,2)*pow(vy2,4)+1000*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+500*pow(roughness,2)*pow(vx2,4))*pow(vz0,8)+((2200*pow(roughness,2)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+1200*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(3800*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+1200*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+1600*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+((3400*pow(roughness,2)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1820*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(3200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+3120*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(5200*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+10520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(3120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+1820*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+3400*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,4)+((2200*pow(roughness,2)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+3840*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(3000*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+3000*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(2640*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+840*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+3840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz0,2)+(500*pow(roughness,2)*pow(vy0,8)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1820*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+840*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,8))*pow(vy2,4)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+3840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1820*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vz2,4)+((800*pow(roughness,2)*vy0*pow(vy2,5)+800*pow(roughness,2)*vx0*vx2*pow(vy2,4)+1600*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+800*pow(roughness,2)*pow(vx2,4)*vy0*vy2+800*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,7)+((2480*pow(roughness,2)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(4800*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(5120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(2320*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+2320*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,5)+((2560*pow(roughness,2)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+2240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(3200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+5440*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(4800*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+10240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+5440*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(5440*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+10240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+4800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(2240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+3200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+2240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,3)+((880*pow(roughness,2)*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,6)+3120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+2640*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+2640*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+3120*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*vz0)*pow(vz2,3)+((200*pow(roughness,2)*pow(vy2,6)+600*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+600*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+200*pow(roughness,2)*pow(vx2,6))*pow(vz0,8)+((1160*pow(roughness,2)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(3000*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+1920*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(2520*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+680*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+1160*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,6)+((2200*pow(roughness,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(5200*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+3840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(5440*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(3840*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(2640*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+840*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,4)+((1720*pow(roughness,2)*pow(vy0,6)+3800*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+440*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(2720*pow(roughness,2)*vx0*vx2*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+2400*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(3800*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+10520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+2520*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(5120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+10240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+5120*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(2520*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+10520*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+3800*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(2400*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+2720*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+440*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+1720*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vz0,2)+(480*pow(roughness,2)*pow(vy0,8)+1480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1600*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+680*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,8))*pow(vy2,6)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,5)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,4)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+680*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+1600*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+1480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,8)*pow(vx2,6))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,7)+240*pow(roughness,2)*vx0*vx2*pow(vy2,6)+720*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+720*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,6)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,7)+((800*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(2320*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(2640*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(2240*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(2400*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,5)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(2480*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(3120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(2320*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+3120*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(2640*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,7)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,6)+(880*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,4)+(800*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+2720*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,7))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,8)+80*pow(roughness,2)*pow(vx2,2)*pow(vy2,6)+120*pow(roughness,2)*pow(vx2,4)*pow(vy2,4)+80*pow(roughness,2)*pow(vx2,6)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,8))*pow(vz0,8)+((200*pow(roughness,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,2))*pow(vy2,8)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,7)+(680*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+440*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,6)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,5)+(840*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,4)+720*pow(roughness,2)*vx0*pow(vx2,5)*vy0*pow(vy2,3)+(440*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,7)*vy0*vy2+80*pow(roughness,2)*pow(vx2,8)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,8))*pow(vz0,6)+((500*pow(roughness,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4))*pow(vy2,8)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,7)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,6)+(2320*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,5)+(1820*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+3840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1820*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,4)+(2240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*pow(vy2,3)+(840*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,8)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,8))*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,6)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,8)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,7)+(1480*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,6)+(2480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,5)+(1600*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,4)+(2320*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*pow(vy2,3)+(680*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+1480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,8)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,8)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,8))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+200*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vy2,8)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,7)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+1160*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,6)+(880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,5)+(500*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+2200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+3400*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+2200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,4)+(800*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*pow(vy2,3)+(200*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+1160*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+1720*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,8)*pow(vx2,6))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,7)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,7)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,8)*pow(vy0,8)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,8)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,8)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,8)*pow(vx2,8),-1) ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
// y1  :  
poly.add(Term6f( (((24*vy0*pow(vz0,5)+(24*pow(vy0,3)+24*pow(vx0,2)*vy0)*pow(vz0,3)+(3*pow(vy0,5)+6*pow(vx0,2)*pow(vy0,3)+3*pow(vx0,4)*vy0)*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,7)+((24*vy2+24*vy0)*pow(vz0,6)+((72*pow(vy0,2)+48*pow(vx0,2))*vy2+48*pow(vy0,3)+(24*vx0*vx2+48*pow(vx0,2))*vy0)*pow(vz0,4)+((39*pow(vy0,4)+66*pow(vx0,2)*pow(vy0,2)+27*pow(vx0,4))*vy2+27*pow(vy0,5)+(12*vx0*vx2+54*pow(vx0,2))*pow(vy0,3)+(12*pow(vx0,3)*vx2+27*pow(vx0,4))*vy0)*pow(vz0,2)+(3*pow(vx0,2)*pow(vy0,4)+6*pow(vx0,4)*pow(vy0,2)+3*pow(vx0,6))*vy2+3*pow(vy0,7)+(9*pow(vx0,2)-3*vx0*vx2)*pow(vy0,5)+(9*pow(vx0,4)-6*pow(vx0,3)*vx2)*pow(vy0,3)+(3*pow(vx0,6)-3*pow(vx0,5)*vx2)*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+(24*vy2*pow(vz0,7)+(96*vy0*pow(vy2,2)+(96*pow(vy0,2)+24*vx0*vx2+72*pow(vx0,2))*vy2+(72*pow(vx2,2)+24*vx0*vx2)*vy0)*pow(vz0,5)+((126*pow(vy0,3)+120*pow(vx0,2)*vy0)*pow(vy2,2)+(111*pow(vy0,4)+(54*vx0*vx2+180*pow(vx0,2))*pow(vy0,2)+42*pow(vx0,3)*vx2+69*pow(vx0,4))*vy2+(78*pow(vx2,2)+42*vx0*vx2)*pow(vy0,3)+(84*pow(vx0,2)*pow(vx2,2)+42*pow(vx0,3)*vx2)*vy0)*pow(vz0,3)+((27*pow(vy0,5)+57*pow(vx0,2)*pow(vy0,3)+30*pow(vx0,4)*vy0)*pow(vy2,2)+(39*pow(vy0,6)+(12*vx0*vx2+99*pow(vx0,2))*pow(vy0,4)+(30*pow(vx0,3)*vx2+81*pow(vx0,4))*pow(vy0,2)+18*pow(vx0,5)*vx2+21*pow(vx0,6))*vy2+(12*pow(vx2,2)+18*vx0*vx2)*pow(vy0,5)+(21*pow(vx0,2)*pow(vx2,2)+36*pow(vx0,3)*vx2)*pow(vy0,3)+(9*pow(vx0,4)*pow(vx2,2)+18*pow(vx0,5)*vx2)*vy0)*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((48*pow(vy2,3)+72*vy0*pow(vy2,2)+(48*pow(vx2,2)+24*vx0*vx2)*vy2+48*pow(vx2,2)*vy0)*pow(vz0,6)+((174*pow(vy0,2)+96*pow(vx0,2))*pow(vy2,3)+(174*pow(vy0,3)+(84*vx0*vx2+168*pow(vx0,2))*vy0)*pow(vy2,2)+((168*pow(vx2,2)+84*vx0*vx2)*pow(vy0,2)+102*pow(vx0,2)*pow(vx2,2)+72*pow(vx0,3)*vx2)*vy2+96*pow(vx2,2)*pow(vy0,3)+(72*vx0*pow(vx2,3)+102*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,4)+((111*pow(vy0,4)+165*pow(vx0,2)*pow(vy0,2)+54*pow(vx0,4))*pow(vy2,3)+(126*pow(vy0,5)+(66*vx0*vx2+243*pow(vx0,2))*pow(vy0,3)+(66*pow(vx0,3)*vx2+117*pow(vx0,4))*vy0)*pow(vy2,2)+((102*pow(vx2,2)+81*vx0*vx2)*pow(vy0,4)+(165*pow(vx0,2)*pow(vx2,2)+144*pow(vx0,3)*vx2)*pow(vy0,2)+63*pow(vx0,4)*pow(vx2,2)+63*pow(vx0,5)*vx2)*vy2+54*pow(vx2,2)*pow(vy0,5)+(48*vx0*pow(vx2,3)+117*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(48*pow(vx0,3)*pow(vx2,3)+63*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vz0,2)+(9*pow(vx0,2)*pow(vy0,4)+15*pow(vx0,4)*pow(vy0,2)+6*pow(vx0,6))*pow(vy2,3)+(24*pow(vy0,7)+(69*pow(vx0,2)-9*vx0*vx2)*pow(vy0,5)+(66*pow(vx0,4)-9*pow(vx0,3)*vx2)*pow(vy0,3)+21*pow(vx0,6)*vy0)*pow(vy2,2)+(21*vx0*vx2*pow(vy0,6)+57*pow(vx0,3)*vx2*pow(vy0,4)+(9*pow(vx0,4)*pow(vx2,2)+51*pow(vx0,5)*vx2)*pow(vy0,2)+9*pow(vx0,6)*pow(vx2,2)+15*pow(vx0,7)*vx2)*vy2+6*pow(vx2,2)*pow(vy0,7)+(21*pow(vx0,2)*pow(vx2,2)-6*vx0*pow(vx2,3))*pow(vy0,5)+(24*pow(vx0,4)*pow(vx2,2)-15*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(9*pow(vx0,6)*pow(vx2,2)-9*pow(vx0,5)*pow(vx2,3))*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((24*pow(vy2,3)+24*pow(vx2,2)*vy2)*pow(vz0,7)+(111*vy0*pow(vy2,4)+(126*pow(vy0,2)+42*vx0*vx2+78*pow(vx0,2))*pow(vy2,3)+(180*pow(vx2,2)+54*vx0*vx2)*vy0*pow(vy2,2)+(120*pow(vx2,2)*pow(vy0,2)+42*vx0*pow(vx2,3)+84*pow(vx0,2)*pow(vx2,2))*vy2+(69*pow(vx2,4)+42*vx0*pow(vx2,3))*vy0)*pow(vz0,5)+((174*pow(vy0,3)+153*pow(vx0,2)*vy0)*pow(vy2,4)+(174*pow(vy0,4)+(114*vx0*vx2+255*pow(vx0,2))*pow(vy0,2)+72*pow(vx0,3)*vx2+81*pow(vx0,4))*pow(vy2,3)+((255*pow(vx2,2)+114*vx0*vx2)*pow(vy0,3)+(255*pow(vx0,2)*pow(vx2,2)+114*pow(vx0,3)*vx2)*vy0)*pow(vy2,2)+(153*pow(vx2,2)*pow(vy0,4)+(114*vx0*pow(vx2,3)+255*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+72*pow(vx0,3)*pow(vx2,3)+102*pow(vx0,4)*pow(vx2,2))*vy2+(81*pow(vx2,4)+72*vx0*pow(vx2,3))*pow(vy0,3)+(102*pow(vx0,2)*pow(vx2,4)+72*pow(vx0,3)*pow(vx2,3))*vy0)*pow(vz0,3)+((48*pow(vy0,5)+93*pow(vx0,2)*pow(vy0,3)+45*pow(vx0,4)*vy0)*pow(vy2,4)+(72*pow(vy0,6)+(36*vx0*vx2+171*pow(vx0,2))*pow(vy0,4)+(66*pow(vx0,3)*vx2+126*pow(vx0,4))*pow(vy0,2)+30*pow(vx0,5)*vx2+27*pow(vx0,6))*pow(vy2,3)+((63*pow(vx2,2)+60*vx0*vx2)*pow(vy0,5)+(126*pow(vx0,2)*pow(vx2,2)+120*pow(vx0,3)*vx2)*pow(vy0,3)+(63*pow(vx0,4)*pow(vx2,2)+60*pow(vx0,5)*vx2)*vy0)*pow(vy2,2)+(57*pow(vx2,2)*pow(vy0,6)+(36*vx0*pow(vx2,3)+156*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+(66*pow(vx0,3)*pow(vx2,3)+141*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+30*pow(vx0,5)*pow(vx2,3)+42*pow(vx0,6)*pow(vx2,2))*vy2+(15*pow(vx2,4)+30*vx0*pow(vx2,3))*pow(vy0,5)+(33*pow(vx0,2)*pow(vx2,4)+60*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(18*pow(vx0,4)*pow(vx2,4)+30*pow(vx0,5)*pow(vx2,3))*vy0)*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((27*pow(vy2,5)+39*vy0*pow(vy2,4)+(54*pow(vx2,2)+12*vx0*vx2)*pow(vy2,3)+66*pow(vx2,2)*vy0*pow(vy2,2)+(27*pow(vx2,4)+12*vx0*pow(vx2,3))*vy2+27*pow(vx2,4)*vy0)*pow(vz0,6)+((126*pow(vy0,2)+54*pow(vx0,2))*pow(vy2,5)+(111*pow(vy0,3)+(81*vx0*vx2+102*pow(vx0,2))*vy0)*pow(vy2,4)+((243*pow(vx2,2)+66*vx0*vx2)*pow(vy0,2)+117*pow(vx0,2)*pow(vx2,2)+48*pow(vx0,3)*vx2)*pow(vy2,3)+(165*pow(vx2,2)*pow(vy0,3)+(144*vx0*pow(vx2,3)+165*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vy2,2)+((117*pow(vx2,4)+66*vx0*pow(vx2,3))*pow(vy0,2)+63*pow(vx0,2)*pow(vx2,4)+48*pow(vx0,3)*pow(vx2,3))*vy2+54*pow(vx2,4)*pow(vy0,3)+(63*vx0*pow(vx2,5)+63*pow(vx0,2)*pow(vx2,4))*vy0)*pow(vz0,4)+((96*pow(vy0,4)+123*pow(vx0,2)*pow(vy0,2)+30*pow(vx0,4))*pow(vy2,5)+(96*pow(vy0,5)+(84*vx0*vx2+177*pow(vx0,2))*pow(vy0,3)+(75*pow(vx0,3)*vx2+81*pow(vx0,4))*vy0)*pow(vy2,4)+((177*pow(vx2,2)+84*vx0*vx2)*pow(vy0,4)+(255*pow(vx0,2)*pow(vx2,2)+135*pow(vx0,3)*vx2)*pow(vy0,2)+72*pow(vx0,4)*pow(vx2,2)+51*pow(vx0,5)*vx2)*pow(vy2,3)+(123*pow(vx2,2)*pow(vy0,5)+(135*vx0*pow(vx2,3)+255*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(129*pow(vx0,3)*pow(vx2,3)+132*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vy2,2)+((81*pow(vx2,4)+75*vx0*pow(vx2,3))*pow(vy0,4)+(132*pow(vx0,2)*pow(vx2,4)+129*pow(vx0,3)*pow(vx2,3))*pow(vy0,2)+42*pow(vx0,4)*pow(vx2,4)+54*pow(vx0,5)*pow(vx2,3))*vy2+30*pow(vx2,4)*pow(vy0,5)+(51*vx0*pow(vx2,5)+72*pow(vx0,2)*pow(vx2,4))*pow(vy0,3)+(54*pow(vx0,3)*pow(vx2,5)+42*pow(vx0,4)*pow(vx2,4))*vy0)*pow(vz0,2)+(6*pow(vx0,2)*pow(vy0,4)+9*pow(vx0,4)*pow(vy0,2)+3*pow(vx0,6))*pow(vy2,5)+(24*pow(vy0,7)+(66*pow(vx0,2)-6*vx0*vx2)*pow(vy0,5)+(60*pow(vx0,4)-3*pow(vx0,3)*vx2)*pow(vy0,3)+(3*pow(vx0,5)*vx2+18*pow(vx0,6))*vy0)*pow(vy2,4)+(30*vx0*vx2*pow(vy0,6)+(3*pow(vx0,2)*pow(vx2,2)+75*pow(vx0,3)*vx2)*pow(vy0,4)+(12*pow(vx0,4)*pow(vx2,2)+60*pow(vx0,5)*vx2)*pow(vy0,2)+9*pow(vx0,6)*pow(vx2,2)+15*pow(vx0,7)*vx2)*pow(vy2,3)+(24*pow(vx2,2)*pow(vy0,7)+(81*pow(vx0,2)*pow(vx2,2)-9*vx0*pow(vx2,3))*pow(vy0,5)+(90*pow(vx0,4)*pow(vx2,2)-12*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(33*pow(vx0,6)*pow(vx2,2)-3*pow(vx0,5)*pow(vx2,3))*vy0)*pow(vy2,2)+(21*vx0*pow(vx2,3)*pow(vy0,6)+(60*pow(vx0,3)*pow(vx2,3)-3*pow(vx0,2)*pow(vx2,4))*pow(vy0,4)+(3*pow(vx0,4)*pow(vx2,4)+57*pow(vx0,5)*pow(vx2,3))*pow(vy0,2)+6*pow(vx0,6)*pow(vx2,4)+18*pow(vx0,7)*pow(vx2,3))*vy2+3*pow(vx2,4)*pow(vy0,7)+(12*pow(vx0,2)*pow(vx2,4)-3*vx0*pow(vx2,5))*pow(vy0,5)+(15*pow(vx0,4)*pow(vx2,4)-9*pow(vx0,3)*pow(vx2,5))*pow(vy0,3)+(6*pow(vx0,6)*pow(vx2,4)-6*pow(vx0,5)*pow(vx2,5))*vy0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((3*pow(vy2,5)+6*pow(vx2,2)*pow(vy2,3)+3*pow(vx2,4)*vy2)*pow(vz0,7)+(39*vy0*pow(vy2,6)+(27*pow(vy0,2)+18*vx0*vx2+12*pow(vx0,2))*pow(vy2,5)+(99*pow(vx2,2)+12*vx0*vx2)*vy0*pow(vy2,4)+(57*pow(vx2,2)*pow(vy0,2)+36*vx0*pow(vx2,3)+21*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(81*pow(vx2,4)+30*vx0*pow(vx2,3))*vy0*pow(vy2,2)+(30*pow(vx2,4)*pow(vy0,2)+18*vx0*pow(vx2,5)+9*pow(vx0,2)*pow(vx2,4))*vy2+(21*pow(vx2,6)+18*vx0*pow(vx2,5))*vy0)*pow(vz0,5)+((72*pow(vy0,3)+57*pow(vx0,2)*vy0)*pow(vy2,6)+(48*pow(vy0,4)+(60*vx0*vx2+63*pow(vx0,2))*pow(vy0,2)+30*pow(vx0,3)*vx2+15*pow(vx0,4))*pow(vy2,5)+((171*pow(vx2,2)+36*vx0*vx2)*pow(vy0,3)+(156*pow(vx0,2)*pow(vx2,2)+36*pow(vx0,3)*vx2)*vy0)*pow(vy2,4)+(93*pow(vx2,2)*pow(vy0,4)+(120*vx0*pow(vx2,3)+126*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+60*pow(vx0,3)*pow(vx2,3)+33*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+((126*pow(vx2,4)+66*vx0*pow(vx2,3))*pow(vy0,3)+(141*pow(vx0,2)*pow(vx2,4)+66*pow(vx0,3)*pow(vx2,3))*vy0)*pow(vy2,2)+(45*pow(vx2,4)*pow(vy0,4)+(60*vx0*pow(vx2,5)+63*pow(vx0,2)*pow(vx2,4))*pow(vy0,2)+30*pow(vx0,3)*pow(vx2,5)+18*pow(vx0,4)*pow(vx2,4))*vy2+(27*pow(vx2,6)+30*vx0*pow(vx2,5))*pow(vy0,3)+(42*pow(vx0,2)*pow(vx2,6)+30*pow(vx0,3)*pow(vx2,5))*vy0)*pow(vz0,3)+((24*pow(vy0,5)+42*pow(vx0,2)*pow(vy0,3)+18*pow(vx0,4)*vy0)*pow(vy2,6)+(24*pow(vy0,6)+(24*vx0*vx2+54*pow(vx0,2))*pow(vy0,4)+(36*pow(vx0,3)*vx2+36*pow(vx0,4))*pow(vy0,2)+12*pow(vx0,5)*vx2+6*pow(vx0,6))*pow(vy2,5)+((54*pow(vx2,2)+24*vx0*vx2)*pow(vy0,5)+(102*pow(vx0,2)*pow(vx2,2)+48*pow(vx0,3)*vx2)*pow(vy0,3)+(48*pow(vx0,4)*pow(vx2,2)+24*pow(vx0,5)*vx2)*vy0)*pow(vy2,4)+(42*pow(vx2,2)*pow(vy0,6)+(48*vx0*pow(vx2,3)+102*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+(72*pow(vx0,3)*pow(vx2,3)+78*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+24*pow(vx0,5)*pow(vx2,3)+18*pow(vx0,6)*pow(vx2,2))*pow(vy2,3)+((36*pow(vx2,4)+36*vx0*pow(vx2,3))*pow(vy0,5)+(78*pow(vx0,2)*pow(vx2,4)+72*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(42*pow(vx0,4)*pow(vx2,4)+36*pow(vx0,5)*pow(vx2,3))*vy0)*pow(vy2,2)+(18*pow(vx2,4)*pow(vy0,6)+(24*vx0*pow(vx2,5)+48*pow(vx0,2)*pow(vx2,4))*pow(vy0,4)+(36*pow(vx0,3)*pow(vx2,5)+42*pow(vx0,4)*pow(vx2,4))*pow(vy0,2)+12*pow(vx0,5)*pow(vx2,5)+12*pow(vx0,6)*pow(vx2,4))*vy2+(6*pow(vx2,6)+12*vx0*pow(vx2,5))*pow(vy0,5)+(18*pow(vx0,2)*pow(vx2,6)+24*pow(vx0,3)*pow(vx2,5))*pow(vy0,3)+(12*pow(vx0,4)*pow(vx2,6)+12*pow(vx0,5)*pow(vx2,5))*vy0)*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((3*pow(vy2,7)+(9*pow(vx2,2)-3*vx0*vx2)*pow(vy2,5)+3*pow(vx2,2)*vy0*pow(vy2,4)+(9*pow(vx2,4)-6*vx0*pow(vx2,3))*pow(vy2,3)+6*pow(vx2,4)*vy0*pow(vy2,2)+(3*pow(vx2,6)-3*vx0*pow(vx2,5))*vy2+3*pow(vx2,6)*vy0)*pow(vz0,6)+((24*pow(vy0,2)+6*pow(vx0,2))*pow(vy2,7)+21*vx0*vx2*vy0*pow(vy2,6)+((69*pow(vx2,2)-9*vx0*vx2)*pow(vy0,2)+21*pow(vx0,2)*pow(vx2,2)-6*pow(vx0,3)*vx2)*pow(vy2,5)+(9*pow(vx2,2)*pow(vy0,3)+57*vx0*pow(vx2,3)*vy0)*pow(vy2,4)+((66*pow(vx2,4)-9*vx0*pow(vx2,3))*pow(vy0,2)+24*pow(vx0,2)*pow(vx2,4)-15*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(15*pow(vx2,4)*pow(vy0,3)+(51*vx0*pow(vx2,5)+9*pow(vx0,2)*pow(vx2,4))*vy0)*pow(vy2,2)+(21*pow(vx2,6)*pow(vy0,2)+9*pow(vx0,2)*pow(vx2,6)-9*pow(vx0,3)*pow(vx2,5))*vy2+6*pow(vx2,6)*pow(vy0,3)+(15*vx0*pow(vx2,7)+9*pow(vx0,2)*pow(vx2,6))*vy0)*pow(vz0,4)+((24*pow(vy0,4)+24*pow(vx0,2)*pow(vy0,2)+3*pow(vx0,4))*pow(vy2,7)+(30*vx0*vx2*pow(vy0,3)+21*pow(vx0,3)*vx2*vy0)*pow(vy2,6)+((66*pow(vx2,2)-6*vx0*vx2)*pow(vy0,4)+(81*pow(vx0,2)*pow(vx2,2)-9*pow(vx0,3)*vx2)*pow(vy0,2)+12*pow(vx0,4)*pow(vx2,2)-3*pow(vx0,5)*vx2)*pow(vy2,5)+(6*pow(vx2,2)*pow(vy0,5)+(75*vx0*pow(vx2,3)+3*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(60*pow(vx0,3)*pow(vx2,3)-3*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vy2,4)+((60*pow(vx2,4)-3*vx0*pow(vx2,3))*pow(vy0,4)+(90*pow(vx0,2)*pow(vx2,4)-12*pow(vx0,3)*pow(vx2,3))*pow(vy0,2)+15*pow(vx0,4)*pow(vx2,4)-9*pow(vx0,5)*pow(vx2,3))*pow(vy2,3)+(9*pow(vx2,4)*pow(vy0,5)+(60*vx0*pow(vx2,5)+12*pow(vx0,2)*pow(vx2,4))*pow(vy0,3)+(57*pow(vx0,3)*pow(vx2,5)+3*pow(vx0,4)*pow(vx2,4))*vy0)*pow(vy2,2)+((18*pow(vx2,6)+3*vx0*pow(vx2,5))*pow(vy0,4)+(33*pow(vx0,2)*pow(vx2,6)-3*pow(vx0,3)*pow(vx2,5))*pow(vy0,2)+6*pow(vx0,4)*pow(vx2,6)-6*pow(vx0,5)*pow(vx2,5))*vy2+3*pow(vx2,6)*pow(vy0,5)+(15*vx0*pow(vx2,7)+9*pow(vx0,2)*pow(vx2,6))*pow(vy0,3)+(18*pow(vx0,3)*pow(vx2,7)+6*pow(vx0,4)*pow(vx2,6))*vy0)*pow(vz0,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(24*vy0*pow(vz0,6)+(36*pow(vy0,3)+36*pow(vx0,2)*vy0)*pow(vz0,4)+(12*pow(vy0,5)+24*pow(vx0,2)*pow(vy0,3)+12*pow(vx0,4)*vy0)*pow(vz0,2))*pow(vz2,8)+((24*vy2+24*vy0)*pow(vz0,7)+((84*pow(vy0,2)+60*pow(vx0,2))*vy2+60*pow(vy0,3)+(24*vx0*vx2+60*pow(vx0,2))*vy0)*pow(vz0,5)+((72*pow(vy0,4)+120*pow(vx0,2)*pow(vy0,2)+48*pow(vx0,4))*vy2+48*pow(vy0,5)+(24*vx0*vx2+96*pow(vx0,2))*pow(vy0,3)+(24*pow(vx0,3)*vx2+48*pow(vx0,4))*vy0)*pow(vz0,3)+((12*pow(vy0,6)+36*pow(vx0,2)*pow(vy0,4)+36*pow(vx0,4)*pow(vy0,2)+12*pow(vx0,6))*vy2+12*pow(vy0,7)+36*pow(vx0,2)*pow(vy0,5)+36*pow(vx0,4)*pow(vy0,3)+12*pow(vx0,6)*vy0)*vz0)*pow(vz2,7)+(24*vy2*pow(vz0,8)+(108*vy0*pow(vy2,2)+(108*pow(vy0,2)+24*vx0*vx2+84*pow(vx0,2))*vy2+(84*pow(vx2,2)+24*vx0*vx2)*vy0)*pow(vz0,6)+((192*pow(vy0,3)+186*pow(vx0,2)*vy0)*pow(vy2,2)+(156*pow(vy0,4)+(66*vx0*vx2+258*pow(vx0,2))*pow(vy0,2)+54*pow(vx0,3)*vx2+102*pow(vx0,4))*vy2+(132*pow(vx2,2)+54*vx0*vx2)*pow(vy0,3)+(138*pow(vx0,2)*pow(vx2,2)+54*pow(vx0,3)*vx2)*vy0)*pow(vz0,4)+((84*pow(vy0,5)+168*pow(vx0,2)*pow(vy0,3)+84*pow(vx0,4)*vy0)*pow(vy2,2)+(84*pow(vy0,6)+(36*vx0*vx2+216*pow(vx0,2))*pow(vy0,4)+(72*pow(vx0,3)*vx2+180*pow(vx0,4))*pow(vy0,2)+36*pow(vx0,5)*vx2+48*pow(vx0,6))*vy2+(48*pow(vx2,2)+36*vx0*vx2)*pow(vy0,5)+(96*pow(vx0,2)*pow(vx2,2)+72*pow(vx0,3)*vx2)*pow(vy0,3)+(48*pow(vx0,4)*pow(vx2,2)+36*pow(vx0,5)*vx2)*vy0)*pow(vz0,2)+(6*pow(vx0,2)*pow(vy0,5)+12*pow(vx0,4)*pow(vy0,3)+6*pow(vx0,6)*vy0)*pow(vy2,2)+(12*pow(vy0,8)+(42*pow(vx0,2)-6*vx0*vx2)*pow(vy0,6)+(54*pow(vx0,4)-6*pow(vx0,3)*vx2)*pow(vy0,4)+(6*pow(vx0,5)*vx2+30*pow(vx0,6))*pow(vy0,2)+6*pow(vx0,7)*vx2+6*pow(vx0,8))*vy2+6*vx0*vx2*pow(vy0,7)+(18*pow(vx0,3)*vx2-6*pow(vx0,2)*pow(vx2,2))*pow(vy0,5)+(18*pow(vx0,5)*vx2-12*pow(vx0,4)*pow(vx2,2))*pow(vy0,3)+(6*pow(vx0,7)*vx2-6*pow(vx0,6)*pow(vx2,2))*vy0)*pow(vz2,6)+((60*pow(vy2,3)+84*vy0*pow(vy2,2)+(60*pow(vx2,2)+24*vx0*vx2)*vy2+60*pow(vx2,2)*vy0)*pow(vz0,7)+((240*pow(vy0,2)+150*pow(vx0,2))*pow(vy2,3)+(240*pow(vy0,3)+(96*vx0*vx2+234*pow(vx0,2))*vy0)*pow(vy2,2)+((234*pow(vx2,2)+96*vx0*vx2)*pow(vy0,2)+156*pow(vx0,2)*pow(vx2,2)+84*pow(vx0,3)*vx2)*vy2+150*pow(vx2,2)*pow(vy0,3)+(84*vx0*pow(vx2,3)+156*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,5)+((228*pow(vy0,4)+348*pow(vx0,2)*pow(vy0,2)+120*pow(vx0,4))*pow(vy2,3)+(228*pow(vy0,5)+(120*vx0*vx2+444*pow(vx0,2))*pow(vy0,3)+(120*pow(vx0,3)*vx2+216*pow(vx0,4))*vy0)*pow(vy2,2)+((216*pow(vx2,2)+120*vx0*vx2)*pow(vy0,4)+(348*pow(vx0,2)*pow(vx2,2)+216*pow(vx0,3)*vx2)*pow(vy0,2)+132*pow(vx0,4)*pow(vx2,2)+96*pow(vx0,5)*vx2)*vy2+120*pow(vx2,2)*pow(vy0,5)+(96*vx0*pow(vx2,3)+252*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(96*pow(vx0,3)*pow(vx2,3)+132*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vz0,3)+((48*pow(vy0,6)+126*pow(vx0,2)*pow(vy0,4)+108*pow(vx0,4)*pow(vy0,2)+30*pow(vx0,6))*pow(vy2,3)+(72*pow(vy0,7)+(24*vx0*vx2+210*pow(vx0,2))*pow(vy0,5)+(48*pow(vx0,3)*vx2+204*pow(vx0,4))*pow(vy0,3)+(24*pow(vx0,5)*vx2+66*pow(vx0,6))*vy0)*pow(vy2,2)+((42*pow(vx2,2)+48*vx0*vx2)*pow(vy0,6)+(120*pow(vx0,2)*pow(vx2,2)+132*pow(vx0,3)*vx2)*pow(vy0,4)+(114*pow(vx0,4)*pow(vx2,2)+120*pow(vx0,5)*vx2)*pow(vy0,2)+36*pow(vx0,6)*pow(vx2,2)+36*pow(vx0,7)*vx2)*vy2+30*pow(vx2,2)*pow(vy0,7)+(12*vx0*pow(vx2,3)+96*pow(vx0,2)*pow(vx2,2))*pow(vy0,5)+(24*pow(vx0,3)*pow(vx2,3)+102*pow(vx0,4)*pow(vx2,2))*pow(vy0,3)+(12*pow(vx0,5)*pow(vx2,3)+36*pow(vx0,6)*pow(vx2,2))*vy0)*vz0)*pow(vz2,5)+((36*pow(vy2,3)+36*pow(vx2,2)*vy2)*pow(vz0,8)+(156*vy0*pow(vy2,4)+(192*pow(vy0,2)+54*vx0*vx2+132*pow(vx0,2))*pow(vy2,3)+(258*pow(vx2,2)+66*vx0*vx2)*vy0*pow(vy2,2)+(186*pow(vx2,2)*pow(vy0,2)+54*vx0*pow(vx2,3)+138*pow(vx0,2)*pow(vx2,2))*vy2+(102*pow(vx2,4)+54*vx0*pow(vx2,3))*vy0)*pow(vz0,6)+((312*pow(vy0,3)+288*pow(vx0,2)*vy0)*pow(vy2,4)+(312*pow(vy0,4)+(168*vx0*vx2+480*pow(vx0,2))*pow(vy0,2)+120*pow(vx0,3)*vx2+168*pow(vx0,4))*pow(vy2,3)+((480*pow(vx2,2)+168*vx0*vx2)*pow(vy0,3)+(480*pow(vx0,2)*pow(vx2,2)+168*pow(vx0,3)*vx2)*vy0)*pow(vy2,2)+(288*pow(vx2,2)*pow(vy0,4)+(168*vx0*pow(vx2,3)+480*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+120*pow(vx0,3)*pow(vx2,3)+192*pow(vx0,4)*pow(vx2,2))*vy2+(168*pow(vx2,4)+120*vx0*pow(vx2,3))*pow(vy0,3)+(192*pow(vx0,2)*pow(vx2,4)+120*pow(vx0,3)*pow(vx2,3))*vy0)*pow(vz0,4)+((156*pow(vy0,5)+300*pow(vx0,2)*pow(vy0,3)+144*pow(vx0,4)*vy0)*pow(vy2,4)+(192*pow(vy0,6)+(102*vx0*vx2+468*pow(vx0,2))*pow(vy0,4)+(180*pow(vx0,3)*vx2+360*pow(vx0,4))*pow(vy0,2)+78*pow(vx0,5)*vx2+84*pow(vx0,6))*pow(vy2,3)+((222*pow(vx2,2)+138*vx0*vx2)*pow(vy0,5)+(444*pow(vx0,2)*pow(vx2,2)+276*pow(vx0,3)*vx2)*pow(vy0,3)+(222*pow(vx0,4)*pow(vx2,2)+138*pow(vx0,5)*vx2)*vy0)*pow(vy2,2)+(162*pow(vx2,2)*pow(vy0,6)+(102*vx0*pow(vx2,3)+438*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+(180*pow(vx0,3)*pow(vx2,3)+390*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+78*pow(vx0,5)*pow(vx2,3)+114*pow(vx0,6)*pow(vx2,2))*vy2+(66*pow(vx2,4)+78*vx0*pow(vx2,3))*pow(vy0,5)+(144*pow(vx0,2)*pow(vx2,4)+156*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(78*pow(vx0,4)*pow(vx2,4)+78*pow(vx0,5)*pow(vx2,3))*vy0)*pow(vz0,2)+(12*pow(vx0,2)*pow(vy0,5)+24*pow(vx0,4)*pow(vy0,3)+12*pow(vx0,6)*vy0)*pow(vy2,4)+(36*pow(vy0,8)+(120*pow(vx0,2)-12*vx0*vx2)*pow(vy0,6)+(144*pow(vx0,4)-12*pow(vx0,3)*vx2)*pow(vy0,4)+(12*pow(vx0,5)*vx2+72*pow(vx0,6))*pow(vy0,2)+12*pow(vx0,7)*vx2+12*pow(vx0,8))*pow(vy2,3)+(36*vx0*vx2*pow(vy0,7)+108*pow(vx0,3)*vx2*pow(vy0,5)+108*pow(vx0,5)*vx2*pow(vy0,3)+36*pow(vx0,7)*vx2*vy0)*pow(vy2,2)+(24*pow(vx2,2)*pow(vy0,8)+(96*pow(vx0,2)*pow(vx2,2)-12*vx0*pow(vx2,3))*pow(vy0,6)+(144*pow(vx0,4)*pow(vx2,2)-12*pow(vx0,3)*pow(vx2,3))*pow(vy0,4)+(12*pow(vx0,5)*pow(vx2,3)+96*pow(vx0,6)*pow(vx2,2))*pow(vy0,2)+12*pow(vx0,7)*pow(vx2,3)+24*pow(vx0,8)*pow(vx2,2))*vy2+12*vx0*pow(vx2,3)*pow(vy0,7)+(36*pow(vx0,3)*pow(vx2,3)-12*pow(vx0,2)*pow(vx2,4))*pow(vy0,5)+(36*pow(vx0,5)*pow(vx2,3)-24*pow(vx0,4)*pow(vx2,4))*pow(vy0,3)+(12*pow(vx0,7)*pow(vx2,3)-12*pow(vx0,6)*pow(vx2,4))*vy0)*pow(vz2,4)+((48*pow(vy2,5)+72*vy0*pow(vy2,4)+(96*pow(vx2,2)+24*vx0*vx2)*pow(vy2,3)+120*pow(vx2,2)*vy0*pow(vy2,2)+(48*pow(vx2,4)+24*vx0*pow(vx2,3))*vy2+48*pow(vx2,4)*vy0)*pow(vz0,7)+((228*pow(vy0,2)+120*pow(vx0,2))*pow(vy2,5)+(228*pow(vy0,3)+(120*vx0*vx2+216*pow(vx0,2))*vy0)*pow(vy2,4)+((444*pow(vx2,2)+120*vx0*vx2)*pow(vy0,2)+252*pow(vx0,2)*pow(vx2,2)+96*pow(vx0,3)*vx2)*pow(vy2,3)+(348*pow(vx2,2)*pow(vy0,3)+(216*vx0*pow(vx2,3)+348*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vy2,2)+((216*pow(vx2,4)+120*vx0*pow(vx2,3))*pow(vy0,2)+132*pow(vx0,2)*pow(vx2,4)+96*pow(vx0,3)*pow(vx2,3))*vy2+120*pow(vx2,4)*pow(vy0,3)+(96*vx0*pow(vx2,5)+132*pow(vx0,2)*pow(vx2,4))*vy0)*pow(vz0,5)+((240*pow(vy0,4)+336*pow(vx0,2)*pow(vy0,2)+96*pow(vx0,4))*pow(vy2,5)+(240*pow(vy0,5)+(168*vx0*vx2+456*pow(vx0,2))*pow(vy0,3)+(168*pow(vx0,3)*vx2+216*pow(vx0,4))*vy0)*pow(vy2,4)+((456*pow(vx2,2)+168*vx0*vx2)*pow(vy0,4)+(672*pow(vx0,2)*pow(vx2,2)+288*pow(vx0,3)*vx2)*pow(vy0,2)+216*pow(vx0,4)*pow(vx2,2)+120*pow(vx0,5)*vx2)*pow(vy2,3)+(336*pow(vx2,2)*pow(vy0,5)+(288*vx0*pow(vx2,3)+672*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(288*pow(vx0,3)*pow(vx2,3)+336*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vy2,2)+((216*pow(vx2,4)+168*vx0*pow(vx2,3))*pow(vy0,4)+(336*pow(vx0,2)*pow(vx2,4)+288*pow(vx0,3)*pow(vx2,3))*pow(vy0,2)+120*pow(vx0,4)*pow(vx2,4)+120*pow(vx0,5)*pow(vx2,3))*vy2+96*pow(vx2,4)*pow(vy0,5)+(120*vx0*pow(vx2,5)+216*pow(vx0,2)*pow(vx2,4))*pow(vy0,3)+(120*pow(vx0,3)*pow(vx2,5)+120*pow(vx0,4)*pow(vx2,4))*vy0)*pow(vz0,3)+((60*pow(vy0,6)+144*pow(vx0,2)*pow(vy0,4)+108*pow(vx0,4)*pow(vy0,2)+24*pow(vx0,6))*pow(vy2,5)+(84*pow(vy0,7)+(48*vx0*vx2+240*pow(vx0,2))*pow(vy0,5)+(96*pow(vx0,3)*vx2+228*pow(vx0,4))*pow(vy0,3)+(48*pow(vx0,5)*vx2+72*pow(vx0,6))*vy0)*pow(vy2,4)+((108*pow(vx2,2)+72*vx0*vx2)*pow(vy0,6)+(276*pow(vx0,2)*pow(vx2,2)+192*pow(vx0,3)*vx2)*pow(vy0,4)+(228*pow(vx0,4)*pow(vx2,2)+168*pow(vx0,5)*vx2)*pow(vy0,2)+60*pow(vx0,6)*pow(vx2,2)+48*pow(vx0,7)*vx2)*pow(vy2,3)+(108*pow(vx2,2)*pow(vy0,7)+(72*vx0*pow(vx2,3)+324*pow(vx0,2)*pow(vx2,2))*pow(vy0,5)+(144*pow(vx0,3)*pow(vx2,3)+324*pow(vx0,4)*pow(vx2,2))*pow(vy0,3)+(72*pow(vx0,5)*pow(vx2,3)+108*pow(vx0,6)*pow(vx2,2))*vy0)*pow(vy2,2)+((48*pow(vx2,4)+72*vx0*pow(vx2,3))*pow(vy0,6)+(132*pow(vx0,2)*pow(vx2,4)+192*pow(vx0,3)*pow(vx2,3))*pow(vy0,4)+(120*pow(vx0,4)*pow(vx2,4)+168*pow(vx0,5)*pow(vx2,3))*pow(vy0,2)+36*pow(vx0,6)*pow(vx2,4)+48*pow(vx0,7)*pow(vx2,3))*vy2+24*pow(vx2,4)*pow(vy0,7)+(24*vx0*pow(vx2,5)+84*pow(vx0,2)*pow(vx2,4))*pow(vy0,5)+(48*pow(vx0,3)*pow(vx2,5)+96*pow(vx0,4)*pow(vx2,4))*pow(vy0,3)+(24*pow(vx0,5)*pow(vx2,5)+36*pow(vx0,6)*pow(vx2,4))*vy0)*vz0)*pow(vz2,3)+((12*pow(vy2,5)+24*pow(vx2,2)*pow(vy2,3)+12*pow(vx2,4)*vy2)*pow(vz0,8)+(84*vy0*pow(vy2,6)+(84*pow(vy0,2)+36*vx0*vx2+48*pow(vx0,2))*pow(vy2,5)+(216*pow(vx2,2)+36*vx0*vx2)*vy0*pow(vy2,4)+(168*pow(vx2,2)*pow(vy0,2)+72*vx0*pow(vx2,3)+96*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(180*pow(vx2,4)+72*vx0*pow(vx2,3))*vy0*pow(vy2,2)+(84*pow(vx2,4)*pow(vy0,2)+36*vx0*pow(vx2,5)+48*pow(vx0,2)*pow(vx2,4))*vy2+(48*pow(vx2,6)+36*vx0*pow(vx2,5))*vy0)*pow(vz0,6)+((192*pow(vy0,3)+162*pow(vx0,2)*vy0)*pow(vy2,6)+(156*pow(vy0,4)+(138*vx0*vx2+222*pow(vx0,2))*pow(vy0,2)+78*pow(vx0,3)*vx2+66*pow(vx0,4))*pow(vy2,5)+((468*pow(vx2,2)+102*vx0*vx2)*pow(vy0,3)+(438*pow(vx0,2)*pow(vx2,2)+102*pow(vx0,3)*vx2)*vy0)*pow(vy2,4)+(300*pow(vx2,2)*pow(vy0,4)+(276*vx0*pow(vx2,3)+444*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+156*pow(vx0,3)*pow(vx2,3)+144*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+((360*pow(vx2,4)+180*vx0*pow(vx2,3))*pow(vy0,3)+(390*pow(vx0,2)*pow(vx2,4)+180*pow(vx0,3)*pow(vx2,3))*vy0)*pow(vy2,2)+(144*pow(vx2,4)*pow(vy0,4)+(138*vx0*pow(vx2,5)+222*pow(vx0,2)*pow(vx2,4))*pow(vy0,2)+78*pow(vx0,3)*pow(vx2,5)+78*pow(vx0,4)*pow(vx2,4))*vy2+(84*pow(vx2,6)+78*vx0*pow(vx2,5))*pow(vy0,3)+(114*pow(vx0,2)*pow(vx2,6)+78*pow(vx0,3)*pow(vx2,5))*vy0)*pow(vz0,4)+((108*pow(vy0,5)+192*pow(vx0,2)*pow(vy0,3)+84*pow(vx0,4)*vy0)*pow(vy2,6)+(108*pow(vy0,6)+(96*vx0*vx2+252*pow(vx0,2))*pow(vy0,4)+(144*pow(vx0,3)*vx2+180*pow(vx0,4))*pow(vy0,2)+48*pow(vx0,5)*vx2+36*pow(vx0,6))*pow(vy2,5)+((252*pow(vx2,2)+96*vx0*vx2)*pow(vy0,5)+(480*pow(vx0,2)*pow(vx2,2)+192*pow(vx0,3)*vx2)*pow(vy0,3)+(228*pow(vx0,4)*pow(vx2,2)+96*pow(vx0,5)*vx2)*vy0)*pow(vy2,4)+(192*pow(vx2,2)*pow(vy0,6)+(192*vx0*pow(vx2,3)+480*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+(288*pow(vx0,3)*pow(vx2,3)+384*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+96*pow(vx0,5)*pow(vx2,3)+96*pow(vx0,6)*pow(vx2,2))*pow(vy2,3)+((180*pow(vx2,4)+144*vx0*pow(vx2,3))*pow(vy0,5)+(384*pow(vx0,2)*pow(vx2,4)+288*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(204*pow(vx0,4)*pow(vx2,4)+144*pow(vx0,5)*pow(vx2,3))*vy0)*pow(vy2,2)+(84*pow(vx2,4)*pow(vy0,6)+(96*vx0*pow(vx2,5)+228*pow(vx0,2)*pow(vx2,4))*pow(vy0,4)+(144*pow(vx0,3)*pow(vx2,5)+204*pow(vx0,4)*pow(vx2,4))*pow(vy0,2)+48*pow(vx0,5)*pow(vx2,5)+60*pow(vx0,6)*pow(vx2,4))*vy2+(36*pow(vx2,6)+48*vx0*pow(vx2,5))*pow(vy0,5)+(96*pow(vx0,2)*pow(vx2,6)+96*pow(vx0,3)*pow(vx2,5))*pow(vy0,3)+(60*pow(vx0,4)*pow(vx2,6)+48*pow(vx0,5)*pow(vx2,5))*vy0)*pow(vz0,2)+(6*pow(vx0,2)*pow(vy0,5)+12*pow(vx0,4)*pow(vy0,3)+6*pow(vx0,6)*vy0)*pow(vy2,6)+(24*pow(vy0,8)+(78*pow(vx0,2)-6*vx0*vx2)*pow(vy0,6)+(90*pow(vx0,4)-6*pow(vx0,3)*vx2)*pow(vy0,4)+(6*pow(vx0,5)*vx2+42*pow(vx0,6))*pow(vy0,2)+6*pow(vx0,7)*vx2+6*pow(vx0,8))*pow(vy2,5)+(30*vx0*vx2*pow(vy0,7)+(6*pow(vx0,2)*pow(vx2,2)+90*pow(vx0,3)*vx2)*pow(vy0,5)+(12*pow(vx0,4)*pow(vx2,2)+90*pow(vx0,5)*vx2)*pow(vy0,3)+(6*pow(vx0,6)*pow(vx2,2)+30*pow(vx0,7)*vx2)*vy0)*pow(vy2,4)+(36*pow(vx2,2)*pow(vy0,8)+(132*pow(vx0,2)*pow(vx2,2)-12*vx0*pow(vx2,3))*pow(vy0,6)+(180*pow(vx0,4)*pow(vx2,2)-12*pow(vx0,3)*pow(vx2,3))*pow(vy0,4)+(12*pow(vx0,5)*pow(vx2,3)+108*pow(vx0,6)*pow(vx2,2))*pow(vy0,2)+12*pow(vx0,7)*pow(vx2,3)+24*pow(vx0,8)*pow(vx2,2))*pow(vy2,3)+(36*vx0*pow(vx2,3)*pow(vy0,7)+(108*pow(vx0,3)*pow(vx2,3)-6*pow(vx0,2)*pow(vx2,4))*pow(vy0,5)+(108*pow(vx0,5)*pow(vx2,3)-12*pow(vx0,4)*pow(vx2,4))*pow(vy0,3)+(36*pow(vx0,7)*pow(vx2,3)-6*pow(vx0,6)*pow(vx2,4))*vy0)*pow(vy2,2)+(12*pow(vx2,4)*pow(vy0,8)+(54*pow(vx0,2)*pow(vx2,4)-6*vx0*pow(vx2,5))*pow(vy0,6)+(90*pow(vx0,4)*pow(vx2,4)-6*pow(vx0,3)*pow(vx2,5))*pow(vy0,4)+(6*pow(vx0,5)*pow(vx2,5)+66*pow(vx0,6)*pow(vx2,4))*pow(vy0,2)+6*pow(vx0,7)*pow(vx2,5)+18*pow(vx0,8)*pow(vx2,4))*vy2+6*vx0*pow(vx2,5)*pow(vy0,7)+(18*pow(vx0,3)*pow(vx2,5)-6*pow(vx0,2)*pow(vx2,6))*pow(vy0,5)+(18*pow(vx0,5)*pow(vx2,5)-12*pow(vx0,4)*pow(vx2,6))*pow(vy0,3)+(6*pow(vx0,7)*pow(vx2,5)-6*pow(vx0,6)*pow(vx2,6))*vy0)*pow(vz2,2)+((12*pow(vy2,7)+12*vy0*pow(vy2,6)+36*pow(vx2,2)*pow(vy2,5)+36*pow(vx2,2)*vy0*pow(vy2,4)+36*pow(vx2,4)*pow(vy2,3)+36*pow(vx2,4)*vy0*pow(vy2,2)+12*pow(vx2,6)*vy2+12*pow(vx2,6)*vy0)*pow(vz0,7)+((72*pow(vy0,2)+30*pow(vx0,2))*pow(vy2,7)+(48*pow(vy0,3)+(48*vx0*vx2+42*pow(vx0,2))*vy0)*pow(vy2,6)+((210*pow(vx2,2)+24*vx0*vx2)*pow(vy0,2)+96*pow(vx0,2)*pow(vx2,2)+12*pow(vx0,3)*vx2)*pow(vy2,5)+(126*pow(vx2,2)*pow(vy0,3)+(132*vx0*pow(vx2,3)+120*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vy2,4)+((204*pow(vx2,4)+48*vx0*pow(vx2,3))*pow(vy0,2)+102*pow(vx0,2)*pow(vx2,4)+24*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(108*pow(vx2,4)*pow(vy0,3)+(120*vx0*pow(vx2,5)+114*pow(vx0,2)*pow(vx2,4))*vy0)*pow(vy2,2)+((66*pow(vx2,6)+24*vx0*pow(vx2,5))*pow(vy0,2)+36*pow(vx0,2)*pow(vx2,6)+12*pow(vx0,3)*pow(vx2,5))*vy2+30*pow(vx2,6)*pow(vy0,3)+(36*vx0*pow(vx2,7)+36*pow(vx0,2)*pow(vx2,6))*vy0)*pow(vz0,5)+((84*pow(vy0,4)+108*pow(vx0,2)*pow(vy0,2)+24*pow(vx0,4))*pow(vy2,7)+(60*pow(vy0,5)+(72*vx0*vx2+108*pow(vx0,2))*pow(vy0,3)+(72*pow(vx0,3)*vx2+48*pow(vx0,4))*vy0)*pow(vy2,6)+((240*pow(vx2,2)+48*vx0*vx2)*pow(vy0,4)+(324*pow(vx0,2)*pow(vx2,2)+72*pow(vx0,3)*vx2)*pow(vy0,2)+84*pow(vx0,4)*pow(vx2,2)+24*pow(vx0,5)*vx2)*pow(vy2,5)+(144*pow(vx2,2)*pow(vy0,5)+(192*vx0*pow(vx2,3)+276*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+(192*pow(vx0,3)*pow(vx2,3)+132*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vy2,4)+((228*pow(vx2,4)+96*vx0*pow(vx2,3))*pow(vy0,4)+(324*pow(vx0,2)*pow(vx2,4)+144*pow(vx0,3)*pow(vx2,3))*pow(vy0,2)+96*pow(vx0,4)*pow(vx2,4)+48*pow(vx0,5)*pow(vx2,3))*pow(vy2,3)+(108*pow(vx2,4)*pow(vy0,5)+(168*vx0*pow(vx2,5)+228*pow(vx0,2)*pow(vx2,4))*pow(vy0,3)+(168*pow(vx0,3)*pow(vx2,5)+120*pow(vx0,4)*pow(vx2,4))*vy0)*pow(vy2,2)+((72*pow(vx2,6)+48*vx0*pow(vx2,5))*pow(vy0,4)+(108*pow(vx0,2)*pow(vx2,6)+72*pow(vx0,3)*pow(vx2,5))*pow(vy0,2)+36*pow(vx0,4)*pow(vx2,6)+24*pow(vx0,5)*pow(vx2,5))*vy2+24*pow(vx2,6)*pow(vy0,5)+(48*vx0*pow(vx2,7)+60*pow(vx0,2)*pow(vx2,6))*pow(vy0,3)+(48*pow(vx0,3)*pow(vx2,7)+36*pow(vx0,4)*pow(vx2,6))*vy0)*pow(vz0,3)+((24*pow(vy0,6)+54*pow(vx0,2)*pow(vy0,4)+36*pow(vx0,4)*pow(vy0,2)+6*pow(vx0,6))*pow(vy2,7)+(24*pow(vy0,7)+(24*vx0*vx2+66*pow(vx0,2))*pow(vy0,5)+(48*pow(vx0,3)*vx2+60*pow(vx0,4))*pow(vy0,3)+(24*pow(vx0,5)*vx2+18*pow(vx0,6))*vy0)*pow(vy2,6)+((66*pow(vx2,2)+24*vx0*vx2)*pow(vy0,6)+(156*pow(vx0,2)*pow(vx2,2)+60*pow(vx0,3)*vx2)*pow(vy0,4)+(114*pow(vx0,4)*pow(vx2,2)+48*pow(vx0,5)*vx2)*pow(vy0,2)+24*pow(vx0,6)*pow(vx2,2)+12*pow(vx0,7)*vx2)*pow(vy2,5)+(54*pow(vx2,2)*pow(vy0,7)+(60*vx0*pow(vx2,3)+156*pow(vx0,2)*pow(vx2,2))*pow(vy0,5)+(120*pow(vx0,3)*pow(vx2,3)+150*pow(vx0,4)*pow(vx2,2))*pow(vy0,3)+(60*pow(vx0,5)*pow(vx2,3)+48*pow(vx0,6)*pow(vx2,2))*vy0)*pow(vy2,4)+((60*pow(vx2,4)+48*vx0*pow(vx2,3))*pow(vy0,6)+(150*pow(vx0,2)*pow(vx2,4)+120*pow(vx0,3)*pow(vx2,3))*pow(vy0,4)+(120*pow(vx0,4)*pow(vx2,4)+96*pow(vx0,5)*pow(vx2,3))*pow(vy0,2)+30*pow(vx0,6)*pow(vx2,4)+24*pow(vx0,7)*pow(vx2,3))*pow(vy2,3)+(36*pow(vx2,4)*pow(vy0,7)+(48*vx0*pow(vx2,5)+114*pow(vx0,2)*pow(vx2,4))*pow(vy0,5)+(96*pow(vx0,3)*pow(vx2,5)+120*pow(vx0,4)*pow(vx2,4))*pow(vy0,3)+(48*pow(vx0,5)*pow(vx2,5)+42*pow(vx0,6)*pow(vx2,4))*vy0)*pow(vy2,2)+((18*pow(vx2,6)+24*vx0*pow(vx2,5))*pow(vy0,6)+(48*pow(vx0,2)*pow(vx2,6)+60*pow(vx0,3)*pow(vx2,5))*pow(vy0,4)+(42*pow(vx0,4)*pow(vx2,6)+48*pow(vx0,5)*pow(vx2,5))*pow(vy0,2)+12*pow(vx0,6)*pow(vx2,6)+12*pow(vx0,7)*pow(vx2,5))*vy2+6*pow(vx2,6)*pow(vy0,7)+(12*vx0*pow(vx2,7)+24*pow(vx0,2)*pow(vx2,6))*pow(vy0,5)+(24*pow(vx0,3)*pow(vx2,7)+30*pow(vx0,4)*pow(vx2,6))*pow(vy0,3)+(12*pow(vx0,5)*pow(vx2,7)+12*pow(vx0,6)*pow(vx2,6))*vy0)*vz0)*vz2+(12*vy0*pow(vy2,8)+6*vx0*vx2*pow(vy2,7)+(42*pow(vx2,2)-6*vx0*vx2)*vy0*pow(vy2,6)+(6*pow(vx2,2)*pow(vy0,2)+18*vx0*pow(vx2,3)-6*pow(vx0,2)*pow(vx2,2))*pow(vy2,5)+(54*pow(vx2,4)-6*vx0*pow(vx2,3))*vy0*pow(vy2,4)+(12*pow(vx2,4)*pow(vy0,2)+18*vx0*pow(vx2,5)-12*pow(vx0,2)*pow(vx2,4))*pow(vy2,3)+(30*pow(vx2,6)+6*vx0*pow(vx2,5))*vy0*pow(vy2,2)+(6*pow(vx2,6)*pow(vy0,2)+6*vx0*pow(vx2,7)-6*pow(vx0,2)*pow(vx2,6))*vy2+(6*pow(vx2,8)+6*vx0*pow(vx2,7))*vy0)*pow(vz0,6)+((36*pow(vy0,3)+24*pow(vx0,2)*vy0)*pow(vy2,8)+(36*vx0*vx2*pow(vy0,2)+12*pow(vx0,3)*vx2)*pow(vy2,7)+((120*pow(vx2,2)-12*vx0*vx2)*pow(vy0,3)+(96*pow(vx0,2)*pow(vx2,2)-12*pow(vx0,3)*vx2)*vy0)*pow(vy2,6)+(12*pow(vx2,2)*pow(vy0,4)+108*vx0*pow(vx2,3)*pow(vy0,2)+36*pow(vx0,3)*pow(vx2,3)-12*pow(vx0,4)*pow(vx2,2))*pow(vy2,5)+((144*pow(vx2,4)-12*vx0*pow(vx2,3))*pow(vy0,3)+(144*pow(vx0,2)*pow(vx2,4)-12*pow(vx0,3)*pow(vx2,3))*vy0)*pow(vy2,4)+(24*pow(vx2,4)*pow(vy0,4)+108*vx0*pow(vx2,5)*pow(vy0,2)+36*pow(vx0,3)*pow(vx2,5)-24*pow(vx0,4)*pow(vx2,4))*pow(vy2,3)+((72*pow(vx2,6)+12*vx0*pow(vx2,5))*pow(vy0,3)+(96*pow(vx0,2)*pow(vx2,6)+12*pow(vx0,3)*pow(vx2,5))*vy0)*pow(vy2,2)+(12*pow(vx2,6)*pow(vy0,4)+36*vx0*pow(vx2,7)*pow(vy0,2)+12*pow(vx0,3)*pow(vx2,7)-12*pow(vx0,4)*pow(vx2,6))*vy2+(12*pow(vx2,8)+12*vx0*pow(vx2,7))*pow(vy0,3)+(24*pow(vx0,2)*pow(vx2,8)+12*pow(vx0,3)*pow(vx2,7))*vy0)*pow(vz0,4)+((24*pow(vy0,5)+36*pow(vx0,2)*pow(vy0,3)+12*pow(vx0,4)*vy0)*pow(vy2,8)+(30*vx0*vx2*pow(vy0,4)+36*pow(vx0,3)*vx2*pow(vy0,2)+6*pow(vx0,5)*vx2)*pow(vy2,7)+((78*pow(vx2,2)-6*vx0*vx2)*pow(vy0,5)+(132*pow(vx0,2)*pow(vx2,2)-12*pow(vx0,3)*vx2)*pow(vy0,3)+(54*pow(vx0,4)*pow(vx2,2)-6*pow(vx0,5)*vx2)*vy0)*pow(vy2,6)+(6*pow(vx2,2)*pow(vy0,6)+(90*vx0*pow(vx2,3)+6*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+(108*pow(vx0,3)*pow(vx2,3)-6*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+18*pow(vx0,5)*pow(vx2,3)-6*pow(vx0,6)*pow(vx2,2))*pow(vy2,5)+((90*pow(vx2,4)-6*vx0*pow(vx2,3))*pow(vy0,5)+(180*pow(vx0,2)*pow(vx2,4)-12*pow(vx0,3)*pow(vx2,3))*pow(vy0,3)+(90*pow(vx0,4)*pow(vx2,4)-6*pow(vx0,5)*pow(vx2,3))*vy0)*pow(vy2,4)+(12*pow(vx2,4)*pow(vy0,6)+(90*vx0*pow(vx2,5)+12*pow(vx0,2)*pow(vx2,4))*pow(vy0,4)+(108*pow(vx0,3)*pow(vx2,5)-12*pow(vx0,4)*pow(vx2,4))*pow(vy0,2)+18*pow(vx0,5)*pow(vx2,5)-12*pow(vx0,6)*pow(vx2,4))*pow(vy2,3)+((42*pow(vx2,6)+6*vx0*pow(vx2,5))*pow(vy0,5)+(108*pow(vx0,2)*pow(vx2,6)+12*pow(vx0,3)*pow(vx2,5))*pow(vy0,3)+(66*pow(vx0,4)*pow(vx2,6)+6*pow(vx0,5)*pow(vx2,5))*vy0)*pow(vy2,2)+(6*pow(vx2,6)*pow(vy0,6)+(30*vx0*pow(vx2,7)+6*pow(vx0,2)*pow(vx2,6))*pow(vy0,4)+(36*pow(vx0,3)*pow(vx2,7)-6*pow(vx0,4)*pow(vx2,6))*pow(vy0,2)+6*pow(vx0,5)*pow(vx2,7)-6*pow(vx0,6)*pow(vx2,6))*vy2+(6*pow(vx2,8)+6*vx0*pow(vx2,7))*pow(vy0,5)+(24*pow(vx0,2)*pow(vx2,8)+12*pow(vx0,3)*pow(vx2,7))*pow(vy0,3)+(18*pow(vx0,4)*pow(vx2,8)+6*pow(vx0,5)*pow(vx2,7))*vy0)*pow(vz0,2))*pow(((160*pow(roughness,2)*pow(vz0,7)+(400*pow(roughness,2)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2))*pow(vz0,5)+(320*pow(roughness,2)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,7)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,6)+((720*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,7)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,6)+240*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+((400*pow(roughness,2)*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2))*pow(vz0,7)+((1280*pow(roughness,2)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+((1360*pow(roughness,2)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+880*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+((480*pow(roughness,2)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+240*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((720*pow(roughness,2)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*vx2*pow(vy2,2)+720*pow(roughness,2)*pow(vx2,2)*vy0*vy2+720*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,6)+((1760*pow(roughness,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1680*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1760*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,4)+((1360*pow(roughness,2)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(1680*pow(roughness,2)*vx0*vx2*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((320*pow(roughness,2)*pow(vy2,4)+640*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,4))*pow(vz0,7)+((1360*pow(roughness,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(2240*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+880*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+((1760*pow(roughness,2)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(2560*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+800*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1760*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,3)+((720*pow(roughness,2)*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((480*pow(roughness,2)*vy0*pow(vy2,5)+480*pow(roughness,2)*vx0*vx2*pow(vy2,4)+960*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,4)*vy0*vy2+480*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,6)+((1360*pow(roughness,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(2560*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(2880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(1200*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+1200*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+1360*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,4)+((1280*pow(roughness,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(1920*pow(roughness,2)*vx0*vx2*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(2240*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2880*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(2880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(960*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+1280*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,2)+(400*pow(roughness,2)*pow(vy0,7)+1040*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+1040*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6))*pow(vz0,7)+((480*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(1200*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+240*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,5)+((720*pow(roughness,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2880*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,7)+80*pow(roughness,2)*vx0*vx2*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,6)+((320*pow(roughness,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(880*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(800*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(1040*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(1680*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(880*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,7)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,7)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,6)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+1280*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+1360*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+1760*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+1360*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,4)+(320*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+1360*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+1760*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+1360*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+1280*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,6)*pow(vy0,7)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,7)*pow(vx2,7))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,8)+(480*pow(roughness,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2))*pow(vz0,6)+(500*pow(roughness,2)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4))*pow(vz0,4)+(200*pow(roughness,2)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,8)+80*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+120*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+80*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vz2,8)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,7)+((880*pow(roughness,2)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+880*pow(roughness,2)*vx0*vx2*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,5)+((800*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*vz0)*pow(vz2,7)+((480*pow(roughness,2)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,2))*pow(vz0,8)+((1720*pow(roughness,2)*pow(vy0,2)+1480*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1480*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+((2200*pow(roughness,2)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+((1160*pow(roughness,2)*pow(vy0,6)+3000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+680*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+1160*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz0,2)+(200*pow(roughness,2)*pow(vy0,8)+680*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+840*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+440*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,8))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,7)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*vy2+80*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+440*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+680*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vz2,6)+((880*pow(roughness,2)*vy0*pow(vy2,3)+880*pow(roughness,2)*vx0*vx2*pow(vy2,2)+880*pow(roughness,2)*pow(vx2,2)*vy0*vy2+880*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,7)+((2560*pow(roughness,2)*pow(vy0,3)+2480*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(2720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(2480*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+2480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,5)+((2480*pow(roughness,2)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(2320*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+2320*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vz0,3)+((800*pow(roughness,2)*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,6)+2640*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2640*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*vz0)*pow(vz2,5)+((500*pow(roughness,2)*pow(vy2,4)+1000*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+500*pow(roughness,2)*pow(vx2,4))*pow(vz0,8)+((2200*pow(roughness,2)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+1200*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(3800*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+1200*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+1600*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+((3400*pow(roughness,2)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+1820*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(3200*pow(roughness,2)*vx0*vx2*pow(vy0,3)+3120*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(5200*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+10520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(3120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+1820*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+3400*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,4)+((2200*pow(roughness,2)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+3840*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(3000*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+3000*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(2640*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+840*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+3840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz0,2)+(500*pow(roughness,2)*pow(vy0,8)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1820*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+840*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,8))*pow(vy2,4)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+2240*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+3840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+1820*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vz2,4)+((800*pow(roughness,2)*vy0*pow(vy2,5)+800*pow(roughness,2)*vx0*vx2*pow(vy2,4)+1600*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+800*pow(roughness,2)*pow(vx2,4)*vy0*vy2+800*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,7)+((2480*pow(roughness,2)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(4800*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(5120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(2320*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+2320*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,5)+((2560*pow(roughness,2)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+2240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(3200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+5440*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(4800*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+10240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+5440*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(5440*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+10240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+4800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(2240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+3200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+2240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+2560*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vz0,3)+((880*pow(roughness,2)*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,6)+3120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+2640*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+2640*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+3120*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*vz0)*pow(vz2,3)+((200*pow(roughness,2)*pow(vy2,6)+600*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+600*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+200*pow(roughness,2)*pow(vx2,6))*pow(vz0,8)+((1160*pow(roughness,2)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(3000*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+1920*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(2520*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+680*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+1160*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,6)+((2200*pow(roughness,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(2800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(5200*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+3840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(5440*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(3840*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(2640*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+2800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+840*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,4)+((1720*pow(roughness,2)*pow(vy0,6)+3800*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+440*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(2720*pow(roughness,2)*vx0*vx2*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+2400*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(3800*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+10520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+9240*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+2520*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(5120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+10240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+5120*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(2520*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+9240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+10520*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+3800*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(2400*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+2720*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+440*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+1720*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vz0,2)+(480*pow(roughness,2)*pow(vy0,8)+1480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+1600*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+680*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,8))*pow(vy2,6)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+2320*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,5)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,4)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+680*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+1600*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+1480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,8)*pow(vx2,6))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,7)+240*pow(roughness,2)*vx0*vx2*pow(vy2,6)+720*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+720*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,6)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,7)+((800*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(2320*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(2640*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(2240*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(2400*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,5)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(2480*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+5120*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+2640*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(3120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+5440*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(2320*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+5440*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+3120*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(2640*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+5120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,2)*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,4)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,6)*vy0)*pow(vy2,7)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,6)+1200*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,7)*vx2)*pow(vy2,6)+(880*pow(roughness,2)*pow(vx2,2)*pow(vy0,7)+2720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,5)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,6)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3))*pow(vy2,4)+(800*pow(roughness,2)*pow(vx2,4)*pow(vy0,7)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,4)+2720*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,7)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,5)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,6)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,7))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,8)+80*pow(roughness,2)*pow(vx2,2)*pow(vy2,6)+120*pow(roughness,2)*pow(vx2,4)*pow(vy2,4)+80*pow(roughness,2)*pow(vx2,6)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,8))*pow(vz0,8)+((200*pow(roughness,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,2))*pow(vy2,8)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,7)+(680*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+440*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,6)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,5)+(840*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,4)+720*pow(roughness,2)*vx0*pow(vx2,5)*vy0*pow(vy2,3)+(440*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,7)*vy0*vy2+80*pow(roughness,2)*pow(vx2,8)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,8))*pow(vz0,6)+((500*pow(roughness,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4))*pow(vy2,8)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,7)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+840*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,6)+(2320*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+2240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,5)+(1820*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+3840*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1820*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,4)+(2240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*pow(vy2,3)+(840*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+2520*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,8)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,8))*pow(vz0,4)+((480*pow(roughness,2)*pow(vy0,6)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,8)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,7)+(1480*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+3800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+3000*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+680*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,6)+(2480*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+2320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,5)+(1600*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+5200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+5200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,4)+(2320*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+4800*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*pow(vy2,3)+(680*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+3000*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+3800*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+1480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,8)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,8)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,8))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,8)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vy0,4)+200*pow(roughness,2)*pow(vx0,6)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,8))*pow(vy2,8)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,7)+880*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,5)+800*pow(roughness,2)*pow(vx0,5)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,7)*vx2*vy0)*pow(vy2,7)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,8)+1720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,6)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,4)+1160*pow(roughness,2)*pow(vx0,6)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,8)*pow(vx2,2))*pow(vy2,6)+(880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,7)+2560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,5)+2480*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,7)*pow(vx2,3)*vy0)*pow(vy2,5)+(500*pow(roughness,2)*pow(vx2,4)*pow(vy0,8)+2200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,6)+3400*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,4)+2200*pow(roughness,2)*pow(vx0,6)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,8)*pow(vx2,4))*pow(vy2,4)+(800*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,7)+2480*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,5)+2560*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,7)*pow(vx2,5)*vy0)*pow(vy2,3)+(200*pow(roughness,2)*pow(vx2,6)*pow(vy0,8)+1160*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,6)+2200*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,4)+1720*pow(roughness,2)*pow(vx0,6)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,8)*pow(vx2,6))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,7)+800*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,7)*pow(vx2,7)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,8)*pow(vy0,8)+200*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,6)+500*pow(roughness,2)*pow(vx0,4)*pow(vx2,8)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,8)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,8)*pow(vx2,8),-1) ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
// x2  :  
poly.add(Term6f( -((((24*bx2*vy0+24*ax2*vx0)*pow(vz0,4)+(24*bx2*pow(vy0,3)+24*ax2*vx0*pow(vy0,2)+24*bx2*pow(vx0,2)*vy0+24*ax2*pow(vx0,3))*pow(vz0,2)+3*bx2*pow(vy0,5)+3*ax2*vx0*pow(vy0,4)+6*bx2*pow(vx0,2)*pow(vy0,3)+6*ax2*pow(vx0,3)*pow(vy0,2)+3*bx2*pow(vx0,4)*vy0+3*ax2*pow(vx0,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+((24*bx2*vy2+24*ax2*vx2)*pow(vz0,5)+(-24*nx2*vy0*vy2-24*nx2*vx0*vx2)*pow(vz0,4)+((72*bx2*pow(vy0,2)+24*ax2*vx0*vy0+48*bx2*pow(vx0,2))*vy2+48*ax2*vx2*pow(vy0,2)+24*bx2*vx0*vx2*vy0+72*ax2*pow(vx0,2)*vx2)*pow(vz0,3)+((-24*nx2*pow(vy0,3)-24*nx2*pow(vx0,2)*vy0)*vy2-24*nx2*vx0*vx2*pow(vy0,2)-24*nx2*pow(vx0,3)*vx2)*pow(vz0,2)+((39*bx2*pow(vy0,4)+18*ax2*vx0*pow(vy0,3)+60*bx2*pow(vx0,2)*pow(vy0,2)+18*ax2*pow(vx0,3)*vy0+21*bx2*pow(vx0,4))*vy2+21*ax2*vx2*pow(vy0,4)+18*bx2*vx0*vx2*pow(vy0,3)+60*ax2*pow(vx0,2)*vx2*pow(vy0,2)+18*bx2*pow(vx0,3)*vx2*vy0+39*ax2*pow(vx0,4)*vx2)*vz0+(-3*nx2*pow(vy0,5)-6*nx2*pow(vx0,2)*pow(vy0,3)-3*nx2*pow(vx0,4)*vy0)*vy2-3*nx2*vx0*vx2*pow(vy0,4)-6*nx2*pow(vx0,3)*vx2*pow(vy0,2)-3*nx2*pow(vx0,5)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((-24*nx2*pow(vy2,2)-24*nx2*pow(vx2,2))*pow(vz0,5)+((72*bx2*vy0+48*ax2*vx0)*pow(vy2,2)+(24*ax2*vx2*vy0+24*bx2*vx0*vx2)*vy2+48*bx2*pow(vx2,2)*vy0+72*ax2*vx0*pow(vx2,2))*pow(vz0,4)+((-72*nx2*pow(vy0,2)-48*nx2*pow(vx0,2))*pow(vy2,2)-48*nx2*vx0*vx2*vy0*vy2-48*nx2*pow(vx2,2)*pow(vy0,2)-72*nx2*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+((102*bx2*pow(vy0,3)+54*ax2*vx0*pow(vy0,2)+96*bx2*pow(vx0,2)*vy0+48*ax2*pow(vx0,3))*pow(vy2,2)+(48*ax2*vx2*pow(vy0,3)+60*bx2*vx0*vx2*pow(vy0,2)+60*ax2*pow(vx0,2)*vx2*vy0+48*bx2*pow(vx0,3)*vx2)*vy2+48*bx2*pow(vx2,2)*pow(vy0,3)+96*ax2*vx0*pow(vx2,2)*pow(vy0,2)+54*bx2*pow(vx0,2)*pow(vx2,2)*vy0+102*ax2*pow(vx0,3)*pow(vx2,2))*pow(vz0,2)+((-39*nx2*pow(vy0,4)-60*nx2*pow(vx0,2)*pow(vy0,2)-21*nx2*pow(vx0,4))*pow(vy2,2)+(-36*nx2*vx0*vx2*pow(vy0,3)-36*nx2*pow(vx0,3)*vx2*vy0)*vy2-21*nx2*pow(vx2,2)*pow(vy0,4)-60*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-39*nx2*pow(vx0,4)*pow(vx2,2))*vz0+(24*bx2*pow(vy0,5)+9*ax2*vx0*pow(vy0,4)+45*bx2*pow(vx0,2)*pow(vy0,3)+15*ax2*pow(vx0,3)*pow(vy0,2)+21*bx2*pow(vx0,4)*vy0+6*ax2*pow(vx0,5))*pow(vy2,2)+(15*ax2*vx2*pow(vy0,5)+21*bx2*vx0*vx2*pow(vy0,4)+36*ax2*pow(vx0,2)*vx2*pow(vy0,3)+36*bx2*pow(vx0,3)*vx2*pow(vy0,2)+21*ax2*pow(vx0,4)*vx2*vy0+15*bx2*pow(vx0,5)*vx2)*vy2+6*bx2*pow(vx2,2)*pow(vy0,5)+21*ax2*vx0*pow(vx2,2)*pow(vy0,4)+15*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+45*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+9*bx2*pow(vx0,4)*pow(vx2,2)*vy0+24*ax2*pow(vx0,5)*pow(vx2,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((24*bx2*pow(vy2,3)+24*ax2*vx2*pow(vy2,2)+24*bx2*pow(vx2,2)*vy2+24*ax2*pow(vx2,3))*pow(vz0,5)+(-72*nx2*vy0*pow(vy2,3)-72*nx2*vx0*vx2*pow(vy2,2)-72*nx2*pow(vx2,2)*vy0*vy2-72*nx2*vx0*pow(vx2,3))*pow(vz0,4)+((102*bx2*pow(vy0,2)+42*ax2*vx0*vy0+54*bx2*pow(vx0,2))*pow(vy2,3)+(60*ax2*vx2*pow(vy0,2)+54*bx2*vx0*vx2*vy0+96*ax2*pow(vx0,2)*vx2)*pow(vy2,2)+(96*bx2*pow(vx2,2)*pow(vy0,2)+54*ax2*vx0*pow(vx2,2)*vy0+60*bx2*pow(vx0,2)*pow(vx2,2))*vy2+54*ax2*pow(vx2,3)*pow(vy0,2)+42*bx2*vx0*pow(vx2,3)*vy0+102*ax2*pow(vx0,2)*pow(vx2,3))*pow(vz0,3)+((-102*nx2*pow(vy0,3)-96*nx2*pow(vx0,2)*vy0)*pow(vy2,3)+(-114*nx2*vx0*vx2*pow(vy0,2)-96*nx2*pow(vx0,3)*vx2)*pow(vy2,2)+(-96*nx2*pow(vx2,2)*pow(vy0,3)-114*nx2*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-96*nx2*vx0*pow(vx2,3)*pow(vy0,2)-102*nx2*pow(vx0,3)*pow(vx2,3))*pow(vz0,2)+((72*bx2*pow(vy0,4)+30*ax2*vx0*pow(vy0,3)+99*bx2*pow(vx0,2)*pow(vy0,2)+30*ax2*pow(vx0,3)*vy0+27*bx2*pow(vx0,4))*pow(vy2,3)+(42*ax2*vx2*pow(vy0,4)+60*bx2*vx0*vx2*pow(vy0,3)+99*ax2*pow(vx0,2)*vx2*pow(vy0,2)+60*bx2*pow(vx0,3)*vx2*vy0+57*ax2*pow(vx0,4)*vx2)*pow(vy2,2)+(57*bx2*pow(vx2,2)*pow(vy0,4)+60*ax2*vx0*pow(vx2,2)*pow(vy0,3)+99*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+60*ax2*pow(vx0,3)*pow(vx2,2)*vy0+42*bx2*pow(vx0,4)*pow(vx2,2))*vy2+27*ax2*pow(vx2,3)*pow(vy0,4)+30*bx2*vx0*pow(vx2,3)*pow(vy0,3)+99*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+30*bx2*pow(vx0,3)*pow(vx2,3)*vy0+72*ax2*pow(vx0,4)*pow(vx2,3))*vz0+(-24*nx2*pow(vy0,5)-45*nx2*pow(vx0,2)*pow(vy0,3)-21*nx2*pow(vx0,4)*vy0)*pow(vy2,3)+(-30*nx2*vx0*vx2*pow(vy0,4)-51*nx2*pow(vx0,3)*vx2*pow(vy0,2)-21*nx2*pow(vx0,5)*vx2)*pow(vy2,2)+(-21*nx2*pow(vx2,2)*pow(vy0,5)-51*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-30*nx2*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-21*nx2*vx0*pow(vx2,3)*pow(vy0,4)-45*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-24*nx2*pow(vx0,5)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((-24*nx2*pow(vy2,4)-48*nx2*pow(vx2,2)*pow(vy2,2)-24*nx2*pow(vx2,4))*pow(vz0,5)+((39*bx2*vy0+27*ax2*vx0)*pow(vy2,4)+(12*ax2*vx2*vy0+12*bx2*vx0*vx2)*pow(vy2,3)+(66*bx2*pow(vx2,2)*vy0+66*ax2*vx0*pow(vx2,2))*pow(vy2,2)+(12*ax2*pow(vx2,3)*vy0+12*bx2*vx0*pow(vx2,3))*vy2+27*bx2*pow(vx2,4)*vy0+39*ax2*vx0*pow(vx2,4))*pow(vz0,4)+((-102*nx2*pow(vy0,2)-54*nx2*pow(vx0,2))*pow(vy2,4)-96*nx2*vx0*vx2*vy0*pow(vy2,3)+(-156*nx2*pow(vx2,2)*pow(vy0,2)-156*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-96*nx2*vx0*pow(vx2,3)*vy0*vy2-54*nx2*pow(vx2,4)*pow(vy0,2)-102*nx2*pow(vx0,2)*pow(vx2,4))*pow(vz0,3)+((72*bx2*pow(vy0,3)+36*ax2*vx0*pow(vy0,2)+63*bx2*pow(vx0,2)*vy0+27*ax2*pow(vx0,3))*pow(vy2,4)+(36*ax2*vx2*pow(vy0,3)+54*bx2*vx0*vx2*pow(vy0,2)+54*ax2*pow(vx0,2)*vx2*vy0+36*bx2*pow(vx0,3)*vx2)*pow(vy2,3)+(99*bx2*pow(vx2,2)*pow(vy0,3)+99*ax2*vx0*pow(vx2,2)*pow(vy0,2)+99*bx2*pow(vx0,2)*pow(vx2,2)*vy0+99*ax2*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(36*ax2*pow(vx2,3)*pow(vy0,3)+54*bx2*vx0*pow(vx2,3)*pow(vy0,2)+54*ax2*pow(vx0,2)*pow(vx2,3)*vy0+36*bx2*pow(vx0,3)*pow(vx2,3))*vy2+27*bx2*pow(vx2,4)*pow(vy0,3)+63*ax2*vx0*pow(vx2,4)*pow(vy0,2)+36*bx2*pow(vx0,2)*pow(vx2,4)*vy0+72*ax2*pow(vx0,3)*pow(vx2,4))*pow(vz0,2)+((-72*nx2*pow(vy0,4)-99*nx2*pow(vx0,2)*pow(vy0,2)-27*nx2*pow(vx0,4))*pow(vy2,4)+(-90*nx2*vx0*vx2*pow(vy0,3)-90*nx2*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-99*nx2*pow(vx2,2)*pow(vy0,4)-198*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-99*nx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-90*nx2*vx0*pow(vx2,3)*pow(vy0,3)-90*nx2*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-27*nx2*pow(vx2,4)*pow(vy0,4)-99*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-72*nx2*pow(vx0,4)*pow(vx2,4))*vz0+(24*bx2*pow(vy0,5)+6*ax2*vx0*pow(vy0,4)+42*bx2*pow(vx0,2)*pow(vy0,3)+9*ax2*pow(vx0,3)*pow(vy0,2)+18*bx2*pow(vx0,4)*vy0+3*ax2*pow(vx0,5))*pow(vy2,4)+(18*ax2*vx2*pow(vy0,5)+30*bx2*vx0*vx2*pow(vy0,4)+39*ax2*pow(vx0,2)*vx2*pow(vy0,3)+45*bx2*pow(vx0,3)*vx2*pow(vy0,2)+21*ax2*pow(vx0,4)*vx2*vy0+15*bx2*pow(vx0,5)*vx2)*pow(vy2,3)+(24*bx2*pow(vx2,2)*pow(vy0,5)+33*ax2*vx0*pow(vx2,2)*pow(vy0,4)+57*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+57*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+33*bx2*pow(vx0,4)*pow(vx2,2)*vy0+24*ax2*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(15*ax2*pow(vx2,3)*pow(vy0,5)+21*bx2*vx0*pow(vx2,3)*pow(vy0,4)+45*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+39*bx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+30*ax2*pow(vx0,4)*pow(vx2,3)*vy0+18*bx2*pow(vx0,5)*pow(vx2,3))*vy2+3*bx2*pow(vx2,4)*pow(vy0,5)+18*ax2*vx0*pow(vx2,4)*pow(vy0,4)+9*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+42*ax2*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+6*bx2*pow(vx0,4)*pow(vx2,4)*vy0+24*ax2*pow(vx0,5)*pow(vx2,4))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((3*bx2*pow(vy2,5)+3*ax2*vx2*pow(vy2,4)+6*bx2*pow(vx2,2)*pow(vy2,3)+6*ax2*pow(vx2,3)*pow(vy2,2)+3*bx2*pow(vx2,4)*vy2+3*ax2*pow(vx2,5))*pow(vz0,5)+(-39*nx2*vy0*pow(vy2,5)-39*nx2*vx0*vx2*pow(vy2,4)-78*nx2*pow(vx2,2)*vy0*pow(vy2,3)-78*nx2*vx0*pow(vx2,3)*pow(vy2,2)-39*nx2*pow(vx2,4)*vy0*vy2-39*nx2*vx0*pow(vx2,5))*pow(vz0,4)+((24*bx2*pow(vy0,2)+18*ax2*vx0*vy0+9*bx2*pow(vx0,2))*pow(vy2,5)+(6*ax2*vx2*pow(vy0,2)+12*bx2*vx0*vx2*vy0+27*ax2*pow(vx0,2)*vx2)*pow(vy2,4)+(51*bx2*pow(vx2,2)*pow(vy0,2)+30*ax2*vx0*pow(vx2,2)*vy0+15*bx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(15*ax2*pow(vx2,3)*pow(vy0,2)+30*bx2*vx0*pow(vx2,3)*vy0+51*ax2*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(27*bx2*pow(vx2,4)*pow(vy0,2)+12*ax2*vx0*pow(vx2,4)*vy0+6*bx2*pow(vx0,2)*pow(vx2,4))*vy2+9*ax2*pow(vx2,5)*pow(vy0,2)+18*bx2*vx0*pow(vx2,5)*vy0+24*ax2*pow(vx0,2)*pow(vx2,5))*pow(vz0,3)+((-72*nx2*pow(vy0,3)-63*nx2*pow(vx0,2)*vy0)*pow(vy2,5)+(-90*nx2*vx0*vx2*pow(vy0,2)-63*nx2*pow(vx0,3)*vx2)*pow(vy2,4)+(-135*nx2*pow(vx2,2)*pow(vy0,3)-153*nx2*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-153*nx2*vx0*pow(vx2,3)*pow(vy0,2)-135*nx2*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-63*nx2*pow(vx2,4)*pow(vy0,3)-90*nx2*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-63*nx2*vx0*pow(vx2,5)*pow(vy0,2)-72*nx2*pow(vx0,3)*pow(vx2,5))*pow(vz0,2)+((24*bx2*pow(vy0,4)+12*ax2*vx0*pow(vy0,3)+30*bx2*pow(vx0,2)*pow(vy0,2)+12*ax2*pow(vx0,3)*vy0+6*bx2*pow(vx0,4))*pow(vy2,5)+(12*ax2*vx2*pow(vy0,4)+24*bx2*vx0*vx2*pow(vy0,3)+30*ax2*pow(vx0,2)*vx2*pow(vy0,2)+24*bx2*pow(vx0,3)*vx2*vy0+18*ax2*pow(vx0,4)*vx2)*pow(vy2,4)+(42*bx2*pow(vx2,2)*pow(vy0,4)+36*ax2*vx0*pow(vx2,2)*pow(vy0,3)+60*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+36*ax2*pow(vx0,3)*pow(vx2,2)*vy0+18*bx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(18*ax2*pow(vx2,3)*pow(vy0,4)+36*bx2*vx0*pow(vx2,3)*pow(vy0,3)+60*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+36*bx2*pow(vx0,3)*pow(vx2,3)*vy0+42*ax2*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(18*bx2*pow(vx2,4)*pow(vy0,4)+24*ax2*vx0*pow(vx2,4)*pow(vy0,3)+30*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+24*ax2*pow(vx0,3)*pow(vx2,4)*vy0+12*bx2*pow(vx0,4)*pow(vx2,4))*vy2+6*ax2*pow(vx2,5)*pow(vy0,4)+12*bx2*vx0*pow(vx2,5)*pow(vy0,3)+30*ax2*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+12*bx2*pow(vx0,3)*pow(vx2,5)*vy0+24*ax2*pow(vx0,4)*pow(vx2,5))*vz0+(-24*nx2*pow(vy0,5)-42*nx2*pow(vx0,2)*pow(vy0,3)-18*nx2*pow(vx0,4)*vy0)*pow(vy2,5)+(-36*nx2*vx0*vx2*pow(vy0,4)-54*nx2*pow(vx0,3)*vx2*pow(vy0,2)-18*nx2*pow(vx0,5)*vx2)*pow(vy2,4)+(-42*nx2*pow(vx2,2)*pow(vy0,5)-96*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-54*nx2*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-54*nx2*vx0*pow(vx2,3)*pow(vy0,4)-96*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-42*nx2*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-18*nx2*pow(vx2,4)*pow(vy0,5)-54*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-36*nx2*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-18*nx2*vx0*pow(vx2,5)*pow(vy0,4)-42*nx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-24*nx2*pow(vx0,5)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((-3*nx2*pow(vy2,6)-9*nx2*pow(vx2,2)*pow(vy2,4)-9*nx2*pow(vx2,4)*pow(vy2,2)-3*nx2*pow(vx2,6))*pow(vz0,5)+(3*ax2*vx0*pow(vy2,6)+(-3*ax2*vx2*vy0-3*bx2*vx0*vx2)*pow(vy2,5)+(3*bx2*pow(vx2,2)*vy0+6*ax2*vx0*pow(vx2,2))*pow(vy2,4)+(-6*ax2*pow(vx2,3)*vy0-6*bx2*vx0*pow(vx2,3))*pow(vy2,3)+(6*bx2*pow(vx2,4)*vy0+3*ax2*vx0*pow(vx2,4))*pow(vy2,2)+(-3*ax2*pow(vx2,5)*vy0-3*bx2*vx0*pow(vx2,5))*vy2+3*bx2*pow(vx2,6)*vy0)*pow(vz0,4)+((-24*nx2*pow(vy0,2)-9*nx2*pow(vx0,2))*pow(vy2,6)-30*nx2*vx0*vx2*vy0*pow(vy2,5)+(-57*nx2*pow(vx2,2)*pow(vy0,2)-42*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)-60*nx2*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(-42*nx2*pow(vx2,4)*pow(vy0,2)-57*nx2*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)-30*nx2*vx0*pow(vx2,5)*vy0*vy2-9*nx2*pow(vx2,6)*pow(vy0,2)-24*nx2*pow(vx0,2)*pow(vx2,6))*pow(vz0,3)+((6*ax2*vx0*pow(vy0,2)+3*ax2*pow(vx0,3))*pow(vy2,6)+(-6*ax2*vx2*pow(vy0,3)-6*bx2*vx0*vx2*pow(vy0,2)+3*ax2*pow(vx0,2)*vx2*vy0-3*bx2*pow(vx0,3)*vx2)*pow(vy2,5)+(6*bx2*pow(vx2,2)*pow(vy0,3)+3*ax2*vx0*pow(vx2,2)*pow(vy0,2)-3*bx2*pow(vx0,2)*pow(vx2,2)*vy0+9*ax2*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+(-9*ax2*pow(vx2,3)*pow(vy0,3)-3*bx2*vx0*pow(vx2,3)*pow(vy0,2)-3*ax2*pow(vx0,2)*pow(vx2,3)*vy0-9*bx2*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(9*bx2*pow(vx2,4)*pow(vy0,3)-3*ax2*vx0*pow(vx2,4)*pow(vy0,2)+3*bx2*pow(vx0,2)*pow(vx2,4)*vy0+6*ax2*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+(-3*ax2*pow(vx2,5)*pow(vy0,3)+3*bx2*vx0*pow(vx2,5)*pow(vy0,2)-6*ax2*pow(vx0,2)*pow(vx2,5)*vy0-6*bx2*pow(vx0,3)*pow(vx2,5))*vy2+3*bx2*pow(vx2,6)*pow(vy0,3)+6*bx2*pow(vx0,2)*pow(vx2,6)*vy0)*pow(vz0,2)+((-24*nx2*pow(vy0,4)-30*nx2*pow(vx0,2)*pow(vy0,2)-6*nx2*pow(vx0,4))*pow(vy2,6)+(-36*nx2*vx0*vx2*pow(vy0,3)-36*nx2*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(-54*nx2*pow(vx2,2)*pow(vy0,4)-90*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-36*nx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(-72*nx2*vx0*pow(vx2,3)*pow(vy0,3)-72*nx2*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(-36*nx2*pow(vx2,4)*pow(vy0,4)-90*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-54*nx2*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(-36*nx2*vx0*pow(vx2,5)*pow(vy0,3)-36*nx2*pow(vx0,3)*pow(vx2,5)*vy0)*vy2-6*nx2*pow(vx2,6)*pow(vy0,4)-30*nx2*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)-24*nx2*pow(vx0,4)*pow(vx2,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+((24*bx2*vy0+24*ax2*vx0)*pow(vz0,5)+(36*bx2*pow(vy0,3)+36*ax2*vx0*pow(vy0,2)+36*bx2*pow(vx0,2)*vy0+36*ax2*pow(vx0,3))*pow(vz0,3)+(12*bx2*pow(vy0,5)+12*ax2*vx0*pow(vy0,4)+24*bx2*pow(vx0,2)*pow(vy0,3)+24*ax2*pow(vx0,3)*pow(vy0,2)+12*bx2*pow(vx0,4)*vy0+12*ax2*pow(vx0,5))*vz0)*pow(vz2,7)+((24*bx2*vy2+24*ax2*vx2)*pow(vz0,6)+(-24*nx2*vy0*vy2-24*nx2*vx0*vx2)*pow(vz0,5)+((84*bx2*pow(vy0,2)+24*ax2*vx0*vy0+60*bx2*pow(vx0,2))*vy2+60*ax2*vx2*pow(vy0,2)+24*bx2*vx0*vx2*vy0+84*ax2*pow(vx0,2)*vx2)*pow(vz0,4)+((-36*nx2*pow(vy0,3)-36*nx2*pow(vx0,2)*vy0)*vy2-36*nx2*vx0*vx2*pow(vy0,2)-36*nx2*pow(vx0,3)*vx2)*pow(vz0,3)+((72*bx2*pow(vy0,4)+30*ax2*vx0*pow(vy0,3)+114*bx2*pow(vx0,2)*pow(vy0,2)+30*ax2*pow(vx0,3)*vy0+42*bx2*pow(vx0,4))*vy2+42*ax2*vx2*pow(vy0,4)+30*bx2*vx0*vx2*pow(vy0,3)+114*ax2*pow(vx0,2)*vx2*pow(vy0,2)+30*bx2*pow(vx0,3)*vx2*vy0+72*ax2*pow(vx0,4)*vx2)*pow(vz0,2)+((-12*nx2*pow(vy0,5)-24*nx2*pow(vx0,2)*pow(vy0,3)-12*nx2*pow(vx0,4)*vy0)*vy2-12*nx2*vx0*vx2*pow(vy0,4)-24*nx2*pow(vx0,3)*vx2*pow(vy0,2)-12*nx2*pow(vx0,5)*vx2)*vz0+(12*bx2*pow(vy0,6)+6*ax2*vx0*pow(vy0,5)+30*bx2*pow(vx0,2)*pow(vy0,4)+12*ax2*pow(vx0,3)*pow(vy0,3)+24*bx2*pow(vx0,4)*pow(vy0,2)+6*ax2*pow(vx0,5)*vy0+6*bx2*pow(vx0,6))*vy2+6*ax2*vx2*pow(vy0,6)+6*bx2*vx0*vx2*pow(vy0,5)+24*ax2*pow(vx0,2)*vx2*pow(vy0,4)+12*bx2*pow(vx0,3)*vx2*pow(vy0,3)+30*ax2*pow(vx0,4)*vx2*pow(vy0,2)+6*bx2*pow(vx0,5)*vx2*vy0+12*ax2*pow(vx0,6)*vx2)*pow(vz2,6)+((-24*nx2*pow(vy2,2)-24*nx2*pow(vx2,2))*pow(vz0,6)+((84*bx2*vy0+60*ax2*vx0)*pow(vy2,2)+(24*ax2*vx2*vy0+24*bx2*vx0*vx2)*vy2+60*bx2*pow(vx2,2)*vy0+84*ax2*vx0*pow(vx2,2))*pow(vz0,5)+((-84*nx2*pow(vy0,2)-60*nx2*pow(vx0,2))*pow(vy2,2)-48*nx2*vx0*vx2*vy0*vy2-60*nx2*pow(vx2,2)*pow(vy0,2)-84*nx2*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((156*bx2*pow(vy0,3)+96*ax2*vx0*pow(vy0,2)+150*bx2*pow(vx0,2)*vy0+90*ax2*pow(vx0,3))*pow(vy2,2)+(60*ax2*vx2*pow(vy0,3)+72*bx2*vx0*vx2*pow(vy0,2)+72*ax2*pow(vx0,2)*vx2*vy0+60*bx2*pow(vx0,3)*vx2)*vy2+90*bx2*pow(vx2,2)*pow(vy0,3)+150*ax2*vx0*pow(vx2,2)*pow(vy0,2)+96*bx2*pow(vx0,2)*pow(vx2,2)*vy0+156*ax2*pow(vx0,3)*pow(vx2,2))*pow(vz0,3)+((-72*nx2*pow(vy0,4)-114*nx2*pow(vx0,2)*pow(vy0,2)-42*nx2*pow(vx0,4))*pow(vy2,2)+(-60*nx2*vx0*vx2*pow(vy0,3)-60*nx2*pow(vx0,3)*vx2*vy0)*vy2-42*nx2*pow(vx2,2)*pow(vy0,4)-114*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-72*nx2*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((72*bx2*pow(vy0,5)+36*ax2*vx0*pow(vy0,4)+138*bx2*pow(vx0,2)*pow(vy0,3)+66*ax2*pow(vx0,3)*pow(vy0,2)+66*bx2*pow(vx0,4)*vy0+30*ax2*pow(vx0,5))*pow(vy2,2)+(36*ax2*vx2*pow(vy0,5)+48*bx2*vx0*vx2*pow(vy0,4)+84*ax2*pow(vx0,2)*vx2*pow(vy0,3)+84*bx2*pow(vx0,3)*vx2*pow(vy0,2)+48*ax2*pow(vx0,4)*vx2*vy0+36*bx2*pow(vx0,5)*vx2)*vy2+30*bx2*pow(vx2,2)*pow(vy0,5)+66*ax2*vx0*pow(vx2,2)*pow(vy0,4)+66*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+138*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+36*bx2*pow(vx0,4)*pow(vx2,2)*vy0+72*ax2*pow(vx0,5)*pow(vx2,2))*vz0+(-12*nx2*pow(vy0,6)-30*nx2*pow(vx0,2)*pow(vy0,4)-24*nx2*pow(vx0,4)*pow(vy0,2)-6*nx2*pow(vx0,6))*pow(vy2,2)+(-12*nx2*vx0*vx2*pow(vy0,5)-24*nx2*pow(vx0,3)*vx2*pow(vy0,3)-12*nx2*pow(vx0,5)*vx2*vy0)*vy2-6*nx2*pow(vx2,2)*pow(vy0,6)-24*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-30*nx2*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-12*nx2*pow(vx0,6)*pow(vx2,2))*pow(vz2,5)+((36*bx2*pow(vy2,3)+36*ax2*vx2*pow(vy2,2)+36*bx2*pow(vx2,2)*vy2+36*ax2*pow(vx2,3))*pow(vz0,6)+(-84*nx2*vy0*pow(vy2,3)-84*nx2*vx0*vx2*pow(vy2,2)-84*nx2*pow(vx2,2)*vy0*vy2-84*nx2*vx0*pow(vx2,3))*pow(vz0,5)+((156*bx2*pow(vy0,2)+54*ax2*vx0*vy0+96*bx2*pow(vx0,2))*pow(vy2,3)+(102*ax2*vx2*pow(vy0,2)+66*bx2*vx0*vx2*vy0+150*ax2*pow(vx0,2)*vx2)*pow(vy2,2)+(150*bx2*pow(vx2,2)*pow(vy0,2)+66*ax2*vx0*pow(vx2,2)*vy0+102*bx2*pow(vx0,2)*pow(vx2,2))*vy2+96*ax2*pow(vx2,3)*pow(vy0,2)+54*bx2*vx0*pow(vx2,3)*vy0+156*ax2*pow(vx0,2)*pow(vx2,3))*pow(vz0,4)+((-156*nx2*pow(vy0,3)-150*nx2*pow(vx0,2)*vy0)*pow(vy2,3)+(-168*nx2*vx0*vx2*pow(vy0,2)-150*nx2*pow(vx0,3)*vx2)*pow(vy2,2)+(-150*nx2*pow(vx2,2)*pow(vy0,3)-168*nx2*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-150*nx2*vx0*pow(vx2,3)*pow(vy0,2)-156*nx2*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+((156*bx2*pow(vy0,4)+66*ax2*vx0*pow(vy0,3)+228*bx2*pow(vx0,2)*pow(vy0,2)+66*ax2*pow(vx0,3)*vy0+72*bx2*pow(vx0,4))*pow(vy2,3)+(90*ax2*vx2*pow(vy0,4)+102*bx2*vx0*vx2*pow(vy0,3)+228*ax2*pow(vx0,2)*vx2*pow(vy0,2)+102*bx2*pow(vx0,3)*vx2*vy0+138*ax2*pow(vx0,4)*vx2)*pow(vy2,2)+(138*bx2*pow(vx2,2)*pow(vy0,4)+102*ax2*vx0*pow(vx2,2)*pow(vy0,3)+228*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+102*ax2*pow(vx0,3)*pow(vx2,2)*vy0+90*bx2*pow(vx0,4)*pow(vx2,2))*vy2+72*ax2*pow(vx2,3)*pow(vy0,4)+66*bx2*vx0*pow(vx2,3)*pow(vy0,3)+228*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+66*bx2*pow(vx0,3)*pow(vx2,3)*vy0+156*ax2*pow(vx0,4)*pow(vx2,3))*pow(vz0,2)+((-72*nx2*pow(vy0,5)-138*nx2*pow(vx0,2)*pow(vy0,3)-66*nx2*pow(vx0,4)*vy0)*pow(vy2,3)+(-84*nx2*vx0*vx2*pow(vy0,4)-150*nx2*pow(vx0,3)*vx2*pow(vy0,2)-66*nx2*pow(vx0,5)*vx2)*pow(vy2,2)+(-66*nx2*pow(vx2,2)*pow(vy0,5)-150*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-84*nx2*pow(vx0,4)*pow(vx2,2)*vy0)*vy2-66*nx2*vx0*pow(vx2,3)*pow(vy0,4)-138*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-72*nx2*pow(vx0,5)*pow(vx2,3))*vz0+(36*bx2*pow(vy0,6)+12*ax2*vx0*pow(vy0,5)+84*bx2*pow(vx0,2)*pow(vy0,4)+24*ax2*pow(vx0,3)*pow(vy0,3)+60*bx2*pow(vx0,4)*pow(vy0,2)+12*ax2*pow(vx0,5)*vy0+12*bx2*pow(vx0,6))*pow(vy2,3)+(24*ax2*vx2*pow(vy0,6)+36*bx2*vx0*vx2*pow(vy0,5)+72*ax2*pow(vx0,2)*vx2*pow(vy0,4)+72*bx2*pow(vx0,3)*vx2*pow(vy0,3)+72*ax2*pow(vx0,4)*vx2*pow(vy0,2)+36*bx2*pow(vx0,5)*vx2*vy0+24*ax2*pow(vx0,6)*vx2)*pow(vy2,2)+(24*bx2*pow(vx2,2)*pow(vy0,6)+36*ax2*vx0*pow(vx2,2)*pow(vy0,5)+72*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+72*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+72*bx2*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+36*ax2*pow(vx0,5)*pow(vx2,2)*vy0+24*bx2*pow(vx0,6)*pow(vx2,2))*vy2+12*ax2*pow(vx2,3)*pow(vy0,6)+12*bx2*vx0*pow(vx2,3)*pow(vy0,5)+60*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+24*bx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+84*ax2*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+12*bx2*pow(vx0,5)*pow(vx2,3)*vy0+36*ax2*pow(vx0,6)*pow(vx2,3))*pow(vz2,4)+((-36*nx2*pow(vy2,4)-72*nx2*pow(vx2,2)*pow(vy2,2)-36*nx2*pow(vx2,4))*pow(vz0,6)+((72*bx2*vy0+48*ax2*vx0)*pow(vy2,4)+(24*ax2*vx2*vy0+24*bx2*vx0*vx2)*pow(vy2,3)+(120*bx2*pow(vx2,2)*vy0+120*ax2*vx0*pow(vx2,2))*pow(vy2,2)+(24*ax2*pow(vx2,3)*vy0+24*bx2*vx0*pow(vx2,3))*vy2+48*bx2*pow(vx2,4)*vy0+72*ax2*vx0*pow(vx2,4))*pow(vz0,5)+((-156*nx2*pow(vy0,2)-96*nx2*pow(vx0,2))*pow(vy2,4)-120*nx2*vx0*vx2*vy0*pow(vy2,3)+(-252*nx2*pow(vx2,2)*pow(vy0,2)-252*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-120*nx2*vx0*pow(vx2,3)*vy0*vy2-96*nx2*pow(vx2,4)*pow(vy0,2)-156*nx2*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+((156*bx2*pow(vy0,3)+84*ax2*vx0*pow(vy0,2)+144*bx2*pow(vx0,2)*vy0+72*ax2*pow(vx0,3))*pow(vy2,4)+(72*ax2*vx2*pow(vy0,3)+96*bx2*vx0*vx2*pow(vy0,2)+96*ax2*pow(vx0,2)*vx2*vy0+72*bx2*pow(vx0,3)*vx2)*pow(vy2,3)+(228*bx2*pow(vx2,2)*pow(vy0,3)+228*ax2*vx0*pow(vx2,2)*pow(vy0,2)+228*bx2*pow(vx0,2)*pow(vx2,2)*vy0+228*ax2*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(72*ax2*pow(vx2,3)*pow(vy0,3)+96*bx2*vx0*pow(vx2,3)*pow(vy0,2)+96*ax2*pow(vx0,2)*pow(vx2,3)*vy0+72*bx2*pow(vx0,3)*pow(vx2,3))*vy2+72*bx2*pow(vx2,4)*pow(vy0,3)+144*ax2*vx0*pow(vx2,4)*pow(vy0,2)+84*bx2*pow(vx0,2)*pow(vx2,4)*vy0+156*ax2*pow(vx0,3)*pow(vx2,4))*pow(vz0,3)+((-156*nx2*pow(vy0,4)-228*nx2*pow(vx0,2)*pow(vy0,2)-72*nx2*pow(vx0,4))*pow(vy2,4)+(-168*nx2*vx0*vx2*pow(vy0,3)-168*nx2*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-228*nx2*pow(vx2,2)*pow(vy0,4)-456*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-228*nx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-168*nx2*vx0*pow(vx2,3)*pow(vy0,3)-168*nx2*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-72*nx2*pow(vx2,4)*pow(vy0,4)-228*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-156*nx2*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+((84*bx2*pow(vy0,5)+36*ax2*vx0*pow(vy0,4)+156*bx2*pow(vx0,2)*pow(vy0,3)+60*ax2*pow(vx0,3)*pow(vy0,2)+72*bx2*pow(vx0,4)*vy0+24*ax2*pow(vx0,5))*pow(vy2,4)+(48*ax2*vx2*pow(vy0,5)+72*bx2*vx0*vx2*pow(vy0,4)+120*ax2*pow(vx0,2)*vx2*pow(vy0,3)+120*bx2*pow(vx0,3)*vx2*pow(vy0,2)+72*ax2*pow(vx0,4)*vx2*vy0+48*bx2*pow(vx0,5)*vx2)*pow(vy2,3)+(108*bx2*pow(vx2,2)*pow(vy0,5)+108*ax2*vx0*pow(vx2,2)*pow(vy0,4)+216*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+216*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+108*bx2*pow(vx0,4)*pow(vx2,2)*vy0+108*ax2*pow(vx0,5)*pow(vx2,2))*pow(vy2,2)+(48*ax2*pow(vx2,3)*pow(vy0,5)+72*bx2*vx0*pow(vx2,3)*pow(vy0,4)+120*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+120*bx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+72*ax2*pow(vx0,4)*pow(vx2,3)*vy0+48*bx2*pow(vx0,5)*pow(vx2,3))*vy2+24*bx2*pow(vx2,4)*pow(vy0,5)+72*ax2*vx0*pow(vx2,4)*pow(vy0,4)+60*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+156*ax2*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+36*bx2*pow(vx0,4)*pow(vx2,4)*vy0+84*ax2*pow(vx0,5)*pow(vx2,4))*vz0+(-36*nx2*pow(vy0,6)-84*nx2*pow(vx0,2)*pow(vy0,4)-60*nx2*pow(vx0,4)*pow(vy0,2)-12*nx2*pow(vx0,6))*pow(vy2,4)+(-48*nx2*vx0*vx2*pow(vy0,5)-96*nx2*pow(vx0,3)*vx2*pow(vy0,3)-48*nx2*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(-48*nx2*pow(vx2,2)*pow(vy0,6)-144*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-144*nx2*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-48*nx2*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(-48*nx2*vx0*pow(vx2,3)*pow(vy0,5)-96*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-48*nx2*pow(vx0,5)*pow(vx2,3)*vy0)*vy2-12*nx2*pow(vx2,4)*pow(vy0,6)-60*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-84*nx2*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-36*nx2*pow(vx0,6)*pow(vx2,4))*pow(vz2,3)+((12*bx2*pow(vy2,5)+12*ax2*vx2*pow(vy2,4)+24*bx2*pow(vx2,2)*pow(vy2,3)+24*ax2*pow(vx2,3)*pow(vy2,2)+12*bx2*pow(vx2,4)*vy2+12*ax2*pow(vx2,5))*pow(vz0,6)+(-72*nx2*vy0*pow(vy2,5)-72*nx2*vx0*vx2*pow(vy2,4)-144*nx2*pow(vx2,2)*vy0*pow(vy2,3)-144*nx2*vx0*pow(vx2,3)*pow(vy2,2)-72*nx2*pow(vx2,4)*vy0*vy2-72*nx2*vx0*pow(vx2,5))*pow(vz0,5)+((72*bx2*pow(vy0,2)+36*ax2*vx0*vy0+36*bx2*pow(vx0,2))*pow(vy2,5)+(36*ax2*vx2*pow(vy0,2)+36*bx2*vx0*vx2*vy0+72*ax2*pow(vx0,2)*vx2)*pow(vy2,4)+(144*bx2*pow(vx2,2)*pow(vy0,2)+72*ax2*vx0*pow(vx2,2)*vy0+72*bx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(72*ax2*pow(vx2,3)*pow(vy0,2)+72*bx2*vx0*pow(vx2,3)*vy0+144*ax2*pow(vx0,2)*pow(vx2,3))*pow(vy2,2)+(72*bx2*pow(vx2,4)*pow(vy0,2)+36*ax2*vx0*pow(vx2,4)*vy0+36*bx2*pow(vx0,2)*pow(vx2,4))*vy2+36*ax2*pow(vx2,5)*pow(vy0,2)+36*bx2*vx0*pow(vx2,5)*vy0+72*ax2*pow(vx0,2)*pow(vx2,5))*pow(vz0,4)+((-156*nx2*pow(vy0,3)-144*nx2*pow(vx0,2)*vy0)*pow(vy2,5)+(-180*nx2*vx0*vx2*pow(vy0,2)-144*nx2*pow(vx0,3)*vx2)*pow(vy2,4)+(-300*nx2*pow(vx2,2)*pow(vy0,3)-324*nx2*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-324*nx2*vx0*pow(vx2,3)*pow(vy0,2)-300*nx2*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-144*nx2*pow(vx2,4)*pow(vy0,3)-180*nx2*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-144*nx2*vx0*pow(vx2,5)*pow(vy0,2)-156*nx2*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+((84*bx2*pow(vy0,4)+42*ax2*vx0*pow(vy0,3)+114*bx2*pow(vx0,2)*pow(vy0,2)+42*ax2*pow(vx0,3)*vy0+30*bx2*pow(vx0,4))*pow(vy2,5)+(42*ax2*vx2*pow(vy0,4)+66*bx2*vx0*vx2*pow(vy0,3)+114*ax2*pow(vx0,2)*vx2*pow(vy0,2)+66*bx2*pow(vx0,3)*vx2*vy0+72*ax2*pow(vx0,4)*vx2)*pow(vy2,4)+(156*bx2*pow(vx2,2)*pow(vy0,4)+108*ax2*vx0*pow(vx2,2)*pow(vy0,3)+228*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+108*ax2*pow(vx0,3)*pow(vx2,2)*vy0+72*bx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,3)+(72*ax2*pow(vx2,3)*pow(vy0,4)+108*bx2*vx0*pow(vx2,3)*pow(vy0,3)+228*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+108*bx2*pow(vx0,3)*pow(vx2,3)*vy0+156*ax2*pow(vx0,4)*pow(vx2,3))*pow(vy2,2)+(72*bx2*pow(vx2,4)*pow(vy0,4)+66*ax2*vx0*pow(vx2,4)*pow(vy0,3)+114*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+66*ax2*pow(vx0,3)*pow(vx2,4)*vy0+42*bx2*pow(vx0,4)*pow(vx2,4))*vy2+30*ax2*pow(vx2,5)*pow(vy0,4)+42*bx2*vx0*pow(vx2,5)*pow(vy0,3)+114*ax2*pow(vx0,2)*pow(vx2,5)*pow(vy0,2)+42*bx2*pow(vx0,3)*pow(vx2,5)*vy0+84*ax2*pow(vx0,4)*pow(vx2,5))*pow(vz0,2)+((-84*nx2*pow(vy0,5)-156*nx2*pow(vx0,2)*pow(vy0,3)-72*nx2*pow(vx0,4)*vy0)*pow(vy2,5)+(-108*nx2*vx0*vx2*pow(vy0,4)-180*nx2*pow(vx0,3)*vx2*pow(vy0,2)-72*nx2*pow(vx0,5)*vx2)*pow(vy2,4)+(-156*nx2*pow(vx2,2)*pow(vy0,5)-336*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-180*nx2*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(-180*nx2*vx0*pow(vx2,3)*pow(vy0,4)-336*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-156*nx2*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(-72*nx2*pow(vx2,4)*pow(vy0,5)-180*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-108*nx2*pow(vx0,4)*pow(vx2,4)*vy0)*vy2-72*nx2*vx0*pow(vx2,5)*pow(vy0,4)-156*nx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-84*nx2*pow(vx0,5)*pow(vx2,5))*vz0+(24*bx2*pow(vy0,6)+6*ax2*vx0*pow(vy0,5)+54*bx2*pow(vx0,2)*pow(vy0,4)+12*ax2*pow(vx0,3)*pow(vy0,3)+36*bx2*pow(vx0,4)*pow(vy0,2)+6*ax2*pow(vx0,5)*vy0+6*bx2*pow(vx0,6))*pow(vy2,5)+(18*ax2*vx2*pow(vy0,6)+30*bx2*vx0*vx2*pow(vy0,5)+48*ax2*pow(vx0,2)*vx2*pow(vy0,4)+60*bx2*pow(vx0,3)*vx2*pow(vy0,3)+42*ax2*pow(vx0,4)*vx2*pow(vy0,2)+30*bx2*pow(vx0,5)*vx2*vy0+12*ax2*pow(vx0,6)*vx2)*pow(vy2,4)+(36*bx2*pow(vx2,2)*pow(vy0,6)+36*ax2*vx0*pow(vx2,2)*pow(vy0,5)+96*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+72*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,3)+84*bx2*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+36*ax2*pow(vx0,5)*pow(vx2,2)*vy0+24*bx2*pow(vx0,6)*pow(vx2,2))*pow(vy2,3)+(24*ax2*pow(vx2,3)*pow(vy0,6)+36*bx2*vx0*pow(vx2,3)*pow(vy0,5)+84*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,4)+72*bx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+96*ax2*pow(vx0,4)*pow(vx2,3)*pow(vy0,2)+36*bx2*pow(vx0,5)*pow(vx2,3)*vy0+36*ax2*pow(vx0,6)*pow(vx2,3))*pow(vy2,2)+(12*bx2*pow(vx2,4)*pow(vy0,6)+30*ax2*vx0*pow(vx2,4)*pow(vy0,5)+42*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+60*ax2*pow(vx0,3)*pow(vx2,4)*pow(vy0,3)+48*bx2*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+30*ax2*pow(vx0,5)*pow(vx2,4)*vy0+18*bx2*pow(vx0,6)*pow(vx2,4))*vy2+6*ax2*pow(vx2,5)*pow(vy0,6)+6*bx2*vx0*pow(vx2,5)*pow(vy0,5)+36*ax2*pow(vx0,2)*pow(vx2,5)*pow(vy0,4)+12*bx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+54*ax2*pow(vx0,4)*pow(vx2,5)*pow(vy0,2)+6*bx2*pow(vx0,5)*pow(vx2,5)*vy0+24*ax2*pow(vx0,6)*pow(vx2,5))*pow(vz2,2)+((-12*nx2*pow(vy2,6)-36*nx2*pow(vx2,2)*pow(vy2,4)-36*nx2*pow(vx2,4)*pow(vy2,2)-12*nx2*pow(vx2,6))*pow(vz0,6)+((12*bx2*vy0+12*ax2*vx0)*pow(vy2,6)+(36*bx2*pow(vx2,2)*vy0+36*ax2*vx0*pow(vx2,2))*pow(vy2,4)+(36*bx2*pow(vx2,4)*vy0+36*ax2*vx0*pow(vx2,4))*pow(vy2,2)+12*bx2*pow(vx2,6)*vy0+12*ax2*vx0*pow(vx2,6))*pow(vz0,5)+((-72*nx2*pow(vy0,2)-36*nx2*pow(vx0,2))*pow(vy2,6)-72*nx2*vx0*vx2*vy0*pow(vy2,5)+(-180*nx2*pow(vx2,2)*pow(vy0,2)-144*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)-144*nx2*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(-144*nx2*pow(vx2,4)*pow(vy0,2)-180*nx2*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)-72*nx2*vx0*pow(vx2,5)*vy0*vy2-36*nx2*pow(vx2,6)*pow(vy0,2)-72*nx2*pow(vx0,2)*pow(vx2,6))*pow(vz0,4)+((36*bx2*pow(vy0,3)+24*ax2*vx0*pow(vy0,2)+30*bx2*pow(vx0,2)*vy0+18*ax2*pow(vx0,3))*pow(vy2,6)+(12*ax2*vx2*pow(vy0,3)+24*bx2*vx0*vx2*pow(vy0,2)+24*ax2*pow(vx0,2)*vx2*vy0+12*bx2*pow(vx0,3)*vx2)*pow(vy2,5)+(90*bx2*pow(vx2,2)*pow(vy0,3)+78*ax2*vx0*pow(vx2,2)*pow(vy0,2)+84*bx2*pow(vx0,2)*pow(vx2,2)*vy0+72*ax2*pow(vx0,3)*pow(vx2,2))*pow(vy2,4)+(24*ax2*pow(vx2,3)*pow(vy0,3)+48*bx2*vx0*pow(vx2,3)*pow(vy0,2)+48*ax2*pow(vx0,2)*pow(vx2,3)*vy0+24*bx2*pow(vx0,3)*pow(vx2,3))*pow(vy2,3)+(72*bx2*pow(vx2,4)*pow(vy0,3)+84*ax2*vx0*pow(vx2,4)*pow(vy0,2)+78*bx2*pow(vx0,2)*pow(vx2,4)*vy0+90*ax2*pow(vx0,3)*pow(vx2,4))*pow(vy2,2)+(12*ax2*pow(vx2,5)*pow(vy0,3)+24*bx2*vx0*pow(vx2,5)*pow(vy0,2)+24*ax2*pow(vx0,2)*pow(vx2,5)*vy0+12*bx2*pow(vx0,3)*pow(vx2,5))*vy2+18*bx2*pow(vx2,6)*pow(vy0,3)+30*ax2*vx0*pow(vx2,6)*pow(vy0,2)+24*bx2*pow(vx0,2)*pow(vx2,6)*vy0+36*ax2*pow(vx0,3)*pow(vx2,6))*pow(vz0,3)+((-84*nx2*pow(vy0,4)-114*nx2*pow(vx0,2)*pow(vy0,2)-30*nx2*pow(vx0,4))*pow(vy2,6)+(-108*nx2*vx0*vx2*pow(vy0,3)-108*nx2*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(-198*nx2*pow(vx2,2)*pow(vy0,4)-342*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-144*nx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(-216*nx2*vx0*pow(vx2,3)*pow(vy0,3)-216*nx2*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(-144*nx2*pow(vx2,4)*pow(vy0,4)-342*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-198*nx2*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(-108*nx2*vx0*pow(vx2,5)*pow(vy0,3)-108*nx2*pow(vx0,3)*pow(vx2,5)*vy0)*vy2-30*nx2*pow(vx2,6)*pow(vy0,4)-114*nx2*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)-84*nx2*pow(vx0,4)*pow(vx2,6))*pow(vz0,2)+((24*bx2*pow(vy0,5)+12*ax2*vx0*pow(vy0,4)+42*bx2*pow(vx0,2)*pow(vy0,3)+18*ax2*pow(vx0,3)*pow(vy0,2)+18*bx2*pow(vx0,4)*vy0+6*ax2*pow(vx0,5))*pow(vy2,6)+(12*ax2*vx2*pow(vy0,5)+24*bx2*vx0*vx2*pow(vy0,4)+36*ax2*pow(vx0,2)*vx2*pow(vy0,3)+36*bx2*pow(vx0,3)*vx2*pow(vy0,2)+24*ax2*pow(vx0,4)*vx2*vy0+12*bx2*pow(vx0,5)*vx2)*pow(vy2,5)+(54*bx2*pow(vx2,2)*pow(vy0,5)+42*ax2*vx0*pow(vx2,2)*pow(vy0,4)+102*bx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+78*ax2*pow(vx0,3)*pow(vx2,2)*pow(vy0,2)+48*bx2*pow(vx0,4)*pow(vx2,2)*vy0+36*ax2*pow(vx0,5)*pow(vx2,2))*pow(vy2,4)+(24*ax2*pow(vx2,3)*pow(vy0,5)+48*bx2*vx0*pow(vx2,3)*pow(vy0,4)+72*ax2*pow(vx0,2)*pow(vx2,3)*pow(vy0,3)+72*bx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+48*ax2*pow(vx0,4)*pow(vx2,3)*vy0+24*bx2*pow(vx0,5)*pow(vx2,3))*pow(vy2,3)+(36*bx2*pow(vx2,4)*pow(vy0,5)+48*ax2*vx0*pow(vx2,4)*pow(vy0,4)+78*bx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+102*ax2*pow(vx0,3)*pow(vx2,4)*pow(vy0,2)+42*bx2*pow(vx0,4)*pow(vx2,4)*vy0+54*ax2*pow(vx0,5)*pow(vx2,4))*pow(vy2,2)+(12*ax2*pow(vx2,5)*pow(vy0,5)+24*bx2*vx0*pow(vx2,5)*pow(vy0,4)+36*ax2*pow(vx0,2)*pow(vx2,5)*pow(vy0,3)+36*bx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+24*ax2*pow(vx0,4)*pow(vx2,5)*vy0+12*bx2*pow(vx0,5)*pow(vx2,5))*vy2+6*bx2*pow(vx2,6)*pow(vy0,5)+18*ax2*vx0*pow(vx2,6)*pow(vy0,4)+18*bx2*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+42*ax2*pow(vx0,3)*pow(vx2,6)*pow(vy0,2)+12*bx2*pow(vx0,4)*pow(vx2,6)*vy0+24*ax2*pow(vx0,5)*pow(vx2,6))*vz0+(-24*nx2*pow(vy0,6)-54*nx2*pow(vx0,2)*pow(vy0,4)-36*nx2*pow(vx0,4)*pow(vy0,2)-6*nx2*pow(vx0,6))*pow(vy2,6)+(-36*nx2*vx0*vx2*pow(vy0,5)-72*nx2*pow(vx0,3)*vx2*pow(vy0,3)-36*nx2*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(-54*nx2*pow(vx2,2)*pow(vy0,6)-144*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)-126*nx2*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)-36*nx2*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(-72*nx2*vx0*pow(vx2,3)*pow(vy0,5)-144*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)-72*nx2*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(-36*nx2*pow(vx2,4)*pow(vy0,6)-126*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)-144*nx2*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)-54*nx2*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(-36*nx2*vx0*pow(vx2,5)*pow(vy0,5)-72*nx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)-36*nx2*pow(vx0,5)*pow(vx2,5)*vy0)*vy2-6*nx2*pow(vx2,6)*pow(vy0,6)-36*nx2*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)-54*nx2*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)-24*nx2*pow(vx0,6)*pow(vx2,6))*vz2+(-12*nx2*vy0*pow(vy2,7)-12*nx2*vx0*vx2*pow(vy2,6)-36*nx2*pow(vx2,2)*vy0*pow(vy2,5)-36*nx2*vx0*pow(vx2,3)*pow(vy2,4)-36*nx2*pow(vx2,4)*vy0*pow(vy2,3)-36*nx2*vx0*pow(vx2,5)*pow(vy2,2)-12*nx2*pow(vx2,6)*vy0*vy2-12*nx2*vx0*pow(vx2,7))*pow(vz0,5)+(6*ax2*vx0*vy0*pow(vy2,7)+(-6*ax2*vx2*pow(vy0,2)-6*bx2*vx0*vx2*vy0+6*ax2*pow(vx0,2)*vx2)*pow(vy2,6)+(6*bx2*pow(vx2,2)*pow(vy0,2)+6*ax2*vx0*pow(vx2,2)*vy0-6*bx2*pow(vx0,2)*pow(vx2,2))*pow(vy2,5)+(-12*ax2*pow(vx2,3)*pow(vy0,2)-6*bx2*vx0*pow(vx2,3)*vy0+12*ax2*pow(vx0,2)*pow(vx2,3))*pow(vy2,4)+(12*bx2*pow(vx2,4)*pow(vy0,2)-6*ax2*vx0*pow(vx2,4)*vy0-12*bx2*pow(vx0,2)*pow(vx2,4))*pow(vy2,3)+(-6*ax2*pow(vx2,5)*pow(vy0,2)+6*bx2*vx0*pow(vx2,5)*vy0+6*ax2*pow(vx0,2)*pow(vx2,5))*pow(vy2,2)+(6*bx2*pow(vx2,6)*pow(vy0,2)-6*ax2*vx0*pow(vx2,6)*vy0-6*bx2*pow(vx0,2)*pow(vx2,6))*vy2+6*bx2*vx0*pow(vx2,7)*vy0)*pow(vz0,4)+((-36*nx2*pow(vy0,3)-30*nx2*pow(vx0,2)*vy0)*pow(vy2,7)+(-48*nx2*vx0*vx2*pow(vy0,2)-30*nx2*pow(vx0,3)*vx2)*pow(vy2,6)+(-102*nx2*pow(vx2,2)*pow(vy0,3)-108*nx2*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(-126*nx2*vx0*pow(vx2,3)*pow(vy0,2)-96*nx2*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(-96*nx2*pow(vx2,4)*pow(vy0,3)-126*nx2*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(-108*nx2*vx0*pow(vx2,5)*pow(vy0,2)-102*nx2*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(-30*nx2*pow(vx2,6)*pow(vy0,3)-48*nx2*pow(vx0,2)*pow(vx2,6)*vy0)*vy2-30*nx2*vx0*pow(vx2,7)*pow(vy0,2)-36*nx2*pow(vx0,3)*pow(vx2,7))*pow(vz0,3)+((6*ax2*vx0*pow(vy0,3)+6*ax2*pow(vx0,3)*vy0)*pow(vy2,7)+(-6*ax2*vx2*pow(vy0,4)-6*bx2*vx0*vx2*pow(vy0,3)-6*bx2*pow(vx0,3)*vx2*vy0+6*ax2*pow(vx0,4)*vx2)*pow(vy2,6)+(6*bx2*pow(vx2,2)*pow(vy0,4)+6*ax2*vx0*pow(vx2,2)*pow(vy0,3)+6*ax2*pow(vx0,3)*pow(vx2,2)*vy0-6*bx2*pow(vx0,4)*pow(vx2,2))*pow(vy2,5)+(-12*ax2*pow(vx2,3)*pow(vy0,4)-6*bx2*vx0*pow(vx2,3)*pow(vy0,3)-6*bx2*pow(vx0,3)*pow(vx2,3)*vy0+12*ax2*pow(vx0,4)*pow(vx2,3))*pow(vy2,4)+(12*bx2*pow(vx2,4)*pow(vy0,4)-6*ax2*vx0*pow(vx2,4)*pow(vy0,3)-6*ax2*pow(vx0,3)*pow(vx2,4)*vy0-12*bx2*pow(vx0,4)*pow(vx2,4))*pow(vy2,3)+(-6*ax2*pow(vx2,5)*pow(vy0,4)+6*bx2*vx0*pow(vx2,5)*pow(vy0,3)+6*bx2*pow(vx0,3)*pow(vx2,5)*vy0+6*ax2*pow(vx0,4)*pow(vx2,5))*pow(vy2,2)+(6*bx2*pow(vx2,6)*pow(vy0,4)-6*ax2*vx0*pow(vx2,6)*pow(vy0,3)-6*ax2*pow(vx0,3)*pow(vx2,6)*vy0-6*bx2*pow(vx0,4)*pow(vx2,6))*vy2+6*bx2*vx0*pow(vx2,7)*pow(vy0,3)+6*bx2*pow(vx0,3)*pow(vx2,7)*vy0)*pow(vz0,2)+((-24*nx2*pow(vy0,5)-42*nx2*pow(vx0,2)*pow(vy0,3)-18*nx2*pow(vx0,4)*vy0)*pow(vy2,7)+(-36*nx2*vx0*vx2*pow(vy0,4)-54*nx2*pow(vx0,3)*vx2*pow(vy0,2)-18*nx2*pow(vx0,5)*vx2)*pow(vy2,6)+(-66*nx2*pow(vx2,2)*pow(vy0,5)-138*nx2*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)-72*nx2*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(-90*nx2*vx0*pow(vx2,3)*pow(vy0,4)-150*nx2*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)-60*nx2*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(-60*nx2*pow(vx2,4)*pow(vy0,5)-150*nx2*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)-90*nx2*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(-72*nx2*vx0*pow(vx2,5)*pow(vy0,4)-138*nx2*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)-66*nx2*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(-18*nx2*pow(vx2,6)*pow(vy0,5)-54*nx2*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)-36*nx2*pow(vx0,4)*pow(vx2,6)*vy0)*vy2-18*nx2*vx0*pow(vx2,7)*pow(vy0,4)-42*nx2*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)-24*nx2*pow(vx0,5)*pow(vx2,7))*vz0)*pow(((160*pow(roughness,2)*pow(vz0,5)+(240*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vz0,3)+(80*pow(roughness,2)*pow(vy0,4)+160*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,7)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,4)+((400*pow(roughness,2)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+400*pow(roughness,2)*vx0*vx2*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,5)+160*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+80*pow(roughness,2)*vx0*vx2*pow(vy0,4)+160*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,5)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,6)+((400*pow(roughness,2)*pow(vy2,2)+400*pow(roughness,2)*pow(vx2,2))*pow(vz0,5)+((880*pow(roughness,2)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+640*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+((480*pow(roughness,2)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+240*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((720*pow(roughness,2)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*vx2*pow(vy2,2)+720*pow(roughness,2)*pow(vx2,2)*vy0*vy2+720*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,4)+((1040*pow(roughness,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,2)+(320*pow(roughness,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((320*pow(roughness,2)*pow(vy2,4)+640*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+320*pow(roughness,2)*pow(vx2,4))*pow(vz0,5)+((1040*pow(roughness,2)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+560*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1040*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,3)+((720*pow(roughness,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+240*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((480*pow(roughness,2)*vy0*pow(vy2,5)+480*pow(roughness,2)*vx0*vx2*pow(vy2,4)+960*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,4)*vy0*vy2+480*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,4)+((880*pow(roughness,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,2)+(400*pow(roughness,2)*pow(vy0,5)+640*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((80*pow(roughness,2)*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6))*pow(vz0,5)+((400*pow(roughness,2)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+480*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(960*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+960*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+480*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+160*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+1200*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((80*pow(roughness,2)*vy0*pow(vy2,7)+80*pow(roughness,2)*vx0*vx2*pow(vy2,6)+240*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+240*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+80*pow(roughness,2)*pow(vx2,6)*vy0*vy2+80*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,4)+((240*pow(roughness,2)*pow(vy0,3)+160*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(400*pow(roughness,2)*vx0*vx2*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(640*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(960*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(560*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+640*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(160*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+160*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,5)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+80*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,4)+400*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(400*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+880*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1040*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(320*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+1040*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+880*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+400*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(80*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+400*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+80*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(160*pow(roughness,2)*pow(vz0,6)+(320*pow(roughness,2)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,2))*pow(vz0,4)+(180*pow(roughness,2)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,4))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,6)+60*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+60*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,6))*pow(vz2,8)+((320*pow(roughness,2)*vy0*vy2+320*pow(roughness,2)*vx0*vx2)*pow(vz0,5)+((560*pow(roughness,2)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+560*pow(roughness,2)*vx0*vx2*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,5)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*vy2+240*pow(roughness,2)*vx0*vx2*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*vz0)*pow(vz2,7)+((480*pow(roughness,2)*pow(vy2,2)+480*pow(roughness,2)*pow(vx2,2))*pow(vz0,6)+((1240*pow(roughness,2)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+480*pow(roughness,2)*vx0*vx2*vy0*vy2+1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((960*pow(roughness,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+600*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+(200*pow(roughness,2)*pow(vy0,6)+480*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,2)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,5)+480*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*vy2+80*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vz2,6)+((880*pow(roughness,2)*vy0*pow(vy2,3)+880*pow(roughness,2)*vx0*vx2*pow(vy2,2)+880*pow(roughness,2)*pow(vx2,2)*vy0*vy2+880*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,5)+((1680*pow(roughness,2)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vz0,3)+((800*pow(roughness,2)*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1680*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*vz0)*pow(vz2,5)+((500*pow(roughness,2)*pow(vy2,4)+1000*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+500*pow(roughness,2)*pow(vx2,4))*pow(vz0,6)+((1700*pow(roughness,2)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+1200*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(2800*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+1200*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+1100*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,4)+((1700*pow(roughness,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(2400*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+720*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz0,2)+(500*pow(roughness,2)*pow(vy0,6)+1100*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,6))*pow(vy2,4)+(800*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*vy2+120*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vz2,4)+((800*pow(roughness,2)*vy0*pow(vy2,5)+800*pow(roughness,2)*vx0*vx2*pow(vy2,4)+1600*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+800*pow(roughness,2)*pow(vx2,4)*vy0*vy2+800*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,5)+((1680*pow(roughness,2)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(2000*pow(roughness,2)*vx0*vx2*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(3200*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(3520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(1520*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+2000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+1520*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vz0,3)+((880*pow(roughness,2)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*vx2*pow(vy0,4)+1920*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,4)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+3520*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,2)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*vy2+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*vz0)*pow(vz2,3)+((200*pow(roughness,2)*pow(vy2,6)+600*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+600*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+200*pow(roughness,2)*pow(vx2,6))*pow(vz0,6)+((960*pow(roughness,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+960*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(2400*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+1920*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(1920*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+960*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+480*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,4)+((1240*pow(roughness,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(1840*pow(roughness,2)*vx0*vx2*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(2800*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+1920*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(3520*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+3520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(1920*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+4920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(1680*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+1840*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+360*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+1240*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vz0,2)+(480*pow(roughness,2)*pow(vy0,6)+1000*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+600*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,6))*pow(vy2,6)+(880*pow(roughness,2)*vx0*vx2*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,5)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+2800*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,4)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+3200*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,3)+(600*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+2800*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,2)+(720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*vy2+80*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vz2,2)+((240*pow(roughness,2)*vy0*pow(vy2,7)+240*pow(roughness,2)*vx0*vx2*pow(vy2,6)+720*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,5)+720*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,4)+720*pow(roughness,2)*pow(vx2,4)*vy0*pow(vy2,3)+720*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,6)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,7))*pow(vz0,5)+((560*pow(roughness,2)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,7)+(720*pow(roughness,2)*vx0*vx2*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,6)+(1600*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+1680*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,5)+(1920*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,4)+(1520*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+1920*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*pow(vy2,3)+(1680*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(vy2,2)+(480*pow(roughness,2)*pow(vx2,6)*pow(vy0,3)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*vy0)*vy2+480*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,2)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,7))*pow(vz0,3)+((320*pow(roughness,2)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,4)*vy0)*pow(vy2,7)+(480*pow(roughness,2)*vx0*vx2*pow(vy0,4)+720*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,5)*vx2)*pow(vy2,6)+(880*pow(roughness,2)*pow(vx2,2)*pow(vy0,5)+1840*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,3)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*vy0)*pow(vy2,5)+(1200*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,4)+2000*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,2)+800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3))*pow(vy2,4)+(800*pow(roughness,2)*pow(vx2,4)*pow(vy0,5)+2000*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,3)+1200*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*vy0)*pow(vy2,3)+(960*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,4)+1840*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,2)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5))*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,6)*pow(vy0,5)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,4)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,7))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,8)+80*pow(roughness,2)*pow(vx2,2)*pow(vy2,6)+120*pow(roughness,2)*pow(vx2,4)*pow(vy2,4)+80*pow(roughness,2)*pow(vx2,6)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,8))*pow(vz0,6)+((180*pow(roughness,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,2))*pow(vy2,8)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,7)+(600*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,6)+720*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,5)+(720*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+720*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,4)+720*pow(roughness,2)*vx0*pow(vx2,5)*vy0*pow(vy2,3)+(360*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+600*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,7)*vy0*vy2+60*pow(roughness,2)*pow(vx2,8)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,8))*pow(vz0,4)+((320*pow(roughness,2)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,4))*pow(vy2,8)+(560*pow(roughness,2)*vx0*vx2*pow(vy0,3)+480*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,7)+(1000*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,6)+(1600*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+1520*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,5)+(1100*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+2400*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+1100*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,4)+(1520*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+1600*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*pow(vy2,3)+(480*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+1560*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+1000*pow(roughness,2)*pow(vx0,4)*pow(vx2,6))*pow(vy2,2)+(480*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,3)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,8)*pow(vy0,4)+360*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,2)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,8))*pow(vz0,2)+(160*pow(roughness,2)*pow(vy0,6)+320*pow(roughness,2)*pow(vx0,2)*pow(vy0,4)+180*pow(roughness,2)*pow(vx0,4)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,6))*pow(vy2,8)+(320*pow(roughness,2)*vx0*vx2*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,3)*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,5)*vx2*vy0)*pow(vy2,7)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,6)+1240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,4)+960*pow(roughness,2)*pow(vx0,4)*pow(vx2,2)*pow(vy0,2)+200*pow(roughness,2)*pow(vx0,6)*pow(vx2,2))*pow(vy2,6)+(880*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*pow(vy0,3)+800*pow(roughness,2)*pow(vx0,5)*pow(vx2,3)*vy0)*pow(vy2,5)+(500*pow(roughness,2)*pow(vx2,4)*pow(vy0,6)+1700*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,4)+1700*pow(roughness,2)*pow(vx0,4)*pow(vx2,4)*pow(vy0,2)+500*pow(roughness,2)*pow(vx0,6)*pow(vx2,4))*pow(vy2,4)+(800*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,5)+1680*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*pow(vy0,3)+880*pow(roughness,2)*pow(vx0,5)*pow(vx2,5)*vy0)*pow(vy2,3)+(200*pow(roughness,2)*pow(vx2,6)*pow(vy0,6)+960*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,4)+1240*pow(roughness,2)*pow(vx0,4)*pow(vx2,6)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,6)*pow(vx2,6))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,7)*pow(vy0,5)+560*pow(roughness,2)*pow(vx0,3)*pow(vx2,7)*pow(vy0,3)+320*pow(roughness,2)*pow(vx0,5)*pow(vx2,7)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,8)*pow(vy0,6)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,8)*pow(vy0,4)+320*pow(roughness,2)*pow(vx0,4)*pow(vx2,8)*pow(vy0,2)+160*pow(roughness,2)*pow(vx0,6)*pow(vx2,8),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
// y2  :  
poly.add(Term6f( -((((12*by2*vy0+12*ay2*vx0)*pow(vz0,2)+3*by2*pow(vy0,3)+3*ay2*vx0*pow(vy0,2)+3*by2*pow(vx0,2)*vy0+3*ay2*pow(vx0,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((12*by2*vy2+12*ay2*vx2)*pow(vz0,3)+(-12*ny2*vy0*vy2-12*ny2*vx0*vx2)*pow(vz0,2)+((21*by2*pow(vy0,2)+6*ay2*vx0*vy0+15*by2*pow(vx0,2))*vy2+15*ay2*vx2*pow(vy0,2)+6*by2*vx0*vx2*vy0+21*ay2*pow(vx0,2)*vx2)*vz0+(-3*ny2*pow(vy0,3)-3*ny2*pow(vx0,2)*vy0)*vy2-3*ny2*vx0*vx2*pow(vy0,2)-3*ny2*pow(vx0,3)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((-12*ny2*pow(vy2,2)-12*ny2*pow(vx2,2))*pow(vz0,3)+((21*by2*vy0+15*ay2*vx0)*pow(vy2,2)+(6*ay2*vx2*vy0+6*by2*vx0*vx2)*vy2+15*by2*pow(vx2,2)*vy0+21*ay2*vx0*pow(vx2,2))*pow(vz0,2)+((-21*ny2*pow(vy0,2)-15*ny2*pow(vx0,2))*pow(vy2,2)-12*ny2*vx0*vx2*vy0*vy2-15*ny2*pow(vx2,2)*pow(vy0,2)-21*ny2*pow(vx0,2)*pow(vx2,2))*vz0+(12*by2*pow(vy0,3)+3*ay2*vx0*pow(vy0,2)+12*by2*pow(vx0,2)*vy0+3*ay2*pow(vx0,3))*pow(vy2,2)+(9*ay2*vx2*pow(vy0,3)+9*by2*vx0*vx2*pow(vy0,2)+9*ay2*pow(vx0,2)*vx2*vy0+9*by2*pow(vx0,3)*vx2)*vy2+3*by2*pow(vx2,2)*pow(vy0,3)+12*ay2*vx0*pow(vx2,2)*pow(vy0,2)+3*by2*pow(vx0,2)*pow(vx2,2)*vy0+12*ay2*pow(vx0,3)*pow(vx2,2))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((3*by2*pow(vy2,3)+3*ay2*vx2*pow(vy2,2)+3*by2*pow(vx2,2)*vy2+3*ay2*pow(vx2,3))*pow(vz0,3)+(-21*ny2*vy0*pow(vy2,3)-21*ny2*vx0*vx2*pow(vy2,2)-21*ny2*pow(vx2,2)*vy0*vy2-21*ny2*vx0*pow(vx2,3))*pow(vz0,2)+((12*by2*pow(vy0,2)+6*ay2*vx0*vy0+6*by2*pow(vx0,2))*pow(vy2,3)+(6*ay2*vx2*pow(vy0,2)+6*by2*vx0*vx2*vy0+12*ay2*pow(vx0,2)*vx2)*pow(vy2,2)+(12*by2*pow(vx2,2)*pow(vy0,2)+6*ay2*vx0*pow(vx2,2)*vy0+6*by2*pow(vx0,2)*pow(vx2,2))*vy2+6*ay2*pow(vx2,3)*pow(vy0,2)+6*by2*vx0*pow(vx2,3)*vy0+12*ay2*pow(vx0,2)*pow(vx2,3))*vz0+(-12*ny2*pow(vy0,3)-12*ny2*pow(vx0,2)*vy0)*pow(vy2,3)+(-12*ny2*vx0*vx2*pow(vy0,2)-12*ny2*pow(vx0,3)*vx2)*pow(vy2,2)+(-12*ny2*pow(vx2,2)*pow(vy0,3)-12*ny2*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-12*ny2*vx0*pow(vx2,3)*pow(vy0,2)-12*ny2*pow(vx0,3)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((-3*ny2*pow(vy2,4)-6*ny2*pow(vx2,2)*pow(vy2,2)-3*ny2*pow(vx2,4))*pow(vz0,3)+(3*ay2*vx0*pow(vy2,4)+(-3*ay2*vx2*vy0-3*by2*vx0*vx2)*pow(vy2,3)+(3*by2*pow(vx2,2)*vy0+3*ay2*vx0*pow(vx2,2))*pow(vy2,2)+(-3*ay2*pow(vx2,3)*vy0-3*by2*vx0*pow(vx2,3))*vy2+3*by2*pow(vx2,4)*vy0)*pow(vz0,2)+((-12*ny2*pow(vy0,2)-6*ny2*pow(vx0,2))*pow(vy2,4)-12*ny2*vx0*vx2*vy0*pow(vy2,3)+(-18*ny2*pow(vx2,2)*pow(vy0,2)-18*ny2*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-12*ny2*vx0*pow(vx2,3)*vy0*vy2-6*ny2*pow(vx2,4)*pow(vy0,2)-12*ny2*pow(vx0,2)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+((12*by2*vy0+12*ay2*vx0)*pow(vz0,3)+(9*by2*pow(vy0,3)+9*ay2*vx0*pow(vy0,2)+9*by2*pow(vx0,2)*vy0+9*ay2*pow(vx0,3))*vz0)*pow(vz2,5)+((12*by2*vy2+12*ay2*vx2)*pow(vz0,4)+(-12*ny2*vy0*vy2-12*ny2*vx0*vx2)*pow(vz0,3)+((27*by2*pow(vy0,2)+6*ay2*vx0*vy0+21*by2*pow(vx0,2))*vy2+21*ay2*vx2*pow(vy0,2)+6*by2*vx0*vx2*vy0+27*ay2*pow(vx0,2)*vx2)*pow(vz0,2)+((-9*ny2*pow(vy0,3)-9*ny2*pow(vx0,2)*vy0)*vy2-9*ny2*vx0*vx2*pow(vy0,2)-9*ny2*pow(vx0,3)*vx2)*vz0+(9*by2*pow(vy0,4)+3*ay2*vx0*pow(vy0,3)+15*by2*pow(vx0,2)*pow(vy0,2)+3*ay2*pow(vx0,3)*vy0+6*by2*pow(vx0,4))*vy2+6*ay2*vx2*pow(vy0,4)+3*by2*vx0*vx2*pow(vy0,3)+15*ay2*pow(vx0,2)*vx2*pow(vy0,2)+3*by2*pow(vx0,3)*vx2*vy0+9*ay2*pow(vx0,4)*vx2)*pow(vz2,4)+((-12*ny2*pow(vy2,2)-12*ny2*pow(vx2,2))*pow(vz0,4)+((27*by2*vy0+21*ay2*vx0)*pow(vy2,2)+(6*ay2*vx2*vy0+6*by2*vx0*vx2)*vy2+21*by2*pow(vx2,2)*vy0+27*ay2*vx0*pow(vx2,2))*pow(vz0,3)+((-27*ny2*pow(vy0,2)-21*ny2*pow(vx0,2))*pow(vy2,2)-12*ny2*vx0*vx2*vy0*vy2-21*ny2*pow(vx2,2)*pow(vy0,2)-27*ny2*pow(vx0,2)*pow(vx2,2))*pow(vz0,2)+((27*by2*pow(vy0,3)+15*ay2*vx0*pow(vy0,2)+27*by2*pow(vx0,2)*vy0+15*ay2*pow(vx0,3))*pow(vy2,2)+(12*ay2*vx2*pow(vy0,3)+12*by2*vx0*vx2*pow(vy0,2)+12*ay2*pow(vx0,2)*vx2*vy0+12*by2*pow(vx0,3)*vx2)*vy2+15*by2*pow(vx2,2)*pow(vy0,3)+27*ay2*vx0*pow(vx2,2)*pow(vy0,2)+15*by2*pow(vx0,2)*pow(vx2,2)*vy0+27*ay2*pow(vx0,3)*pow(vx2,2))*vz0+(-9*ny2*pow(vy0,4)-15*ny2*pow(vx0,2)*pow(vy0,2)-6*ny2*pow(vx0,4))*pow(vy2,2)+(-6*ny2*vx0*vx2*pow(vy0,3)-6*ny2*pow(vx0,3)*vx2*vy0)*vy2-6*ny2*pow(vx2,2)*pow(vy0,4)-15*ny2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-9*ny2*pow(vx0,4)*pow(vx2,2))*pow(vz2,3)+((9*by2*pow(vy2,3)+9*ay2*vx2*pow(vy2,2)+9*by2*pow(vx2,2)*vy2+9*ay2*pow(vx2,3))*pow(vz0,4)+(-27*ny2*vy0*pow(vy2,3)-27*ny2*vx0*vx2*pow(vy2,2)-27*ny2*pow(vx2,2)*vy0*vy2-27*ny2*vx0*pow(vx2,3))*pow(vz0,3)+((27*by2*pow(vy0,2)+9*ay2*vx0*vy0+18*by2*pow(vx0,2))*pow(vy2,3)+(18*ay2*vx2*pow(vy0,2)+9*by2*vx0*vx2*vy0+27*ay2*pow(vx0,2)*vx2)*pow(vy2,2)+(27*by2*pow(vx2,2)*pow(vy0,2)+9*ay2*vx0*pow(vx2,2)*vy0+18*by2*pow(vx0,2)*pow(vx2,2))*vy2+18*ay2*pow(vx2,3)*pow(vy0,2)+9*by2*vx0*pow(vx2,3)*vy0+27*ay2*pow(vx0,2)*pow(vx2,3))*pow(vz0,2)+((-27*ny2*pow(vy0,3)-27*ny2*pow(vx0,2)*vy0)*pow(vy2,3)+(-27*ny2*vx0*vx2*pow(vy0,2)-27*ny2*pow(vx0,3)*vx2)*pow(vy2,2)+(-27*ny2*pow(vx2,2)*pow(vy0,3)-27*ny2*pow(vx0,2)*pow(vx2,2)*vy0)*vy2-27*ny2*vx0*pow(vx2,3)*pow(vy0,2)-27*ny2*pow(vx0,3)*pow(vx2,3))*vz0+(12*by2*pow(vy0,4)+3*ay2*vx0*pow(vy0,3)+18*by2*pow(vx0,2)*pow(vy0,2)+3*ay2*pow(vx0,3)*vy0+6*by2*pow(vx0,4))*pow(vy2,3)+(9*ay2*vx2*pow(vy0,4)+9*by2*vx0*vx2*pow(vy0,3)+18*ay2*pow(vx0,2)*vx2*pow(vy0,2)+9*by2*pow(vx0,3)*vx2*vy0+9*ay2*pow(vx0,4)*vx2)*pow(vy2,2)+(9*by2*pow(vx2,2)*pow(vy0,4)+9*ay2*vx0*pow(vx2,2)*pow(vy0,3)+18*by2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+9*ay2*pow(vx0,3)*pow(vx2,2)*vy0+9*by2*pow(vx0,4)*pow(vx2,2))*vy2+6*ay2*pow(vx2,3)*pow(vy0,4)+3*by2*vx0*pow(vx2,3)*pow(vy0,3)+18*ay2*pow(vx0,2)*pow(vx2,3)*pow(vy0,2)+3*by2*pow(vx0,3)*pow(vx2,3)*vy0+12*ay2*pow(vx0,4)*pow(vx2,3))*pow(vz2,2)+((-9*ny2*pow(vy2,4)-18*ny2*pow(vx2,2)*pow(vy2,2)-9*ny2*pow(vx2,4))*pow(vz0,4)+((9*by2*vy0+9*ay2*vx0)*pow(vy2,4)+(18*by2*pow(vx2,2)*vy0+18*ay2*vx0*pow(vx2,2))*pow(vy2,2)+9*by2*pow(vx2,4)*vy0+9*ay2*vx0*pow(vx2,4))*pow(vz0,3)+((-27*ny2*pow(vy0,2)-18*ny2*pow(vx0,2))*pow(vy2,4)-18*ny2*vx0*vx2*vy0*pow(vy2,3)+(-45*ny2*pow(vx2,2)*pow(vy0,2)-45*ny2*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)-18*ny2*vx0*pow(vx2,3)*vy0*vy2-18*ny2*pow(vx2,4)*pow(vy0,2)-27*ny2*pow(vx0,2)*pow(vx2,4))*pow(vz0,2)+((12*by2*pow(vy0,3)+6*ay2*vx0*pow(vy0,2)+12*by2*pow(vx0,2)*vy0+6*ay2*pow(vx0,3))*pow(vy2,4)+(6*ay2*vx2*pow(vy0,3)+6*by2*vx0*vx2*pow(vy0,2)+6*ay2*pow(vx0,2)*vx2*vy0+6*by2*pow(vx0,3)*vx2)*pow(vy2,3)+(18*by2*pow(vx2,2)*pow(vy0,3)+18*ay2*vx0*pow(vx2,2)*pow(vy0,2)+18*by2*pow(vx0,2)*pow(vx2,2)*vy0+18*ay2*pow(vx0,3)*pow(vx2,2))*pow(vy2,2)+(6*ay2*pow(vx2,3)*pow(vy0,3)+6*by2*vx0*pow(vx2,3)*pow(vy0,2)+6*ay2*pow(vx0,2)*pow(vx2,3)*vy0+6*by2*pow(vx0,3)*pow(vx2,3))*vy2+6*by2*pow(vx2,4)*pow(vy0,3)+12*ay2*vx0*pow(vx2,4)*pow(vy0,2)+6*by2*pow(vx0,2)*pow(vx2,4)*vy0+12*ay2*pow(vx0,3)*pow(vx2,4))*vz0+(-12*ny2*pow(vy0,4)-18*ny2*pow(vx0,2)*pow(vy0,2)-6*ny2*pow(vx0,4))*pow(vy2,4)+(-12*ny2*vx0*vx2*pow(vy0,3)-12*ny2*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(-18*ny2*pow(vx2,2)*pow(vy0,4)-36*ny2*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)-18*ny2*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(-12*ny2*vx0*pow(vx2,3)*pow(vy0,3)-12*ny2*pow(vx0,3)*pow(vx2,3)*vy0)*vy2-6*ny2*pow(vx2,4)*pow(vy0,4)-18*ny2*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)-12*ny2*pow(vx0,4)*pow(vx2,4))*vz2+(-9*ny2*vy0*pow(vy2,5)-9*ny2*vx0*vx2*pow(vy2,4)-18*ny2*pow(vx2,2)*vy0*pow(vy2,3)-18*ny2*vx0*pow(vx2,3)*pow(vy2,2)-9*ny2*pow(vx2,4)*vy0*vy2-9*ny2*vx0*pow(vx2,5))*pow(vz0,3)+(3*ay2*vx0*vy0*pow(vy2,5)+(-3*ay2*vx2*pow(vy0,2)-3*by2*vx0*vx2*vy0+3*ay2*pow(vx0,2)*vx2)*pow(vy2,4)+(3*by2*pow(vx2,2)*pow(vy0,2)-3*by2*pow(vx0,2)*pow(vx2,2))*pow(vy2,3)+(3*ay2*pow(vx0,2)*pow(vx2,3)-3*ay2*pow(vx2,3)*pow(vy0,2))*pow(vy2,2)+(3*by2*pow(vx2,4)*pow(vy0,2)-3*ay2*vx0*pow(vx2,4)*vy0-3*by2*pow(vx0,2)*pow(vx2,4))*vy2+3*by2*vx0*pow(vx2,5)*vy0)*pow(vz0,2)+((-12*ny2*pow(vy0,3)-12*ny2*pow(vx0,2)*vy0)*pow(vy2,5)+(-12*ny2*vx0*vx2*pow(vy0,2)-12*ny2*pow(vx0,3)*vx2)*pow(vy2,4)+(-24*ny2*pow(vx2,2)*pow(vy0,3)-24*ny2*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(-24*ny2*vx0*pow(vx2,3)*pow(vy0,2)-24*ny2*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(-12*ny2*pow(vx2,4)*pow(vy0,3)-12*ny2*pow(vx0,2)*pow(vx2,4)*vy0)*vy2-12*ny2*vx0*pow(vx2,5)*pow(vy0,2)-12*ny2*pow(vx0,3)*pow(vx2,5))*vz0)*pow(((80*pow(roughness,2)*pow(vz0,3)+(60*pow(roughness,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,5)+((120*pow(roughness,2)*vy0*vy2+120*pow(roughness,2)*vx0*vx2)*pow(vz0,2)+(60*pow(roughness,2)*pow(vy0,3)+60*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+60*pow(roughness,2)*vx0*vx2*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,3)*vx2)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,4)+((140*pow(roughness,2)*pow(vy2,2)+140*pow(roughness,2)*pow(vx2,2))*pow(vz0,3)+((180*pow(roughness,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+120*pow(roughness,2)*vx0*vx2*vy0*vy2+120*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,3)+((180*pow(roughness,2)*vy0*pow(vy2,3)+180*pow(roughness,2)*vx0*vx2*pow(vy2,2)+180*pow(roughness,2)*pow(vx2,2)*vy0*vy2+180*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,2)+(140*pow(roughness,2)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(180*pow(roughness,2)*vx0*vx2*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(120*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+140*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*pow(vz2,2)+((60*pow(roughness,2)*pow(vy2,4)+120*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+60*pow(roughness,2)*pow(vx2,4))*pow(vz0,3)+((120*pow(roughness,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+120*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(180*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+120*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+60*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*vz0)*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2)*vz2+((60*pow(roughness,2)*vy0*pow(vy2,5)+60*pow(roughness,2)*vx0*vx2*pow(vy2,4)+120*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+120*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+60*pow(roughness,2)*pow(vx2,4)*vy0*vy2+60*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,3)+60*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(120*pow(roughness,2)*vx0*vx2*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(140*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(180*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+140*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(60*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+60*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*pow(pow(vz0,2)+pow(vy0,2)+pow(vx0,2),1/2))*pow(pow(vz2,2)+pow(vy2,2)+pow(vx2,2),1/2)+(80*pow(roughness,2)*pow(vz0,4)+(100*pow(roughness,2)*pow(vy0,2)+100*pow(roughness,2)*pow(vx0,2))*pow(vz0,2)+20*pow(roughness,2)*pow(vy0,4)+40*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,4))*pow(vz2,6)+((120*pow(roughness,2)*vy0*vy2+120*pow(roughness,2)*vx0*vx2)*pow(vz0,3)+((120*pow(roughness,2)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,2)*vy0)*vy2+120*pow(roughness,2)*vx0*vx2*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,3)*vx2)*vz0)*pow(vz2,5)+((180*pow(roughness,2)*pow(vy2,2)+180*pow(roughness,2)*pow(vx2,2))*pow(vz0,4)+((300*pow(roughness,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2))*pow(vy2,2)+120*pow(roughness,2)*vx0*vx2*vy0*vy2+240*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vz0,2)+(120*pow(roughness,2)*pow(vy0,4)+180*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,4))*pow(vy2,2)+(120*pow(roughness,2)*vx0*vx2*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*vy2+60*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vz2,4)+((240*pow(roughness,2)*vy0*pow(vy2,3)+240*pow(roughness,2)*vx0*vx2*pow(vy2,2)+240*pow(roughness,2)*pow(vx2,2)*vy0*vy2+240*pow(roughness,2)*vx0*pow(vx2,3))*pow(vz0,3)+((240*pow(roughness,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,3)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,2)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*vy2+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*vz0)*pow(vz2,3)+((120*pow(roughness,2)*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,2)*pow(vy2,2)+120*pow(roughness,2)*pow(vx2,4))*pow(vz0,4)+((300*pow(roughness,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2))*pow(vy2,4)+240*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,3)+(480*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+240*pow(roughness,2)*vx0*pow(vx2,3)*vy0*vy2+180*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vz0,2)+(180*pow(roughness,2)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+60*pow(roughness,2)*pow(vx0,4))*pow(vy2,4)+(240*pow(roughness,2)*vx0*vx2*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,3)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+480*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*vy2+60*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vz2,2)+((120*pow(roughness,2)*vy0*pow(vy2,5)+120*pow(roughness,2)*vx0*vx2*pow(vy2,4)+240*pow(roughness,2)*pow(vx2,2)*vy0*pow(vy2,3)+240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy2,2)+120*pow(roughness,2)*pow(vx2,4)*vy0*vy2+120*pow(roughness,2)*vx0*pow(vx2,5))*pow(vz0,3)+((120*pow(roughness,2)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,2)*vy0)*pow(vy2,5)+(120*pow(roughness,2)*vx0*vx2*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,3)*vx2)*pow(vy2,4)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*vy0)*pow(vy2,3)+(240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3))*pow(vy2,2)+(120*pow(roughness,2)*pow(vx2,4)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*vy0)*vy2+120*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5))*vz0)*vz2+(20*pow(roughness,2)*pow(vy2,6)+60*pow(roughness,2)*pow(vx2,2)*pow(vy2,4)+60*pow(roughness,2)*pow(vx2,4)*pow(vy2,2)+20*pow(roughness,2)*pow(vx2,6))*pow(vz0,4)+((100*pow(roughness,2)*pow(vy0,2)+40*pow(roughness,2)*pow(vx0,2))*pow(vy2,6)+120*pow(roughness,2)*vx0*vx2*vy0*pow(vy2,5)+(240*pow(roughness,2)*pow(vx2,2)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,2)*pow(vx2,2))*pow(vy2,4)+240*pow(roughness,2)*vx0*pow(vx2,3)*vy0*pow(vy2,3)+(180*pow(roughness,2)*pow(vx2,4)*pow(vy0,2)+240*pow(roughness,2)*pow(vx0,2)*pow(vx2,4))*pow(vy2,2)+120*pow(roughness,2)*vx0*pow(vx2,5)*vy0*vy2+40*pow(roughness,2)*pow(vx2,6)*pow(vy0,2)+100*pow(roughness,2)*pow(vx0,2)*pow(vx2,6))*pow(vz0,2)+(80*pow(roughness,2)*pow(vy0,4)+100*pow(roughness,2)*pow(vx0,2)*pow(vy0,2)+20*pow(roughness,2)*pow(vx0,4))*pow(vy2,6)+(120*pow(roughness,2)*vx0*vx2*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,3)*vx2*vy0)*pow(vy2,5)+(180*pow(roughness,2)*pow(vx2,2)*pow(vy0,4)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+120*pow(roughness,2)*pow(vx0,4)*pow(vx2,2))*pow(vy2,4)+(240*pow(roughness,2)*vx0*pow(vx2,3)*pow(vy0,3)+240*pow(roughness,2)*pow(vx0,3)*pow(vx2,3)*vy0)*pow(vy2,3)+(120*pow(roughness,2)*pow(vx2,4)*pow(vy0,4)+300*pow(roughness,2)*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+180*pow(roughness,2)*pow(vx0,4)*pow(vx2,4))*pow(vy2,2)+(120*pow(roughness,2)*vx0*pow(vx2,5)*pow(vy0,3)+120*pow(roughness,2)*pow(vx0,3)*pow(vx2,5)*vy0)*vy2+20*pow(roughness,2)*pow(vx2,6)*pow(vy0,4)+100*pow(roughness,2)*pow(vx0,2)*pow(vx2,6)*pow(vy0,2)+80*pow(roughness,2)*pow(vx0,4)*pow(vx2,6),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
#endif

#if 0 // no propagation, beckmann
// 1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(n2,2)-2*n1*n2+pow(n1,2))*vz0*pow(vz2,5)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vz0,2)*pow(vz2,4)+((6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vz0,3)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*vy2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,3)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vz0,4)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vy0*vy2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy0,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0*vx2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,2))*pow(vz0,2))*pow(vz2,2)+((pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vz0,5)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*vy2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vy2,4)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*pow(vy2,3)+((6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy0,3)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0*vx2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,2))*vy0)*vy2+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vy0,4)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vx2,4)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*pow(vx2,3)+(6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,3)*vx2+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vx0,4))*vz0)*vz2)*pow(((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz2,4)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,3)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,3)+((6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,4)+((6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,2))*pow(vz2,2)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,5)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,3))*vz2+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,6)+((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,4),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// x0  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*vz0+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz2,6)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,3)+((10*pow(roughness,2)*bx0*pow(n2,2)-20*pow(roughness,2)*bx0*n1*n2+10*pow(roughness,2)*bx0*pow(n1,2))*vy0+(10*pow(roughness,2)*ax0*pow(n2,2)-20*pow(roughness,2)*ax0*n1*n2+10*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vz0,2)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,4)+(((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*vy2+((16*pow(roughness,2)+2)*bx0*pow(n2,2)+(-32*pow(roughness,2)-4)*bx0*n1*n2+(16*pow(roughness,2)+2)*bx0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+((16*pow(roughness,2)+2)*ax0*pow(n2,2)+(-32*pow(roughness,2)-4)*ax0*n1*n2+(16*pow(roughness,2)+2)*ax0*pow(n1,2))*vx0)*pow(vz0,3)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vy0*vy2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,2)+(((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+(((4*pow(roughness,2)+2)*bx0*pow(n2,2)+(-8*pow(roughness,2)-4)*bx0*n1*n2+(4*pow(roughness,2)+2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+(2*bx0*pow(n2,2)-4*bx0*n1*n2+2*bx0*pow(n1,2))*pow(vy0,3)+(((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+(2*ax0*pow(n2,2)-4*ax0*n1*n2+2*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(2*bx0*pow(n2,2)-4*bx0*n1*n2+2*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((4*pow(roughness,2)+2)*ax0*pow(n2,2)+(-8*pow(roughness,2)-4)*ax0*n1*n2+(4*pow(roughness,2)+2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(2*ax0*pow(n2,2)-4*ax0*n1*n2+2*ax0*pow(n1,2))*pow(vx0,3))*vz0+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,5)+(((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*vy2+((8*pow(roughness,2)+6)*bx0*pow(n2,2)+(-16*pow(roughness,2)-12)*bx0*n1*n2+(8*pow(roughness,2)+6)*bx0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+((8*pow(roughness,2)+6)*ax0*pow(n2,2)+(-16*pow(roughness,2)-12)*ax0*n1*n2+(8*pow(roughness,2)+6)*ax0*pow(n1,2))*vx0)*pow(vz0,4)+(((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx0*vy0*vy2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx0*pow(vy0,2)+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,3)+(((12*pow(roughness,2)*bx0*pow(n2,2)-24*pow(roughness,2)*bx0*n1*n2+12*pow(roughness,2)*bx0*pow(n1,2))*vy0+(12*pow(roughness,2)*ax0*pow(n2,2)-24*pow(roughness,2)*ax0*n1*n2+12*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+(((12*pow(roughness,2)+6)*bx0*pow(n2,2)+(-24*pow(roughness,2)-12)*bx0*n1*n2+(12*pow(roughness,2)+6)*bx0*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ax0*pow(n2,2)-48*pow(roughness,2)*ax0*n1*n2+24*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+(6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vy0,3)+(((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((12*pow(roughness,2)*bx0*pow(n2,2)-24*pow(roughness,2)*bx0*n1*n2+12*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*bx0*pow(n2,2)-48*pow(roughness,2)*bx0*n1*n2+24*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vx0,2))*vy0+(12*pow(roughness,2)*ax0*pow(n2,2)-24*pow(roughness,2)*ax0*n1*n2+12*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((12*pow(roughness,2)+6)*ax0*pow(n2,2)+(-24*pow(roughness,2)-12)*ax0*n1*n2+(12*pow(roughness,2)+6)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,3)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,4))*vz0)*pow(vz2,3)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,6)+(((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*vy2+((6-2*pow(roughness,2))*bx0*pow(n2,2)+(4*pow(roughness,2)-12)*bx0*n1*n2+(6-2*pow(roughness,2))*bx0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+((6-2*pow(roughness,2))*ax0*pow(n2,2)+(4*pow(roughness,2)-12)*ax0*n1*n2+(6-2*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vz0,5)+(((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx0*vy0*vy2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx0*vx0*vx2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,4)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy2,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+((6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vy0,2)+((-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2+(16*pow(roughness,2)*ax0*pow(n2,2)-32*pow(roughness,2)*ax0*n1*n2+16*pow(roughness,2)*ax0*pow(n1,2))*vx0)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((6-16*pow(roughness,2))*bx0*pow(n2,2)+(32*pow(roughness,2)-12)*bx0*n1*n2+(6-16*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+((6-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-12)*bx0*n1*n2+(6-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,3)+(((6-16*pow(roughness,2))*ax0*pow(n2,2)+(32*pow(roughness,2)-12)*ax0*n1*n2+(6-16*pow(roughness,2))*ax0*pow(n1,2))*vx2+((6-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-12)*ax0*n1*n2+(6-4*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*bx0*pow(n2,2)-32*pow(roughness,2)*bx0*n1*n2+16*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((6-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-12)*bx0*n1*n2+(6-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*pow(vx0,2)*vx2+((6-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-12)*ax0*n1*n2+(6-4*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,2))*vy0)*vy2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*vx0*vx2+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz0,2)+(((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,5))*vz0+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,5)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*pow(vx2,3)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,5)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,6))*pow(vz2,2)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,7)+(((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*vy2+((2-2*pow(roughness,2))*bx0*pow(n2,2)+(4*pow(roughness,2)-4)*bx0*n1*n2+(2-2*pow(roughness,2))*bx0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+((2-2*pow(roughness,2))*ax0*pow(n2,2)+(4*pow(roughness,2)-4)*ax0*n1*n2+(2-2*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vz0,6)+(((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx0*vy0*vy2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,5)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy2,3)+((-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2)*pow(vy2,2)+(((2-8*pow(roughness,2))*bx0*pow(n2,2)+(16*pow(roughness,2)-4)*bx0*n1*n2+(2-8*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((2-8*pow(roughness,2))*bx0*pow(n2,2)+(16*pow(roughness,2)-4)*bx0*n1*n2+(2-8*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,3)+(((2-8*pow(roughness,2))*ax0*pow(n2,2)+(16*pow(roughness,2)-4)*ax0*n1*n2+(2-8*pow(roughness,2))*ax0*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2)*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((2-8*pow(roughness,2))*ax0*pow(n2,2)+(16*pow(roughness,2)-4)*ax0*n1*n2+(2-8*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,2)*vx2+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,3)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*vy0)*vy2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz0,3)+(((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+(((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,5)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,6)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,4)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*pow(vx2,3)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,5)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,6))*vz0)*vz2)*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2))*pow(vz2,3)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3))*pow(vz2,2)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4))*vz2+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,9)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,7)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,5),-1) ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// y0  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*vz0+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz2,6)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,3)+((10*pow(roughness,2)*by0*pow(n2,2)-20*pow(roughness,2)*by0*n1*n2+10*pow(roughness,2)*by0*pow(n1,2))*vy0+(10*pow(roughness,2)*ay0*pow(n2,2)-20*pow(roughness,2)*ay0*n1*n2+10*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vz0,2)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,4)+(((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*vy2+((16*pow(roughness,2)+2)*by0*pow(n2,2)+(-32*pow(roughness,2)-4)*by0*n1*n2+(16*pow(roughness,2)+2)*by0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+((16*pow(roughness,2)+2)*ay0*pow(n2,2)+(-32*pow(roughness,2)-4)*ay0*n1*n2+(16*pow(roughness,2)+2)*ay0*pow(n1,2))*vx0)*pow(vz0,3)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vy0*vy2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,2)+(((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+(((4*pow(roughness,2)+2)*by0*pow(n2,2)+(-8*pow(roughness,2)-4)*by0*n1*n2+(4*pow(roughness,2)+2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+(2*by0*pow(n2,2)-4*by0*n1*n2+2*by0*pow(n1,2))*pow(vy0,3)+(((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+(2*ay0*pow(n2,2)-4*ay0*n1*n2+2*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(2*by0*pow(n2,2)-4*by0*n1*n2+2*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((4*pow(roughness,2)+2)*ay0*pow(n2,2)+(-8*pow(roughness,2)-4)*ay0*n1*n2+(4*pow(roughness,2)+2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(2*ay0*pow(n2,2)-4*ay0*n1*n2+2*ay0*pow(n1,2))*pow(vx0,3))*vz0+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,5)+(((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*vy2+((8*pow(roughness,2)+6)*by0*pow(n2,2)+(-16*pow(roughness,2)-12)*by0*n1*n2+(8*pow(roughness,2)+6)*by0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+((8*pow(roughness,2)+6)*ay0*pow(n2,2)+(-16*pow(roughness,2)-12)*ay0*n1*n2+(8*pow(roughness,2)+6)*ay0*pow(n1,2))*vx0)*pow(vz0,4)+(((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny0*vy0*vy2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny0*pow(vy0,2)+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,3)+(((12*pow(roughness,2)*by0*pow(n2,2)-24*pow(roughness,2)*by0*n1*n2+12*pow(roughness,2)*by0*pow(n1,2))*vy0+(12*pow(roughness,2)*ay0*pow(n2,2)-24*pow(roughness,2)*ay0*n1*n2+12*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+(((12*pow(roughness,2)+6)*by0*pow(n2,2)+(-24*pow(roughness,2)-12)*by0*n1*n2+(12*pow(roughness,2)+6)*by0*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ay0*pow(n2,2)-48*pow(roughness,2)*ay0*n1*n2+24*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+(6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vy0,3)+(((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((12*pow(roughness,2)*by0*pow(n2,2)-24*pow(roughness,2)*by0*n1*n2+12*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*by0*pow(n2,2)-48*pow(roughness,2)*by0*n1*n2+24*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vx0,2))*vy0+(12*pow(roughness,2)*ay0*pow(n2,2)-24*pow(roughness,2)*ay0*n1*n2+12*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((12*pow(roughness,2)+6)*ay0*pow(n2,2)+(-24*pow(roughness,2)-12)*ay0*n1*n2+(12*pow(roughness,2)+6)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,3)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,4))*vz0)*pow(vz2,3)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,6)+(((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*vy2+((6-2*pow(roughness,2))*by0*pow(n2,2)+(4*pow(roughness,2)-12)*by0*n1*n2+(6-2*pow(roughness,2))*by0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+((6-2*pow(roughness,2))*ay0*pow(n2,2)+(4*pow(roughness,2)-12)*ay0*n1*n2+(6-2*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vz0,5)+(((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny0*vy0*vy2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny0*vx0*vx2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,4)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy2,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+((6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vy0,2)+((-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2+(16*pow(roughness,2)*ay0*pow(n2,2)-32*pow(roughness,2)*ay0*n1*n2+16*pow(roughness,2)*ay0*pow(n1,2))*vx0)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((6-16*pow(roughness,2))*by0*pow(n2,2)+(32*pow(roughness,2)-12)*by0*n1*n2+(6-16*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+((6-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-12)*by0*n1*n2+(6-4*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,3)+(((6-16*pow(roughness,2))*ay0*pow(n2,2)+(32*pow(roughness,2)-12)*ay0*n1*n2+(6-16*pow(roughness,2))*ay0*pow(n1,2))*vx2+((6-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-12)*ay0*n1*n2+(6-4*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*by0*pow(n2,2)-32*pow(roughness,2)*by0*n1*n2+16*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((6-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-12)*by0*n1*n2+(6-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*pow(vx0,2)*vx2+((6-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-12)*ay0*n1*n2+(6-4*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,2))*vy0)*vy2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*vx0*vx2+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz0,2)+(((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,5))*vz0+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,5)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*pow(vx2,3)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,5)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,6))*pow(vz2,2)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,7)+(((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*vy2+((2-2*pow(roughness,2))*by0*pow(n2,2)+(4*pow(roughness,2)-4)*by0*n1*n2+(2-2*pow(roughness,2))*by0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+((2-2*pow(roughness,2))*ay0*pow(n2,2)+(4*pow(roughness,2)-4)*ay0*n1*n2+(2-2*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vz0,6)+(((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny0*vy0*vy2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,5)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy2,3)+((-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2)*pow(vy2,2)+(((2-8*pow(roughness,2))*by0*pow(n2,2)+(16*pow(roughness,2)-4)*by0*n1*n2+(2-8*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((2-8*pow(roughness,2))*by0*pow(n2,2)+(16*pow(roughness,2)-4)*by0*n1*n2+(2-8*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,3)+(((2-8*pow(roughness,2))*ay0*pow(n2,2)+(16*pow(roughness,2)-4)*ay0*n1*n2+(2-8*pow(roughness,2))*ay0*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2)*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((2-8*pow(roughness,2))*ay0*pow(n2,2)+(16*pow(roughness,2)-4)*ay0*n1*n2+(2-8*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,2)*vx2+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,3)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*vy0)*vy2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz0,3)+(((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+(((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,5)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,6)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,4)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*pow(vx2,3)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,5)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,6))*vz0)*vz2)*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2))*pow(vz2,3)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3))*pow(vz2,2)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4))*vz2+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,9)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,7)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,5),-1) ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
// x1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vz0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*vz0+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz2,5)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vz0,2)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,4)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,4)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vx2+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vx0)*pow(vz0,3)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vx0,2)*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*vz0+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vy0)*vy2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,4)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vz2,3)+(((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,5)+(((16-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-32)*n1*n2+(16-32*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vz0,4)+(((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+(((16-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-32)*n1*n2+(16-32*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*pow(n2,2)-32*n1*n2+16*pow(n1,2))*pow(vx0,2)*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(10*k1*pow(roughness,2)*pow(n2,2)-20*k1*pow(roughness,2)*n1*n2+10*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vz0)*pow(vz2,2)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,6)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vx2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vx0)*pow(vz0,5)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(10*k1*pow(roughness,2)*pow(n2,2)-20*k1*pow(roughness,2)*n1*n2+10*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+(((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+((-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+(((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(-40*pow(roughness,2)*pow(n2,2)+80*pow(roughness,2)*n1*n2-40*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((8-48*pow(roughness,2))*pow(n2,2)+(96*pow(roughness,2)-16)*n1*n2+(8-48*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,3)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,4)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,3)+(((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(-40*pow(roughness,2)*pow(n2,2)+80*pow(roughness,2)*n1*n2-40*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(-48*pow(roughness,2)*pow(n2,2)+96*pow(roughness,2)*n1*n2-48*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vz0+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vy0)*pow(vy2,3)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,4)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,5)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,6)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,4)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,7))*vz2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vz0,7)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*vy0*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,5)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*vy0*pow(vy2,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*vy0)*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,3)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2*vy0)*pow(vy2,3)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,5)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,6)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,4)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2)*vz0)*pow(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,3)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2))*pow(vz2,2)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3))*vz2+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4),-1) ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
// y1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vz0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vy0*vz0+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vz2,5)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vy0*pow(vz0,2)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*vz0)*pow(vz2,4)+(((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,4)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vy2+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vy0)*pow(vz0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,2)+((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vy0,2)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*vz0+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,2)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2))*vy2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vz2,3)+(((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,5)+(((16-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-32)*n1*n2+(16-32*pow(roughness,2))*pow(n1,2))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vy0)*pow(vz0,4)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((16*pow(n2,2)-32*n1*n2+16*pow(n1,2))*pow(vy0,2)+((16-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-32)*n1*n2+(16-32*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,2)+(((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,2)+((10*k2*pow(roughness,2)*pow(n2,2)-20*k2*pow(roughness,2)*n1*n2+10*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*vz0)*pow(vz2,2)+(((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,6)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vy2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vy0)*pow(vz0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(10*k2*pow(roughness,2)*pow(n2,2)-20*k2*pow(roughness,2)*n1*n2+10*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,4)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(-40*pow(roughness,2)*pow(n2,2)+80*pow(roughness,2)*n1*n2-40*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+(((8-48*pow(roughness,2))*pow(n2,2)+(96*pow(roughness,2)-16)*n1*n2+(8-48*pow(roughness,2))*pow(n1,2))*pow(vy0,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vz0,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy2,3)+((-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-40*pow(roughness,2)*pow(n2,2)+80*pow(roughness,2)*n1*n2-40*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(-48*pow(roughness,2)*pow(n2,2)+96*pow(roughness,2)*n1*n2-48*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*vz0+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2))*pow(vy2,3)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2))*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,7)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,5)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy0)*vz2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy2*pow(vz0,7)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2)*pow(vz0,5)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy2,5)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2)*pow(vz0,3)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,4)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy2)*vz0)*pow(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,3)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2))*pow(vz2,2)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3))*vz2+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4),-1) ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
// x2  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*vz0*pow(vz2,5)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,2)+(((4*pow(roughness,2)-2)*bx2*pow(n2,2)+(4-8*pow(roughness,2))*bx2*n1*n2+(4*pow(roughness,2)-2)*bx2*pow(n1,2))*vy2+((4*pow(roughness,2)-2)*bx2*pow(n2,2)+(4-8*pow(roughness,2))*bx2*n1*n2+(4*pow(roughness,2)-2)*bx2*pow(n1,2))*vy0+((4*pow(roughness,2)-2)*ax2*pow(n2,2)+(4-8*pow(roughness,2))*ax2*n1*n2+(4*pow(roughness,2)-2)*ax2*pow(n1,2))*vx2+((4*pow(roughness,2)-2)*ax2*pow(n2,2)+(4-8*pow(roughness,2))*ax2*n1*n2+(4*pow(roughness,2)-2)*ax2*pow(n1,2))*vx0)*vz0)*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,3)+(((12*pow(roughness,2)-6)*bx2*pow(n2,2)+(12-24*pow(roughness,2))*bx2*n1*n2+(12*pow(roughness,2)-6)*bx2*pow(n1,2))*vy2+((12*pow(roughness,2)-6)*bx2*pow(n2,2)+(12-24*pow(roughness,2))*bx2*n1*n2+(12*pow(roughness,2)-6)*bx2*pow(n1,2))*vy0+((12*pow(roughness,2)-6)*ax2*pow(n2,2)+(12-24*pow(roughness,2))*ax2*n1*n2+(12*pow(roughness,2)-6)*ax2*pow(n1,2))*vx2+((12*pow(roughness,2)-6)*ax2*pow(n2,2)+(12-24*pow(roughness,2))*ax2*n1*n2+(12*pow(roughness,2)-6)*ax2*pow(n1,2))*vx0)*pow(vz0,2)+(((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vy2,2)+((4-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-8)*n1*n2+(4-4*pow(roughness,2))*pow(n1,2))*nx2*vy0*vy2+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vy0,2)+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vx2,2)+((4-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-8)*n1*n2+(4-4*pow(roughness,2))*pow(n1,2))*nx2*vx0*vx2+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vx0,2))*vz0)*pow(vz2,3)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,4)+(((12*pow(roughness,2)-6)*bx2*pow(n2,2)+(12-24*pow(roughness,2))*bx2*n1*n2+(12*pow(roughness,2)-6)*bx2*pow(n1,2))*vy2+((12*pow(roughness,2)-6)*bx2*pow(n2,2)+(12-24*pow(roughness,2))*bx2*n1*n2+(12*pow(roughness,2)-6)*bx2*pow(n1,2))*vy0+((12*pow(roughness,2)-6)*ax2*pow(n2,2)+(12-24*pow(roughness,2))*ax2*n1*n2+(12*pow(roughness,2)-6)*ax2*pow(n1,2))*vx2+((12*pow(roughness,2)-6)*ax2*pow(n2,2)+(12-24*pow(roughness,2))*ax2*n1*n2+(12*pow(roughness,2)-6)*ax2*pow(n1,2))*vx0)*pow(vz0,3)+(((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vy2,2)+((8-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-16)*n1*n2+(8-4*pow(roughness,2))*pow(n1,2))*nx2*vy0*vy2+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vy0,2)+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vx2,2)+((8-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-16)*n1*n2+(8-4*pow(roughness,2))*pow(n1,2))*nx2*vx0*vx2+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,2)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy2,3)+((12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*vy0+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,2)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx0)*vy0+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+((4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,3)+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,2)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,5)+(((4*pow(roughness,2)-2)*bx2*pow(n2,2)+(4-8*pow(roughness,2))*bx2*n1*n2+(4*pow(roughness,2)-2)*bx2*pow(n1,2))*vy2+((4*pow(roughness,2)-2)*bx2*pow(n2,2)+(4-8*pow(roughness,2))*bx2*n1*n2+(4*pow(roughness,2)-2)*bx2*pow(n1,2))*vy0+((4*pow(roughness,2)-2)*ax2*pow(n2,2)+(4-8*pow(roughness,2))*ax2*n1*n2+(4*pow(roughness,2)-2)*ax2*pow(n1,2))*vx2+((4*pow(roughness,2)-2)*ax2*pow(n2,2)+(4-8*pow(roughness,2))*ax2*n1*n2+(4*pow(roughness,2)-2)*ax2*pow(n1,2))*vx0)*pow(vz0,4)+(((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*nx2*pow(vy2,2)+((4*pow(roughness,2)+4)*pow(n2,2)+(-8*pow(roughness,2)-8)*n1*n2+(4*pow(roughness,2)+4)*pow(n1,2))*nx2*vy0*vy2+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*nx2*pow(vy0,2)+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*nx2*pow(vx2,2)+((4*pow(roughness,2)+4)*pow(n2,2)+(-8*pow(roughness,2)-8)*n1*n2+(4*pow(roughness,2)+4)*pow(n1,2))*nx2*vx0*vx2+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,3)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy2,3)+((12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*vy0+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,2)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx0)*vy0+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+((4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,3)+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,4)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*vy0*pow(vy2,3)+((-18*pow(roughness,2)*pow(n2,2)+36*pow(roughness,2)*n1*n2-18*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*vy2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(-18*pow(roughness,2)*pow(n2,2)+36*pow(roughness,2)*n1*n2-18*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*vz0)*vz2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,6)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vy0*vy2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,4)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vy0*pow(vy2,3)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*vy2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*pow(vz0,2))*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz2,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2))*pow(vz2,3)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3))*pow(vz2,2)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4))*vz2+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
// y2  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*vz0*pow(vz2,5)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,2)+(((4*pow(roughness,2)-2)*by2*pow(n2,2)+(4-8*pow(roughness,2))*by2*n1*n2+(4*pow(roughness,2)-2)*by2*pow(n1,2))*vy2+((4*pow(roughness,2)-2)*by2*pow(n2,2)+(4-8*pow(roughness,2))*by2*n1*n2+(4*pow(roughness,2)-2)*by2*pow(n1,2))*vy0+((4*pow(roughness,2)-2)*ay2*pow(n2,2)+(4-8*pow(roughness,2))*ay2*n1*n2+(4*pow(roughness,2)-2)*ay2*pow(n1,2))*vx2+((4*pow(roughness,2)-2)*ay2*pow(n2,2)+(4-8*pow(roughness,2))*ay2*n1*n2+(4*pow(roughness,2)-2)*ay2*pow(n1,2))*vx0)*vz0)*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,3)+(((12*pow(roughness,2)-6)*by2*pow(n2,2)+(12-24*pow(roughness,2))*by2*n1*n2+(12*pow(roughness,2)-6)*by2*pow(n1,2))*vy2+((12*pow(roughness,2)-6)*by2*pow(n2,2)+(12-24*pow(roughness,2))*by2*n1*n2+(12*pow(roughness,2)-6)*by2*pow(n1,2))*vy0+((12*pow(roughness,2)-6)*ay2*pow(n2,2)+(12-24*pow(roughness,2))*ay2*n1*n2+(12*pow(roughness,2)-6)*ay2*pow(n1,2))*vx2+((12*pow(roughness,2)-6)*ay2*pow(n2,2)+(12-24*pow(roughness,2))*ay2*n1*n2+(12*pow(roughness,2)-6)*ay2*pow(n1,2))*vx0)*pow(vz0,2)+(((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vy2,2)+((4-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-8)*n1*n2+(4-4*pow(roughness,2))*pow(n1,2))*ny2*vy0*vy2+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vy0,2)+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vx2,2)+((4-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-8)*n1*n2+(4-4*pow(roughness,2))*pow(n1,2))*ny2*vx0*vx2+((2-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-4)*n1*n2+(2-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vx0,2))*vz0)*pow(vz2,3)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,4)+(((12*pow(roughness,2)-6)*by2*pow(n2,2)+(12-24*pow(roughness,2))*by2*n1*n2+(12*pow(roughness,2)-6)*by2*pow(n1,2))*vy2+((12*pow(roughness,2)-6)*by2*pow(n2,2)+(12-24*pow(roughness,2))*by2*n1*n2+(12*pow(roughness,2)-6)*by2*pow(n1,2))*vy0+((12*pow(roughness,2)-6)*ay2*pow(n2,2)+(12-24*pow(roughness,2))*ay2*n1*n2+(12*pow(roughness,2)-6)*ay2*pow(n1,2))*vx2+((12*pow(roughness,2)-6)*ay2*pow(n2,2)+(12-24*pow(roughness,2))*ay2*n1*n2+(12*pow(roughness,2)-6)*ay2*pow(n1,2))*vx0)*pow(vz0,3)+(((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vy2,2)+((8-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-16)*n1*n2+(8-4*pow(roughness,2))*pow(n1,2))*ny2*vy0*vy2+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vy0,2)+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vx2,2)+((8-4*pow(roughness,2))*pow(n2,2)+(8*pow(roughness,2)-16)*n1*n2+(8-4*pow(roughness,2))*pow(n1,2))*ny2*vx0*vx2+((4-2*pow(roughness,2))*pow(n2,2)+(4*pow(roughness,2)-8)*n1*n2+(4-2*pow(roughness,2))*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,2)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy2,3)+((12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*vy0+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,2)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx0)*vy0+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+((4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,3)+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,2)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,5)+(((4*pow(roughness,2)-2)*by2*pow(n2,2)+(4-8*pow(roughness,2))*by2*n1*n2+(4*pow(roughness,2)-2)*by2*pow(n1,2))*vy2+((4*pow(roughness,2)-2)*by2*pow(n2,2)+(4-8*pow(roughness,2))*by2*n1*n2+(4*pow(roughness,2)-2)*by2*pow(n1,2))*vy0+((4*pow(roughness,2)-2)*ay2*pow(n2,2)+(4-8*pow(roughness,2))*ay2*n1*n2+(4*pow(roughness,2)-2)*ay2*pow(n1,2))*vx2+((4*pow(roughness,2)-2)*ay2*pow(n2,2)+(4-8*pow(roughness,2))*ay2*n1*n2+(4*pow(roughness,2)-2)*ay2*pow(n1,2))*vx0)*pow(vz0,4)+(((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*ny2*pow(vy2,2)+((4*pow(roughness,2)+4)*pow(n2,2)+(-8*pow(roughness,2)-8)*n1*n2+(4*pow(roughness,2)+4)*pow(n1,2))*ny2*vy0*vy2+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*ny2*pow(vy0,2)+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*ny2*pow(vx2,2)+((4*pow(roughness,2)+4)*pow(n2,2)+(-8*pow(roughness,2)-8)*n1*n2+(4*pow(roughness,2)+4)*pow(n1,2))*ny2*vx0*vx2+((2*pow(roughness,2)+2)*pow(n2,2)+(-4*pow(roughness,2)-4)*n1*n2+(2*pow(roughness,2)+2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,3)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy2,3)+((12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*vy0+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,2)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx0)*vy0+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+((4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,3)+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,4)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*vy0*pow(vy2,3)+((-18*pow(roughness,2)*pow(n2,2)+36*pow(roughness,2)*n1*n2-18*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*vy2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(-18*pow(roughness,2)*pow(n2,2)+36*pow(roughness,2)*n1*n2-18*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*vz0)*vz2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,6)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vy0*vy2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,4)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vy0*pow(vy2,3)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*vy2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*pow(vz0,2))*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz2,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2))*pow(vz2,3)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3))*pow(vz2,2)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4))*vz2+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
#endif

#if 0 // phong
  // conversion beckmann roughness -> phong exponent following walter 2007
  const float exponent = 2.0f/(roughness*roughness) - 2.0f;
  assert(!(exponent != exponent));
   
// 1  :  
poly.add(Term6f( ((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*vz0*vz2*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz2,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2),-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// x0  :  
poly.add(Term6f( -((((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx0*
pow(vz0,2)+(((2*exponent+4)*bx0*
pow(n2,2)+(-4*exponent-8)*bx0*n1*n2+(2*exponent+4)*bx0*
pow(n1,2))*vy0+((2*exponent+4)*ax0*
pow(n2,2)+(-4*exponent-8)*ax0*n1*n2+(2*exponent+4)*ax0*
pow(n1,2))*vx0)*vz0+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vz2,4)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*nx0*
pow(vz0,3)+(((6*exponent+12)*bx0*
pow(n2,2)+(-12*exponent-24)*bx0*n1*n2+(6*exponent+12)*bx0*
pow(n1,2))*vy0+((6*exponent+12)*ax0*
pow(n2,2)+(-12*exponent-24)*ax0*n1*n2+(6*exponent+12)*ax0*
pow(n1,2))*vx0)*
pow(vz0,2)+(((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*nx0*
pow(vy0,2)+((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*nx0*
pow(vx0,2))*vz0)*
pow(vz2,3)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*nx0*
pow(vz0,4)+(((
pow(exponent,2)+2*exponent)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx0*n1*n2+(
pow(exponent,2)+2*exponent)*bx0*
pow(n1,2))*vy2+((
pow(exponent,2)+8*exponent+12)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*bx0*n1*n2+(
pow(exponent,2)+8*exponent+12)*bx0*
pow(n1,2))*vy0+((
pow(exponent,2)+2*exponent)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax0*n1*n2+(
pow(exponent,2)+2*exponent)*ax0*
pow(n1,2))*vx2+((
pow(exponent,2)+8*exponent+12)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*ax0*n1*n2+(
pow(exponent,2)+8*exponent+12)*ax0*
pow(n1,2))*vx0)*
pow(vz0,3)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx0*
pow(vy2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*nx0*vy0*vy2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vy0,2)+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx0*
pow(vx2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*nx0*vx0*vx2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vz0,2)+((((2*exponent+4)*bx0*
pow(n2,2)+(-4*exponent-8)*bx0*n1*n2+(2*exponent+4)*bx0*
pow(n1,2))*vy0+((2*exponent+4)*ax0*
pow(n2,2)+(-4*exponent-8)*ax0*n1*n2+(2*exponent+4)*ax0*
pow(n1,2))*vx0)*
pow(vy2,2)+(((
pow(exponent,2)+6*exponent+8)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*bx0*n1*n2+(
pow(exponent,2)+6*exponent+8)*bx0*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ax0*
pow(n2,2)+(-8*exponent-16)*ax0*n1*n2+(4*exponent+8)*ax0*
pow(n1,2))*vx0*vy0+((
pow(exponent,2)+2*exponent)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx0*n1*n2+(
pow(exponent,2)+2*exponent)*bx0*
pow(n1,2))*
pow(vx0,2))*vy2+((
pow(exponent,2)+4*exponent+4)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx0*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx0*
pow(n1,2))*
pow(vy0,3)+(((
pow(exponent,2)+2*exponent)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax0*n1*n2+(
pow(exponent,2)+2*exponent)*ax0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax0*
pow(n1,2))*vx0)*
pow(vy0,2)+(((2*exponent+4)*bx0*
pow(n2,2)+(-4*exponent-8)*bx0*n1*n2+(2*exponent+4)*bx0*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*bx0*
pow(n2,2)+(-8*exponent-16)*bx0*n1*n2+(4*exponent+8)*bx0*
pow(n1,2))*vx0*vx2+((
pow(exponent,2)+4*exponent+4)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx0*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx0*
pow(n1,2))*
pow(vx0,2))*vy0+((2*exponent+4)*ax0*
pow(n2,2)+(-4*exponent-8)*ax0*n1*n2+(2*exponent+4)*ax0*
pow(n1,2))*vx0*
pow(vx2,2)+((
pow(exponent,2)+6*exponent+8)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ax0*n1*n2+(
pow(exponent,2)+6*exponent+8)*ax0*
pow(n1,2))*
pow(vx0,2)*vx2+((
pow(exponent,2)+4*exponent+4)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax0*
pow(n1,2))*
pow(vx0,3))*vz0+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vy2,2)+(((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vy0,3)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,2)*vy0)*vy2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vy0,4)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*vx0*vx2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vx0,2)*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,3)*vx2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx0*
pow(vx0,4))*
pow(vz2,2)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx0*
pow(vz0,5)+(((
pow(exponent,2)+2*exponent)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx0*n1*n2+(
pow(exponent,2)+2*exponent)*bx0*
pow(n1,2))*vy2+((
pow(exponent,2)+4*exponent+4)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx0*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx0*
pow(n1,2))*vy0+((
pow(exponent,2)+2*exponent)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax0*n1*n2+(
pow(exponent,2)+2*exponent)*ax0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax0*
pow(n1,2))*vx0)*
pow(vz0,4)+(((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*nx0*
pow(vy2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*nx0*vy0*vy2+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*nx0*
pow(vy0,2)+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*nx0*
pow(vx2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*nx0*vx0*vx2+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vz0,3)+((((2*exponent+4)*bx0*
pow(n2,2)+(-4*exponent-8)*bx0*n1*n2+(2*exponent+4)*bx0*
pow(n1,2))*vy0+((2*exponent+4)*ax0*
pow(n2,2)+(-4*exponent-8)*ax0*n1*n2+(2*exponent+4)*ax0*
pow(n1,2))*vx0)*
pow(vy2,2)+(((
pow(exponent,2)+6*exponent+8)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*bx0*n1*n2+(
pow(exponent,2)+6*exponent+8)*bx0*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ax0*
pow(n2,2)+(-8*exponent-16)*ax0*n1*n2+(4*exponent+8)*ax0*
pow(n1,2))*vx0*vy0+((
pow(exponent,2)+2*exponent)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx0*n1*n2+(
pow(exponent,2)+2*exponent)*bx0*
pow(n1,2))*
pow(vx0,2))*vy2+((
pow(exponent,2)+4*exponent+4)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx0*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx0*
pow(n1,2))*
pow(vy0,3)+(((
pow(exponent,2)+2*exponent)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax0*n1*n2+(
pow(exponent,2)+2*exponent)*ax0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax0*
pow(n1,2))*vx0)*
pow(vy0,2)+(((2*exponent+4)*bx0*
pow(n2,2)+(-4*exponent-8)*bx0*n1*n2+(2*exponent+4)*bx0*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*bx0*
pow(n2,2)+(-8*exponent-16)*bx0*n1*n2+(4*exponent+8)*bx0*
pow(n1,2))*vx0*vx2+((
pow(exponent,2)+4*exponent+4)*bx0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx0*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx0*
pow(n1,2))*
pow(vx0,2))*vy0+((2*exponent+4)*ax0*
pow(n2,2)+(-4*exponent-8)*ax0*n1*n2+(2*exponent+4)*ax0*
pow(n1,2))*vx0*
pow(vx2,2)+((
pow(exponent,2)+6*exponent+8)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ax0*n1*n2+(
pow(exponent,2)+6*exponent+8)*ax0*
pow(n1,2))*
pow(vx0,2)*vx2+((
pow(exponent,2)+4*exponent+4)*ax0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax0*
pow(n1,2))*
pow(vx0,3))*
pow(vz0,2)+((((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vy2,2)+(((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx0*
pow(vy0,3)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,2)*vy0)*vy2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vy0,4)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx0*vx0*vx2+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,2))*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vx0,2)*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx0*
pow(vx0,3)*vx2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx0*
pow(vx0,4))*vz0)*vz2)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,5)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,4)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vz2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,7)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,5)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*vz0)*
pow(vz2,2)+(((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,4)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*vz2+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,7)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,5)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,3)+(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*vz0,-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// y0  :  
poly.add(Term6f( -((((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny0*
pow(vz0,2)+(((2*exponent+4)*by0*
pow(n2,2)+(-4*exponent-8)*by0*n1*n2+(2*exponent+4)*by0*
pow(n1,2))*vy0+((2*exponent+4)*ay0*
pow(n2,2)+(-4*exponent-8)*ay0*n1*n2+(2*exponent+4)*ay0*
pow(n1,2))*vx0)*vz0+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vz2,4)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*ny0*
pow(vz0,3)+(((6*exponent+12)*by0*
pow(n2,2)+(-12*exponent-24)*by0*n1*n2+(6*exponent+12)*by0*
pow(n1,2))*vy0+((6*exponent+12)*ay0*
pow(n2,2)+(-12*exponent-24)*ay0*n1*n2+(6*exponent+12)*ay0*
pow(n1,2))*vx0)*
pow(vz0,2)+(((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*ny0*
pow(vy0,2)+((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*ny0*
pow(vx0,2))*vz0)*
pow(vz2,3)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*ny0*
pow(vz0,4)+(((
pow(exponent,2)+2*exponent)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by0*n1*n2+(
pow(exponent,2)+2*exponent)*by0*
pow(n1,2))*vy2+((
pow(exponent,2)+8*exponent+12)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*by0*n1*n2+(
pow(exponent,2)+8*exponent+12)*by0*
pow(n1,2))*vy0+((
pow(exponent,2)+2*exponent)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay0*n1*n2+(
pow(exponent,2)+2*exponent)*ay0*
pow(n1,2))*vx2+((
pow(exponent,2)+8*exponent+12)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*ay0*n1*n2+(
pow(exponent,2)+8*exponent+12)*ay0*
pow(n1,2))*vx0)*
pow(vz0,3)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny0*
pow(vy2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*ny0*vy0*vy2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vy0,2)+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny0*
pow(vx2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*ny0*vx0*vx2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vz0,2)+((((2*exponent+4)*by0*
pow(n2,2)+(-4*exponent-8)*by0*n1*n2+(2*exponent+4)*by0*
pow(n1,2))*vy0+((2*exponent+4)*ay0*
pow(n2,2)+(-4*exponent-8)*ay0*n1*n2+(2*exponent+4)*ay0*
pow(n1,2))*vx0)*
pow(vy2,2)+(((
pow(exponent,2)+6*exponent+8)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*by0*n1*n2+(
pow(exponent,2)+6*exponent+8)*by0*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ay0*
pow(n2,2)+(-8*exponent-16)*ay0*n1*n2+(4*exponent+8)*ay0*
pow(n1,2))*vx0*vy0+((
pow(exponent,2)+2*exponent)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by0*n1*n2+(
pow(exponent,2)+2*exponent)*by0*
pow(n1,2))*
pow(vx0,2))*vy2+((
pow(exponent,2)+4*exponent+4)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by0*n1*n2+(
pow(exponent,2)+4*exponent+4)*by0*
pow(n1,2))*
pow(vy0,3)+(((
pow(exponent,2)+2*exponent)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay0*n1*n2+(
pow(exponent,2)+2*exponent)*ay0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay0*
pow(n1,2))*vx0)*
pow(vy0,2)+(((2*exponent+4)*by0*
pow(n2,2)+(-4*exponent-8)*by0*n1*n2+(2*exponent+4)*by0*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*by0*
pow(n2,2)+(-8*exponent-16)*by0*n1*n2+(4*exponent+8)*by0*
pow(n1,2))*vx0*vx2+((
pow(exponent,2)+4*exponent+4)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by0*n1*n2+(
pow(exponent,2)+4*exponent+4)*by0*
pow(n1,2))*
pow(vx0,2))*vy0+((2*exponent+4)*ay0*
pow(n2,2)+(-4*exponent-8)*ay0*n1*n2+(2*exponent+4)*ay0*
pow(n1,2))*vx0*
pow(vx2,2)+((
pow(exponent,2)+6*exponent+8)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ay0*n1*n2+(
pow(exponent,2)+6*exponent+8)*ay0*
pow(n1,2))*
pow(vx0,2)*vx2+((
pow(exponent,2)+4*exponent+4)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay0*
pow(n1,2))*
pow(vx0,3))*vz0+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vy2,2)+(((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vy0,3)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,2)*vy0)*vy2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vy0,4)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*vx0*vx2+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vx0,2)*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,3)*vx2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny0*
pow(vx0,4))*
pow(vz2,2)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny0*
pow(vz0,5)+(((
pow(exponent,2)+2*exponent)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by0*n1*n2+(
pow(exponent,2)+2*exponent)*by0*
pow(n1,2))*vy2+((
pow(exponent,2)+4*exponent+4)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by0*n1*n2+(
pow(exponent,2)+4*exponent+4)*by0*
pow(n1,2))*vy0+((
pow(exponent,2)+2*exponent)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay0*n1*n2+(
pow(exponent,2)+2*exponent)*ay0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay0*
pow(n1,2))*vx0)*
pow(vz0,4)+(((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*ny0*
pow(vy2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*ny0*vy0*vy2+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*ny0*
pow(vy0,2)+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*ny0*
pow(vx2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*ny0*vx0*vx2+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vz0,3)+((((2*exponent+4)*by0*
pow(n2,2)+(-4*exponent-8)*by0*n1*n2+(2*exponent+4)*by0*
pow(n1,2))*vy0+((2*exponent+4)*ay0*
pow(n2,2)+(-4*exponent-8)*ay0*n1*n2+(2*exponent+4)*ay0*
pow(n1,2))*vx0)*
pow(vy2,2)+(((
pow(exponent,2)+6*exponent+8)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*by0*n1*n2+(
pow(exponent,2)+6*exponent+8)*by0*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ay0*
pow(n2,2)+(-8*exponent-16)*ay0*n1*n2+(4*exponent+8)*ay0*
pow(n1,2))*vx0*vy0+((
pow(exponent,2)+2*exponent)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by0*n1*n2+(
pow(exponent,2)+2*exponent)*by0*
pow(n1,2))*
pow(vx0,2))*vy2+((
pow(exponent,2)+4*exponent+4)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by0*n1*n2+(
pow(exponent,2)+4*exponent+4)*by0*
pow(n1,2))*
pow(vy0,3)+(((
pow(exponent,2)+2*exponent)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay0*n1*n2+(
pow(exponent,2)+2*exponent)*ay0*
pow(n1,2))*vx2+((
pow(exponent,2)+4*exponent+4)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay0*
pow(n1,2))*vx0)*
pow(vy0,2)+(((2*exponent+4)*by0*
pow(n2,2)+(-4*exponent-8)*by0*n1*n2+(2*exponent+4)*by0*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*by0*
pow(n2,2)+(-8*exponent-16)*by0*n1*n2+(4*exponent+8)*by0*
pow(n1,2))*vx0*vx2+((
pow(exponent,2)+4*exponent+4)*by0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by0*n1*n2+(
pow(exponent,2)+4*exponent+4)*by0*
pow(n1,2))*
pow(vx0,2))*vy0+((2*exponent+4)*ay0*
pow(n2,2)+(-4*exponent-8)*ay0*n1*n2+(2*exponent+4)*ay0*
pow(n1,2))*vx0*
pow(vx2,2)+((
pow(exponent,2)+6*exponent+8)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ay0*n1*n2+(
pow(exponent,2)+6*exponent+8)*ay0*
pow(n1,2))*
pow(vx0,2)*vx2+((
pow(exponent,2)+4*exponent+4)*ay0*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay0*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay0*
pow(n1,2))*
pow(vx0,3))*
pow(vz0,2)+((((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vy2,2)+(((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny0*
pow(vy0,3)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,2)*vy0)*vy2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vy0,4)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny0*vx0*vx2+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,2))*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vx0,2)*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny0*
pow(vx0,3)*vx2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny0*
pow(vx0,4))*vz0)*vz2)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,5)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,4)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vz2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,7)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,5)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*vz0)*
pow(vz2,2)+(((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,4)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*vz2+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,7)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,5)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,3)+(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*vz0,-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
// x1  :  
poly.add(Term6f( ((((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vz0,2)+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vx0*vz0+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vz2,5)+((((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vz0,3)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vz0,2)+((((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*vz0)*
pow(vz2,4)+((((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0)*
pow(vz0,4)+(((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx0)*
pow(vz0,3)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vy2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*vy0*vy2+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vz0,2)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vy2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vy0*vy2+(((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx0)*
pow(vy0,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vx2,2)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx0,2)*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,3))*vz0+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy2,2)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*vy0)*vy2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vy0,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,4)*vx2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,5))*
pow(vz2,3)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vz0,5)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx2*
pow(vz0,4)+((((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vy2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*vy0*vy2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vz0,3)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vy2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx2*
pow(vy0,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vz0,2)+(((((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy2,2)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2*vy0)*vy2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vy0,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,4)*vx2)*vz0)*
pow(vz2,2)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vx2*
pow(vz0,5)+((((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0)*
pow(vy2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vx2,2))*
pow(vz0,4)+((((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx0)*
pow(vy2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx2*vy0*vy2+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx2*
pow(vy0,2)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx2,3)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*vx0*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vz0,3)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vy2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*vy0*
pow(vy2,3)+((((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0)*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)*vy0*vy2+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2))*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vx2,4)+((4*exponent+8)*k1*
pow(n2,2)+(-8*exponent-16)*k1*n1*n2+(4*exponent+8)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2))*
pow(vz0,2)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vx0*
pow(vy2,4)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vy0*
pow(vy2,3)+((((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vx0)*
pow(vy0,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vx2,2)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx0,2)*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,3))*
pow(vy2,2)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx2*
pow(vy0,3)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2)*vx2)*vy0)*vy2+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vx2*
pow(vy0,4)+(((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx2,3)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*vx0*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vy0,2)+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vx0*
pow(vx2,4)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*
pow(vx0,4)*vx2)*vz0+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy2,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*vy0)*
pow(vy2,3)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vy0,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3))*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,4)*vx2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,5))*
pow(vy2,2)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)*vy0)*vy2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vx2,2)*
pow(vy0,4)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx0*
pow(vx2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2))*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,3)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,2))*vz2+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vy2,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx2,3))*
pow(vz0,5)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vy2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*vy0*
pow(vy2,3)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vy2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)*vy0*vy2+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx2,5)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3))*
pow(vz0,3)+((((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vy2,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx2*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2*vy0)*
pow(vy2,3)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*vx2*
pow(vy0,4)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vy0,2)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,4)*vx2)*
pow(vy2,2)+(((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx2,3)*
pow(vy0,3)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3)*vy0)*vy2+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx2,3)*
pow(vy0,4)+(((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx2,5)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*vx0*
pow(vx2,4)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,3))*
pow(vy0,2)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,5)+((2*exponent+4)*k1*
pow(n2,2)+(-4*exponent-8)*k1*n1*n2+(2*exponent+4)*k1*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,4)+((exponent+2)*k1*
pow(n2,2)+(-2*exponent-4)*k1*n1*n2+(exponent+2)*k1*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,3))*vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vz0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vz0,5)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vz0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,4)+(((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,2)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vz2,4)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,5)+(((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,3)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vz0)*
pow(vz2,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,4)+(((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,4)+((48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*vx2+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,2)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,6)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*
pow(vz2,2)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,5)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz0)*vz2+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,4)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*
pow(vz0,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,4),-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
// y1  :  
poly.add(Term6f( ((((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy0*
pow(vz0,2)+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vy0*vz0+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vz2,5)+((((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vy0)*
pow(vz0,3)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vy0*
pow(vz0,2)+((((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,2)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*vz0)*
pow(vz2,4)+((((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vy2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy0)*
pow(vz0,4)+(((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vy2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*vy0)*
pow(vz0,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vy0*
pow(vy2,2)+(((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*vy0)*
pow(vz0,2)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vy0*
pow(vy2,2)+(((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vy0,2)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,2))*vy2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vy0,3)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vx2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,2))*vy0)*vz0+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,2)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2))*vy2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,5)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*vx2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4))*vy0)*
pow(vz2,3)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy2*
pow(vz0,5)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vy2*
pow(vz0,4)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy2,3)+((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*vy0*
pow(vy2,2)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*vy0)*
pow(vz0,3)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vy0*
pow(vy2,2)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vy0,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2))*vy2+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx2,2)*vy0)*
pow(vz0,2)+((((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,3)+(((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*
pow(vy0,3)+((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,2)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,4)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*vx2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vz0)*
pow(vz2,2)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vy2*
pow(vz0,5)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy2,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy0*
pow(vy2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*vy2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,2)*vy0)*
pow(vz0,4)+(((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vy2,3)+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*vy0*
pow(vy2,2)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vy0,2)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vx2+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2))*vy2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx2,2)*vy0)*
pow(vz0,3)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*vy0*
pow(vy2,4)+(((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*vy0)*
pow(vy2,2)+(((4*exponent+8)*k2*
pow(n2,2)+(-8*exponent-16)*k2*n1*n2+(4*exponent+8)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*vy2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*
pow(vx2,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*vy0)*
pow(vz0,2)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*vy0*
pow(vy2,4)+(((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vy0,2)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,3)+(((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vy0,3)+(((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vx2+((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx0,2))*vy0)*
pow(vy2,2)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*
pow(vy0,4)+(((4*
pow(exponent,2)+20*exponent+24)*
pow(n2,2)+(-8*
pow(exponent,2)-40*exponent-48)*n1*n2+(4*
pow(exponent,2)+20*exponent+24)*
pow(n1,2))*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*vx2+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*
pow(vx0,3)*vx2+((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*
pow(vx0,4))*vy2+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(((4*exponent+8)*
pow(n2,2)+(-8*exponent-16)*n1*n2+(4*exponent+8)*
pow(n1,2))*
pow(vx2,4)+((8*exponent+16)*
pow(n2,2)+(-16*exponent-32)*n1*n2+(8*exponent+16)*
pow(n1,2))*vx0*
pow(vx2,3)+((4*
pow(exponent,2)+12*exponent+8)*
pow(n2,2)+(-8*
pow(exponent,2)-24*exponent-16)*n1*n2+(4*
pow(exponent,2)+12*exponent+8)*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*vy0)*vz0+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,4)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2))*
pow(vy2,3)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,5)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*vx2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4))*vy0)*
pow(vy2,2)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,2))*vy2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*
pow(vx2,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,3)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vy0)*vz2+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy2,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,2)*vy2)*
pow(vz0,5)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy2,5)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vy0*
pow(vy2,4)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,2)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*
pow(vx2,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*vy2)*
pow(vz0,3)+((((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,2)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,5)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vy0,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,4)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vy0,4)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*vx2+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,3)+(((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,2)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+(((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*vx0*
pow(vx2,3)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+((2*exponent+4)*k2*
pow(n2,2)+(-4*exponent-8)*k2*n1*n2+(2*exponent+4)*k2*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+((exponent+2)*k2*
pow(n2,2)+(-2*exponent-4)*k2*n1*n2+(exponent+2)*k2*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vy2)*vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vz0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vz0,5)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vz0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,4)+(((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz0,2)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vz2,4)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,5)+(((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,3)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vz0)*
pow(vz2,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,4)+(((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vy0,4)+((48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*vx2+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz0,2)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,6)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*
pow(vz2,2)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,5)+(((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,3)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz0)*vz2+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,4)+(((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*vx2+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(48*M_PI*
pow(n2,2)+96*M_PI*n1*n2+48*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*
pow(vz0,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,6)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*vy0)*
pow(vy2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6))*
pow(vy2,4)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,5)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*
pow(vy0,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,6)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,4)+((24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(32*M_PI*
pow(n2,2)+64*M_PI*n1*n2+32*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(24*M_PI*
pow(n2,2)+48*M_PI*n1*n2+24*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,2))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,5)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*
pow(vy0,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,6)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,5)*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,6)*
pow(vx2,4),-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
// x2  :  
poly.add(Term6f( -(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx2*vz0*
pow(vz2,5)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*nx2*
pow(vz0,2)+(((
pow(exponent,2)+4*exponent+4)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx2*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx2*
pow(n1,2))*vy2+((
pow(exponent,2)+2*exponent)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx2*n1*n2+(
pow(exponent,2)+2*exponent)*bx2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax2*n1*n2+(
pow(exponent,2)+2*exponent)*ax2*
pow(n1,2))*vx0)*vz0)*
pow(vz2,4)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*nx2*
pow(vz0,3)+(((
pow(exponent,2)+8*exponent+12)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*bx2*n1*n2+(
pow(exponent,2)+8*exponent+12)*bx2*
pow(n1,2))*vy2+((
pow(exponent,2)+2*exponent)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx2*n1*n2+(
pow(exponent,2)+2*exponent)*bx2*
pow(n1,2))*vy0+((
pow(exponent,2)+8*exponent+12)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*ax2*n1*n2+(
pow(exponent,2)+8*exponent+12)*ax2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax2*n1*n2+(
pow(exponent,2)+2*exponent)*ax2*
pow(n1,2))*vx0)*
pow(vz0,2)+(((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*nx2*
pow(vy2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*nx2*vy0*vy2+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*nx2*
pow(vy0,2)+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*nx2*
pow(vx2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*nx2*vx0*vx2+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*nx2*
pow(vx0,2))*vz0)*
pow(vz2,3)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx2*
pow(vz0,4)+(((6*exponent+12)*bx2*
pow(n2,2)+(-12*exponent-24)*bx2*n1*n2+(6*exponent+12)*bx2*
pow(n1,2))*vy2+((6*exponent+12)*ax2*
pow(n2,2)+(-12*exponent-24)*ax2*n1*n2+(6*exponent+12)*ax2*
pow(n1,2))*vx2)*
pow(vz0,3)+(((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*
pow(vy2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*nx2*vy0*vy2+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx2*
pow(vy0,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*
pow(vx2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*nx2*vx0*vx2+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*nx2*
pow(vx0,2))*
pow(vz0,2)+(((
pow(exponent,2)+4*exponent+4)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx2*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx2*
pow(n1,2))*
pow(vy2,3)+(((
pow(exponent,2)+6*exponent+8)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*bx2*n1*n2+(
pow(exponent,2)+6*exponent+8)*bx2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax2*n1*n2+(
pow(exponent,2)+2*exponent)*ax2*
pow(n1,2))*vx0)*
pow(vy2,2)+(((2*exponent+4)*bx2*
pow(n2,2)+(-4*exponent-8)*bx2*n1*n2+(2*exponent+4)*bx2*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ax2*
pow(n2,2)+(-8*exponent-16)*ax2*n1*n2+(4*exponent+8)*ax2*
pow(n1,2))*vx2*vy0+((
pow(exponent,2)+4*exponent+4)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx2*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx2*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*bx2*
pow(n2,2)+(-8*exponent-16)*bx2*n1*n2+(4*exponent+8)*bx2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*bx2*
pow(n2,2)+(-4*exponent-8)*bx2*n1*n2+(2*exponent+4)*bx2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*ax2*
pow(n2,2)+(-4*exponent-8)*ax2*n1*n2+(2*exponent+4)*ax2*
pow(n1,2))*vx2*
pow(vy0,2)+((
pow(exponent,2)+2*exponent)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx2*n1*n2+(
pow(exponent,2)+2*exponent)*bx2*
pow(n1,2))*
pow(vx2,2)*vy0+((
pow(exponent,2)+4*exponent+4)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax2*
pow(n1,2))*
pow(vx2,3)+((
pow(exponent,2)+6*exponent+8)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ax2*n1*n2+(
pow(exponent,2)+6*exponent+8)*ax2*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*ax2*
pow(n2,2)+(-4*exponent-8)*ax2*n1*n2+(2*exponent+4)*ax2*
pow(n1,2))*
pow(vx0,2)*vx2)*vz0)*
pow(vz2,2)+((((2*exponent+4)*bx2*
pow(n2,2)+(-4*exponent-8)*bx2*n1*n2+(2*exponent+4)*bx2*
pow(n1,2))*vy2+((2*exponent+4)*ax2*
pow(n2,2)+(-4*exponent-8)*ax2*n1*n2+(2*exponent+4)*ax2*
pow(n1,2))*vx2)*
pow(vz0,4)+(((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*nx2*
pow(vy2,2)+((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*nx2*
pow(vx2,2))*
pow(vz0,3)+(((
pow(exponent,2)+4*exponent+4)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx2*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx2*
pow(n1,2))*
pow(vy2,3)+(((
pow(exponent,2)+6*exponent+8)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*bx2*n1*n2+(
pow(exponent,2)+6*exponent+8)*bx2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ax2*n1*n2+(
pow(exponent,2)+2*exponent)*ax2*
pow(n1,2))*vx0)*
pow(vy2,2)+(((2*exponent+4)*bx2*
pow(n2,2)+(-4*exponent-8)*bx2*n1*n2+(2*exponent+4)*bx2*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ax2*
pow(n2,2)+(-8*exponent-16)*ax2*n1*n2+(4*exponent+8)*ax2*
pow(n1,2))*vx2*vy0+((
pow(exponent,2)+4*exponent+4)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*bx2*n1*n2+(
pow(exponent,2)+4*exponent+4)*bx2*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*bx2*
pow(n2,2)+(-8*exponent-16)*bx2*n1*n2+(4*exponent+8)*bx2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*bx2*
pow(n2,2)+(-4*exponent-8)*bx2*n1*n2+(2*exponent+4)*bx2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*ax2*
pow(n2,2)+(-4*exponent-8)*ax2*n1*n2+(2*exponent+4)*ax2*
pow(n1,2))*vx2*
pow(vy0,2)+((
pow(exponent,2)+2*exponent)*bx2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*bx2*n1*n2+(
pow(exponent,2)+2*exponent)*bx2*
pow(n1,2))*
pow(vx2,2)*vy0+((
pow(exponent,2)+4*exponent+4)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ax2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ax2*
pow(n1,2))*
pow(vx2,3)+((
pow(exponent,2)+6*exponent+8)*ax2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ax2*n1*n2+(
pow(exponent,2)+6*exponent+8)*ax2*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*ax2*
pow(n2,2)+(-4*exponent-8)*ax2*n1*n2+(2*exponent+4)*ax2*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vz0,2)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vy2,4)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx2*vy0*
pow(vy2,3)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vy0,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx2*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx2*vx0*vx2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vx0,2))*
pow(vy2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx2*
pow(vx2,2)*vy0*vy2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vx2,2)*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vx2,4)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*nx2*vx0*
pow(vx2,3)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*nx2*
pow(vx0,2)*
pow(vx2,2))*vz0)*vz2+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vy2,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vx2,2))*
pow(vz0,4)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vy2,4)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*vy0*
pow(vy2,3)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vy0,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*vx0*vx2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vx0,2))*
pow(vy2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*
pow(vx2,2)*vy0*vy2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vx2,2)*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vx2,4)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*nx2*vx0*
pow(vx2,3)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*nx2*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,2))*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz2,7)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*vz0)*
pow(vz2,6)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,5)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+(((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,4)+(((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,2)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz2,3)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,5)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,3)+(((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+((20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vz0)*
pow(vz2,2)+(((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,4)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz2+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,3)+(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz0,-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
// y2  :  
poly.add(Term6f( -(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny2*vz0*
pow(vz2,5)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*ny2*
pow(vz0,2)+(((
pow(exponent,2)+4*exponent+4)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by2*n1*n2+(
pow(exponent,2)+4*exponent+4)*by2*
pow(n1,2))*vy2+((
pow(exponent,2)+2*exponent)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by2*n1*n2+(
pow(exponent,2)+2*exponent)*by2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay2*n1*n2+(
pow(exponent,2)+2*exponent)*ay2*
pow(n1,2))*vx0)*vz0)*
pow(vz2,4)+(((3*exponent+6)*
pow(n2,2)+(-6*exponent-12)*n1*n2+(3*exponent+6)*
pow(n1,2))*ny2*
pow(vz0,3)+(((
pow(exponent,2)+8*exponent+12)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*by2*n1*n2+(
pow(exponent,2)+8*exponent+12)*by2*
pow(n1,2))*vy2+((
pow(exponent,2)+2*exponent)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by2*n1*n2+(
pow(exponent,2)+2*exponent)*by2*
pow(n1,2))*vy0+((
pow(exponent,2)+8*exponent+12)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-16*exponent-24)*ay2*n1*n2+(
pow(exponent,2)+8*exponent+12)*ay2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay2*n1*n2+(
pow(exponent,2)+2*exponent)*ay2*
pow(n1,2))*vx0)*
pow(vz0,2)+(((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*ny2*
pow(vy2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*ny2*vy0*vy2+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*ny2*
pow(vy0,2)+((-
pow(exponent,2)-2*exponent)*
pow(n2,2)+(2*
pow(exponent,2)+4*exponent)*n1*n2+(-
pow(exponent,2)-2*exponent)*
pow(n1,2))*ny2*
pow(vx2,2)+((-2*
pow(exponent,2)-2*exponent+4)*
pow(n2,2)+(4*
pow(exponent,2)+4*exponent-8)*n1*n2+(-2*
pow(exponent,2)-2*exponent+4)*
pow(n1,2))*ny2*vx0*vx2+((-
pow(exponent,2)-exponent+2)*
pow(n2,2)+(2*
pow(exponent,2)+2*exponent-4)*n1*n2+(-
pow(exponent,2)-exponent+2)*
pow(n1,2))*ny2*
pow(vx0,2))*vz0)*
pow(vz2,3)+(((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny2*
pow(vz0,4)+(((6*exponent+12)*by2*
pow(n2,2)+(-12*exponent-24)*by2*n1*n2+(6*exponent+12)*by2*
pow(n1,2))*vy2+((6*exponent+12)*ay2*
pow(n2,2)+(-12*exponent-24)*ay2*n1*n2+(6*exponent+12)*ay2*
pow(n1,2))*vx2)*
pow(vz0,3)+(((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*
pow(vy2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*ny2*vy0*vy2+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny2*
pow(vy0,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*
pow(vx2,2)+((2*exponent+4)*
pow(n2,2)+(-4*exponent-8)*n1*n2+(2*exponent+4)*
pow(n1,2))*ny2*vx0*vx2+((exponent+2)*
pow(n2,2)+(-2*exponent-4)*n1*n2+(exponent+2)*
pow(n1,2))*ny2*
pow(vx0,2))*
pow(vz0,2)+(((
pow(exponent,2)+4*exponent+4)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by2*n1*n2+(
pow(exponent,2)+4*exponent+4)*by2*
pow(n1,2))*
pow(vy2,3)+(((
pow(exponent,2)+6*exponent+8)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*by2*n1*n2+(
pow(exponent,2)+6*exponent+8)*by2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay2*n1*n2+(
pow(exponent,2)+2*exponent)*ay2*
pow(n1,2))*vx0)*
pow(vy2,2)+(((2*exponent+4)*by2*
pow(n2,2)+(-4*exponent-8)*by2*n1*n2+(2*exponent+4)*by2*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ay2*
pow(n2,2)+(-8*exponent-16)*ay2*n1*n2+(4*exponent+8)*ay2*
pow(n1,2))*vx2*vy0+((
pow(exponent,2)+4*exponent+4)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by2*n1*n2+(
pow(exponent,2)+4*exponent+4)*by2*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*by2*
pow(n2,2)+(-8*exponent-16)*by2*n1*n2+(4*exponent+8)*by2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*by2*
pow(n2,2)+(-4*exponent-8)*by2*n1*n2+(2*exponent+4)*by2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*ay2*
pow(n2,2)+(-4*exponent-8)*ay2*n1*n2+(2*exponent+4)*ay2*
pow(n1,2))*vx2*
pow(vy0,2)+((
pow(exponent,2)+2*exponent)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by2*n1*n2+(
pow(exponent,2)+2*exponent)*by2*
pow(n1,2))*
pow(vx2,2)*vy0+((
pow(exponent,2)+4*exponent+4)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay2*
pow(n1,2))*
pow(vx2,3)+((
pow(exponent,2)+6*exponent+8)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ay2*n1*n2+(
pow(exponent,2)+6*exponent+8)*ay2*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*ay2*
pow(n2,2)+(-4*exponent-8)*ay2*n1*n2+(2*exponent+4)*ay2*
pow(n1,2))*
pow(vx0,2)*vx2)*vz0)*
pow(vz2,2)+((((2*exponent+4)*by2*
pow(n2,2)+(-4*exponent-8)*by2*n1*n2+(2*exponent+4)*by2*
pow(n1,2))*vy2+((2*exponent+4)*ay2*
pow(n2,2)+(-4*exponent-8)*ay2*n1*n2+(2*exponent+4)*ay2*
pow(n1,2))*vx2)*
pow(vz0,4)+(((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*ny2*
pow(vy2,2)+((-3*exponent-6)*
pow(n2,2)+(6*exponent+12)*n1*n2+(-3*exponent-6)*
pow(n1,2))*ny2*
pow(vx2,2))*
pow(vz0,3)+(((
pow(exponent,2)+4*exponent+4)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by2*n1*n2+(
pow(exponent,2)+4*exponent+4)*by2*
pow(n1,2))*
pow(vy2,3)+(((
pow(exponent,2)+6*exponent+8)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*by2*n1*n2+(
pow(exponent,2)+6*exponent+8)*by2*
pow(n1,2))*vy0+((
pow(exponent,2)+4*exponent+4)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay2*
pow(n1,2))*vx2+((
pow(exponent,2)+2*exponent)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*ay2*n1*n2+(
pow(exponent,2)+2*exponent)*ay2*
pow(n1,2))*vx0)*
pow(vy2,2)+(((2*exponent+4)*by2*
pow(n2,2)+(-4*exponent-8)*by2*n1*n2+(2*exponent+4)*by2*
pow(n1,2))*
pow(vy0,2)+((4*exponent+8)*ay2*
pow(n2,2)+(-8*exponent-16)*ay2*n1*n2+(4*exponent+8)*ay2*
pow(n1,2))*vx2*vy0+((
pow(exponent,2)+4*exponent+4)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*by2*n1*n2+(
pow(exponent,2)+4*exponent+4)*by2*
pow(n1,2))*
pow(vx2,2)+((4*exponent+8)*by2*
pow(n2,2)+(-8*exponent-16)*by2*n1*n2+(4*exponent+8)*by2*
pow(n1,2))*vx0*vx2+((2*exponent+4)*by2*
pow(n2,2)+(-4*exponent-8)*by2*n1*n2+(2*exponent+4)*by2*
pow(n1,2))*
pow(vx0,2))*vy2+((2*exponent+4)*ay2*
pow(n2,2)+(-4*exponent-8)*ay2*n1*n2+(2*exponent+4)*ay2*
pow(n1,2))*vx2*
pow(vy0,2)+((
pow(exponent,2)+2*exponent)*by2*
pow(n2,2)+(-2*
pow(exponent,2)-4*exponent)*by2*n1*n2+(
pow(exponent,2)+2*exponent)*by2*
pow(n1,2))*
pow(vx2,2)*vy0+((
pow(exponent,2)+4*exponent+4)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-8*exponent-8)*ay2*n1*n2+(
pow(exponent,2)+4*exponent+4)*ay2*
pow(n1,2))*
pow(vx2,3)+((
pow(exponent,2)+6*exponent+8)*ay2*
pow(n2,2)+(-2*
pow(exponent,2)-12*exponent-16)*ay2*n1*n2+(
pow(exponent,2)+6*exponent+8)*ay2*
pow(n1,2))*vx0*
pow(vx2,2)+((2*exponent+4)*ay2*
pow(n2,2)+(-4*exponent-8)*ay2*n1*n2+(2*exponent+4)*ay2*
pow(n1,2))*
pow(vx0,2)*vx2)*
pow(vz0,2)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vy2,4)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny2*vy0*
pow(vy2,3)+(((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vy0,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny2*
pow(vx2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny2*vx0*vx2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vx0,2))*
pow(vy2,2)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny2*
pow(vx2,2)*vy0*vy2+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vx2,2)*
pow(vy0,2)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vx2,4)+((-2*
pow(exponent,2)-6*exponent-4)*
pow(n2,2)+(4*
pow(exponent,2)+12*exponent+8)*n1*n2+(-2*
pow(exponent,2)-6*exponent-4)*
pow(n1,2))*ny2*vx0*
pow(vx2,3)+((-
pow(exponent,2)-3*exponent-2)*
pow(n2,2)+(2*
pow(exponent,2)+6*exponent+4)*n1*n2+(-
pow(exponent,2)-3*exponent-2)*
pow(n1,2))*ny2*
pow(vx0,2)*
pow(vx2,2))*vz0)*vz2+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vy2,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vx2,2))*
pow(vz0,4)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vy2,4)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*vy0*
pow(vy2,3)+(((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vy0,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*
pow(vx2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*vx0*vx2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vx0,2))*
pow(vy2,2)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*
pow(vx2,2)*vy0*vy2+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vx2,2)*
pow(vy0,2)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vx2,4)+((-2*exponent-4)*
pow(n2,2)+(4*exponent+8)*n1*n2+(-2*exponent-4)*
pow(n1,2))*ny2*vx0*
pow(vx2,3)+((-exponent-2)*
pow(n2,2)+(2*exponent+4)*n1*n2+(-exponent-2)*
pow(n1,2))*ny2*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,2))*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),exponent/2)*
pow(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz2,7)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,3)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*vz0)*
pow(vz2,6)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vz0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,2)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vz2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vz0,5)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vz0,3)+(((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(14*M_PI*
pow(n2,2)+28*M_PI*n1*n2+14*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*vz0)*
pow(vz2,4)+(((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vy2,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,2)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+((12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vz2,3)+(((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2))*
pow(vz0,5)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vy0*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vz0,3)+(((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,4)+((20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(20*M_PI*
pow(n2,2)+40*M_PI*n1*n2+20*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,2)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(10*M_PI*
pow(n2,2)+20*M_PI*n1*n2+10*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*vz0)*
pow(vz2,2)+(((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vy2,4)+(12*M_PI*
pow(n2,2)+24*M_PI*n1*n2+12*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(16*M_PI*
pow(n2,2)+32*M_PI*n1*n2+16*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,2)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz2+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,4)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy2,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4))*
pow(vz0,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vy0*
pow(vy2,5)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*vy0*
pow(vy2,3)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*vy0*vy2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vz0,3)+(((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy2,6)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*vy0)*
pow(vy2,5)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*vx2+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*vx2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4))*
pow(vy2,4)+((8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2)*vy0)*
pow(vy2,3)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,2)*
pow(vy0,4)+((6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*vx0*
pow(vx2,3)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,2))*
pow(vy0,2)+(6*M_PI*
pow(n2,2)+12*M_PI*n1*n2+6*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)+(8*M_PI*
pow(n2,2)+16*M_PI*n1*n2+8*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,2))*
pow(vy2,2)+((4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,3)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4)*vy0)*vy2+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,4)*
pow(vy0,4)+((2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*vx0*
pow(vx2,5)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,4))*
pow(vy0,2)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,2)*
pow(vx2,6)+(4*M_PI*
pow(n2,2)+8*M_PI*n1*n2+4*M_PI*
pow(n1,2))*
pow(vx0,3)*
pow(vx2,5)+(2*M_PI*
pow(n2,2)+4*M_PI*n1*n2+2*M_PI*
pow(n1,2))*
pow(vx0,4)*
pow(vx2,4))*vz0,-1)*
pow((vz2+vz0)*
pow(
pow(vz2,2)+2*vz0*vz2+
pow(vz0,2)+
pow(vy2,2)+2*vy0*vy2+
pow(vy0,2)+
pow(vx2,2)+2*vx0*vx2+
pow(vx0,2),-1),exponent) ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
#endif

#if 0 // beckmann
// 1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(n2,2)-2*n1*n2+pow(n1,2))*vz0*pow(vz2,5)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vz0,2)*pow(vz2,4)+((6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vz0,3)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*vy2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,3)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vz0,4)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vy0*vy2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy0,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0*vx2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,2))*pow(vz0,2))*pow(vz2,2)+((pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vz0,5)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*vy2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vy2,4)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vy0*pow(vy2,3)+((6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vy0,2)+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vy0,3)+((4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx2,2)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0*vx2+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,2))*vy0)*vy2+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vy0,4)+((2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*vx2+(2*pow(n2,2)-4*n1*n2+2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vx2,4)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*vx0*pow(vx2,3)+(6*pow(n2,2)-12*n1*n2+6*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*pow(n2,2)-8*n1*n2+4*pow(n1,2))*pow(vx0,3)*vx2+(pow(n2,2)-2*n1*n2+pow(n1,2))*pow(vx0,4))*vz0)*vz2)*pow(((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz2,6)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,3)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,5)+((6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,4)+((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy2,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz2,4)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,5)+((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+(((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vz0)*pow(vz2,3)+((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vz0,6)+((6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+(((6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(6*M_PI*pow(roughness,2)*pow(n2,2)+12*M_PI*pow(roughness,2)*n1*n2+6*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,2))*pow(vz2,2)+(((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2))*pow(vz0,5)+(((4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,2)*pow(n2,2)+8*M_PI*pow(roughness,2)*n1*n2+4*M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,3))*vz2+((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2))*pow(vz0,6)+(((M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,2)*pow(n2,2)+2*M_PI*pow(roughness,2)*n1*n2+M_PI*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,4),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// x0  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*vz0+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz2,6)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,3)+((10*pow(roughness,2)*bx0*pow(n2,2)-20*pow(roughness,2)*bx0*n1*n2+10*pow(roughness,2)*bx0*pow(n1,2))*vy0+(10*pow(roughness,2)*ax0*pow(n2,2)-20*pow(roughness,2)*ax0*n1*n2+10*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vz0,2)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,4)+(((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*vy2+((16*pow(roughness,2)+2)*bx0*pow(n2,2)+(-32*pow(roughness,2)-4)*bx0*n1*n2+(16*pow(roughness,2)+2)*bx0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+((16*pow(roughness,2)+2)*ax0*pow(n2,2)+(-32*pow(roughness,2)-4)*ax0*n1*n2+(16*pow(roughness,2)+2)*ax0*pow(n1,2))*vx0)*pow(vz0,3)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vy0*vy2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,2)+(((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+(((4*pow(roughness,2)+2)*bx0*pow(n2,2)+(-8*pow(roughness,2)-4)*bx0*n1*n2+(4*pow(roughness,2)+2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+(2*bx0*pow(n2,2)-4*bx0*n1*n2+2*bx0*pow(n1,2))*pow(vy0,3)+(((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+(2*ax0*pow(n2,2)-4*ax0*n1*n2+2*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(2*bx0*pow(n2,2)-4*bx0*n1*n2+2*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((4*pow(roughness,2)+2)*ax0*pow(n2,2)+(-8*pow(roughness,2)-4)*ax0*n1*n2+(4*pow(roughness,2)+2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(2*ax0*pow(n2,2)-4*ax0*n1*n2+2*ax0*pow(n1,2))*pow(vx0,3))*vz0+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,5)+(((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*vy2+((8*pow(roughness,2)+6)*bx0*pow(n2,2)+(-16*pow(roughness,2)-12)*bx0*n1*n2+(8*pow(roughness,2)+6)*bx0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+((8*pow(roughness,2)+6)*ax0*pow(n2,2)+(-16*pow(roughness,2)-12)*ax0*n1*n2+(8*pow(roughness,2)+6)*ax0*pow(n1,2))*vx0)*pow(vz0,4)+(((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx0*vy0*vy2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx0*pow(vy0,2)+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,3)+(((12*pow(roughness,2)*bx0*pow(n2,2)-24*pow(roughness,2)*bx0*n1*n2+12*pow(roughness,2)*bx0*pow(n1,2))*vy0+(12*pow(roughness,2)*ax0*pow(n2,2)-24*pow(roughness,2)*ax0*n1*n2+12*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+(((12*pow(roughness,2)+6)*bx0*pow(n2,2)+(-24*pow(roughness,2)-12)*bx0*n1*n2+(12*pow(roughness,2)+6)*bx0*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ax0*pow(n2,2)-48*pow(roughness,2)*ax0*n1*n2+24*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+(6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vy0,3)+(((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((12*pow(roughness,2)*bx0*pow(n2,2)-24*pow(roughness,2)*bx0*n1*n2+12*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*bx0*pow(n2,2)-48*pow(roughness,2)*bx0*n1*n2+24*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vx0,2))*vy0+(12*pow(roughness,2)*ax0*pow(n2,2)-24*pow(roughness,2)*ax0*n1*n2+12*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((12*pow(roughness,2)+6)*ax0*pow(n2,2)+(-24*pow(roughness,2)-12)*ax0*n1*n2+(12*pow(roughness,2)+6)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,3)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,4))*vz0)*pow(vz2,3)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,6)+(((6-12*pow(roughness,2))*bx0*pow(n2,2)+(24*pow(roughness,2)-12)*bx0*n1*n2+(6-12*pow(roughness,2))*bx0*pow(n1,2))*vy2+((6-2*pow(roughness,2))*bx0*pow(n2,2)+(4*pow(roughness,2)-12)*bx0*n1*n2+(6-2*pow(roughness,2))*bx0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ax0*pow(n2,2)+(24*pow(roughness,2)-12)*ax0*n1*n2+(6-12*pow(roughness,2))*ax0*pow(n1,2))*vx2+((6-2*pow(roughness,2))*ax0*pow(n2,2)+(4*pow(roughness,2)-12)*ax0*n1*n2+(6-2*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vz0,5)+(((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx0*vy0*vy2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx0*vx0*vx2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,4)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy2,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,2)+((6*bx0*pow(n2,2)-12*bx0*n1*n2+6*bx0*pow(n1,2))*pow(vy0,2)+((-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2+(16*pow(roughness,2)*ax0*pow(n2,2)-32*pow(roughness,2)*ax0*n1*n2+16*pow(roughness,2)*ax0*pow(n1,2))*vx0)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((6-16*pow(roughness,2))*bx0*pow(n2,2)+(32*pow(roughness,2)-12)*bx0*n1*n2+(6-16*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+((6-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-12)*bx0*n1*n2+(6-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,3)+(((6-16*pow(roughness,2))*ax0*pow(n2,2)+(32*pow(roughness,2)-12)*ax0*n1*n2+(6-16*pow(roughness,2))*ax0*pow(n1,2))*vx2+((6-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-12)*ax0*n1*n2+(6-4*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*bx0*pow(n2,2)-32*pow(roughness,2)*bx0*n1*n2+16*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((6-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-12)*bx0*n1*n2+(6-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(6*ax0*pow(n2,2)-12*ax0*n1*n2+6*ax0*pow(n1,2))*pow(vx0,2)*vx2+((6-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-12)*ax0*n1*n2+(6-4*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,2))*vy0)*vy2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*vx0*vx2+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz0,2)+(((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,5))*vz0+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,5)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*pow(vx2,3)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,5)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,6))*pow(vz2,2)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx0*pow(vz0,7)+(((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*vy2+((2-2*pow(roughness,2))*bx0*pow(n2,2)+(4*pow(roughness,2)-4)*bx0*n1*n2+(2-2*pow(roughness,2))*bx0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx2+((2-2*pow(roughness,2))*ax0*pow(n2,2)+(4*pow(roughness,2)-4)*ax0*n1*n2+(2-2*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vz0,6)+(((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx0*vy0*vy2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vz0,5)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy2,3)+((-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2)*pow(vy2,2)+(((2-8*pow(roughness,2))*bx0*pow(n2,2)+(16*pow(roughness,2)-4)*bx0*n1*n2+(2-8*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+((2-8*pow(roughness,2))*bx0*pow(n2,2)+(16*pow(roughness,2)-4)*bx0*n1*n2+(2-8*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2))*vy2+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vy0,3)+(((2-8*pow(roughness,2))*ax0*pow(n2,2)+(16*pow(roughness,2)-4)*ax0*n1*n2+(2-8*pow(roughness,2))*ax0*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((2-4*pow(roughness,2))*bx0*pow(n2,2)+(8*pow(roughness,2)-4)*bx0*n1*n2+(2-4*pow(roughness,2))*bx0*pow(n1,2))*pow(vx0,2)*vy0+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+((2-8*pow(roughness,2))*ax0*pow(n2,2)+(16*pow(roughness,2)-4)*ax0*n1*n2+(2-8*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,2)*vx2+((2-4*pow(roughness,2))*ax0*pow(n2,2)+(8*pow(roughness,2)-4)*ax0*n1*n2+(2-4*pow(roughness,2))*ax0*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vy2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,2)+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,2)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vy0,3)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*vy0)*vy2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx0*pow(vx0,3)*vx2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vz0,3)+(((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*bx0*pow(n2,2)-8*pow(roughness,2)*bx0*n1*n2+4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax0*pow(n2,2)-16*pow(roughness,2)*ax0*n1*n2+8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*bx0*pow(n2,2)+16*pow(roughness,2)*bx0*n1*n2-8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*bx0*pow(n2,2)+8*pow(roughness,2)*bx0*n1*n2-4*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax0*pow(n2,2)+16*pow(roughness,2)*ax0*n1*n2-8*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*bx0*pow(n2,2)-4*pow(roughness,2)*bx0*n1*n2+2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*bx0*pow(n2,2)-16*pow(roughness,2)*bx0*n1*n2+8*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*bx0*pow(n2,2)+4*pow(roughness,2)*bx0*n1*n2-2*pow(roughness,2)*bx0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ax0*pow(n2,2)-4*pow(roughness,2)*ax0*n1*n2+2*pow(roughness,2)*ax0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax0*pow(n2,2)-8*pow(roughness,2)*ax0*n1*n2+4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ax0*pow(n2,2)+8*pow(roughness,2)*ax0*n1*n2-4*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ax0*pow(n2,2)+4*pow(roughness,2)*ax0*n1*n2-2*pow(roughness,2)*ax0*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+(((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*vy0)*pow(vy2,3)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,2)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,5)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vy0,6)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2))*pow(vy0,4)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,3)*pow(vx2,3)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,4)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,5)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx0*pow(vx0,6))*vz0)*vz2)*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,7)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,6)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz2,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vz0)*pow(vz2,4)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+(((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,2))*pow(vz2,3)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,9)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,7)+(((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,3))*pow(vz2,2)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,8)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,4))*vz2+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,9)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,7)+(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,5),-1) ,  1 ,  0 ,  0 ,  0 ,  0 ,  0 )); 
// y0  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*vz0+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz2,6)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,3)+((10*pow(roughness,2)*by0*pow(n2,2)-20*pow(roughness,2)*by0*n1*n2+10*pow(roughness,2)*by0*pow(n1,2))*vy0+(10*pow(roughness,2)*ay0*pow(n2,2)-20*pow(roughness,2)*ay0*n1*n2+10*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vz0,2)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,4)+(((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*vy2+((16*pow(roughness,2)+2)*by0*pow(n2,2)+(-32*pow(roughness,2)-4)*by0*n1*n2+(16*pow(roughness,2)+2)*by0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+((16*pow(roughness,2)+2)*ay0*pow(n2,2)+(-32*pow(roughness,2)-4)*ay0*n1*n2+(16*pow(roughness,2)+2)*ay0*pow(n1,2))*vx0)*pow(vz0,3)+((2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vy0*vy2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,2)+(((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+(((4*pow(roughness,2)+2)*by0*pow(n2,2)+(-8*pow(roughness,2)-4)*by0*n1*n2+(4*pow(roughness,2)+2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+(2*by0*pow(n2,2)-4*by0*n1*n2+2*by0*pow(n1,2))*pow(vy0,3)+(((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+(2*ay0*pow(n2,2)-4*ay0*n1*n2+2*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(2*by0*pow(n2,2)-4*by0*n1*n2+2*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((4*pow(roughness,2)+2)*ay0*pow(n2,2)+(-8*pow(roughness,2)-4)*ay0*n1*n2+(4*pow(roughness,2)+2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(2*ay0*pow(n2,2)-4*ay0*n1*n2+2*ay0*pow(n1,2))*pow(vx0,3))*vz0+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz2,4)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,5)+(((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*vy2+((8*pow(roughness,2)+6)*by0*pow(n2,2)+(-16*pow(roughness,2)-12)*by0*n1*n2+(8*pow(roughness,2)+6)*by0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+((8*pow(roughness,2)+6)*ay0*pow(n2,2)+(-16*pow(roughness,2)-12)*ay0*n1*n2+(8*pow(roughness,2)+6)*ay0*pow(n1,2))*vx0)*pow(vz0,4)+(((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny0*vy0*vy2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny0*pow(vy0,2)+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,3)+(((12*pow(roughness,2)*by0*pow(n2,2)-24*pow(roughness,2)*by0*n1*n2+12*pow(roughness,2)*by0*pow(n1,2))*vy0+(12*pow(roughness,2)*ay0*pow(n2,2)-24*pow(roughness,2)*ay0*n1*n2+12*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+(((12*pow(roughness,2)+6)*by0*pow(n2,2)+(-24*pow(roughness,2)-12)*by0*n1*n2+(12*pow(roughness,2)+6)*by0*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ay0*pow(n2,2)-48*pow(roughness,2)*ay0*n1*n2+24*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+(6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vy0,3)+(((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((12*pow(roughness,2)*by0*pow(n2,2)-24*pow(roughness,2)*by0*n1*n2+12*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*by0*pow(n2,2)-48*pow(roughness,2)*by0*n1*n2+24*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vx0,2))*vy0+(12*pow(roughness,2)*ay0*pow(n2,2)-24*pow(roughness,2)*ay0*n1*n2+12*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((12*pow(roughness,2)+6)*ay0*pow(n2,2)+(-24*pow(roughness,2)-12)*ay0*n1*n2+(12*pow(roughness,2)+6)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,3)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,4))*vz0)*pow(vz2,3)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,6)+(((6-12*pow(roughness,2))*by0*pow(n2,2)+(24*pow(roughness,2)-12)*by0*n1*n2+(6-12*pow(roughness,2))*by0*pow(n1,2))*vy2+((6-2*pow(roughness,2))*by0*pow(n2,2)+(4*pow(roughness,2)-12)*by0*n1*n2+(6-2*pow(roughness,2))*by0*pow(n1,2))*vy0+((6-12*pow(roughness,2))*ay0*pow(n2,2)+(24*pow(roughness,2)-12)*ay0*n1*n2+(6-12*pow(roughness,2))*ay0*pow(n1,2))*vx2+((6-2*pow(roughness,2))*ay0*pow(n2,2)+(4*pow(roughness,2)-12)*ay0*n1*n2+(6-2*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vz0,5)+(((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny0*vy0*vy2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny0*vx0*vx2+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,4)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy2,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,2)+((6*by0*pow(n2,2)-12*by0*n1*n2+6*by0*pow(n1,2))*pow(vy0,2)+((-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2+(16*pow(roughness,2)*ay0*pow(n2,2)-32*pow(roughness,2)*ay0*n1*n2+16*pow(roughness,2)*ay0*pow(n1,2))*vx0)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((6-16*pow(roughness,2))*by0*pow(n2,2)+(32*pow(roughness,2)-12)*by0*n1*n2+(6-16*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+((6-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-12)*by0*n1*n2+(6-4*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,3)+(((6-16*pow(roughness,2))*ay0*pow(n2,2)+(32*pow(roughness,2)-12)*ay0*n1*n2+(6-16*pow(roughness,2))*ay0*pow(n1,2))*vx2+((6-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-12)*ay0*n1*n2+(6-4*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*by0*pow(n2,2)-32*pow(roughness,2)*by0*n1*n2+16*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((6-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-12)*by0*n1*n2+(6-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(6*ay0*pow(n2,2)-12*ay0*n1*n2+6*ay0*pow(n1,2))*pow(vx0,2)*vx2+((6-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-12)*ay0*n1*n2+(6-4*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,2)+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,2))*vy0)*vy2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*vx0*vx2+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz0,2)+(((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,5))*vz0+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,5)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*pow(vx2,3)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,5)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,6))*pow(vz2,2)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny0*pow(vz0,7)+(((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*vy2+((2-2*pow(roughness,2))*by0*pow(n2,2)+(4*pow(roughness,2)-4)*by0*n1*n2+(2-2*pow(roughness,2))*by0*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx2+((2-2*pow(roughness,2))*ay0*pow(n2,2)+(4*pow(roughness,2)-4)*ay0*n1*n2+(2-2*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vz0,6)+(((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny0*vy0*vy2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vz0,5)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy2,3)+((-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2)*pow(vy2,2)+(((2-8*pow(roughness,2))*by0*pow(n2,2)+(16*pow(roughness,2)-4)*by0*n1*n2+(2-8*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+((2-8*pow(roughness,2))*by0*pow(n2,2)+(16*pow(roughness,2)-4)*by0*n1*n2+(2-8*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2))*vy2+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vy0,3)+(((2-8*pow(roughness,2))*ay0*pow(n2,2)+(16*pow(roughness,2)-4)*ay0*n1*n2+(2-8*pow(roughness,2))*ay0*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((2-4*pow(roughness,2))*by0*pow(n2,2)+(8*pow(roughness,2)-4)*by0*n1*n2+(2-4*pow(roughness,2))*by0*pow(n1,2))*pow(vx0,2)*vy0+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+((2-8*pow(roughness,2))*ay0*pow(n2,2)+(16*pow(roughness,2)-4)*ay0*n1*n2+(2-8*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,2)*vx2+((2-4*pow(roughness,2))*ay0*pow(n2,2)+(8*pow(roughness,2)-4)*ay0*n1*n2+(2-4*pow(roughness,2))*ay0*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vy2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,2)+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,2)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vy0,3)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*vy0)*vy2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny0*pow(vx0,3)*vx2+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vz0,3)+(((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy2,4)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,2)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*vy0+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+((-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,4)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx2*pow(vy0,3)+((4*pow(roughness,2)*by0*pow(n2,2)-8*pow(roughness,2)*by0*n1*n2+4*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*vx2+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay0*pow(n2,2)-16*pow(roughness,2)*ay0*n1*n2+8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-8*pow(roughness,2)*by0*pow(n2,2)+16*pow(roughness,2)*by0*n1*n2-8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,3)*vx2+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy2+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vy0,5)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*vx0)*pow(vy0,4)+(-4*pow(roughness,2)*by0*pow(n2,2)+8*pow(roughness,2)*by0*n1*n2-4*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vy0,3)+((-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay0*pow(n2,2)+16*pow(roughness,2)*ay0*n1*n2-8*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+((2*pow(roughness,2)*by0*pow(n2,2)-4*pow(roughness,2)*by0*n1*n2+2*pow(roughness,2)*by0*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(roughness,2)*by0*pow(n2,2)-16*pow(roughness,2)*by0*n1*n2+8*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-2*pow(roughness,2)*by0*pow(n2,2)+4*pow(roughness,2)*by0*n1*n2-2*pow(roughness,2)*by0*pow(n1,2))*pow(vx0,4))*vy0+(2*pow(roughness,2)*ay0*pow(n2,2)-4*pow(roughness,2)*ay0*n1*n2+2*pow(roughness,2)*ay0*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay0*pow(n2,2)-8*pow(roughness,2)*ay0*n1*n2+4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-4*pow(roughness,2)*ay0*pow(n2,2)+8*pow(roughness,2)*ay0*n1*n2-4*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,4)*vx2+(-2*pow(roughness,2)*ay0*pow(n2,2)+4*pow(roughness,2)*ay0*n1*n2-2*pow(roughness,2)*ay0*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+(((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*vy0)*pow(vy2,3)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,2)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,5)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vy0,6)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2))*pow(vy0,4)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*vx2+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,3)*pow(vx2,3)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,4)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,5)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny0*pow(vx0,6))*vz0)*vz2)*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,7)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,6)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz2,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vz0)*pow(vz2,4)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+(((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,2))*pow(vz2,3)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,9)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,7)+(((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,3))*pow(vz2,2)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,8)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,4))*vz2+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,9)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,7)+(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,5),-1) ,  0 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
// x1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vz0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*vz0+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz2,7)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vz0,2)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,6)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,4)+(((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vx2+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vx0)*pow(vz0,3)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+(((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((4*pow(roughness,2)+8)*pow(n2,2)+(-8*pow(roughness,2)-16)*n1*n2+(4*pow(roughness,2)+8)*pow(n1,2))*pow(vx0,2)*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*vz0+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vy0)*vy2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vz2,5)+(((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,5)+(((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vz0,4)+(((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vy0*vy2+(((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx0,2)*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+((((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(10*k1*pow(roughness,2)*pow(n2,2)-20*k1*pow(roughness,2)*n1*n2+10*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vz0)*pow(vz2,4)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vz0,6)+(((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vx2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vx0)*pow(vz0,5)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vz0,4)+((((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*vx2+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0)*pow(vy2,2)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*vy0*vy2+((8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vx0*pow(vx2,2)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vz0,3)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,3)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(24*k1*pow(roughness,2)*pow(n2,2)-48*k1*pow(roughness,2)*n1*n2+24*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vz0,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,4)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,3)+((((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*vx2+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx0)*pow(vy0,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vx0,2)*vx2+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+(((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*vy2+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+(((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*vx0*pow(vx2,2)+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vz0+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vy0)*pow(vy2,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,4)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,5)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,6)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,7))*pow(vz2,3)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vz0,7)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx2*pow(vz0,6)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*vy0*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,5)+((((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx2*vy0*vy2+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*vx0*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,4)+(((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*pow(vy2,3)+(((16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*vy0)*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+((12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(24*k1*pow(roughness,2)*pow(n2,2)-48*k1*pow(roughness,2)*n1*n2+24*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,4)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,3)+((((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx0,2)*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*vy0)*vy2+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+(((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*vx0*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,2)+((((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,4)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*pow(vy2,3)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,5)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,6)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2)*vz0)*pow(vz2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx2*pow(vz0,7)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2))*pow(vz0,6)+((((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*vx2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx2*vy0*vy2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((4*pow(roughness,2)+8)*pow(n2,2)+(-8*pow(roughness,2)-16)*n1*n2+(4*pow(roughness,2)+8)*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,5)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0)*vy0*pow(vy2,3)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2))*vy0*vy2+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2))*pow(vy0,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(10*k1*pow(roughness,2)*pow(n2,2)-20*k1*pow(roughness,2)*n1*n2+10*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2))*pow(vz0,4)+(((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy2,4)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*vy0*pow(vy2,3)+(((8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*vx0)*pow(vy0,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*vy0)*vy2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx2,3)+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vx0*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,3)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,6)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,5)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,4)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*pow(vy2,3)+(((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(20*k1*pow(roughness,2)*pow(n2,2)-40*k1*pow(roughness,2)*n1*n2+20*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,2)+(((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2))*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2))*vy0)*vy2+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2))*pow(vy0,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,6)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2))*pow(vz0,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*pow(vy2,6)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vy0*pow(vy2,5)+(((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*vx2+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,4)+(((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*vy0)*pow(vy2,3)+(((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*vx0)*pow(vy0,4)+((-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,2)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,5)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*vy0)*vy2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,6)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,4)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,6)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2)*vz0+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy2,6)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vy0)*pow(vy2,5)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy2,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,5)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*vy0)*pow(vy2,3)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vy0,6)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3))*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5))*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,7))*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)*pow(vy0,5)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2))*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2))*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)*pow(vy0,6)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2))*pow(vy0,4)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,6)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,6)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,5)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*pow(vx2,3)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,7)*pow(vx2,2))*vz2+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy2,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3))*pow(vz0,7)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*vy0*pow(vy2,3)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*vy0*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*pow(vy0,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3))*pow(vz0,5)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy2,6)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*vy0*pow(vy2,5)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,4)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*vy0)*pow(vy2,3)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+((12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(12*k1*pow(roughness,2)*pow(n2,2)-24*k1*pow(roughness,2)*n1*n2+12*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy2,2)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3))*vy0)*vy2+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*pow(vy0,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,7)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,6)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3))*pow(vz0,3)+(((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy2,6)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2*vy0)*pow(vy2,5)+((6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy2,4)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,5)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,3)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*vy0)*pow(vy2,3)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*vx2*pow(vy0,6)+((8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,4)+((3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(16*k1*pow(roughness,2)*pow(n2,2)-32*k1*pow(roughness,2)*n1*n2+16*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*vx2)*pow(vy0,2)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*vx2)*pow(vy2,2)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*pow(vy0,5)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3))*pow(vy0,3)+((4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3))*vy0)*vy2+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,3)*pow(vy0,6)+((2*k1*pow(roughness,2)*pow(n2,2)-4*k1*pow(roughness,2)*n1*n2+2*k1*pow(roughness,2)*pow(n1,2))*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,4)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,3))*pow(vy0,4)+((k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx2,7)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,6)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,5)+(8*k1*pow(roughness,2)*pow(n2,2)-16*k1*pow(roughness,2)*n1*n2+8*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,4)+(3*k1*pow(roughness,2)*pow(n2,2)-6*k1*pow(roughness,2)*n1*n2+3*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,3))*pow(vy0,2)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,7)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,6)+(6*k1*pow(roughness,2)*pow(n2,2)-12*k1*pow(roughness,2)*n1*n2+6*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,5)+(4*k1*pow(roughness,2)*pow(n2,2)-8*k1*pow(roughness,2)*n1*n2+4*k1*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,4)+(k1*pow(roughness,2)*pow(n2,2)-2*k1*pow(roughness,2)*n1*n2+k1*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*pow(vx2,3))*vz0)*pow(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,8)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,7)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz2,6)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+(((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vz0)*pow(vz2,5)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz2,4)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,7)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*vz0)*pow(vz2,3)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,8)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+(((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+(((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,2))*pow(vz2,2)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,7)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,3))*vz2+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,8)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,4),-1) ,  0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
// y1  :  
poly.add(Term6f( expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vz0,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vy0*vz0+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vz2,7)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,3)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vy0*pow(vz0,2)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*vz0)*pow(vz2,6)+(((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,4)+(((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vy2+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vy0)*pow(vz0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+(((4*pow(roughness,2)+8)*pow(n2,2)+(-8*pow(roughness,2)-16)*n1*n2+(4*pow(roughness,2)+8)*pow(n1,2))*pow(vy0,2)+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*vz0+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,2)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2))*vy2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vz2,5)+(((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,5)+(((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*vy0)*pow(vz0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+(((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vy0,2)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,2)+(((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,2)+((10*k2*pow(roughness,2)*pow(n2,2)-20*k2*pow(roughness,2)*n1*n2+10*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*vz0)*pow(vz2,4)+(((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0)*pow(vz0,6)+(((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*vy2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*vy0)*pow(vz0,5)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,4)+(((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vy2,3)+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*vy0*pow(vy2,2)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vy0,2)+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vx0,2))*vy2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*pow(vz0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((24*k2*pow(roughness,2)*pow(n2,2)-48*k2*pow(roughness,2)*n1*n2+24*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vz0,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+(((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vy0,2)+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vy0,3)+((24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*vz0+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2))*pow(vy2,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2))*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,7)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,5)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy0)*pow(vz2,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy2*pow(vz0,7)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vy2*pow(vz0,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*vy0)*pow(vz0,5)+(((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vy2,3)+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*vy0*pow(vy2,2)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx2,2)*vy0)*pow(vz0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,5)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((24*k2*pow(roughness,2)*pow(n2,2)-48*k2*pow(roughness,2)*n1*n2+24*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+(((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vy0,2)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vy0,3)+((32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(((16*pow(roughness,2)+16)*pow(n2,2)+(-32*pow(roughness,2)-32)*n1*n2+(16*pow(roughness,2)+16)*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+((16-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-32)*n1*n2+(16-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,2)+(((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,5)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,4)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy0)*vz0)*pow(vz2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vy2*pow(vz0,7)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*vy0)*pow(vz0,6)+(((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vy2,3)+((4*pow(roughness,2)+8)*pow(n2,2)+(-8*pow(roughness,2)-16)*n1*n2+(4*pow(roughness,2)+8)*pow(n1,2))*vy0*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+((8-8*pow(roughness,2))*pow(n2,2)+(16*pow(roughness,2)-16)*n1*n2+(8-8*pow(roughness,2))*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy2+((8-12*pow(roughness,2))*pow(n2,2)+(24*pow(roughness,2)-16)*n1*n2+(8-12*pow(roughness,2))*pow(n1,2))*pow(vx2,2)*vy0)*pow(vz0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,5)+(10*k2*pow(roughness,2)*pow(n2,2)-20*k2*pow(roughness,2)*n1*n2+10*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,4)+((-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vy2,5)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+(((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vy0,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8-16*pow(roughness,2))*pow(n2,2)+(32*pow(roughness,2)-16)*n1*n2+(8-16*pow(roughness,2))*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vy0,3)+((-32*pow(roughness,2)*pow(n2,2)+64*pow(roughness,2)*n1*n2-32*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+((8*pow(roughness,2)+8)*pow(n2,2)+(-16*pow(roughness,2)-16)*n1*n2+(8*pow(roughness,2)+8)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(((8-32*pow(roughness,2))*pow(n2,2)+(64*pow(roughness,2)-16)*n1*n2+(8-32*pow(roughness,2))*pow(n1,2))*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*vx2+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*pow(n2,2)-16*n1*n2+8*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy2+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+((8-24*pow(roughness,2))*pow(n2,2)+(48*pow(roughness,2)-16)*n1*n2+(8-24*pow(roughness,2))*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vz0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,5)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,4)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(20*k2*pow(roughness,2)*pow(n2,2)-40*k2*pow(roughness,2)*n1*n2+20*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,5)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vz0,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,6)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,5)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,4)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(32*pow(roughness,2)*pow(n2,2)-64*pow(roughness,2)*n1*n2+32*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy2+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,5)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(-24*pow(roughness,2)*pow(n2,2)+48*pow(roughness,2)*n1*n2-24*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*pow(vx2,6)+(16*pow(roughness,2)*pow(n2,2)-32*pow(roughness,2)*n1*n2+16*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,5)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(-16*pow(roughness,2)*pow(n2,2)+32*pow(roughness,2)*n1*n2-16*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(-12*pow(roughness,2)*pow(n2,2)+24*pow(roughness,2)*n1*n2-12*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy0)*vz0+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,6)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vy0,2))*pow(vy2,5)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2))*pow(vy2,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,7)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,5)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,3)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*vy0)*pow(vy2,2)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,6)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy0,2))*vy2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,7)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,5)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,5)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy0,3)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,5)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*pow(vx2,2))*vy0)*vz2+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy2,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*vy2)*pow(vz0,7)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vy2,5)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*vy0*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy2)*pow(vz0,5)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy2,7)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vy0*pow(vy2,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,5)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*vy0)*pow(vy2,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(12*k2*pow(roughness,2)*pow(n2,2)-24*k2*pow(roughness,2)*n1*n2+12*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0)*pow(vy2,2)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,5)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy2)*pow(vz0,3)+(((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy2,7)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*vy0)*pow(vy2,6)+((6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy2,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vy0,5)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,3)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*vy0)*pow(vy2,4)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vy0,6)+((8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2))*pow(vy0,4)+((3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(16*k2*pow(roughness,2)*pow(n2,2)-32*k2*pow(roughness,2)*n1*n2+16*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*vx2+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4))*pow(vy0,2)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*vx2+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6))*pow(vy2,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,5)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,3)+((4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vy0)*pow(vy2,2)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,2)*pow(vy0,6)+((2*k2*pow(roughness,2)*pow(n2,2)-4*k2*pow(roughness,2)*n1*n2+2*k2*pow(roughness,2)*pow(n1,2))*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,4)+((k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*vx0*pow(vx2,5)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(8*k2*pow(roughness,2)*pow(n2,2)-16*k2*pow(roughness,2)*n1*n2+8*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,3)+(3*k2*pow(roughness,2)*pow(n2,2)-6*k2*pow(roughness,2)*n1*n2+3*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy0,2)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,2)*pow(vx2,6)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,3)*pow(vx2,5)+(6*k2*pow(roughness,2)*pow(n2,2)-12*k2*pow(roughness,2)*n1*n2+6*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,4)*pow(vx2,4)+(4*k2*pow(roughness,2)*pow(n2,2)-8*k2*pow(roughness,2)*n1*n2+4*k2*pow(roughness,2)*pow(n1,2))*pow(vx0,5)*pow(vx2,3)+(k2*pow(roughness,2)*pow(n2,2)-2*k2*pow(roughness,2)*n1*n2+k2*pow(roughness,2)*pow(n1,2))*pow(vx0,6)*pow(vx2,2))*vy2)*vz0)*pow(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz2,8)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*vz0)*pow(vz2,7)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,2)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz2,6)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+(((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,3)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*vz0)*pow(vz2,5)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,8)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,6)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vz0,4)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,2)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz2,4)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,7)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,3)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*vz0)*pow(vz2,3)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,8)+((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,6)+(((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+((48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vz0,4)+(((12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(48*M_PI*pow(roughness,4)*pow(n2,2)+96*M_PI*pow(roughness,4)*n1*n2+48*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(24*M_PI*pow(roughness,4)*pow(n2,2)+48*M_PI*pow(roughness,4)*n1*n2+24*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(12*M_PI*pow(roughness,4)*pow(n2,2)+24*M_PI*pow(roughness,4)*n1*n2+12*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,2))*pow(vz2,2)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,7)+(((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,5)+(((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(32*M_PI*pow(roughness,4)*pow(n2,2)+64*M_PI*pow(roughness,4)*n1*n2+32*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(16*M_PI*pow(roughness,4)*pow(n2,2)+32*M_PI*pow(roughness,4)*n1*n2+16*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,3))*vz2+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,8)+(((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,6)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4))*pow(vy2,4)+((4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,4)+(8*M_PI*pow(roughness,4)*pow(n2,2)+16*M_PI*pow(roughness,4)*n1*n2+8*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2)*pow(vy0,2)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,4)+(4*M_PI*pow(roughness,4)*pow(n2,2)+8*M_PI*pow(roughness,4)*n1*n2+4*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,4)*pow(vx2,4))*pow(vz0,4),-1) ,  0 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
// x2  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*vz0*pow(vz2,7)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,2)+(((2-2*pow(roughness,2))*bx2*pow(n2,2)+(4*pow(roughness,2)-4)*bx2*n1*n2+(2-2*pow(roughness,2))*bx2*pow(n1,2))*vy2+((2-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-4)*bx2*n1*n2+(2-4*pow(roughness,2))*bx2*pow(n1,2))*vy0+((2-2*pow(roughness,2))*ax2*pow(n2,2)+(4*pow(roughness,2)-4)*ax2*n1*n2+(2-2*pow(roughness,2))*ax2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-4)*ax2*n1*n2+(2-4*pow(roughness,2))*ax2*pow(n1,2))*vx0)*vz0)*pow(vz2,6)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,3)+(((6-2*pow(roughness,2))*bx2*pow(n2,2)+(4*pow(roughness,2)-12)*bx2*n1*n2+(6-2*pow(roughness,2))*bx2*pow(n1,2))*vy2+((6-12*pow(roughness,2))*bx2*pow(n2,2)+(24*pow(roughness,2)-12)*bx2*n1*n2+(6-12*pow(roughness,2))*bx2*pow(n1,2))*vy0+((6-2*pow(roughness,2))*ax2*pow(n2,2)+(4*pow(roughness,2)-12)*ax2*n1*n2+(6-2*pow(roughness,2))*ax2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ax2*pow(n2,2)+(24*pow(roughness,2)-12)*ax2*n1*n2+(6-12*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vz0,2)+(((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx2*vy0*vy2+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy0,2)+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*vx2+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,4)+(((8*pow(roughness,2)+6)*bx2*pow(n2,2)+(-16*pow(roughness,2)-12)*bx2*n1*n2+(8*pow(roughness,2)+6)*bx2*pow(n1,2))*vy2+((6-12*pow(roughness,2))*bx2*pow(n2,2)+(24*pow(roughness,2)-12)*bx2*n1*n2+(6-12*pow(roughness,2))*bx2*pow(n1,2))*vy0+((8*pow(roughness,2)+6)*ax2*pow(n2,2)+(-16*pow(roughness,2)-12)*ax2*n1*n2+(8*pow(roughness,2)+6)*ax2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ax2*pow(n2,2)+(24*pow(roughness,2)-12)*ax2*n1*n2+(6-12*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vz0,3)+(((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx2*vy0*vy2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vy0,2)+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*nx2*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,2)+(((2-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-4)*bx2*n1*n2+(2-4*pow(roughness,2))*bx2*pow(n1,2))*pow(vy2,3)+(((2-8*pow(roughness,2))*bx2*pow(n2,2)+(16*pow(roughness,2)-4)*bx2*n1*n2+(2-8*pow(roughness,2))*bx2*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-4)*ax2*n1*n2+(2-4*pow(roughness,2))*ax2*pow(n1,2))*vx2+((2-8*pow(roughness,2))*ax2*pow(n2,2)+(16*pow(roughness,2)-4)*ax2*n1*n2+(2-8*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vy2,2)+((-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-4)*bx2*n1*n2+(2-4*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2))*vy2+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vy0,2)+(((2-8*pow(roughness,2))*bx2*pow(n2,2)+(16*pow(roughness,2)-4)*bx2*n1*n2+(2-8*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+((2-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-4)*ax2*n1*n2+(2-4*pow(roughness,2))*ax2*pow(n1,2))*pow(vx2,3)+((2-8*pow(roughness,2))*ax2*pow(n2,2)+(16*pow(roughness,2)-4)*ax2*n1*n2+(2-8*pow(roughness,2))*ax2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,5)+(((16*pow(roughness,2)+2)*bx2*pow(n2,2)+(-32*pow(roughness,2)-4)*bx2*n1*n2+(16*pow(roughness,2)+2)*bx2*pow(n1,2))*vy2+((2-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-4)*bx2*n1*n2+(2-4*pow(roughness,2))*bx2*pow(n1,2))*vy0+((16*pow(roughness,2)+2)*ax2*pow(n2,2)+(-32*pow(roughness,2)-4)*ax2*n1*n2+(16*pow(roughness,2)+2)*ax2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-4)*ax2*n1*n2+(2-4*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vz0,4)+((-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx2*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx2*vy0*vy2+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy0,2)+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*nx2*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*vx2+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,3)+(((6-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-12)*bx2*n1*n2+(6-4*pow(roughness,2))*bx2*pow(n1,2))*pow(vy2,3)+((6*bx2*pow(n2,2)-12*bx2*n1*n2+6*bx2*pow(n1,2))*vy0+((6-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-12)*ax2*n1*n2+(6-4*pow(roughness,2))*ax2*pow(n1,2))*vx2+((6-16*pow(roughness,2))*ax2*pow(n2,2)+(32*pow(roughness,2)-12)*ax2*n1*n2+(6-16*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vy2,2)+(((16*pow(roughness,2)*ax2*pow(n2,2)-32*pow(roughness,2)*ax2*n1*n2+16*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0)*vy0+((6-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-12)*bx2*n1*n2+(6-4*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*bx2*pow(n2,2)-32*pow(roughness,2)*bx2*n1*n2+16*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy2+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy0,2)+(((6-16*pow(roughness,2))*bx2*pow(n2,2)+(32*pow(roughness,2)-12)*bx2*n1*n2+(6-16*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+((6-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-12)*ax2*n1*n2+(6-4*pow(roughness,2))*ax2*pow(n1,2))*pow(vx2,3)+(6*ax2*pow(n2,2)-12*ax2*n1*n2+6*ax2*pow(n1,2))*vx0*pow(vx2,2)+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+(((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy2,4)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx2*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*vy2+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx2,4)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*vz0)*pow(vz2,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vz0,6)+((10*pow(roughness,2)*bx2*pow(n2,2)-20*pow(roughness,2)*bx2*n1*n2+10*pow(roughness,2)*bx2*pow(n1,2))*vy2+(10*pow(roughness,2)*ax2*pow(n2,2)-20*pow(roughness,2)*ax2*n1*n2+10*pow(roughness,2)*ax2*pow(n1,2))*vx2)*pow(vz0,5)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vy0*vy2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vz0,4)+((6*bx2*pow(n2,2)-12*bx2*n1*n2+6*bx2*pow(n1,2))*pow(vy2,3)+(((12*pow(roughness,2)+6)*bx2*pow(n2,2)+(-24*pow(roughness,2)-12)*bx2*n1*n2+(12*pow(roughness,2)+6)*bx2*pow(n1,2))*vy0+(6*ax2*pow(n2,2)-12*ax2*n1*n2+6*ax2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ax2*pow(n2,2)+(24*pow(roughness,2)-12)*ax2*n1*n2+(6-12*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ax2*pow(n2,2)-48*pow(roughness,2)*ax2*n1*n2+24*pow(roughness,2)*ax2*pow(n1,2))*vx2*vy0+(6*bx2*pow(n2,2)-12*bx2*n1*n2+6*bx2*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*bx2*pow(n2,2)-48*pow(roughness,2)*bx2*n1*n2+24*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(12*pow(roughness,2)*bx2*pow(n2,2)-24*pow(roughness,2)*bx2*n1*n2+12*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy2+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,2)+((6-12*pow(roughness,2))*bx2*pow(n2,2)+(24*pow(roughness,2)-12)*bx2*n1*n2+(6-12*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2)*vy0+(6*ax2*pow(n2,2)-12*ax2*n1*n2+6*ax2*pow(n1,2))*pow(vx2,3)+((12*pow(roughness,2)+6)*ax2*pow(n2,2)+(-24*pow(roughness,2)-12)*ax2*n1*n2+(12*pow(roughness,2)+6)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ax2*pow(n2,2)-24*pow(roughness,2)*ax2*n1*n2+12*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,3)+(((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vy2,4)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx2*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vy0,2)+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*nx2*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx2*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx2*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*vy2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,4)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*nx2*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*pow(vz0,2)+((-2*pow(roughness,2)*bx2*pow(n2,2)+4*pow(roughness,2)*bx2*n1*n2-2*pow(roughness,2)*bx2*pow(n1,2))*pow(vy2,5)+((-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*vy0+(-2*pow(roughness,2)*ax2*pow(n2,2)+4*pow(roughness,2)*ax2*n1*n2-2*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy2,4)+((-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2))*pow(vy2,3)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy0,2)+((-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((2*pow(roughness,2)*bx2*pow(n2,2)-4*pow(roughness,2)*bx2*n1*n2+2*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,4)+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,3)+((-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-2*pow(roughness,2)*bx2*pow(n2,2)+4*pow(roughness,2)*bx2*n1*n2-2*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,3)*vx2+(2*pow(roughness,2)*bx2*pow(n2,2)-4*pow(roughness,2)*bx2*n1*n2+2*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,4))*vy2+(2*pow(roughness,2)*ax2*pow(n2,2)-4*pow(roughness,2)*ax2*n1*n2+2*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,4)+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+((-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,4)+(-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*vx0*pow(vx2,3)+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0+(-2*pow(roughness,2)*ax2*pow(n2,2)+4*pow(roughness,2)*ax2*n1*n2-2*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,5)+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*pow(roughness,2)*ax2*pow(n2,2)-4*pow(roughness,2)*ax2*n1*n2+2*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,4)*vx2)*vz0)*pow(vz2,2)+(((2*pow(roughness,2)*bx2*pow(n2,2)-4*pow(roughness,2)*bx2*n1*n2+2*pow(roughness,2)*bx2*pow(n1,2))*vy2+(2*pow(roughness,2)*ax2*pow(n2,2)-4*pow(roughness,2)*ax2*n1*n2+2*pow(roughness,2)*ax2*pow(n1,2))*vx2)*pow(vz0,6)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2))*pow(vz0,5)+((2*bx2*pow(n2,2)-4*bx2*n1*n2+2*bx2*pow(n1,2))*pow(vy2,3)+(((4*pow(roughness,2)+2)*bx2*pow(n2,2)+(-8*pow(roughness,2)-4)*bx2*n1*n2+(4*pow(roughness,2)+2)*bx2*pow(n1,2))*vy0+(2*ax2*pow(n2,2)-4*ax2*n1*n2+2*ax2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ax2*pow(n2,2)+(8*pow(roughness,2)-4)*ax2*n1*n2+(2-4*pow(roughness,2))*ax2*pow(n1,2))*vx0)*pow(vy2,2)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2*vy0+(2*bx2*pow(n2,2)-4*bx2*n1*n2+2*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,2)+((2-4*pow(roughness,2))*bx2*pow(n2,2)+(8*pow(roughness,2)-4)*bx2*n1*n2+(2-4*pow(roughness,2))*bx2*pow(n1,2))*pow(vx2,2)*vy0+(2*ax2*pow(n2,2)-4*ax2*n1*n2+2*ax2*pow(n1,2))*pow(vx2,3)+((4*pow(roughness,2)+2)*ax2*pow(n2,2)+(-8*pow(roughness,2)-4)*ax2*n1*n2+(4*pow(roughness,2)+2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy2,4)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx2*vy0*pow(vy2,3)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vy0,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx2*pow(vx2,2)*vy0*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx2,4)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*nx2*vx0*pow(vx2,3)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+((-2*pow(roughness,2)*bx2*pow(n2,2)+4*pow(roughness,2)*bx2*n1*n2-2*pow(roughness,2)*bx2*pow(n1,2))*pow(vy2,5)+((-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*vy0+(-2*pow(roughness,2)*ax2*pow(n2,2)+4*pow(roughness,2)*ax2*n1*n2-2*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy2,4)+((-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2))*pow(vy2,3)+((4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0)*pow(vy0,2)+((-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ax2*pow(n2,2)+16*pow(roughness,2)*ax2*n1*n2-8*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((2*pow(roughness,2)*bx2*pow(n2,2)-4*pow(roughness,2)*bx2*n1*n2+2*pow(roughness,2)*bx2*pow(n1,2))*pow(vy0,4)+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,3)+((-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*bx2*pow(n2,2)-8*pow(roughness,2)*bx2*n1*n2+4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ax2*pow(n2,2)-16*pow(roughness,2)*ax2*n1*n2+8*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-2*pow(roughness,2)*bx2*pow(n2,2)+4*pow(roughness,2)*bx2*n1*n2-2*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*pow(roughness,2)*bx2*pow(n2,2)-16*pow(roughness,2)*bx2*n1*n2+8*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,3)*vx2+(2*pow(roughness,2)*bx2*pow(n2,2)-4*pow(roughness,2)*bx2*n1*n2+2*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,4))*vy2+(2*pow(roughness,2)*ax2*pow(n2,2)-4*pow(roughness,2)*ax2*n1*n2+2*pow(roughness,2)*ax2*pow(n1,2))*vx2*pow(vy0,4)+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+((-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx2,4)+(-8*pow(roughness,2)*bx2*pow(n2,2)+16*pow(roughness,2)*bx2*n1*n2-8*pow(roughness,2)*bx2*pow(n1,2))*vx0*pow(vx2,3)+(-4*pow(roughness,2)*bx2*pow(n2,2)+8*pow(roughness,2)*bx2*n1*n2-4*pow(roughness,2)*bx2*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0+(-2*pow(roughness,2)*ax2*pow(n2,2)+4*pow(roughness,2)*ax2*n1*n2-2*pow(roughness,2)*ax2*pow(n1,2))*pow(vx2,5)+(-4*pow(roughness,2)*ax2*pow(n2,2)+8*pow(roughness,2)*ax2*n1*n2-4*pow(roughness,2)*ax2*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ax2*pow(n2,2)-8*pow(roughness,2)*ax2*n1*n2+4*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*pow(roughness,2)*ax2*pow(n2,2)-4*pow(roughness,2)*ax2*n1*n2+2*pow(roughness,2)*ax2*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,2)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,6)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*vy0*pow(vy2,5)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+((24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*pow(vy2,3)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+((24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,6)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,5)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*pow(vx2,3)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4)*pow(vx2,2))*vz0)*vz2+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2))*pow(vz0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vy0*pow(vy2,3)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*vy0*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vy2,6)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vy0*pow(vy2,5)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,3)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*vy0)*pow(vy2,3)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vy0,4)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2))*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,2)*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,3)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vx2,6)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*vx0*pow(vx2,5)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,3)*pow(vx2,3)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*nx2*pow(vx0,4)*pow(vx2,2))*pow(vz0,2))*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz2,9)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz2,7)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vz0)*pow(vz2,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,2)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz2,5)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*vz0)*pow(vz2,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,6)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,2))*pow(vz2,3)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,7)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,3))*pow(vz2,2)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,6)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,4))*vz2+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,7)+(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,5),-1) ,  0 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
// y2  :  
poly.add(Term6f( -expf(-(pow(vy2,2)+2*vy0*vy2+pow(vy0,2)+pow(vx2,2)+2*vx0*vx2+pow(vx0,2))*pow(pow(roughness,2)*pow(vz2,2)+2*pow(roughness,2)*vz0*vz2+pow(roughness,2)*pow(vz0,2)+pow(roughness,2)*pow(vy2,2)+2*pow(roughness,2)*vy0*vy2+pow(roughness,2)*pow(vy0,2)+pow(roughness,2)*pow(vx2,2)+2*pow(roughness,2)*vx0*vx2+pow(roughness,2)*pow(vx0,2),-1))*((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*vz0*pow(vz2,7)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,2)+(((2-2*pow(roughness,2))*by2*pow(n2,2)+(4*pow(roughness,2)-4)*by2*n1*n2+(2-2*pow(roughness,2))*by2*pow(n1,2))*vy2+((2-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-4)*by2*n1*n2+(2-4*pow(roughness,2))*by2*pow(n1,2))*vy0+((2-2*pow(roughness,2))*ay2*pow(n2,2)+(4*pow(roughness,2)-4)*ay2*n1*n2+(2-2*pow(roughness,2))*ay2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-4)*ay2*n1*n2+(2-4*pow(roughness,2))*ay2*pow(n1,2))*vx0)*vz0)*pow(vz2,6)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,3)+(((6-2*pow(roughness,2))*by2*pow(n2,2)+(4*pow(roughness,2)-12)*by2*n1*n2+(6-2*pow(roughness,2))*by2*pow(n1,2))*vy2+((6-12*pow(roughness,2))*by2*pow(n2,2)+(24*pow(roughness,2)-12)*by2*n1*n2+(6-12*pow(roughness,2))*by2*pow(n1,2))*vy0+((6-2*pow(roughness,2))*ay2*pow(n2,2)+(4*pow(roughness,2)-12)*ay2*n1*n2+(6-2*pow(roughness,2))*ay2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ay2*pow(n2,2)+(24*pow(roughness,2)-12)*ay2*n1*n2+(6-12*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vz0,2)+(((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny2*vy0*vy2+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy0,2)+((5*pow(roughness,2)-2)*pow(n2,2)+(4-10*pow(roughness,2))*n1*n2+(5*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx2,2)+((12*pow(roughness,2)-4)*pow(n2,2)+(8-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*vx2+((6*pow(roughness,2)-2)*pow(n2,2)+(4-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2))*vz0)*pow(vz2,5)+((10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,4)+(((8*pow(roughness,2)+6)*by2*pow(n2,2)+(-16*pow(roughness,2)-12)*by2*n1*n2+(8*pow(roughness,2)+6)*by2*pow(n1,2))*vy2+((6-12*pow(roughness,2))*by2*pow(n2,2)+(24*pow(roughness,2)-12)*by2*n1*n2+(6-12*pow(roughness,2))*by2*pow(n1,2))*vy0+((8*pow(roughness,2)+6)*ay2*pow(n2,2)+(-16*pow(roughness,2)-12)*ay2*n1*n2+(8*pow(roughness,2)+6)*ay2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ay2*pow(n2,2)+(24*pow(roughness,2)-12)*ay2*n1*n2+(6-12*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vz0,3)+(((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vy2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny2*vy0*vy2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vy0,2)+((9*pow(roughness,2)-4)*pow(n2,2)+(8-18*pow(roughness,2))*n1*n2+(9*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)+((28*pow(roughness,2)-8)*pow(n2,2)+(16-56*pow(roughness,2))*n1*n2+(28*pow(roughness,2)-8)*pow(n1,2))*ny2*vx0*vx2+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,2)+(((2-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-4)*by2*n1*n2+(2-4*pow(roughness,2))*by2*pow(n1,2))*pow(vy2,3)+(((2-8*pow(roughness,2))*by2*pow(n2,2)+(16*pow(roughness,2)-4)*by2*n1*n2+(2-8*pow(roughness,2))*by2*pow(n1,2))*vy0+((2-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-4)*ay2*n1*n2+(2-4*pow(roughness,2))*ay2*pow(n1,2))*vx2+((2-8*pow(roughness,2))*ay2*pow(n2,2)+(16*pow(roughness,2)-4)*ay2*n1*n2+(2-8*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vy2,2)+((-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,2)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0*vy0+((2-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-4)*by2*n1*n2+(2-4*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2))*vy2+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vy0,2)+(((2-8*pow(roughness,2))*by2*pow(n2,2)+(16*pow(roughness,2)-4)*by2*n1*n2+(2-8*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+((2-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-4)*ay2*n1*n2+(2-4*pow(roughness,2))*ay2*pow(n1,2))*pow(vx2,3)+((2-8*pow(roughness,2))*ay2*pow(n2,2)+(16*pow(roughness,2)-4)*ay2*n1*n2+(2-8*pow(roughness,2))*ay2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*vz0)*pow(vz2,4)+((5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,5)+(((16*pow(roughness,2)+2)*by2*pow(n2,2)+(-32*pow(roughness,2)-4)*by2*n1*n2+(16*pow(roughness,2)+2)*by2*pow(n1,2))*vy2+((2-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-4)*by2*n1*n2+(2-4*pow(roughness,2))*by2*pow(n1,2))*vy0+((16*pow(roughness,2)+2)*ay2*pow(n2,2)+(-32*pow(roughness,2)-4)*ay2*n1*n2+(16*pow(roughness,2)+2)*ay2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-4)*ay2*n1*n2+(2-4*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vz0,4)+((-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny2*pow(vy2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny2*vy0*vy2+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy0,2)+(-2*pow(n2,2)+4*n1*n2-2*pow(n1,2))*ny2*pow(vx2,2)+((20*pow(roughness,2)-4)*pow(n2,2)+(8-40*pow(roughness,2))*n1*n2+(20*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*vx2+((10*pow(roughness,2)-2)*pow(n2,2)+(4-20*pow(roughness,2))*n1*n2+(10*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,3)+(((6-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-12)*by2*n1*n2+(6-4*pow(roughness,2))*by2*pow(n1,2))*pow(vy2,3)+((6*by2*pow(n2,2)-12*by2*n1*n2+6*by2*pow(n1,2))*vy0+((6-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-12)*ay2*n1*n2+(6-4*pow(roughness,2))*ay2*pow(n1,2))*vx2+((6-16*pow(roughness,2))*ay2*pow(n2,2)+(32*pow(roughness,2)-12)*ay2*n1*n2+(6-16*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vy2,2)+(((16*pow(roughness,2)*ay2*pow(n2,2)-32*pow(roughness,2)*ay2*n1*n2+16*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0)*vy0+((6-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-12)*by2*n1*n2+(6-4*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2)+(16*pow(roughness,2)*by2*pow(n2,2)-32*pow(roughness,2)*by2*n1*n2+16*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy2+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy0,2)+(((6-16*pow(roughness,2))*by2*pow(n2,2)+(32*pow(roughness,2)-12)*by2*n1*n2+(6-16*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2)+(-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+((6-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-12)*ay2*n1*n2+(6-4*pow(roughness,2))*ay2*pow(n1,2))*pow(vx2,3)+(6*ay2*pow(n2,2)-12*ay2*n1*n2+6*ay2*pow(n1,2))*vx0*pow(vx2,2)+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*pow(vz0,2)+(((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy2,4)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny2*vy0*pow(vy2,3)+(((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy0,2)+((14*pow(roughness,2)-4)*pow(n2,2)+(8-28*pow(roughness,2))*n1*n2+(14*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*vx2+((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+((20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+(((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)+(40*pow(roughness,2)*pow(n2,2)-80*pow(roughness,2)*n1*n2+40*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*vy2+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+(((12*pow(roughness,2)-2)*pow(n2,2)+(4-24*pow(roughness,2))*n1*n2+(12*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(10*pow(roughness,2)*pow(n2,2)-20*pow(roughness,2)*n1*n2+10*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+((7*pow(roughness,2)-2)*pow(n2,2)+(4-14*pow(roughness,2))*n1*n2+(7*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx2,4)+((24*pow(roughness,2)-4)*pow(n2,2)+(8-48*pow(roughness,2))*n1*n2+(24*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*pow(vx2,3)+((32*pow(roughness,2)-2)*pow(n2,2)+(4-64*pow(roughness,2))*n1*n2+(32*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(20*pow(roughness,2)*pow(n2,2)-40*pow(roughness,2)*n1*n2+20*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(5*pow(roughness,2)*pow(n2,2)-10*pow(roughness,2)*n1*n2+5*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*vz0)*pow(vz2,3)+((pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vz0,6)+((10*pow(roughness,2)*by2*pow(n2,2)-20*pow(roughness,2)*by2*n1*n2+10*pow(roughness,2)*by2*pow(n1,2))*vy2+(10*pow(roughness,2)*ay2*pow(n2,2)-20*pow(roughness,2)*ay2*n1*n2+10*pow(roughness,2)*ay2*pow(n1,2))*vx2)*pow(vz0,5)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vy0*vy2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vz0,4)+((6*by2*pow(n2,2)-12*by2*n1*n2+6*by2*pow(n1,2))*pow(vy2,3)+(((12*pow(roughness,2)+6)*by2*pow(n2,2)+(-24*pow(roughness,2)-12)*by2*n1*n2+(12*pow(roughness,2)+6)*by2*pow(n1,2))*vy0+(6*ay2*pow(n2,2)-12*ay2*n1*n2+6*ay2*pow(n1,2))*vx2+((6-12*pow(roughness,2))*ay2*pow(n2,2)+(24*pow(roughness,2)-12)*ay2*n1*n2+(6-12*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vy2,2)+((12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,2)+(24*pow(roughness,2)*ay2*pow(n2,2)-48*pow(roughness,2)*ay2*n1*n2+24*pow(roughness,2)*ay2*pow(n1,2))*vx2*vy0+(6*by2*pow(n2,2)-12*by2*n1*n2+6*by2*pow(n1,2))*pow(vx2,2)+(24*pow(roughness,2)*by2*pow(n2,2)-48*pow(roughness,2)*by2*n1*n2+24*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(12*pow(roughness,2)*by2*pow(n2,2)-24*pow(roughness,2)*by2*n1*n2+12*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy2+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,2)+((6-12*pow(roughness,2))*by2*pow(n2,2)+(24*pow(roughness,2)-12)*by2*n1*n2+(6-12*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2)*vy0+(6*ay2*pow(n2,2)-12*ay2*n1*n2+6*ay2*pow(n1,2))*pow(vx2,3)+((12*pow(roughness,2)+6)*ay2*pow(n2,2)+(-24*pow(roughness,2)-12)*ay2*n1*n2+(12*pow(roughness,2)+6)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(12*pow(roughness,2)*ay2*pow(n2,2)-24*pow(roughness,2)*ay2*n1*n2+12*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,3)+(((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vy2,4)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny2*vy0*pow(vy2,3)+(((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vy0,2)+((6*pow(roughness,2)-8)*pow(n2,2)+(16-12*pow(roughness,2))*n1*n2+(6*pow(roughness,2)-8)*pow(n1,2))*ny2*pow(vx2,2)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny2*vx0*vx2+((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+((4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+(((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny2*pow(vx2,2)+(8*pow(roughness,2)*pow(n2,2)-16*pow(roughness,2)*n1*n2+8*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*vy2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+(((4*pow(roughness,2)-4)*pow(n2,2)+(8-8*pow(roughness,2))*n1*n2+(4*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(2*pow(roughness,2)*pow(n2,2)-4*pow(roughness,2)*n1*n2+2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+((3*pow(roughness,2)-4)*pow(n2,2)+(8-6*pow(roughness,2))*n1*n2+(3*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,4)+((8*pow(roughness,2)-8)*pow(n2,2)+(16-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-8)*pow(n1,2))*ny2*vx0*pow(vx2,3)+((8*pow(roughness,2)-4)*pow(n2,2)+(8-16*pow(roughness,2))*n1*n2+(8*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(4*pow(roughness,2)*pow(n2,2)-8*pow(roughness,2)*n1*n2+4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(pow(roughness,2)*pow(n2,2)-2*pow(roughness,2)*n1*n2+pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*pow(vz0,2)+((-2*pow(roughness,2)*by2*pow(n2,2)+4*pow(roughness,2)*by2*n1*n2-2*pow(roughness,2)*by2*pow(n1,2))*pow(vy2,5)+((-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*vy0+(-2*pow(roughness,2)*ay2*pow(n2,2)+4*pow(roughness,2)*ay2*n1*n2-2*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy2,4)+((-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2))*pow(vy2,3)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy0,2)+((-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((2*pow(roughness,2)*by2*pow(n2,2)-4*pow(roughness,2)*by2*n1*n2+2*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,4)+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,3)+((-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-2*pow(roughness,2)*by2*pow(n2,2)+4*pow(roughness,2)*by2*n1*n2-2*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,3)*vx2+(2*pow(roughness,2)*by2*pow(n2,2)-4*pow(roughness,2)*by2*n1*n2+2*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,4))*vy2+(2*pow(roughness,2)*ay2*pow(n2,2)-4*pow(roughness,2)*ay2*n1*n2+2*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,4)+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+((-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,4)+(-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*vx0*pow(vx2,3)+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0+(-2*pow(roughness,2)*ay2*pow(n2,2)+4*pow(roughness,2)*ay2*n1*n2-2*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,5)+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*pow(roughness,2)*ay2*pow(n2,2)-4*pow(roughness,2)*ay2*n1*n2+2*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,4)*vx2)*vz0)*pow(vz2,2)+(((2*pow(roughness,2)*by2*pow(n2,2)-4*pow(roughness,2)*by2*n1*n2+2*pow(roughness,2)*by2*pow(n1,2))*vy2+(2*pow(roughness,2)*ay2*pow(n2,2)-4*pow(roughness,2)*ay2*n1*n2+2*pow(roughness,2)*ay2*pow(n1,2))*vx2)*pow(vz0,6)+((-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,2)+(-5*pow(roughness,2)*pow(n2,2)+10*pow(roughness,2)*n1*n2-5*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2))*pow(vz0,5)+((2*by2*pow(n2,2)-4*by2*n1*n2+2*by2*pow(n1,2))*pow(vy2,3)+(((4*pow(roughness,2)+2)*by2*pow(n2,2)+(-8*pow(roughness,2)-4)*by2*n1*n2+(4*pow(roughness,2)+2)*by2*pow(n1,2))*vy0+(2*ay2*pow(n2,2)-4*ay2*n1*n2+2*ay2*pow(n1,2))*vx2+((2-4*pow(roughness,2))*ay2*pow(n2,2)+(8*pow(roughness,2)-4)*ay2*n1*n2+(2-4*pow(roughness,2))*ay2*pow(n1,2))*vx0)*pow(vy2,2)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,2)+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2*vy0+(2*by2*pow(n2,2)-4*by2*n1*n2+2*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy2+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,2)+((2-4*pow(roughness,2))*by2*pow(n2,2)+(8*pow(roughness,2)-4)*by2*n1*n2+(2-4*pow(roughness,2))*by2*pow(n1,2))*pow(vx2,2)*vy0+(2*ay2*pow(n2,2)-4*ay2*n1*n2+2*ay2*pow(n1,2))*pow(vx2,3)+((4*pow(roughness,2)+2)*ay2*pow(n2,2)+(-8*pow(roughness,2)-4)*ay2*n1*n2+(4*pow(roughness,2)+2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vz0,4)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy2,4)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny2*vy0*pow(vy2,3)+(((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vy0,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*vx2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny2*pow(vx2,2)*vy0*vy2+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,2)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx2,4)+((-4*pow(roughness,2)-4)*pow(n2,2)+(8*pow(roughness,2)+8)*n1*n2+(-4*pow(roughness,2)-4)*pow(n1,2))*ny2*vx0*pow(vx2,3)+((-2*pow(roughness,2)-2)*pow(n2,2)+(4*pow(roughness,2)+4)*n1*n2+(-2*pow(roughness,2)-2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+((-2*pow(roughness,2)*by2*pow(n2,2)+4*pow(roughness,2)*by2*n1*n2-2*pow(roughness,2)*by2*pow(n1,2))*pow(vy2,5)+((-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*vy0+(-2*pow(roughness,2)*ay2*pow(n2,2)+4*pow(roughness,2)*ay2*n1*n2-2*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy2,4)+((-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0*vy0+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2))*pow(vy2,3)+((4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,3)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0)*pow(vy0,2)+((-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*vy0+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,3)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(-8*pow(roughness,2)*ay2*pow(n2,2)+16*pow(roughness,2)*ay2*n1*n2-8*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3))*pow(vy2,2)+((2*pow(roughness,2)*by2*pow(n2,2)-4*pow(roughness,2)*by2*n1*n2+2*pow(roughness,2)*by2*pow(n1,2))*pow(vy0,4)+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,3)+((-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*vx0*vx2+(4*pow(roughness,2)*by2*pow(n2,2)-8*pow(roughness,2)*by2*n1*n2+4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2))*pow(vy0,2)+((8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(8*pow(roughness,2)*ay2*pow(n2,2)-16*pow(roughness,2)*ay2*n1*n2+8*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*vy0+(-2*pow(roughness,2)*by2*pow(n2,2)+4*pow(roughness,2)*by2*n1*n2-2*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,4)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2)*pow(vx2,2)+(8*pow(roughness,2)*by2*pow(n2,2)-16*pow(roughness,2)*by2*n1*n2+8*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,3)*vx2+(2*pow(roughness,2)*by2*pow(n2,2)-4*pow(roughness,2)*by2*n1*n2+2*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,4))*vy2+(2*pow(roughness,2)*ay2*pow(n2,2)-4*pow(roughness,2)*ay2*n1*n2+2*pow(roughness,2)*ay2*pow(n1,2))*vx2*pow(vy0,4)+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,2)*pow(vy0,3)+((4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,2)+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,2)*vx2)*pow(vy0,2)+((-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx2,4)+(-8*pow(roughness,2)*by2*pow(n2,2)+16*pow(roughness,2)*by2*n1*n2-8*pow(roughness,2)*by2*pow(n1,2))*vx0*pow(vx2,3)+(-4*pow(roughness,2)*by2*pow(n2,2)+8*pow(roughness,2)*by2*n1*n2-4*pow(roughness,2)*by2*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vy0+(-2*pow(roughness,2)*ay2*pow(n2,2)+4*pow(roughness,2)*ay2*n1*n2-2*pow(roughness,2)*ay2*pow(n1,2))*pow(vx2,5)+(-4*pow(roughness,2)*ay2*pow(n2,2)+8*pow(roughness,2)*ay2*n1*n2-4*pow(roughness,2)*ay2*pow(n1,2))*vx0*pow(vx2,4)+(4*pow(roughness,2)*ay2*pow(n2,2)-8*pow(roughness,2)*ay2*n1*n2+4*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,3)*pow(vx2,2)+(2*pow(roughness,2)*ay2*pow(n2,2)-4*pow(roughness,2)*ay2*n1*n2+2*pow(roughness,2)*ay2*pow(n1,2))*pow(vx0,4)*vx2)*pow(vz0,2)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,6)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*vy0*pow(vy2,5)+((18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,4)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+((24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*pow(vy2,3)+((3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+((24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+(9*pow(roughness,2)*pow(n2,2)-18*pow(roughness,2)*n1*n2+9*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*pow(vy2,2)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,3)+((12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(24*pow(roughness,2)*pow(n2,2)-48*pow(roughness,2)*n1*n2+24*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*vy0)*vy2+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,4)+((6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(6*pow(roughness,2)*pow(n2,2)-12*pow(roughness,2)*n1*n2+6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,6)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,5)+(18*pow(roughness,2)*pow(n2,2)-36*pow(roughness,2)*n1*n2+18*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,4)+(12*pow(roughness,2)*pow(n2,2)-24*pow(roughness,2)*n1*n2+12*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*pow(vx2,3)+(3*pow(roughness,2)*pow(n2,2)-6*pow(roughness,2)*n1*n2+3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4)*pow(vx2,2))*vz0)*vz2+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2))*pow(vz0,6)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vy0*pow(vy2,3)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*vy0*vy2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,2)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vy2,6)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vy0*pow(vy2,5)+((-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy2,4)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,3)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*vy0)*pow(vy2,3)+((-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vy0,4)+((-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*vx2+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2))*pow(vy0,2)+(-3*pow(roughness,2)*pow(n2,2)+6*pow(roughness,2)*n1*n2-3*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*vx2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4))*pow(vy2,2)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,3)+((-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(-8*pow(roughness,2)*pow(n2,2)+16*pow(roughness,2)*n1*n2-8*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*vy0)*vy2+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,2)*pow(vy0,4)+((-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,3)+(-2*pow(roughness,2)*pow(n2,2)+4*pow(roughness,2)*n1*n2-2*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,2))*pow(vy0,2)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vx2,6)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*vx0*pow(vx2,5)+(-6*pow(roughness,2)*pow(n2,2)+12*pow(roughness,2)*n1*n2-6*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,2)*pow(vx2,4)+(-4*pow(roughness,2)*pow(n2,2)+8*pow(roughness,2)*n1*n2-4*pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,3)*pow(vx2,3)+(-pow(roughness,2)*pow(n2,2)+2*pow(roughness,2)*n1*n2-pow(roughness,2)*pow(n1,2))*ny2*pow(vx0,4)*pow(vx2,2))*pow(vz0,2))*pow(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz2,9)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,3)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*vz0)*pow(vz2,8)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,2)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz2,7)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,5)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,3)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*vz0)*pow(vz2,6)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,6)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,4)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,2)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz2,5)+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vz0,7)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vz0,5)+((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,3)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*vz0)*pow(vz2,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,6)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,4)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,2))*pow(vz2,3)+(((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2))*pow(vz0,7)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vz0,5)+(((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(20*M_PI*pow(roughness,4)*pow(n2,2)+40*M_PI*pow(roughness,4)*n1*n2+20*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,3))*pow(vz2,2)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,6)+(((5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(10*M_PI*pow(roughness,4)*pow(n2,2)+20*M_PI*pow(roughness,4)*n1*n2+10*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(5*M_PI*pow(roughness,4)*pow(n2,2)+10*M_PI*pow(roughness,4)*n1*n2+5*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,4))*vz2+((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy2,4)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4))*pow(vz0,7)+(((M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2))*pow(vy2,4)+((2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,2)*pow(vy0,2)+(2*M_PI*pow(roughness,4)*pow(n2,2)+4*M_PI*pow(roughness,4)*n1*n2+2*M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,2))*pow(vy2,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx2,4)*pow(vy0,2)+(M_PI*pow(roughness,4)*pow(n2,2)+2*M_PI*pow(roughness,4)*n1*n2+M_PI*pow(roughness,4)*pow(n1,2))*pow(vx0,2)*pow(vx2,4))*pow(vz0,5),-1) ,  0 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
#endif

std::cout << " raw reflect " << std::endl;
std::cout << poly << std::endl;
// assert(0);

}

}

#endif

